#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

class Session < ActiveRecord::Base
  # セッションタイムアウトの処理
  # この処理ではcreated_atとupdated_atを1年前に変更して、続くtrim sessionの
  # 処理で実際にセッションを削除する形になっていることに注意。
  # なお、ユーザidの
  def self.expire
    self.transaction do
      expires = self.lock.where('updated_at < ?', Time.current.ago(Sidfm.session_timeout.minutes))
      Rails.logger.info("expire session:")
      expires_count = 0
      user_ids = []
      expires.each do |expire_session|
        expire_session.update_attribute :updated_at, Time.current.ago(1.year)
        expires_count += 1
        begin
          data = JSON.parse(expire_session.data || '{}')
          if data["value"] and data["value"]["user_id"] then
            user_ids << data["value"]["user_id"].to_i
            user = User.where(:id => data["value"]["user_id"])
            if user.count > 0 then
              user = user.first
              Rails.logger.info(" expire user session: user_id=#{data["value"]["user_id"]}(login: #{user.login})")
            else
              Rails.logger.info(" expire user session: user_id=#{data["value"]["user_id"]}")
            end
          end
        rescue JSON::ParserError
        end
      end
      Rails.logger.info(" expired #{expires_count} sessions")
      #不要なOicSessionを削除
      OicSession.where(:user_id => nil).where('updated_at < ?', Time.current.ago(Sidfm.session_timeout.minutes)).destroy_all
      #ログアウトせずにセッションが切れたユーザのOicSessionを削除
      OicSession.where(:user_id => user_ids).destroy_all
    end
  end

  def self.delete_session(user_id)
    Session.all.each do |sesn|
      begin
        data = JSON.parse(sesn.data || '{}')
      rescue JSON::ParserError
        data = {}
      end
      if data["value"] and data["value"]["user_id"] and user_id == data["value"]["user_id"].to_i then
        user = User.where(:id => data["value"]["user_id"])
        if user.count > 0 then
          user = user.first
          Rails.logger.info(" expire user session: user_id=#{data["value"]["user_id"]}(login: #{user.login})")
        else
          Rails.logger.info(" expire user session: user_id=#{data["value"]["user_id"]}")
        end
        sesn.destroy
        break
      end
    end
  end

end
