#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

class SidfmdCtag < ActiveRecord::Base
  belongs_to :sidfmd_category
  has_many :sidfmd_issue_ctags, dependent: :destroy
  has_many :sidfmd_vv_ctags, dependent: :destroy
  has_many :sidfmd_hv_ctags, dependent: :destroy
  has_many :sidfmd_rule_cm_ctags, dependent: :destroy
  has_many :sidfmd_rule_am_ctags, dependent: :destroy

  validates :keyword, uniqueness: true

  # ユーザの操作可能なカスタムタグのidを返す
  # 閲覧はsidfmd_category_id = nilも可能
  scope :user_ctag_ids, -> (role_admin, user_category_ids) {
    role_admin ?
      pluck(:id) :
      where(:sidfmd_category_id => user_category_ids).pluck(:id) 
  }

  # ユーザの操作不可能なカスタムタグのidを返す
  # 閲覧はsidfmd_category_id = nilも可能
  scope :not_user_ctag_ids, -> (role_admin, user_category_ids) {
    role_admin ?
      pluck(:id) :
      where.not(:sidfmd_category_id => user_category_ids).pluck(:id)         
  }

  # ユーザの脆弱性/ルールに紐付け可能なカスタムタグのidを返す
  scope :user_relatable_ctag_ids, -> (role_admin, user_category_ids) {
    role_admin ?
      pluck(:id) :
      where("sidfmd_category_id IS NULL OR sidfmd_category_id IN (?)", user_category_ids).pluck(:id)
  }

  #　同じkeyword(タグ名)が登録されているか否かを返す
  def self.is_same_keyword(keyword, id = '')
    target = self.where(:keyword => keyword)
    if id.to_s.length > 0 then
      target = target.where.not(:id => id)
    end
    ret = target.exists?
  end

  # サマリー表の表示等で使う元データ
  def self.create_content(**opts)
    content = self.joins(%|
      LEFT OUTER JOIN sidfmd_categories ON sidfmd_categories.id = sidfmd_ctags.sidfmd_category_id
    |)
    .select(
      'sidfmd_ctags.id AS id',
      :priority,
      'sidfmd_ctags.keyword AS name',
      :color,
      :bgcolor,
      'sidfmd_categories.id AS category_id',
      'sidfmd_categories.name AS category_name',
      'sidfmd_categories.desc AS category_desc'
    )
    unless opts[:role_admin] then
      content = content.where("sidfmd_category_id IS NULL OR sidfmd_category_id IN (?)", opts[:user_category_ids])
    end

    content
  end

  # 指定したカスタムタグ(のid)のうち、脆弱性と紐づいているカスタムタグのタグ名を返す
  def self.issue_related_ctags(ctag_ids)
    self.where(:id => SidfmdIssueCtag.where(:sidfmd_ctag_id => ctag_ids).select(:sidfmd_ctag_id))
        .order(:priority)
        .order('LOWER(keyword COLLATE "C")')
        .select(:id, :keyword)
  end

  # ルール以外の関連するテーブルの行を削除して、紐づいていたsidfmd_issue_idを返す(View更新のため)
  def delete_ctag_vul_related_rows
    issue_ctags = self.sidfmd_issue_ctags
    ret = issue_ctags.pluck(:sidfmd_issue_id)
    issue_ctags.delete_all
    self.sidfmd_vv_ctags.delete_all
    self.sidfmd_hv_ctags.delete_all
    ret
  end

  # ルール以外の関連するテーブルの行を一括削除
  # 紐づいていたsidfmd_issue_idを返す(View更新のため)
  def self.delete_ctags(ctag_ids)
    issue_ctags = SidfmdIssueCtag.where(:sidfmd_ctag_id => ctag_ids)
    ret = issue_ctags.pluck(:sidfmd_issue_id)
    issue_ctags.delete_all
    SidfmdVvCtag.where(:sidfmd_ctag_id => ctag_ids).delete_all
    SidfmdHvCtag.where(:sidfmd_ctag_id => ctag_ids).delete_all
    SidfmdCtag.where(:id => ctag_ids).destroy_all
    ret
  end

  # カテゴリが、指定されたカテゴリorカテゴリの指定なしのカスタムタグを返す
  def self.category_ctags(category_id, ctag_ids = nil)
    ctags = SidfmdCtag.where(
      "sidfmd_category_id IS NULL OR sidfmd_category_id = ?", category_id.to_i
    )
    if ctag_ids then
      ctags = ctags.where(:id => ctag_ids)
    end
    ctags
  end

end
