#!/bin/bash

# SIDfm VM Server docker images builder v 1.0.0

# docker container dns setting must be set. (dns directive in daemon.json)

cat <<EOF > Dockerfile
FROM	centos:centos7
LABEL	maintainer="SofTek System Inc <sales@softek.co.jp>"
ENV	HTTP_PROXY="" \
	HTTPS_PROXY="" \
	NO_PROXY="" \
	http_proxy="" \
	https_proxy="" \
	no_proxy=""
COPY	rh-ruby23.sh /etc/profile.d/
COPY	railsenv.sh /etc/profile.d/
COPY	timezone.sh /etc/profile.d/
COPY	sidfmvm.tar.gz /var/lib/
COPY	vminit.sh /var/lib/
COPY	database.yml /var/lib/
COPY	redmine.conf /var/lib/
COPY	secrets.yml /var/lib/
COPY	inflections.rb /var/lib/
COPY	ssl.conf.default /var/lib/
COPY	ssl.conf.custom	/var/lib/
COPY	vars /var/lib/
COPY	LICENSE /
RUN	echo "now building..." && \
	export HTTP_PROXY="${HTTP_PROXY}" && \
	export HTTPS_PROXY="${HTTPS_PROXY}" && \
	export NO_PROXY="${NO_PROXY}" && \
	export http_proxy="${http_proxy}" && \
	export https_proxy="${https_proxy}" && \
	export no_proxy="${no_proxy}" && \
	curl -L https://www.softek.co.jp/ >/dev/null 2>&1 && \
	rpm --import /etc/pki/rpm-gpg/RPM-GPG-KEY-CentOS-7 && \
	echo "include_only=.jp" >> /etc/yum/pluginconf.d/fastestmirror.conf && \
	yum clean all && \
	rm -rf /var/cache/yum/* && \
	( yum -y update || yum -y update || yum -y update ) && \
	yum -y install \
	centos-release-scl-rh \
	gcc-c++ \
	git \
	httpd \
	httpd-devel \
	less \
	libcurl-devel \
	libproxy \
	libtool \
	libxml2 \
	libxml2-devel \
	libyaml \
	libyaml-devel \
	logrotate \
	make \
	mod_ssl \
	net-snmp \
	openldap \
	openldap-devel \
	openssl \
	openssl-devel \
	patch \
	postgresql-libs \
	postgresql-devel \
	subversion \
	ImageMagick \
	ImageMagick-devel \
	unzip \
	which \
	zip \
	&& \
	rpm --import /etc/pki/rpm-gpg/RPM-GPG-KEY-CentOS-SIG-SCLo && \
	yum -y update && \
	yum -y install \
	rh-ruby23 \
	rh-ruby23-ruby-devel \
	&& \
	source /etc/profile.d/rh-ruby23.sh && \
	gem install bundler -v 1.17.3 && \
	cd /var/lib && \
	git clone https://github.com/OpenVPN/easy-rsa && \
	cd /var/lib/easy-rsa/easyrsa3/ && \
	mv /var/lib/vars . && \
	mv /var/lib/ssl.conf.default /etc/httpd/conf.d/ && \
	mv /var/lib/ssl.conf.custom /etc/httpd/conf.d/ && \
	cd /var/lib && \
	curl -OL https://www.redmine.org/releases/redmine-3.4.13.tar.gz && \
	tar xzf redmine-3.4.13.tar.gz && \
	mv redmine-3.4.13 redmine && \
	rm redmine-3.4.13.tar.gz && \
	cd /var/lib/redmine && \
	tar xzf ../sidfmvm.tar.gz && \
	rm -f ../sidfmvm.tar.gz && \
	mv -f ../vminit.sh plugins/sidfm-vm-server && \
	mv -f ../database.yml ../secrets.yml ./config/ && \
	mv -f ../inflections.rb ./config/initializers/ && \
	chmod 500 plugins/sidfm-vm-server && \
	chown -R apache:apache /var/lib/redmine && \
	sh plugins/sidfm-vm-server/vminit.sh enable && \
	bundle install --path vendor/bundler && \
	chown -R apache:apache vendor/bundler && \
	gem install passenger && \
	passenger-install-apache2-module --auto && \
	passenger-install-apache2-module --snippet > /etc/httpd/conf.d/redmine.conf && \
	cat ../redmine.conf >> /etc/httpd/conf.d/redmine.conf && \
	rm -f ../redmine.conf
CMD	/vm_server
EOF

sudo docker build -t sidfmd --rm=true --force-rm=true --no-cache=true .
