/* coding: utf-8 */
/* Copyright (C) 2018 SofTek Systems, Inc. */

//ルール設定画面
//BootStrapTableのtdのスタイル。画面ごとに関数を作って設定する
function BootStrapTableCellStyle(value, row, index, field) {
    switch (field){
        case 'chkbox':
            return { classes: 'chk-box' };
        case 'rule_name':
            return { classes: 'rule-name' };
        case 'rule_category':
            return { classes: 'rule-category' };
        case 'trigger':
            return { classes: 'rule-trigger' };
        case 'condition':
            return { classes: 'rule-condition' };
        case 'operation':
            return { classes: 'rule-operation' };
        case 'rule_position':
            return { classes: 'rule-priority' };
        case 'log':
            return { classes: 'rule-log' };
        case 'rule_active':
            return { classes: 'rule-active' };
        case 'rule_edit':
            return { classes: 'rule-edit' };
        default:
            return { classes: '' };
    }
}

function BootStrapTableRowStyle(row, index){
    return {
        classes: 'list_alternative'
    }
}


$(function() {
    const MAX_POSITION_NUM = parseInt($('#rule-max-position-num').val());
    const MIN_POSITION_NUM = parseInt($('#rule-min-position-num').val());

    const summaryTableId = 'rules-summary-table';
    const summaryTableElm = $('#' + summaryTableId);
    const pageName = 'rule';

    //サマリーテーブルで条件・操作のボタンを押した時に内容表示
    const enableDisplayRuleItemDetails = () => {
        $(`#${summaryTableId} .open-RuleTable-detail`).on('click', function(){
            (async () => {
                const roleOrCategoryError = await dynamicSummaryTable.checkRolesAndCatsBeforeAction();
                if( roleOrCategoryError ) return;
                var button = $(this);
                var detail = button.attr('data-detail');
                var rule_id = button.attr('data-ruleid');
                var same_kind_popovers = $('.rule-setting-original-popover[data-detail=' + detail + ']');
                var popovers_rule_id = same_kind_popovers.attr('data-ruleid');
                same_kind_popovers.remove();
                $('.rule-detail-popover').remove();
                if(rule_id === popovers_rule_id){
                    return false;
                }     
                var original_popover = $('<div>').attr({
                                           'class': 'rule-setting-original-popover',
                                           'data-detail': detail,
                                           'data-ruleid': rule_id,
                                        });
                var inner = $('<div>').attr({
                                'class': 'original-popover-inner text-left max-height-170px max-width-250px',
                            });
                if(detail === 'cond'){
                    var rule_param = JSON.parse($('#rule-param-' + rule_id).val());
                    var condition = rule_param['condition'];
                    var content = create_cond_display_content(condition, 'open-detail', original_popover, button);
                }else if(detail === 'ope'){
                    var rule_param = JSON.parse($('#rule-param-' + rule_id).val());
                    var action = rule_param['action'];
                    var content = create_ope_display_content(action, 'open-detail', original_popover, button);
                }
                inner.append(content);
                var triangle = $('<div>').attr('class', 'position-relative triangle');
                original_popover.append(inner, triangle);
                $('body').append(original_popover);
            })(); 
        });
    };

    //ログを表示
    const enableDisplayLogs = () => {
        $(`#${summaryTableId} .open-logs-list`).on('click', function(){
            var button = $(this);
            var log_types = JSON.parse($('#log-types').val());
            var log_kinds = JSON.parse($('#log-kinds').val());
            var logs = JSON.parse(button.attr('data-log'));
            $('#modalLog .modal-title').text(button.attr('data-rulename') + ' ルールのログ一覧');
            var target_tbody = $('#LogList');
            $('#LogList > *').remove();
            for(var i = 0; i < logs.length; i++){
                var log = logs[i];
                var type = log["type"];
                if(type === 'info'){
                    var tr = $('<tr>').attr('class', 'list_alternative2');
                }else if(type === 'warn'){
                    var tr = $('<tr>').attr('class', 'list_alternative bg-warning');
                }else{
                    var tr = $('<tr>').attr('class', 'list_alternative bg-danger text-white');
                }
                var td_type = $('<td>').attr('class', 'log-kind').text(log_types[type]);
                var td_kind = $('<td>').attr('class', 'log-item').text(log_kinds[log["kind"]]);
                var td_content = $('<td>').attr('class', 'log-content').text(log["message"]);
                var td_date = $('<td>').attr('class', 'log-date').text(log["created_at"]);
                tr.append(td_type, td_kind, td_content, td_date);
                target_tbody.append(tr);
            }
            $('#modalLog').modal('show');
        });
    };

    const enableDeleteRule = () => {
        $('#' + summaryTableId + ' .delete-rule').on('click', function(){
            const ruleId = $(this).data('ruleid');
            const form = $('#delete-rule-form-' + ruleId);
            let conf = confirm($(this).attr('data-confirm'));
            if( !conf ) return;
            dynamicSummaryTable.executeProcess({
                paramsData: { delete_id: ruleId },
                path: form.attr('action'),
                additionalAlertMessage: '削除に失敗しました。'
            });
            return false;
        });
    }

    //サマリー表の動的処理のため(bootstrap-tableのため)にbindする。でないとイベントが効かない
    const bindSummaryTableRowEvents = () => {
        enable_bootstrap_actions('#' + summaryTableId);
        enableDisplayRuleItemDetails();
        enableDisplayLogs();
        enableDeleteRule();
    }

    //「既存のルールを元に新規登録」ボタンの状態を変更
    const modifyCheckBoxRelatedContent = () => {
        if( JSON.parse($('#pages_checked').val()).length === 1 ){
            $('#create-copied-rule-btn').removeClass('disabled');
        }else{
            $('#create-copied-rule-btn').addClass('disabled');
        }
    };

    //サマリー表の表示データの変更に合わせてルールの変更等に使うデータを変更
    const modifyPageContent = (data) => {
        $('#rules-count-num').text(data.all_rule_count);
        const summaryTable = $(`#${summaryTableId}`);
        summaryTable.find('[data-ruleparam="true"]').remove();
        data.hidden_params.forEach(param => {
            setTimeout(() => {
                summaryTable.append(
                    $('<input>').attr({
                        type: 'hidden',
                        id: 'rule-param-' + param.rule_id,
                        value: JSON.stringify(param.param),
                        'data-ruleparam': true
                    })
                );
            });
        });
    };

    const dynamicSummaryTable = new DynamicSummaryTable({
        summaryTableId,
        bindSummaryTableRowEventsFunction: bindSummaryTableRowEvents,
        pageName,
        allFilteredIdsId: 'all-rule-ids',
        chkboxSelector: `#${summaryTableId} .SelectRule`,
        allChkboxSelector: `#${summaryTableId} .allRuleSelect`,
        bulkBtnSelector: '.bulk-btn',
        modifyCheckBoxRelatedContent,
        modifyPageContent
    });

    $('#pages_form').on('submit', function(){ set_summary_table_lines(pageName, $('#pages_per_page').val()) });
    sticky_control('SummaryTable', pageName);

    //サマリーテーブルのソート
    //$('.sort-rule-name').on('click', sort_toggle('rule-name'));
    //$('.sort-rule-category').on('click', sort_toggle('rule-category'));
    //$('.sort-rule-active').on('click', sort_toggle('rule-active'));
    //$('.sort-rule-position').on('click', sort_toggle('rule-position'));

    //検索処理
    const summaryTableFilter = new SummaryTableFilter({dynamicSummaryTable});
    summaryTableFilter.enableFilterChangeLink('.filter-change'); //タイミング・条件・操作・有効無効 検索
    summaryTableFilter.enableFilterChangeLink('.category-change'); //カテゴリ検索 
    summaryTableFilter.enableFilterText(
       '#filter_rulename',
       '#filter_rulename_reset',
       { fieldSelector: '#search_rulename_field' },
       'search_rulename'
    ); //ルール名検索

    summaryTableFilter.enableResetAllFilters('#filter-clear'); //全リセット

    //ブラウザバックだったらチェックボックスの表示を修正
    /*$.post($('#check_browser_back_path').val(), {'page_name': 'rule'},
        function(info){
            if(info === 'browser_back'){
                var is_all_checked = $('#pages_all_checked').val();
                if(is_all_checked === 'true'){
                    var all_ids = JSON.parse($('#all-rule-ids').val());
                    $('#RuleTable').find('#allSelect, .SelectRule').prop('checked', true);
                    $('#pages_checked').val(JSON.stringify(all_ids.filter(Boolean)));
                }else{
                    $('#RuleTable').find('#allSelect, .SelectRule').prop('checked', false);
                    var pages_checked = JSON.parse($('#pages_checked').val());
                    var target_elm = '';
                    for (var id of pages_checked){
                        target_elm += '.SelectRule[value=' + id + '],';
                    }
                    target_elm = target_elm.slice(0, -1);
                    $(target_elm).prop('checked', true);
                }
            }
    });*/
    //modal-contentをドラッグ＆ドロップで動かせるようにする
   /* $('.movable-content').on('mousedown', function(e){
        var target = $(this);
        var offset = target.offset();
        var x = e.pageX - offset.left;
        var y = e.pageY - offset.top;
        target.attr({'data-x': x, 'data-y':y, 'data-moving': 'true'});
    });

    $('.movable-content').on('mousemove', function(e){
        var target = $(this);
        if(target.attr('data-moving')){
            $('.original-tooltip, .rule-setting-original-popover').remove();
            $('#modalCreateRule-set-trigger-btn').tooltip('hide');
            target.offset({
                'top': (e.pageY - parseInt(target.attr('data-y'))),
                'left': (e.pageX - parseInt(target.attr('data-x')))
            });
        }
    });

    $('.movable-content').on('mouseup', function(){
        var target = $(this);
        if(target.attr('data-moving')){
            target.removeAttr('data-y data-x data-moving');
        }
    });*/

    //modalをスクロールした時にoriginal-tooltip, popover削除
    $('.modal').on('scroll', function(){
        $('[data-toggle="tooltip"]').tooltip('hide');
        $('.original-tooltip, .rule-setting-original-popover').remove();
    });

    $('.fixed-table-body').on('scroll', function(){
        $('[data-toggle="tooltip"]').tooltip('hide');
        $('.original-tooltip, .rule-setting-original-popover').remove();
    });
 
    init_sidfm();    
    activate_sidebar_item('sidebar-item-rulesetting-manip');
    set_style_for_ActionArea_on_filter_dropdown_event();
 
    //Datepicker日本語化   
    $.datepicker.regional['ja'] = {
        closeText: '閉じる',
        prevText: '<前',
        nextText: '次>',
        currentText: '今日',
        monthNames: ['1月','2月','3月','4月','5月','6月','7月','8月','9月','10月','11月','12月'],
        monthNamesShort: ['1月','2月','3月','4月','5月','6月', '7月','8月','9月','10月','11月','12月'],
        dayNames: ['日曜日','月曜日','火曜日','水曜日','木曜日','金曜日','土曜日'],
        dayNamesShort: ['日','月','火','水','木','金','土'],
        dayNamesMin: ['日','月','火','水','木','金','土'],
        weekHeader: '週',
        dateFormat: 'yy/mm/dd',
        firstDay: 0,
        isRTL: false,
        showMonthAfterYear: true,
        yearSuffix: '年'
    };
    $.datepicker.setDefaults($.datepicker.regional['ja']);

    //以下の一覧を使ってHTMLの要素も作成しています 
    //条件項目名一覧 
    const cond_items_name = JSON.parse($('#rule-all-conds').val());
    //条件詳細名データと項目名対応一覧 
    const cond_detail_items_name = {
        'issue-update-regi'        : '脆弱性の新規登録',
        'issue-update-del'         : '脆弱性影響無への変更',
        'issue-update-regi-and-del': '脆弱性登録 & 影響無変更',
        'cvss-exploitable'         : '攻撃コード指定',
        'sri-above'                : 'SRI指定値以上',
        'sri-below'                : 'SRI指定値以下',
        'sri-warning'              : 'SRI値要対応',
        'cvss-above'               : 'CVSS指定値以上',
        'cvss-below'               : 'CVSS指定値以下',
        'cvss-warning'             : 'CVSS値要対応',
        'cvss-notreg'              : 'CVSS値未定義',
        'cvss-vector'              : 'CVSS指標',
        'status-match'             : 'ステータスの一致',
        'product-match'            : 'プロダクト限定',
        'product-unmatch'          : 'プロダクト除外',
        'due-date-expired'         : '対策予定日超過',
        'due-date-approach'        : '対策予定日接近',
        'due-date-undef'           : '対策予定日未設定',
        'history-create'           : '履歴追加',
        'category-match'           : 'カテゴリ限定',
        'category-unmatch'         : 'カテゴリ除外',
        'host-match'               : 'ホスト限定',   
        'host-unmatch'             : 'ホスト除外',
        'host-syncing'             : '同期中ホスト',
        'timerange'                : '期間指定',
        'ctag'                     : 'カスタムタグ',
    }
    //変更不可の条件詳細一覧 
    const uneditable_cond_details = [ 'due-date-expired', 'due-date-undef' ];

    //操作項目名一覧
    const ope_items_name = JSON.parse($('#rule-all-opes').val());
    //操作詳細名データと項目名対応一覧
    const ope_detail_items_name = {
        'send-mail-user'    : 'メール送信(ユーザ)',
        'send-mail-category': 'メール送信(カテゴリ)',
        'status-change'     : 'ステータス変更',
        'due-date-change'   : '対策予定日設定',
        'send-syslog'       : 'syslog送信',
        'set-ctag'          : 'カスタムタグ設定',
    }
   
    //「処理するデータ」の項目一覧
    var process_values_list = JSON.parse($('#process-values-list').val()); 
    const process_values_with_names = {
        [process_values_list[0]]: '脆弱性一覧のデータ',
        [process_values_list[1]]: 'ホスト一覧のデータ',
        [process_values_list[2]]: '新規登録された、脆弱性一覧のデータ',
        [process_values_list[3]]: '値が更新された、脆弱性一覧のデータ',
        [process_values_list[4]]: 'ステータスが「影響無」に更新された、脆弱性一覧のデータ',
        [process_values_list[5]]: '登録・更新・削除された、脆弱性一覧のデータ',
    }
   
    //「処理するデータ」の項目名作成 他の場所でもjsで同じ値を参照するためjsで管理
    $('#hidden-select-process-data-area').find('[data-id="select-process-data"] option').each(function(){
        var elm = $(this);
        elm.text(process_values_with_names[elm.val()]);
    });

    //新規条件追加modal内のボタンのテキスト作成 他の場所でもjsで同じ値を参照するためjsで管理
    $('#wrap-condition-item-btns').find('[data-cond]')
        .each(function(){
            var cond = $(this).data('cond');
            var text = cond_items_name[cond];
            $(this).text(text);
    });

    //新規操作追加modal内のボタンのテキスト作成 他の場所でもjsで同じ値を参照するためjsで管理
    $('#operation-item-area').find('[data-ope]')
        .each(function(){
            var ope = $(this).data('ope');
            var text = ope_items_name[ope];
            $(this).text(text);
    });
    
    //閲覧者権限の場合、登録・編集・削除ボタンを押せないようにする
    var user_auth = JSON.parse($('#user_auth').val());
    if(!user_auth.includes('admin') && !user_auth.includes('use')){
        $('#create-copied-rule-btn, #create-new-rule-btn, #create-rule-by-preset, .edit-rule, .delete-rule').prop('disabled', true);
    }

    //モーダルが開くときにツールチップがあれば削除
    $('.modal').on('show.bs.modal', function(){
        $('.original-tooltip').remove();
        $('#wrap-ActionArea').css('z-index', '10');
        $('#floating-scroll').css('z-index', '10');
    });

    //モーダルが閉じる時にロードマーク・ツールチップ・ポップオーバー削除
    $('.modal').on('hide.bs.modal', function(){
        $('#keyword_loading').remove();
        if($(this).attr('id') != 'modalEditOperation' && $(this).attr('id') != 'modalEditCondition'){
            $('.original-tooltip').remove();
        }
        $('[data-toggle="tooltip"]').tooltip('hide');
        $('.rule-setting-original-popover').remove();
    });

    //サマリー表のチェックボックス
    /*$('#allSelect').on('change', function(){
        var chk_box = $(this);
        if(chk_box.prop('checked')){
            $('.SelectRule').prop('checked', true);
            $('#pages_checked').val($('#all-rule-ids').val());
            $('#pages_all_checked').val('true');
            $('.bulk-btn').prop('disabled', false);
        }else{
            $('.SelectRule').prop('checked', false);
            $('#pages_checked').val('[]');
            $('#pages_all_checked').val('false');
            $('.bulk-btn').prop('disabled', true);
        }
        $('#create-copied-rule-btn').addClass('disabled');
    });

    $('.SelectRule').on('change', function(){
        var chk_box = $(this);
        var rule_id = parseInt(chk_box.val());
        var selected_ids = JSON.parse($('#pages_checked').val());
        var all_rule_ids = JSON.parse($('#all-rule-ids').val());
        if(chk_box.prop('checked')){
            selected_ids.push(rule_id);
        }else{
            var idx = selected_ids.indexOf(rule_id);
            selected_ids.splice(idx, 1);
        }
        $('#pages_checked').val(JSON.stringify(selected_ids));
        if(all_rule_ids.length === selected_ids.length){
            $('#allSelect').prop('checked', true);
            $('#pages_all_checked').val('true');
        }else{
            $('#allSelect').prop('checked', false);
            $('#pages_all_checked').val('false');
        }    
        if(selected_ids.length === 1){
            $('#create-copied-rule-btn').removeClass('disabled');
        }else{
            $('#create-copied-rule-btn').addClass('disabled');
        }
        if(selected_ids.length === 0){
            $('.bulk-btn').prop('disabled', true);
        }else{
            $('.bulk-btn').prop('disabled', false);
        }
    });*/
 
    function reset_all_modals_related_create_rule(){
         setRulePositionArea(0);
         $('#rule-updated-at').val('');
         $('#modalCreateRule').removeAttr('data-create');
         $('#execute-register-rule').text('登録');
         $('#rule_id_for_modify_rule').val('');
         $('.rule-setting-content').text('未設定').removeClass('text-left');
         $('#form-create-rulename').val('');
         $('#select-enable-rule').val('enable');
         $('#selecting-cond-item').text('- - -');
         $('#periodic-hour-btn, #periodic-interval-btn').text('時間を選択');
         $('#unit-of-periodic-interval').text('単位を選択');
         $('#periodic-interval-area').hide();
         $('#select-periodic-hour, #select-periodic-interval').val('');
         $('.set-rule-trigger').prop('checked', false);
         $('#wrap-search-hostname').tooltip('enable');
         $('#select-trigger-hostname > *').remove();
         $('#trigger-category').removeAttr('data-categoryid');
         $('#trigger-category-menu > *').remove();
         $('#form-create-trigger').val('{}');
         $('#form-create-condition').val('{}');
         $('#form-create-operation').val('{}');
         $('#search-hostname').tooltip('hide');
         $('#search-hostname').val('').removeAttr('data-hostname data-extname');
         $('#select-trigger-hostname > *').remove();
         $('#select-trigger-hostname').hide();
         $('#modalCreateRule-set-trigger-btn').attr('class', 'btn btn-xs btn-secondary rule-modal-setting-btn open-register-modal')
                                              .text('タイミング未設定').tooltip('enable');
         $('#modalCreateRule-set-cond-btn').attr('class', 'btn  btn-xs btn-secondary rule-modal-setting-btn open-register-modal disabled')
                                           .text('条件未設定');
         $('#modalCreateRule-set-operation-btn').attr('class', 'btn btn-xs btn-secondary open-register-modal disabled')
                                                .text('操作未設定');
         $('.rule-setting-content').attr('class', 'rule-setting-content unexpanded-detail');
         //$('.expand-btn').remove(); 
         $('#label-cond-detail').hide();
         $('#cond-detail-body > *').remove();
         $('#wrap-condition-item-btns').find('show')
                                       .siblings('[data-toggle=dropdown]').click();
         $('#set-condition-tbody > *').remove();
         $('#set-operation-tbody > *').remove();
         $('#cond-detail').data('selected', '');
         $('#set-cond-param').val('{}');
         $('#set-ope-param').val('{}');
         $('#search-address-items-area-tbody > *').remove();
         $('#send-mail-title').val('');
         $('#send-mail-content').val('');
         $('#modalSelectItems').find('.modal-body > *').remove();
         $('#modalSelectItems').find('[data-selectitem]').prop('disabled', false);
      /*   $('#ope-process-data').val("5");
         $('#ope-process-flg').val('true');*/
         reset_mail_select_type();
         ctagCondInfo.dataid = 0;
         timerangeCondInfo.dataid = 0;
    }


    //ルールmodalを開いたときにツールチップ表示
    $('#modalCreateRule').on('shown.bs.modal',function(){
        $('#modalCreateRule-set-trigger-btn').tooltip('show');
        $('.rule-setting-content').scrollTop(0);
        show_change_rule_name_message();
    });
 
    //登録ボタンの押下可能・不可能チェック
    $('#modalCreateRule').on('show.bs.modal',function(){
        check_rule_is_registable();
    });

    //登録ボタンの押下可能・不可能チェック
    $('#form-create-rulename').on('input', function(){
        check_rule_is_registable();
        if($('.original-tooltip').length > 0 ){
            $('.original-tooltip').remove();
        }
    });


    //設定ボタンで各種設定modalを表示
    $('#modalCreateRule-set-trigger-btn').on('click', function(){
        $(this).tooltip('hide').tooltip('disable');
        $('#modalRegisterTrigger').modal('show');
    });

    $('#modalCreateRule-set-cond-btn').on('click', function(){
        if(!$(this).hasClass('disabled')){
            $('#modalRegisterCondition').modal('show');
        }
    });

    $('#modalCreateRule-set-operation-btn').on('click', function(){
        if(!$(this).hasClass('disabled')){
            $('#modalRegisterOperation').modal('show');
        }
    });

    //カスタムタグの条件の要素を、カテゴリ変更時に(必要があれば)削除
    function deleteCtagCondElms(ctagDataId){
        $('#set-condition-tbody [data-conddetail="ctag"][data-editdata=' + ctagDataId + ']').closest('tr[data-cond="ctag"]').remove();
        $('#rule-setting-item-cond').find('[data-cond=ctag' + ctagDataId + ']').remove();
    }
   
    //ルールmodalでカテゴリ決定・変更
    $(document).on('click', '#set-main-category', function(){
        var selected_category = $('#select-rule-c-cats').find('.c-category-radio:checked');
        var category_id = selected_category.val();
        var category_name = selected_category.attr('data-text');
        var category_descs = JSON.parse($('#category-descs').val());
        if($('#main-category-id').val() != category_id){
            var ope_param = JSON.parse($('#form-create-operation').val());
            var cond_param = JSON.parse($('#form-create-condition').val());
            var trigger_param = JSON.parse($('#form-create-trigger').val());
            var path = $('#check_change_rule_category_path').val();
            $.ajax({
                type: 'POST',
                url: path, 
                data: { 
                    'new_category_name' : category_name,
                    'new_category_id'   : category_id,
                    'trigger'           : JSON.stringify(trigger_param),
                    'conds'             : JSON.stringify(cond_param),
                    'actions'           : JSON.stringify(ope_param)
                },
            })
            .done(function(data){
                if(data['is_alert']){
                    var conf = confirm(data['message']);
                    if(conf){
                        var new_trigger = data['trigger'];
                        var new_cond = data['cond'];
                        var new_ope = data['action'];
                        var del_conds = data['del_conds'];
                        var del_opes = data['del_actions'];
                        if(new_trigger['period'] === 'vulsync'){
                            var trigger_content = trigger_display_content('vulsync', category_name, null);
                            $('#rule-setting-item-trigger').html(trigger_content).addClass('text-left');
                            //var expand_btn = $('<i>').attr('class', 'expand-btn fas fa-angle-down mr-1')
                            //$('#rule-setting-item-trigger').append(expand_btn);
                            $('#search-hostname').val('');
                        }
                        for(var i = 0; i < del_conds.length; i++){
                            var del_elm = del_conds[i];
                            $('#set-condition-tbody, #rule-setting-item-cond').find('[data-cond=' + del_elm + ']').remove();
                        }
                        for(var i = 0; i < del_opes.length; i++){
                            var del_elm = del_opes[i];
                            $('#set-operation-tbody, #rule-setting-item-ope').find('[data-addope=' + del_elm + ']').remove();
                        }
                        if(data['remove_action_send_mail_to_user_ids'].length > 0){
                            $('#set-operation-tbody, #rule-setting-item-ope').find('[data-addope="send-mail-user"]').remove();
                            var modify_data = new_ope['send-mail-user'];
                            update_modalRegisterOperation_items('send-mail-user', { 'send-mail-user': modify_data } , 'prepare_edit');
                            create_ope_display_content({ 'send-mail-user': modify_data }, 'modify-send-mail-ope');
                        }
                        if( data['is_action_ctag_change'] ) {
                            $('#set-operation-tbody, #rule-setting-item-ope').find('[data-addope="set-ctag"]').remove();
                            update_modalRegisterOperation_items('set-ctag', { 'set-ctag': new_ope['set-ctag'] } , 'prepare_edit');
                            create_ope_display_content({ 'set-ctag': new_ope['set-ctag'] }, 'modify-set-ctag');
                        }
                        if( data['modify_category_items'].length > 0 ){
                            data['modify_category_items'].forEach(function(item){
                                $('#set-condition-tbody, #rule-setting-item-cond').find('[data-cond="' + item + '"]').remove();
                                update_modalRegisterCondition_items(item, { [item]: new_cond[item] }, 'prepare_edit');
                                create_cond_display_content({ [item]: new_cond[item] }, 'modify'); 
                            });
                        }
                        if( data['history_cond_out_of_category_user_ids'].length > 0){
                            $('#set-condition-tbody, #rule-setting-item-cond').find('[data-cond="history-create"]').remove();
                            update_modalRegisterCondition_items('history-create', { 'history-create' : new_cond['history-create'] }, 'prepare_edit');
                            create_cond_display_content({ 'history-create': new_cond['history-create'] }, 'modify');
                        }
                        if( data['del_ctag_data_ids'] ){
                            data['del_ctag_data_ids'].forEach(ctagDataId => deleteCtagCondElms(ctagDataId));
                        }
                        if( data['modify_ctag_data_ids'] ) {
                            const newCtagCond = new_cond['ctag'];
                            data['modify_ctag_data_ids'].forEach(ctagDataId => deleteCtagCondElms(ctagDataId));
                            data['modify_ctag_data_ids'].forEach(ctagDataId => {
                                const targetCtagData = newCtagCond.find(ct => ct.dataid == ctagDataId);
                                update_modalRegisterCondition_items(
                                    'ctag',
                                    { 'ctag' : [ targetCtagData ] },
                                    'prepare_edit'
                                );
                                $('#rule-setting-item-cond').append($(createCtagCondDisplayElm(targetCtagData)));
                            });
                        }
                        $('#form-create-trigger').val(JSON.stringify(new_trigger));
                        $('#form-create-condition, #set-cond-param').val(JSON.stringify(new_cond));
                        $('#form-create-operation, #set-ope-param').val(JSON.stringify(new_ope));
                        if(Object.keys(new_cond).length === 0){
                            $('#modalCreateRule-set-cond-btn').removeClass('btn-success set')
                                                              .addClass('btn-secondary')
                                                              .text('条件未設定');
                            $('#rule-setting-item-cond').html('未設定').removeClass('text-left');
                        }
                        if(Object.keys(new_ope).length === 0){
                            $('#modalCreateRule-set-operation-btn').removeClass('btn-success set')
                                                                   .addClass('btn-secondary')
                                                                   .text('操作未設定');
                            $('#rule-setting-item-ope').html('未設定').removeClass('text-left');
                        }
                        $('#main-category-id').val(category_id).attr('data-text', category_name);
                        $('#rule-main-category').text(category_name);
                        check_rule_is_registable();
                    }else{
                        return false;
                    }
                }else{
                    var new_trigger = data['trigger'];
                    if(new_trigger['period'] === 'vulsync'){
                        $('#form-create-trigger').val(JSON.stringify(new_trigger));
                        var trigger_content = trigger_display_content('vulsync', category_name, null);
                        $('#rule-setting-item-trigger').html(trigger_content).addClass('text-left');
                        //var expand_btn = $('<i>').attr('class', 'expand-btn fas fa-angle-down mr-1')
                        //$('#rule-setting-item-trigger').append(expand_btn);
                        $('#search-hostname').val('');
                    }
                    $('#main-category-id').val(category_id).attr('data-text', category_name);
                    $('#rule-main-category').text(category_name);
                    check_rule_is_registable();
                }
                $('#rule-setting-item-trigger [data-toggle="tooltip"]').tooltip();
                var main_cat_id = $('#main-category-id').val();
                $('#rule-main-category').attr({
                    'data-toggle': 'tooltip',
                    'data-original-title': (category_descs[main_cat_id] ? category_descs[main_cat_id].desc : ''),
                    'data-boundary': 'window'
                }).tooltip();
                $('#modalSelectItems').modal('hide');
                $('#modalCreateRule').modal('show');
            });
        }else{
            $('#modalSelectItems').modal('hide');
            $('#modalCreateRule').modal('show');
        }
    });

    //設定画面開く前にツールチップ削除
    $('.open-register-modal').on('click', function(){
        $('#c-category-selection-label').tooltip('hide');
    });

    //メインカテゴリ選択
   function set_main_category(){
       var target_modal = $('#modalSelectItems');
       target_modal.find('.modal-body > *').remove();
       var description = $('<div>').attr('class', 'font-weight-bold text-left mb-2 ml-3')
                                   .text($('#modalCreateRule').is(':visible') ? 'ルールを適用するカテゴリを選択して下さい' : 'はじめに、ルールを適用するカテゴリを選択して下さい');
       var select_area = $('<div>').attr('class', 'mx-auto');
       var selections = make_selectable_categories('select-rule-c-cats', true);
       select_area.append(selections);
       select_area.find('[data-toggle="tooltip"]').tooltip();
       target_modal.find('.modal-body').append(description, select_area);
       target_modal.find('[data-selectitem]').attr('id', 'set-main-category')
                                             .prop('disabled', true);
       target_modal.find('.modal-title').text('カテゴリ選択');
       $('#modalSelectItems').modal('show');
   }

   $('#change-main-category').on('click', function(){
       $('.open-register-modal').tooltip('hide');
       set_main_category();
   });

   //新規登録時のpath設定
   $('#create-new-rule-btn').on('click', function(){
       var users_categories = JSON.parse($('#users_selectable_categories').val());
       var user_auth = JSON.parse($('#user_auth').val());
       if(Object.keys(users_categories).length === 0 && !user_auth.includes("admin")){
           alert('カテゴリに属していないユーザは登録できません');
           return false;
       }
       reset_all_modals_related_create_rule();
       $('#form-create-rule').attr('action', $('#rule_setting_create_path').val());
       $('#modalCreateRule-title').text('ルール登録');
       set_main_category();
   });

   //各種設定の各項目内でスクロールした時詳細の表示を削除
   $('.rule-setting-content').on('scroll', function(){
       $('.rule-detail-popover').remove();
   });


    function getRegistParamsData(processName){
        const paramsData = {};
        const form = $('#form-create-rule');
        if( processName === '登録' ){
            form.find('[name]:not([disabled])').each((idx, elm) => {
                elm = $(elm);
                switch (elm.attr('name')){
                    case 'form[rulename]':
                        paramsData.rulename = elm.val();
                        break;
                    case 'form[category]':
                        paramsData.category = elm.val();
                        break;
                    case 'form[flg]':
                        paramsData.flg = elm.val();
                        break;
                    case 'form[trigger]':
                        paramsData.trigger = elm.val();
                        break;
                    case 'form[condition]':
                        paramsData.condition = elm.val();
                        break;
                    case 'form[operation]':
                        paramsData.operation = elm.val();
                        break;
                    case 'form[rule_position]':
                        paramsData.rule_position = elm.val();
                        break;
                };
            });
        }else{ //変更
            form.find('[name]:not([disabled])').each((idx, elm) => {
                elm = $(elm);
                switch (elm.attr('name')){
                    case 'form[rule_id]':
                         paramsData.rule_id = elm.val();
                         break;
                    case 'form[rulename]': 
                        paramsData.rulename = elm.val();
                        break;
                    case 'form[category]':
                        paramsData.category = elm.val();
                        break;
                    case 'form[flg]': 
                        paramsData.flg = elm.val();
                        break;
                    case 'form[trigger]':
                        paramsData.trigger = elm.val();
                        break;
                    case 'form[condition]':
                        paramsData.condition = elm.val();
                        break;
                    case 'form[operation]': 
                        paramsData.operation = elm.val(); 
                        break;
                    case 'form[rule_position]':
                        paramsData.rule_position = elm.val();
                        break;
                    case 'form[updated_at]':
                        paramsData.updated_at = elm.val();
                        break;
                    case 'form[only_flg_changed]':
                        paramsData.only_flg_changed = elm.val();
                        break;
                };
            });
        }
        return paramsData;
    }

    //ルール登録modalのバリデーション
    $('#execute-register-rule').on('click', function(){
       var button = $(this);
       button.css('pointer-events','none');
       var rule_id = $('#rule_id_for_modify_rule').val();
       var rule_name = $('#form-create-rulename').val();
       var trigger_param = JSON.parse($('#form-create-trigger').val());
       const rulePosition = $('#rule-position-field').val().trim();
       setRulePositionArea(rulePosition);
       const preventSubmit = (alert_message) => {
           button.css('pointer-events','');
           alert(alert_message);
           return false;
       };
       if( !rulePosition.length || !rulePosition.match(/(\-{1}|)[0-9]/) || rulePosition.match(/\./) ){
           return preventSubmit(`実行優先度に、${MIN_POSITION_NUM}以上の数値(整数)を入力して下さい。`);
       }
       var path = $('#sidfm_internal_api_check_rule_path').val();
       let processName = ''
       if( rule_id.length > 0 ){
           //変更
           processName = '変更';
           var prev_rule_info = $('#rule-param-' + rule_id).val();
           var new_rule_info = {
             'rule_name': $('#form-create-rulename').val(),
             'rule_position': parseInt(rulePosition),
             'is_enable': $('#select-enable-rule').val() === 'enable',
             'category': $('#main-category-id').val(),
             'trigger' : trigger_param,
             'condition': JSON.parse($('#form-create-condition').val()),
             'action' : JSON.parse($('#form-create-operation').val())
           }
           var param = {
               'rule_id': rule_id,
               'rule_name': rule_name,
               'prev_rule_info': prev_rule_info,
               'new_rule_info': JSON.stringify(new_rule_info)
           }
       }else{
           //新規登録
           processName = '登録';
           var param = {
               'rule_id': rule_id,
               'rule_name': rule_name
           }
       }

       (async () => {
           const roleOrCategoryError = await dynamicSummaryTable.checkRolesAndCatsBeforeAction({
               additionalAlertMessage: processName + 'に失敗しました。'
           });
           if( roleOrCategoryError ) return;
           $.post(path, param, function(data){
               if(data.is_same_rule_name){
                   return preventSubmit('同名のルールが登録されています');
               }
               if($('#modalCreateRule-set-trigger-btn:not(.set), #modalCreateRule-set-cond-btn:not(.set), #modalCreateRule-set-operation-btn:not(.set)').length > 0){
                   return preventSubmit('未設定の項目があります');
               }
               if(rule_name === ''){
                   return preventSubmit('ルール名の設定を行ってください');
               }
               if( !data.is_rule_changed ){
                   return preventSubmit('変更がありません');
               }
               if( parseInt(rulePosition) > MAX_POSITION_NUM ){
                   return preventSubmit(`実行優先度に設定可能な最大数値は、${MAX_POSITION_NUM}です。`);
               }else if( parseInt(rulePosition) < MIN_POSITION_NUM ){
                   return preventSubmit(`実行優先度に設定可能な最小値は、${MIN_POSITION_NUM}です。`);
               }
               //if(data.is_updated){
               //    alert('ルールが更新されているため、更新は行われません。画面をリロードします。');
               //    $('#pages_form').submit();
               //    return false;
               //}
               if(trigger_param['period'] === 'vulsync'){
                   alert('タイミング設定が「脆弱性同期時 実行」の場合、脆弱性同期時に値の更新が無い脆弱性は条件の対象となりません。\nご注意下さい。');
               }
               $('#rule-only-flg-changed').val(data.only_flg_changed);

               const form = $('#form-create-rule');
               const paramsData = getRegistParamsData(processName);
               dynamicSummaryTable.executeProcess({
                   paramsData,
                   path: form.attr('action'),
                   additionalAlertMessage: processName + 'に失敗しました。'
               });
               button.css('pointer-events','');
               return false;
               //$('#form-create-rule').submit();
           });
       })();
    });

    //ルール登録modalをキャンセルor×ボタンで閉じる時
    $('.close-modalCreateRule').on('click', function(){
        var conf = confirm('内容は破棄されます。よろしいですか？');
        if(conf){
            $('#modalCreateRule').modal('hide');
        }else{
            return false;
        }
        $('#modalCreateRule').modal('hide');
    });

    //カスタムカテゴリを１つだけ選択させる
    $(document).on('change', '.c-category-radio',function(){
        var radio = $(this);
        var wrap;
        if(radio.closest('#select-rule-c-cats').length === 1){
            wrap = $('#select-rule-c-cats');
        }else if(radio.closest('#select-trigger-category').length === 1){
            wrap = $('#select-trigger-category');
        }else if(radio.closest('#select-mail-category').length === 1){
            wrap = $('#select-mail-category');
        }else if(radio.closest('#select-mail-change-category').length === 1){
            wrap = $('#select-mail-change-category');
        }
        if(wrap){
            wrap.find('.c-category-radio').prop('checked', false);
            radio.prop('checked', true);
          //  $('#search-hostname').prop('disabled', false);
            $('#set-main-category').prop('disabled', false);
        }
    });
   
    $('#modalRegisterTrigger').on('show.bs.modal',function(e){
        var selected_op = $('#main-category-id');
        $('#select-trigger-hostname').hide();
        if(selected_op.val() === 'all'){
            if($('#select-trigger-category').length === 0){
                var selections = make_selectable_categories('select-trigger-category', true);
                $('#trigger-category-menu > *').remove();
                $('#trigger-category-menu').append(selections);
                $('#trigger-category-menu [data-toggle="tooltip"]').tooltip();
                $('#trigger-category').text('カテゴリを選択')
                                      .addClass('dropdown-toggle')
                                      .attr('data-categoryid', '-1')
                                      .prop('disabled', false);
                var trigger_param = JSON.parse($('#form-create-trigger').val()); 
                if(trigger_param['period'] === 'vulsync'){
                    if(trigger_param['category'] === 'all'){
                        $('#trigger-category').text('全体').attr('data-categoryid', 'all');
                        selections.find('[value="all"]').prop('checked', true);
                    }else{
                        var target = selections.find('[value="' + trigger_param['category'] + '"]');
                        target.prop('checked', true);
                        var category_name = target.attr('data-text');
                        $('#trigger-category').text(category_name)
                                              .attr({'data-categoryid': trigger_param['category']});
                    }
                }
            }
        }else{
            var trigger_param = JSON.parse($('#form-create-trigger').val());
            $('#trigger-category-menu > *').remove();
            $('#trigger-category').text(selected_op.attr('data-text'))
                                  .removeClass('dropdown-toggle')
                                  .attr('data-categoryid', selected_op.val())
                                  .prop('disabled', true);
            var category_descs = JSON.parse($('#category-descs').val());
            var desc_cat = null;
            var main_cat_id = selected_op.val(); 
            if( Object.keys(trigger_param).length === 0 ){
                for(category_id in category_descs){
                    if( main_cat_id == category_id ){
                        desc_cat = category_descs[category_id];
                        break;
                    }
                }
            }else{
                desc_cat = category_descs[trigger_param['category']];
            }
            var target = $('#trigger-category').parent();
            target.attr({
                //'data-toggle': 'tooltip',
                //'data-title': (desc ? desc : ''),
                //'data-placement': 'right',
                //'data-boundary': 'window',
                //'data-trigger': 'manual'
                title: (desc_cat ? desc_cat.desc : '')
            });
        }
        if($('#search-hostname').val() === ''){
            $('#select-trigger-hostname > *').remove();
        }
    });

    $('#modalRegisterTrigger').on('shown.bs.modal',function(){
        check_blank_in_trigger_setting_modal();
        //modal-bodyを保存
      /*  var clone_modal_body = $(this).find('.modal-body')
                                      .clone(true);
        clone_modal_body.find('[id]')
            .each(function(){
                var elm = $(this);
                elm.attr('data-tempid', elm.attr('id'))
                   .removeAttr('id');
        });
        $('#hidden-modalRegisterTrigger-zone *').remove();
        $('#hidden-modalRegisterTrigger-zone').append(clone_modal_body);*/
    });

    $('#modalRegisterTrigger').on('hide.bs.modal',function(){
        var param = JSON.parse($('#form-create-trigger').val());
        if(Object.keys(param).length > 0){
            $('#modalCreateRule-set-trigger-btn').removeClass('btn-secondary')
                                                 .addClass('btn-success set')
                                                 .text('タイミング設定済み');
        }
    });

    $('#modalRegisterTrigger').on('hidden.bs.modal',function(){
        setTimeout(function(){
            show_change_rule_name_message();
        },500);
    });
    //タイミング設定のキャンセルで内容を元に戻す
 /*   $('#modalRegisterTrigger [data-dismiss="modal"]').on('click', function(){
        if($(this).attr('id') != 'set-trigger-setting'){
            var clone_modal_body = $('#hidden-modalRegisterTrigger-zone').find('.modal-body');
            clone_modal_body.find('[data-tempid]')
                .each(function(){
                    var elm = $(this);
                    elm.attr('id', elm.attr('data-tempid'))
                       .removeAttr('data-tempid');
            });
            $('#modalRegisterTrigger').find('.modal-body').replaceWith(clone_modal_body);
            $('#modalRegisterTrigger').find('[data-toggle="dropdown"]').dropdown();
        }
    });*/

    //タイミング設定の脆弱性同期時のカテゴリ選択
    $(document).on('click', '#select-trigger-category .c-category-radio, #select-trigger-category label',function(e){
        var op = $('#select-trigger-category input:checked');
        $('#trigger-category').text(op.attr('data-text'))
                              .attr('data-categoryid', op.val())
        if($('#trigger-category-menu').hasClass('show')){
            $('#trigger-category').click();
        }               
        check_blank_in_trigger_setting_modal();
        $('#select-trigger-hostname').hide();
        $('#select-trigger-hostname > *').remove();
        $('#search-hostname').val('');
    }); 
   

    //条件設定modal
    $('#modalRegisterCondition').on('show.bs.modal',function(e){
        toggle_execute_set_btn();
        var button = $(e.relatedTarget);
        if($('#set-condition-tbody').find('tr').length === 0){
            display_no_condition_message();
        }
        reset_add_condition_area();
        var selected_category = $('#main-category-id').val();
        var category_btns = $('#wrap-condition-item-btns .condition-item-btn').find('[data-cond="category-match"], [data-cond="category-unmatch"]');
        var wrap_category_btns = category_btns.closest('.condition-item-btn');
        var trigger = JSON.parse($('#form-create-trigger').val());
        /*var make_category_btns_disabled = function(){
                                              category_btns.addClass('disabled');
                                              category_btns.closest('.condition-item-btn')
                                                  .attr({
                                                      'data-toggle': 'tooltip',
                                                      'data-title':  'タイミング設定でカテゴリまたはホストを指定していない場合に使用できます',
                                              }).tooltip('enable');
                                          };*/
        category_btns.removeClass('disabled');
        if(wrap_category_btns.attr('data-toggle') === 'tooltip'){
            wrap_category_btns.tooltip('disable');
        }
        
        var host_btns = $('#wrap-condition-item-btns').find('[data-cond="host-match"], [data-cond="host-unmatch"]');
        if(trigger['period'] === 'vulsync'){
            var target_btn = $('#wrap-condition-item-btns').find('[data-cond="issue-update"]');
            target_btn.removeClass('disabled');
            target_btn.closest('.condition-item-btn').tooltip('disable');
            if(trigger['hostname'] != ''){
                host_btns.addClass('disabled');
                host_btns.closest('.condition-item-btn').tooltip('enable');
            }else{
                host_btns.removeClass('disabled');
                host_btns.closest('.condition-item-btn').tooltip('disable');
            }
        }else{
            host_btns.removeClass('disabled');
            host_btns.closest('.condition-item-btn').tooltip('disable');
            var target_btn = $('#wrap-condition-item-btns').find('[data-cond="issue-update"]');
            target_btn.addClass('disabled');
            target_btn.closest('.condition-item-btn').tooltip('enable');
        } 
        //テーブルのコピーを保存  
        $('#hidden-modalRegisterCondition-zone > *').remove(); 
        var clone_tbody = $(this).find('#set-condition-tbody')
                                 .clone()
                                 .attr('id', 'clone-set-condition-tbody');
        $('#hidden-modalRegisterCondition-zone').append(clone_tbody);
    });

    $('#modalRegisterCondition').on('shown.bs.modal',function(){
        //操作追加ボタンと被らないようにスペースを確保
        $('#keep-margin-for-cond').height($('#execute-add-cond').outerHeight(true) + 5);
    });
  
    $('#modalRegisterCondition').on('hide.bs.modal',function(){
        $('#wrap-condition-item-btns').scrollTop(0);
        $('#modalRegisterCondition-table').scrollTop(0);
    });

    $('#modalRegisterCondition').on('hidden.bs.modal',function(){
        var form_param = JSON.parse($('#form-create-condition').val());
        if(Object.keys(form_param).length > 0){
            $('#modalCreateRule-set-cond-btn').removeClass('btn-secondary')
                                              .addClass('btn-success set')
                                              .text('条件設定済み');
        }else{
            $('#modalCreateRule-set-cond-btn').removeClass('btn-success set')
                                              .addClass('btn-secondary')
                                              .text('条件未設定'); 
        }
        check_rule_is_registable();
        setTimeout(function(){
             show_change_rule_name_message();
        },500);
    });  

    $('#execute-set-cond').on('click', function(){
        var pass = true;
        if($('#cond-detail-body').find('[data-addcond]').length > 0){
            pass = confirm('編集中の条件詳細項目があります。このまま条件を設定しますか？');
        }
        if(pass){
            $('#form-create-condition').val($('#set-cond-param').val());
            $('.original-tooltip').remove();
            create_cond_display_content(JSON.parse($('#set-cond-param').val())); 
            $('#modalRegisterCondition').modal('hide');
        }
    });

    /*$(document).on('click', '.expand-btn', function(){
        var button = $(this);
        button.closest('.rule-setting-content')
               .addClass('expanded-detail')
               .removeClass('unexpanded-detail');
        button.replaceWith(
            $('<i>').attr('class', 'reset-expand-btn fas fa-angle-up mr-1')
        );
        $('.original-tooltip').remove();
    });

    $(document).on('click', '.reset-expand-btn', function(){
        var button = $(this);
        button.closest('.rule-setting-content')
               .addClass('unexpanded-detail')
               .removeClass('expanded-detail');
        button.replaceWith(
            $('<i>').attr('class', 'expand-btn fas fa-angle-down mr-1')
        );
        $('.original-tooltip').remove();
    });*/

    $('#close-modalRegisterCondition-btn, #cancel-modalRegisterCondition-btn').on('click',function(){
        var conf = confirm_close_RegisterModal('modalRegisterCondition');
        if(conf){
            $('#modalRegisterCondition').modal('hide');
        }
        
    });    


    //新規条件追加modalと新規操作追加modalが開いたとき、window内にできるだけ表示されるように調整
    $('#modalAddCondition, #modalAddOperation').on('shown.bs.modal',function(){
        var target = $('#modalAddCondition-content:visible, #modalAddOperation-content:visible')
        var margin_width = $(window).width() - target.outerWidth(true);
        if(margin_width < 130){
            target.css({
                  'width': target.outerWidth(true) + 'px',
                  'position': 'fixed',
                  'left': 0,
                  });
        }else{
            target.css({
                  'width': '',
                  'position': '',
                  'left': '',
            });
        }
    });

    //新規条件追加modalが消えるとき
    $('#modalAddCondition').on('hide.bs.modal',function(){ 
         $('#wrap-condition-item-btns').get(0).scrollTop = 0;
         $('#scroll-to-cond-item').find('[data-target=update]').prop('selected', true);
    });

    $('#modalAddCondition').on('hidden.bs.modal', function(){
        $('#cond-detail-body > *').remove();
        $('#modalSelectItems').find('.modal-body > *').remove();
    });

    //脆弱性同期時の項目の切り替え
    $(document).on('change', '#switch-issue-update-items', function(){
        var selected_op = $(this).find('option:selected');
        var wrap_elm = $('#cond-detail-body');//:visible, #edit-cond-after:visible');
        wrap_elm.find('[data-addcond]').hide();
        wrap_elm.find('[data-addcond="' + selected_op.val() + '"]')
                .show();
    });
 
    //新規条件modalの条件詳細部分の切り替え
    $('.select-cond').on('click', function(){
        var button = $(this);
        var cond = button.data('cond');
        var displayed_detail_item = $('#cond-detail').find('[data-addcond]');
        var detail_cond = displayed_detail_item.attr('data-addcond');
        if(displayed_detail_item.length > 0 && detail_cond != cond){
            if(cond === 'issue-update' && detail_cond.indexOf('issue-update') != -1){
                return false;
            }
            var conf = confirm('条件詳細の内容は破棄されます。よろしいですか?');
            if(!conf){
                return false;
            }
        }else if(displayed_detail_item.length > 0 && detail_cond === cond){
            return false;
        }
        color_selecting_item_on_AddModal(button);
        var item_name = cond_items_name[cond];
        $('#selecting-cond-item').text(item_name);
        var items = make_cond_detail_item(cond);
        var wrap_items = $('<div>').attr({
            'class':'badge badge-light cond-tr',
            'data-cond':cond,
        });
        wrap_items.append(items);
        $('#cond-detail-body').append(wrap_items);
        $('#execute-add-cond').show();
        $('#cond-detail-body [data-toggle="dropdown"]').dropdown();
        $('#cond-detail-body [data-toggle="tooltip"]').tooltip();
        $('#execute-add-cond').prop('disabled', false);
        $('#modalSelectItems').find('.modal-body > *').remove();
    });

    //条件詳細の項目作成function    
    function cond_issue_update_items(condition){
         var cond = condition;
         var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small')
                                     .html('脆弱性との照合を行う同期タイミングにおいて、対象ホストに新しい脆弱性が検出されチケット登録された場合、あるいは、過去に検出されていた脆弱性が検出されなくなったため、ステータスが「影響無」に変更された場合といった「脆弱性の更新」事象が発生したことを、条件として設定します。');
         $('#cond-detail-body').append(description);
         var wrap = $('<div>');
         var div = $('<div>').attr('class', 'mb-2');
         var select = $('<select>').attr({
                                        'class': 'form-control form-control-sm w-auto g-text-shadow-black',
                                        'id': 'switch-issue-update-items'
                                    });
         var op1 = $('<option>').attr({'value': 'issue-update-regi'}).text('脆弱性が新規に登録された時');
         var op2 = $('<option>').attr({'value': 'issue-update-del'}).text('脆弱性のステータスが「影響無」に変更された時');
         var op3 = $('<option>').attr({'value': 'issue-update-regi-and-del'}).text('両方を対象とする場合');
         select.append(op1, op2, op3);
         div.append(select);
         var item1 = $('<div>').attr({
                               'class': 'm-1',
                               'data-addcond': 'issue-update-regi'
                               })
                               .html('<div class="issue-update-item">'+
                                     '<div class="g-font-size-13 my-2">新しい脆弱性の登録が、</div>' +
																	 	 '<div class="input-group g-width-10em">' +
                                     '<input class="form-control detail-value" type="number" min="0" aria-describedby="iud1">' +
                                     '<div class="input-group-append">' +
                                     '<span class="input-group-text" id="iud1">件以上</span>' + '</div>' +
                                     '</div>'  + 
																		 '</div>'
                                     );
         var item2 = $('<div>').attr({
                               'class': 'm-1',
                               'data-addcond': 'issue-update-del'
                               })
                               .html('<div class="issue-update-item">' +
                                     '<div class="g-font-size-13 my-2">ステータスが「影響無」に変更されたものが、</div>' +
																	 	 '<div class="input-group g-width-10em">' +
                                     '<input class="form-control detail-value" type="number" min="0" aria-describedby="iud2">' +
                                     '<div class="input-group-append">' +
                                     '<span class="input-group-text" id="iud2">件以上</span>' + '</div>' +
                                     '</div>'  + 
																		 '</div>'
                                    ).hide();
         var item3 = $('<div>').attr({
                               'class': 'issue-update-item',
                               'data-addcond': 'issue-update-regi-and-del'
                               })
                               .html('<div>' +
                                     '<div class="g-font-size-13 my-2">新しい脆弱性の登録が、</div>' +
																	 	 '<div class="input-group g-width-10em">' +
                                     '<input class="form-control detail-value" type="number" min="0" id="issue-update-double-cond-regi"  aria-describedby="iud3">' +
                                     '<div class="input-group-append">' +
                                     '<span class="input-group-text" id="iud3">件以上</span>' + '</div>' +
                                     '</div>' + '</div>' +
                                     '<div>' +
                                     '<div class="g-font-size-13 my-2">ステータスが「影響無」に変更されたものが、</div>' +
																	 	 '<div class="input-group g-width-10em">' +
                                     '<input class="form-control detail-value" type="number" min="0" id="issue-update-double-cond-del" aria-describedby="iud4">' +
                                     '<div class="input-group-append">' +
                                     '<span class="input-group-text" id="iud4">件以上</span>' + '</div>' +
                                     '</div>'  + '</div>' 
                               ).hide();
         wrap.append(div, item1, item2, item3);
         return wrap;
    }
    
    function cond_cvss_exploitable(condition){
        var param = JSON.parse($('#form-create-trigger').val());
        if (param['period'] === 'vulsync') {
            var additional_item = '<div class="d-flex justify-content-end" style="margin-top:9px;">' +
                                  '<input type="checkbox" id="cvss-exploitable-is-updated" class="mr-1">' +
                                  '<div>攻撃コードが更新されている脆弱性のみ</div>' +
                                  '<a id="cvss-exploitable-is-update-description" data-toggle="tooltip" title="脆弱性同期時の前後で、攻撃コードが更新されている脆弱性のみをこの条件の対象とする場合には、チェックを付けて下さい">' +
                                  '<i class="fas fa-question-circle ml-1 c_gray_70"></i>' +
                                  '</a>' +
                                  '</div>';
        }else{
           var additional_item = '';
        }
        var item1 = $('<div>').attr({'class': 'm-1', 'data-addcond': condition})
                              .html(
                        '<div class="input-group g-width-21em">' +
										 		'<div class="input-group-prepend">' + 
                        '<label class="input-group-text" for="cvss-exploitable-is-exploitable">攻撃コードが</label>' +
												'</div>' +
                        '<select id="cvss-exploitable-is-exploitable" class="custom-select cond-detail-select">' +
                        '<option class="detail-value" value="true">有り</option>' +
                        '<option class="detail-value" value="false">無し</option>' +
                        '</select>' +
                        '</div>' +
                        additional_item
                    );
         var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small')
                                     .html('攻撃コードの「有」「無」を、条件として設定します。');
         $('#cond-detail-body').append(description);
         return [item1];
    }

    function cond_sri_items(condition){
        var cond = condition;
        var item1 = $('<div>');
        if(cond === 'sri-warning'){
            var text = '脆弱性のSRI値が、要対応条件に「一致」するか、「不一致」であるかを、条件として設定します。';
            item1.html(
                     '<div class="input-group g-width-21em">' +
										 '<div class="input-group-prepend">' + 
                     '<label class="input-group-text" for="sri1">SRI値が</label>' + '</div>' +
                     '<select class="custom-select cond-detail-select" aria-labelledby="sri1">' +
                     '<option class="detail-value" value="true">要対応条件に一致</option>' +
                     '<option class="detail-value" value="false">要対応条件に不一致</option>' +
                     '</select>' +
                     '</div>' 
                     ).attr('data-addcond', 'sri-warning');
        }else{
            var text = '脆弱性のSRI値を、条件として設定します。';
            var select = $('<select>').attr({
																						class: 'custom-select cond-detail-select g-width-7em',
																						id: 'sri2'
																						});
            var values = ['重大', '高', '中', '低'];
            for(var i=0; i < values.length; i++){
                var op = $('<option>').attr('class', 'detail-value');
                switch (i) {
                    case 0:
                        op.val('4');break;
                    case 1:
                        op.val('3');break;
                    case 2:
                        op.val('2');break;
                    case 3:
                        op.val('1');break;
                }
                op.text(values[i]);
                select.append(op);
            }
            item1.html(
                     '<div class="input-group">' +
										 '<div class="input-group-prepend">' + 
                     '<label class="input-group-text" for="sri2">SRI値が</label>' + '</div>' +
										 '<div data-select="risks"></div>' + 
										 '<div class="input-group-append">' + 
										 ' <label class="input-group-text" for="sri2">以上</lable>' + '</div>' +
										 '</div>'
										 );
            if(cond === 'sri-below'){
                item1.html(item1.html().replace('以上', '以下'))
                     .attr('data-addcond', 'sri-below');
            }else if(cond === 'sri-above'){
                item1.attr('data-addcond', 'sri-above');
            }

            item1.find('[data-select=risks]')
                 .append(select);
        }
        var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small')
                                     .text(text);
        $('#cond-detail-body').append(description);
        var items = [item1];
        return items;
    }

    function cond_cvss_items(condition){
        var cond = condition;
        var item1 = $('<div>');
        if(cond === 'cvss-notreg'){
            var text = '脆弱性のCVSS値が定義されているか、否かを、条件として設定します。';
            item1.html(
											 '<div class="input-group g-width-16em">' +
                       '<div class="input-group-prepend">' +
											 '<label class="input-group-text" for="cvss1">CVSS値が</label>' + '</div>' +
                       '<select class="custom-select cond-detail-select" aria-labelledby="cvss1">' +
                       '<option class="detail-value" value="true">未定義</option>' +
                       '<option class="detail-value" value="false">未定義以外</option>' +
                       '</select>' +
											 '</div>'
                      ).attr('data-addcond', 'cvss-notreg');
            var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small')
                                        .text(text);
            $('#cond-detail-body').append(description);
            var items = [item1];
            return items;
        }else if(cond === 'cvss-warning'){
             var text = '脆弱性のCVSS値が要対応条件に「一致」するか、「不一致」であるかを、条件として設定します。';
             item1.html(
											  '<div class="input-group g-width-21em">' +
                        '<div class="input-group-prepend">' +
						 						'<label class="input-group-text" for="cvss2">CVSS値が</label>' + '</div>' +
                        '<select class="custom-select cond-detail-select"  aria-labelledby="cvss2">' +
                        '<option class="detail-value" value="true">要対応条件に一致</option>' +
                        '<option class="detail-value" value="false">要対応条件に不一致</option>' +
                        '</select>' +
												 '</div>'
                        ).attr('data-addcond', 'cvss-warning');
             var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small')
                                         .text(text);
             $('#cond-detail-body').append(description);
             var items = [item1];
             return items;
        }else if( cond === 'cvss-above' || cond === 'cvss-below' ){
            var text = '脆弱性のCVSS値を、条件として設定します。※小数点第一位までの0.0～10.0で指定可能';
            var include_notreg = $('<div>').attr('class', 'text-right optional-cond mb-0 pr-2')
                                 .html( '<div class="mb-1 text-right font-size-small font-weight-bold">' +
																 				'CVSS値が未定義である場合を含めるかどうか' + 
                                        '</div>' +
                                        '<div class="d-flex">' +
                                        '<select class="custom-select custom-select-sm ml-auto w-auto" id="if-include-notreg">' + 
                                        '<option value="false">未定義の脆弱性を含めない</option>' +
                                        '<option value="true">未定義の脆弱性を含める</option>' +
                                        '</select>' +
                                        '</div>'
                                  );
            var flg_exploitable = $('<div>').attr('class', 'text-right optional-cond mb-2 p-2')
                                  .html('<div class="mb-1 text-right font-size-small font-weight-bold">' +
                                        '攻撃コードの有無' +
                                        '<i id="cvss-tooltip" class="ml-1 fas fa-question-circle c_gray_70" data-toggle="tooltip" data-title="CVSS値が定義されている脆弱性に対してのみ、この「攻撃コードの有無」のオプションは適用されます"></i>' + 
                                        '</div>' +
                                        '<div class="d-flex">' +
                                        '<select class="custom-select custom-select-sm  ml-auto w-auto" id="flg-exploitable">' +
                                        '<option value="all">全ての脆弱性が対象</option>' +
                                        '<option value="not_exist">無しの脆弱性が対象</option>' +
                                        '<option value="exist">有りの脆弱性が対象</option>' +
                                        '</select>' +
                                        '</div>'
                                  );
            item1.html(
						           '<div class="input-group g-width-18em">' +
											 '<div class="input-group-prepend">' +
											 '<label class="input-group-text" for="cvss3">CVSS値が</label>' + '</div>' +
											 '<input class="form-control detail-value cvss-value" type="number" max="10" min="0" step="0.1">' +
											 '<div class="input-group-append">' +
											 ' <label class="input-group-text" for="cvss3">以上</lable>' + '</div>' +
											 '</div>'
											 );
            if(cond === 'cvss-below'){
								item1.html(
													 '<div class="input-group g-width-18em">' +
													 '<div class="input-group-prepend">' +
													 '<label class="input-group-text" for="cvss4">CVSS値が</label>' + '</div>' +
													 '<input class="form-control detail-value cvss-value" type="number" max="10" min="0" step="0.1">' +
													 '<div class="input-group-append">' +
													 ' <label class="input-group-text" for="cvss4">以下</lable>' + '</div>' +
													 '</div>'
													 );
                item1.attr('data-addcond', 'cvss-below');
            }else if(cond === 'cvss-above'){
                item1.attr('data-addcond', 'cvss-above');
            }
            var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small')
                                        .html(text);
            $('#cond-detail-body').append(description);
            var items = [item1, flg_exploitable, include_notreg];
            return items;
        }else if( cond == 'cvss-vector' ){
            var vectors = {
                'vectors': {},
                'flg_perfect': 'false',
                'flg_ignore_undef': 'false'
            };
            var cvss = JSON.parse($('#cvss-vectors-data').val());
            $.each(Object.keys(cvss), function(){
                vectors['vectors'][this] = null;
            });
            item1.attr({
                'class': 'cvss-selection-area',
                'data-addcond': cond,
            });
            var vector_elms = [
                $('<input>').attr({ 'type': 'hidden', 'id': 'cvss-vectors' }).val(JSON.stringify(vectors)),
                $('<div>').attr({ 'class': 'd-flex mt-1 mr-1 ml-1' }).append(
                    $('<div>').attr({'class': 'cvss-v-item-name text-center cvss-v-item-name-av'}).html(cvss['AV']['item_name']),
                    $('<div>').attr({
                        'class': 'cvss-v-item cvss-v-item-av text-center unselected-cvss',
                        'data-cvss': 'AV'
                    }).text('---')
                ),
                $('<div>').attr({ 'class': 'd-flex mt-1 mr-1 ml-1' }).append(
                    $('<div>').attr({'class': 'cvss-v-item-name text-center'}).html(cvss['AC']['item_name']),
                    $('<div>').attr({
                        'class': 'cvss-v-item cvss-v-item-ac text-center unselected-cvss',
                        'data-cvss': 'AC'
                    }).text('---')
                ),
                $('<div>').attr({ 'class': 'd-flex mt-1 mr-1 ml-1' }).append(
                    $('<div>').attr({'class': 'cvss-v-item-name text-center'}).html(cvss['PR']['item_name']),
                    $('<div>').attr({
                        'class': 'cvss-v-item cvss-v-item-pr text-center unselected-cvss',
                        'data-cvss': 'PR'
                    }).text('---')
                ),
                $('<div>').attr({ 'class': 'd-flex mt-1 mr-1 ml-1' }).append(
                    $('<div>').attr({'class': 'cvss-v-item-name text-center'}).html(cvss['UI']['item_name']),
                    $('<div>').attr({
                        'class': 'cvss-v-item cvss-v-item-ui text-center unselected-cvss',
                        'data-cvss': 'UI'
                    }).text('---')
                ),
                $('<div>').attr({ 'class': 'd-flex mt-1 mr-1 ml-1' }).append(
                    $('<div>').attr({'class': 'cvss-v-item-name text-center'}).html(cvss['S']['item_name']),
                    $('<div>').attr({
                        'class': 'cvss-v-item cvss-v-item-s text-center unselected-cvss',
                        'data-cvss': 'S'
                    }).text('---')
                ),
                $('<div>').attr({ 'class': 'd-flex mt-1 mr-1 ml-1' }).append(
                    $('<div>').attr({'class': 'cvss-v-item-name text-center'}).html(cvss['C']['item_name']), 
                    $('<div>').attr({
                        'class': 'cvss-v-item cvss-v-item-c text-center unselected-cvss',
                        'data-cvss': 'C'
                    }).text('---')  
                ),
                $('<div>').attr({ 'class': 'd-flex mt-1 mr-1 ml-1' }).append(
                    $('<div>').attr({'class': 'cvss-v-item-name text-center'}).html(cvss['I']['item_name']), 
                    $('<div>').attr({
                        'class': 'cvss-v-item cvss-v-item-i text-center unselected-cvss',
                        'data-cvss': 'I'
                    }).text('---')  
                ),
                $('<div>').attr({ 'class': 'd-flex mt-1 mr-1 ml-1 mb-1' }).append(
                    $('<div>').attr({'class': 'cvss-v-item-name text-center'}).html(cvss['A']['item_name']), 
                    $('<div>').attr({
                       'class': 'cvss-v-item cvss-v-item-a text-center unselected-cvss',
                       'data-cvss': 'A'
                    }).text('---')
                )
            ];
            var set_button = $('<div>').attr('class', 'd-flex mb-1').append(
                                 $('<button>').attr({
                                     'type': 'button',
                                     'class': 'mr-auto btn btn-primary btn-xs',
                                     'id': 'set-cvss-vectors',
                                 }).text('選択する')
                             );
            item1.append(vector_elms); 
            var opts_area = $('<div>').attr('id', 'cvss-vectors-opts-area');
            return [ set_button, item1, opts_area ];
        }
    }

    function cond_due_date_items(condition){
        var cond = condition;
        var item1 = $('<div>');
        if(cond === 'due-date-expired'){
            var text = '対策予定日の超過を、条件として設定します。'
            item1.html('<span>対策予定日を超過</span>')
                 .attr({
                 'class': 'p-2',
                 'data-addcond': 'due-date-expired'
                 });
        }else if(cond === 'due-date-approach'){
           var text = '対策予定日が接近していることを知らせるための条件設定です。対策予定日までの残日数を指定します。'
           item1.html(
                      '<div class="input-group g-width-18em">' +
                      '<div class="input-group-prepend">' +
                      '<span class="input-group-text">対策予定日が</span>' + '</div>' +
                      '<input class="form-control detail-value" type="number" min="0">' +
                      '<div class="input-group-append">' +
                      ' <span class="input-group-text">日以内</span>' + '</div>' +
                      '</div>'
                      )
                .attr('data-addcond', 'due-date-approach');
        }else if(cond === 'due-date-undef'){
            var text = '対策予定日が未設定であることを、条件として設定します。';
            item1.html('<span>対策予定日が未設定</span>')
                 .attr('data-addcond', condition);
        }
        var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small')
                                    .text(text);
        $('#cond-detail-body').append(description);
        var items = [item1];
        return items;
    }

    function cond_history_items(condition){
        $('#mail-setting-users-tbody > *').remove();
        var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small')
                                    .html('履歴の投稿件数を、条件として設定します。<br>ユーザを指定する場合は、チェックボックスのいずれかにチェックを入れた後、ユーザ検索からユーザを指定して下さい。');
        $('#cond-detail-body').append(description);
        var item1 = $('<div>').attr('data-addcond', 'history-create');
				item1.append(
            $('<div>').attr({ 'class': 'input-group g-width-21em' })
                      .append(
                          $('<div>').attr({ 'class': 'input-group-prepend' })
                                    .append($('<span>').attr({ 'class': 'input-group-text' }).text('履歴の投稿が合計'))
                      )
                      .append(
                          $('<input>').attr({ 
                              'class': 'form-control detail-value',
                              'type': 'number',
                              'min': 0,
                              'id': 'history-create-cond-history-num'
                          }),
                          $('<div>').attr({ 'class': 'input-group-append' })
                                    .append($('<span>').attr({ 'class': 'input-group-text' }).text('件以上'))
                      ) 
				);
        var users_area = $('<div>').attr({ 'class': 'disabled history-create-cond-users-area', 'id': 'history-create-cond-users-area' }).append(
            $('<input>').attr({ 'type': 'hidden', 'id': 'history-create-cond-selected-user-ids' }).val('[]'),
            $('<div>').append(
                $('<button>').attr({
                    'type': 'button',
                    'class': 'btn btn-primary btn-xs text-nowrap mb-2',
                    'id': 'history-create-cond-search-user-button'
                }).text('ユーザ検索')
            ),
            $('<div>').attr({ 'class': 'history-create-cond-selected-users-area', 'id': 'history-create-cond-selected-users-area'} ),
        );
        var chkbox_area = $('<div>').append(
            $('<div>').attr('class', 'mt-2 ml-1 d-flex align-items-center').append(
                $('<input>').attr({ 
                    'type': 'radio', 
                    'id': 'history-create-cond-select-none',
                    'name': 'history-create-cond-opts',
                    'checked': true
                }),
                $('<label>').attr({ 'class': 'ml-1 mb-0', 'for': 'history-create-cond-select-none' }).text('ユーザ指定なし')
            ),
            $('<div>').attr('class', 'mt-2 ml-1 d-flex align-items-center history-create-cond-checkbox-select-only-area').append(
                $('<input>').attr({ 
                    'type': 'radio',
                    'id': 'history-create-cond-select-only',
                    'name': 'history-create-cond-opts',
                }),
                $('<label>').attr({ 'class': 'ml-1 mb-0', 'for': 'history-create-cond-select-only' }).text('指定したユーザが作成した履歴のみを対象にする')
            ),
            $('<div>').attr('class', 'mt-2 ml-1 d-flex align-items-center history-create-cond-checkbox-select-except-area').append(
                $('<input>').attr({ 
                    'type': 'radio',
                    'id': 'history-create-cond-select-except',
                    'name': 'history-create-cond-opts',
                }),
                $('<label>').attr({ 'class': 'ml-1 mb-0', 'for': 'history-create-cond-select-except' }).text('指定したユーザが作成した履歴以外を対象にする')
            )
        );
        var items = [ item1, users_area, chkbox_area ];
        return items;
    }
   
    function cond_category_items(condition){
        var cond = condition; 
        var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small');
        var description_text = (cond === 'category-match' ? 'ルールを適用するカテゴリを限定します。' : 'ルールを適用させない(ルールの対象から除外する)カテゴリを設定します。');
        if( $('#main-category-id').val() != 'all' ){
            description_text += '<br><small>※カスタムカテゴリは、「ルールを適用するカスタムカテゴリ」で「全体」を選択している場合にのみ選択できます。</small>';
        }
        $('#cond-detail-body').append(description.html(description_text));
        var item1 = $('<div>').attr({
                              'class': 'mt-1 mb-2', 
                              'data-addcond': cond,
                              }).text(cond === 'category-match' ? 'カテゴリ限定' : 'カテゴリ除外');
        var category_select_area = $('<div>').attr('class', 'optional-cond');
        var desc = $('<div>').attr('class', 'text-right font-weight-normal font-size-small').text('カテゴリ選択※複数選択可');
        var selections =  make_selectable_categories('select-cond-cats', false, true);
        selections.find('input')
                  .attr({
                      'type': 'checkbox',
                      'data-checkbox': 'true'
                  });
        category_select_area.append(desc, selections);
        var items = [item1, category_select_area];
        return items;
    }

    function cond_host_items(condition){
        var cond = condition;
        if( condition === 'host-syncing' ){
            var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small')
                                        .text('ホストの同期状態を、条件として設定します。');
            $('#cond-detail-body').append(description);
            var item = $('<div>').attr({
                'class': 'ml-1',
                'data-addcond': cond,
            });
            var inner = $('<div>').attr('class', 'input-group g-width-21em');
            var label_div = $('<div>').attr('class', 'input-group-prepend').append(
                                $('<label>').attr({'class': 'input-group-text', 'for': 'select-host-syncing'}).text('のホスト')
                            );
            var select = $('<select>').attr({
                'class': 'form-control',
                'id': 'select-host-syncing'
            });
            select.append(
                $('<option>').val('true').text('同期中'),
                $('<option>').val('false').text('非同期')
            );
            inner.append(select, label_div);
            item.append(inner);
            var items = [ item ];
        }else{
            var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small')
                                        .text(cond === 'host-match' ? 'ルールを適用するホストを限定します。' : 'ルールを適用させない(ルールの対象から除外する)ホストを設定します。');
            $('#cond-detail-body').append(description);
            var item1 = $('<div>').attr({
                'class': 'mb-3 ml-1',
                'data-addcond': cond,
            }).text(cond === 'host-match' ? 'ホスト限定' : 'ホスト除外');
            var wrap_elms = $('<div>').attr({'class': 'mt-2', 'id': 'added-hosts-area'});
            var search_hosts_field = $('<div>').attr('class', '');
            var hosts = $('<div>').attr({'id': 'cond-detail-add-hosts'});
            var search_btn = $('<div>').attr('class', 'mb-2')
                                       .append(
                                           $('<button>').attr({
                                               'type': 'button',
                                               'class': 'btn btn-primary btn-xs text-nowrap',
                                               'id': 'cond-detail-search-host-btn',
                                               'data-targetcond': cond
                                           }).text('ホスト検索')
                                       );
            search_hosts_field.append(search_btn, hosts);
            wrap_elms.append(search_hosts_field);
            var items = [item1, wrap_elms];
        }
        return items;
     }

     function cond_status_items(condition){
         var cond = condition;
         var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small')
                                     .text('脆弱性への「対応の状態」が一致している場合を、条件として設定します。');
         $('#cond-detail-body').append(description);
         var item = $('<div>').attr({
                              'data-addcond': cond
                              });
         var select_area = $('<div>').attr({
                                     'class': 'd-flex justify-content-center mb-3',
                                     'id': 'cond-detail-select-statuses'
                                     });
         for(status_id in all_statuses){
             var status_elm = $('<div>').attr({
                                         'class': 'btn unselected-status ml-1',
                                         'data-statusid': status_id,
                                         'data-item': 'status',
                                         }).text(all_statuses[status_id]);
             select_area.append(status_elm);
         }
         var title = $('<div>').attr('class', 'mt-1 mb-3')
                               .text(cond_detail_items_name[cond])
         var caution = $('<div>').attr('class', 'text-right font-weight-normal text-titleblue font-size-small mb-1')
                                 .text('以下のステータスの選択は、２つまで選択可能');
         item.append(title, caution, select_area)
         return item;
    }   
   
    function cond_product_items (condition){
        var cond = condition;
        var item = $('<div>').attr({
                              'data-addcond': cond
                              });
        var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small')
                                    .text(cond === 'product-match' ? 'ルールを適用させるプロダクトを設定します。' : 'ルールを適用させない（ルールの対象から除外する）プロダクトを設定します。');
        $('#cond-detail-body').append(description);
        var title = $('<div>').attr('class', 'mb-3 ml-1')
                              .text(cond_detail_items_name[cond]);
        var search_prod_field = $('<div>').attr('class', '');
        var products_area = $('<div>').attr({
                                          'id': 'cond-detail-products-area',
                                      });
        var search_btn = $('<div>').attr('class', 'mb-2')
                                   .append(
                                       $('<button>').attr({
                                           'type': 'button',
                                           'id': 'cond-detail-search-products', 
                                           'class': 'btn btn-xs btn-primary text-nowrap',
                                       }).text('プロダクト検索')  
                                   );
        search_prod_field.append(search_btn, products_area);
        item.append(title, search_prod_field);
        return item;
    }

    function cond_ctag_items(condition){
        var item = $('<div>').attr({
            'data-addcond': condition
        });
        var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small').html(
            '指定したカスタムタグの有無を条件として設定します。<br>' +
            'この条件は、１つのルールに複数設定可能です。<br>' +
            '同じ条件で設定したカスタムタグはOR条件になり、別条件として設定したカスタムタグはAND条件となります。'
        );
        $('#cond-detail-body').append(description);
        var area = $('<div>');
        var search_btn = $('<div>').append(
                                       $('<button>').attr({
                                           'type': 'button',
                                           'id': 'cond-detail-search-ctags',
                                           'class': 'btn btn-xs btn-primary text-nowrap',
                                       }).text('カスタムタグ検索')
                                   );
        var true_searched_area = $('<div>').attr('class', 'mt-3').append(
            $('<div>').attr('class', 'font-weight-bold pb-1').text('あれば成立'),
            $('<div>').attr({ id: 'cond-detail-ctags-true-area', class: 'detail-ctags-area' })
        );
        var false_searched_area = $('<div>').attr('class', 'mt-2').append(
            $('<div>').attr('class', 'font-weight-bold pb-1').text('なければ成立'),
            $('<div>').attr({ id: 'cond-detail-ctags-false-area', class: 'detail-ctags-area' })
        );
        var no_ctags_elm = $('<div>').attr('class', 'mt-2').append(
            $('<input>').attr({
                type: 'checkbox',
                id: 'match-if-not-ctag'
            }),
            $('<label>').attr({
                'class': 'ml-1 mb-0',
                'for': 'match-if-not-ctag'
            }).text(MATCH_IF_NO_CTAG_LABEL)
        );
        area.append(search_btn, true_searched_area, false_searched_area, no_ctags_elm);
        item.append(area);
        return item;
    }

    function cond_timespan_items(condition){
        var description = $('<div>').attr('class', 'mt-1 mb-2 font-size-small').append(
                                        $('<span>').html('ルールの処理日時を起点とした期間を条件として設定します。<br>ルールの実行日時から、指定した期間をさかのぼった期間に登録・更新された脆弱性と、その脆弱性に関連するホストがルールの対象となります。'),
                                        $('<a>').attr({
                                            'data-toggle': 'tooltip', 
                                            'data-title': '<div class="p-2">' + 
                                                          '<div class="font-weight-bold mb-2" style="font-size: 108%;">例:&nbsp;2022年1月10日 12時00分00秒にルールが実行された場合(期間内)</div>' +
                                                          '<div class="mb-2">期間を1時間に設定した場合&nbsp;→&nbsp;2022年1月10日 11時00分00秒からルールの実行日時までに登録・更新されたものが適合</div>' +
                                                          '<div class="mb-2">期間を1日に設定した場合&nbsp;→&nbsp;2022年1月09日 12時00分00秒からルールの実行日時までに登録・更新されたものが適合</div>' + 
                                                          '<div class="mb-2">期間を1週に設定した場合&nbsp;→&nbsp;2022年1月03日 12時00分00秒からルールの実行日時までに登録・更新されたものが適合</div>' +
                                                          '<div>期間を1カ月に設定した場合&nbsp;→&nbsp;2021年12月10日 12時00分00秒からルールの実行日時までに登録・更新されたものが適合</div>' +
                                                          '</div>',
                                            'data-html': 'true'
                                                            
                                        }).append(
                                            $('<i>').attr('class', 'fas fa-info-circle c_gray_70')
                                        ),
                                        $('<div>').text('「指定した期間を除く期間が対象」にチェックを入れると、指定した期間をさかのぼった期間以外の期間が対象になります。')
                                    );
        $('#cond-detail-body').append(description);
        var item = $('<div>').attr({ 'data-addcond': condition });
        //間隔のスタイルに合わせてdropdownを使用する
        var unit_dropdown = $('<div>').attr('class', 'g-width-21em');
        var unit_btn = $('<button>').attr({
            'type': 'button',
            'class': 'cond-timerange-select-btn g-width-21em',
            'data-toggle': 'dropdown'
        }).html('<span id="cond-timerange-selected-unit-text">時間</span><i class="fas fa-angle-down ml-auto"></i>');
        var unit_d_menu = $('<div>').attr({
            'class': 'dropdown-menu cond-timerange-d-menu'
        });
        var unit_div = $('<div>').attr('class', 'form-group');
        var unit_label = $('<label>').attr({
            'class': 'font-weight-bold ml-1',
            'for': 'cond-timerange-select-unit'
        }).text('単位');
        var select_unit = $('<select>').attr({
            'id': 'cond-timerange-select-unit',
            'class': 'form-control form-control-sm cond-timerange-select',
            'size': '4'
        }).append(
            $('<option>').val('hour').text('時間').prop('selected', true),
            $('<option>').val('day').text('日'),
            $('<option>').val('week').text('週'),
            $('<option>').val('month').text('月')
        );
        unit_d_menu.append(select_unit);
        unit_dropdown.append(unit_btn, unit_d_menu);
        unit_div.append(unit_label, unit_dropdown);
        var value_div = $('<div>').attr('class', 'form-group');
        var value_label = $('<label>').attr({
            'class': 'font-weight-bold ml-1',
            'for': 'cond-timerange-select-value'
        }).text('間隔');
        //selectボックスの大きさが長くなりすぎるのでdropdownを使用
        var dropdown = $('<div>').attr('class', 'g-width-21em');
        var btn = $('<button>').attr({
            'type': 'button', 
            'class': 'cond-timerange-select-btn g-width-21em',
            'data-toggle': 'dropdown'
        }).html('<span id="cond-timerange-selected-value-text">1時間</span><i class="fas fa-angle-down ml-auto"></i>');
        var d_menu = $('<div>').attr({
            'class': 'dropdown-menu cond-timerange-d-menu'
        });
        var select_value = $('<select>').attr({
            'id': 'cond-timerange-select-value',
            'class': 'form-control form-control-sm cond-timerange-select',
            'size': '6'
        });
        var hours = [];
        var days = [];
        var weeks = [];
        var months = [];
        //デフォルトは時間
        for(i = 1; i <= 90; i++){
            if( i <= 12 ){
                months.push($('<option>').val(i + '-month').attr('data-unit', 'month').text(i + 'カ月').hide());
            }
            if( i <= 24 ){
                var op = $('<option>').val(i + '-hour').attr('data-unit', 'hour').text(i + '時間');
                if( i === 1 ) {
                    op.prop('selected', true);
                }
                hours.push(op);
            }
            if( i <= 30 ){
                days.push($('<option>').val(i + '-day').attr('data-unit', 'day').text(i + '日').hide());
            }
            weeks.push($('<option>').val(i + '-week').attr('data-unit', 'week').text(i + '週').hide());
        }
        select_value.append(hours);
        select_value.append(days);
        select_value.append(weeks);
        select_value.append(months);
        d_menu.append(select_value)
        dropdown.append(btn, d_menu);
        value_div.append(value_label, dropdown);

        var except_check_area = $('<div>').append(
            $('<input>').attr({
                type: 'checkbox',
                id: 'cond-timerange-flg-except'
            }),
            $('<label>').attr({ 'for': 'cond-timerange-flg-except', 'class': 'ml-1 mb-0' }).text('指定した期間を除く期間が対象')
        );
        item.append(unit_div, value_div, except_check_area);
        return item;
    }

    //新規条件modalの選択項目作成function
    function make_cond_detail_item(condition){
        var cond = condition;
        var wrap_items = $('<div>');
        $('#cond-detail-body > *').remove();
        if(cond === 'issue-update'){
            var items = cond_issue_update_items(cond);
        }else if(cond === 'cvss-exploitable'){
            var items = cond_cvss_exploitable(cond);
        }else if(cond.indexOf('sri') != -1){
            var items = cond_sri_items(cond);

        }else if(cond.indexOf('cvss') != -1){
            var items = cond_cvss_items(cond);

        }else if(cond.indexOf('due-date') != -1){ 
            var items = cond_due_date_items(cond);

        }else if(cond === 'history-create'){
           var items = cond_history_items(cond);

        }else if(cond === 'category-match' || cond === 'category-unmatch'){
            var items = cond_category_items(cond);
        }else if(cond === 'host-match' || cond === 'host-unmatch' || cond === 'host-syncing'){
            var items = cond_host_items(cond);
        }else if(cond === 'status-match' || cond === 'status-unmatch'){
            var items = cond_status_items(cond);
        }else if(cond === 'product-match' || cond === 'product-unmatch'){
            reset_SetProducts_modal();
            var items = cond_product_items(cond);
        }else if(cond === 'ctag'){
            ruleSelectedCtags.val('{}'); //登録内容をリセット
            ruleSelectedRegistCtags.val('{}');
            var items = cond_ctag_items(cond);
        }else if(cond === 'timerange'){
            var items = cond_timespan_items(cond);
        }
        return items;
    }

    //条件の「履歴追加」で、ラジオボタンを選択した時
    $(document).on('change', '[name="history-create-cond-opts"]', function(){
        var radio_id = $(this).attr('id');
        if( radio_id === 'history-create-cond-select-only' || radio_id === 'history-create-cond-select-except' ){
            $('#history-create-cond-users-area').removeClass('disabled');
        }else{
            $('#history-create-cond-users-area').addClass('disabled');
        }
    });

    //条件の「履歴追加」の選択可能なユーザ一覧で、決定ボタンを押した時
    $(document).on('click', '#decide-history-create-cond-users', function(){
        $('#history-create-cond-selected-users-area *').remove();
        var user_ids = [];
        var user_elms = [];
        var cnt = 0;
        $('#mail-setting-users-tbody').find('.chk-box:checked').each(function(){
            var checked = $(this);
            var id = checked.val();
            var tr = checked.closest('tr');
            var name = tr.find('[data-name]').text();
            var name_elm = $('<div>').text(name)
                                     .attr({
                                         'class': 'min-h-30px pl-1 g-font-size-14 d-flex align-items-center font-weight-normal',
                                         'data-userid': id,
                                     });
            if(cnt % 2 === 0){
                name_elm.addClass('user-name-bg');
            }
            cnt += 1;
            user_elms.push(name_elm);
            user_ids.push(id);
         });
         $('#history-create-cond-selected-user-ids').val(JSON.stringify(user_ids));
         $('#history-create-cond-selected-users-area').append(user_elms);
         $('#modalSearchUsers').modal('hide'); 
    });

    //条件の「期間指定」で、単位・間隔を選択した時
    $(document).on('change', '#cond-timerange-select-unit', function(){
        var op = $(this).find('option:selected');
        $('#cond-timerange-selected-unit-text').text(op.text());
        var target = '[data-unit="' + op.val() + '"]';
        var select_value = $('#cond-timerange-select-value');
        var show_ops = select_value.find(target); 
        show_ops.show();
        select_value.find(':not(' + target + ')').hide();
        var select_value_target_op = show_ops.first();
        select_value_target_op.prop('selected', true);
        $('#cond-timerange-selected-value-text').text(select_value_target_op.text());
    });

    $(document).on('change', '#cond-timerange-select-value', function(){
        var op = $(this).find('option:selected');
        $('#cond-timerange-selected-value-text').text(op.text());
    });

    //条件の「期間指定」の日本語化
    function create_timerange_text(param){
        var timerange_unit = '';
        var timerange_val = '';
        if( param['unit'] === 'hour' ){
            timerange_unit = '時間';
            timerange_val = param['value'] + '時間';
        }else if( param['unit'] === 'day' ){
            timerange_unit = '日';
            timerange_val = param['value'] + '日';
        }else if( param['unit'] === 'week' ){
            timerange_unit = '週';
            timerange_val = param['value'] + '週';
        }else if( param['unit'] === 'month' ){
            timerange_unit = '月';
            timerange_val = param['value'] + 'カ月';
        }
        return { 'unit': timerange_unit, 'value': timerange_val, 'except': (param['except'] ? '期間外が対象' : '期間内が対象') };
    }

    //条件の「ステータスの一致」で、選択したステータスの色を変える
    $(document).on('click', '#cond-detail-select-statuses [data-item="status"]', function(){
       var button = $(this);
       if(button.hasClass('selected-status')){
           button.addClass('unselected-status')
                 .removeClass('selected-status');
       }else{
           button.addClass('selected-status')
                 .removeClass('unselected-status');
       }
   });


   //CVSS指標で項目を選択した時
   $('#cvss-selection-area [data-cvss]').on('click', function(){
       var cvss = JSON.parse($('#selected-cvss-vectors').val());
       var button = $(this);
       var cvss_item = button.attr('data-cvss');
       var cvss_value = button.attr('data-value');
       var target_values = cvss[cvss_item];
       if( button.hasClass('selected-cvss-item') ){
           button.removeClass('selected-cvss-item');
           cvss[cvss_item] = target_values.filter(function(v){
               return v != cvss_value;
           });
           if( cvss[cvss_item].length === 0 ){
               cvss[cvss_item].push('NONE');
               $('#cvss-selection-area [data-cvss="' + cvss_item + '"][data-value="NONE"]').addClass('selected-cvss-item');
           }
       }else{
           if( cvss_value === 'NONE' ){
               target_values = [];
               $('#cvss-selection-area [data-cvss="' + cvss_item + '"]').removeClass('selected-cvss-item');
           }else{
              target_values = target_values.filter(function(v){
                  return v != 'NONE';
              });
              $('#cvss-selection-area [data-cvss="' + cvss_item + '"][data-value="NONE"]').removeClass('selected-cvss-item');
           }
           button.addClass('selected-cvss-item');
           target_values.push(cvss_value);
           cvss[cvss_item] = target_values;
       }
       $('#selected-cvss-vectors').val(JSON.stringify(cvss));
       var is_unselected_item = false;
       $.each(Object.values(cvss), function(){
           if( this.length === 0 ){
               is_unselected_item = true;
               return false;  
           }
       });
       if( is_unselected_item){
           $('#execute-select-cvss-vectors').prop('disabled', true);
       }else{
           $('#execute-select-cvss-vectors').prop('disabled', false);
       }
   });

   //CVSS指標の条件詳細で「選択する」ボタンを押した時に項目を選択するモーダルを開く
   $(document).on('click', '#set-cvss-vectors', function(){
       var cvss = JSON.parse($('#cvss-vectors').val());
       if( Object.values(cvss['vectors']).includes(null) ){
           var temp = {}
           $.each(cvss['vectors'], function(k, v){
               temp[k] = [ "NONE" ];//modalを開いた時の初期値は「未設定」。「選択する」ボタンを押す前はnull。
           });
           cvss['vectors'] = temp;
       }
       $('#selected-cvss-vectors').val(JSON.stringify(cvss['vectors']));
       var target = $('#cvss-selection-area');
       target.find('[data-cvss]').removeClass('selected-cvss-item');
       var is_unselected_item = false;
       $.each(cvss['vectors'], function(k, values){
           $.each(values, function(){
               target.find('[data-cvss="' + k + '"][data-value="' + this + '"]').addClass('selected-cvss-item'); 
           });
       });
       $('#cvss-vectors-flg-perfect-select').val(cvss.flg_perfect);
       $('#cvss-vectors-flg-ignore-undef').val(cvss.flg_ignore_undef);
       if( cvss.flg_perfect === 'true'){
           $('#cvss-vectors-flg-ignore-undef').prop('disabled', false);
       }else{
           $('#cvss-vectors-flg-ignore-undef').prop('disabled', true);
       }
       $('#modalCvssVectors').modal('show');
   });

   function sort_cvss_values(item_name, values){
       //UIの値順にする
       var cvss_item_values = JSON.parse($('#cvss-vectors-data').val());
       var target_values = cvss_item_values[item_name]['values'];
       var target_value_keys = Object.keys(target_values);
       values.sort(function(a, b){
           if( target_value_keys.indexOf(a) > target_value_keys.indexOf(b) ){
               return 1;
           }else{
               return -1;
           }
       });
   };

   function create_cvss_vector_flg_perfect_elm(flg_perfect){
       var cvss_vectors_opts = JSON.parse($('#cvss-vectors-opts').val());
       var elm = $('<div>').attr('class', 'cvss-vectors-opts-item').append(
                     $('<div>').attr('class', 'd-flex').append(
                         $('<div>').attr('class', 'font-weight-bold').text(cvss_vectors_opts['flg_perfect']['label'] + ':'),
                         $('<div>').attr('class', 'font-weight-normal text-center ml-1 w-100').text(cvss_vectors_opts['flg_perfect']['options'][flg_perfect])
                     )
                 );
       return elm;   
   };

   function create_cvss_vector_flg_ignore_undef_elm(flg_ignore_undef){
       var cvss_vectors_opts = JSON.parse($('#cvss-vectors-opts').val());
       var elm = $('<div>').attr('class', 'cvss-vectors-opts-item').append(
                     $('<div>').attr('class', 'd-flex').append(
                         $('<div>').attr('class', 'font-weight-bold').text(cvss_vectors_opts['flg_ignore_undef']['label'] + ':'),
                         $('<div>').attr('class', 'font-weight-normal text-center ml-1 w-100').text(
                             cvss_vectors_opts['flg_ignore_undef']['options'][flg_ignore_undef]
                         )
                     )
                 );
       return elm;
   };

   //CVSS指標を選択するモーダルで「決定」ボタンを押した時
   $('#execute-select-cvss-vectors').on('click', function(){
       var cvss = JSON.parse($('#cvss-vectors').val());
       cvss['vectors'] = JSON.parse($('#selected-cvss-vectors').val());
       cvss['flg_perfect'] = $('#cvss-vectors-flg-perfect-select').val();
       cvss['flg_ignore_undef'] = $('#cvss-vectors-flg-ignore-undef').val();

       $('#cvss-vectors').val(JSON.stringify(cvss));
       var cvss_item_values = JSON.parse($('#cvss-vectors-data').val());
       if( $('#modalEditCondition').is(':visible') ){
           var target = $('#edit-cond-after');
       }else{
           var target = $('#cond-detail-body');
       }
       var selection_area = $('#cvss-selection-area');
       var create_cvss_value_elm = function(item, value, value_name){
           return $('<div>').attr({
               'class': 'selected-cvss-item cvss-v-item cvss-v-item-' + item.toLowerCase(),
               'data-value': value
           }).html(value_name);
       };
       $.each(cvss['vectors'], function(item, values){
           sort_cvss_values(item, values);
           var target_values = cvss_item_values[item]['values'];
           target_item_elm = target.find('[data-cvss="' + item + '"]');
           target_item_elm.attr('class', 'selected-cvss-values-list')
                          .html('');
           if( item === 'AV'){
               target.find('.cvss-v-item-name-av').attr('style', '');
           }
           if( item === 'AV' && ((values.length > 2 && values.includes('N') ) || values.length > 3)) {
               var num = Math.round(values.length / 2); 
               var elm = $('<div>').attr('class', 'd-flex');
               var elm2 = $('<div>').attr('class', 'd-flex');
               $.each(values, function(idx, v){
                   var value_name = target_values[v];
                   var value_elm = create_cvss_value_elm(item, v, value_name);
                   if( idx + 1 <= num ){
                       elm.append(value_elm);
                   }else{
                       elm2.append(value_elm);
                   }
               });
               target.find('.cvss-v-item-name-av').attr('style', 'padding:1.17rem !important;');//cssの!importantを上書き
               target_item_elm.append(elm, elm2);
           }else{
               var selected_values = $.map(values, function(v, idx){
                   var value_name = target_values[v];
                   return create_cvss_value_elm(item, v, value_name);
               });
               target_item_elm.addClass('d-flex').append(selected_values);
           }
       });

       $('#cvss-vectors-opts-area').html('').attr('class', 'mt-2').append(
           create_cvss_vector_flg_perfect_elm($('#cvss-vectors-flg-perfect-select option:selected').val())    
       );
       if( cvss.flg_perfect === 'true' ){
           $('#cvss-vectors-opts-area').append(
               create_cvss_vector_flg_ignore_undef_elm($('#cvss-vectors-flg-ignore-undef option:selected').val()).addClass('mt-2')
           );
       }
       $('#modalCvssVectors').modal('hide');
   });

   $('#cvss-vectors-flg-perfect-select').on('change', function(){
       if( $(this).val() === 'true' ){
           $('#cvss-vectors-flg-ignore-undef').prop('disabled', false);
       }else{
           $('#cvss-vectors-flg-ignore-undef').prop('disabled', true);
       }
   });

   //CVSSの値を小数第２位までに制限する
  /* $(document).on('input', '.cvss-value', function(){
       var input = $(this);
       var value = parseFloat(input.val());
       if(isNaN(value)){
           alert('半角で入力して下さい');
           input.val('');
       }else if(value > 10 || value < 0){
           input.val('');
       }else{
           value = value * 10;
           input.val(Math.round(value) / 10);
       }
   });*/

   //プロダクト名取得
   function get_products_by_ids({ prod_ids, ver_ids, lprod_ids, lver_ids }){
       return new Promise((resolve, reject) => {
            $.ajax({
                type: 'POST',
                url: $('#sidfm_internal_api_get_products_and_versions_by_ids_path').val(),
                data:{
                    'product_ids': JSON.stringify(prod_ids),
                    'version_ids': JSON.stringify(ver_ids),
                    'lproduct_ids': JSON.stringify(lprod_ids),
                    'lversion_ids': JSON.stringify(lver_ids),
                },
            }).done(function(data){
                var names = [];
                data.products.forEach(prd => {
                    names.push(escape_html_special_chars(prd['name']));
                });
                data.versions.forEach(ver => {
                    names.push(escape_html_special_chars(ver['fullname']));
                });
                data.lproducts.forEach(lprd => {
                    names.push(escape_html_special_chars(lprd['name']));
                });
                data.lversions.forEach(lver => {
                    names.push(escape_html_special_chars(lver['fullname']));
                });
                resolve(names);
            });
       });
   }

    //新規条件modalのプロダクト検索
    $(document).on('click', '#cond-detail-search-products', function(){
        var path = $('#sidfm_internal_api_get_products_and_versions_by_ids_path').val();
        var param = $('#set-cond-param').val();
        $('#temp-param-of-search-products').val(param);
        param = JSON.parse(param);
        if($(this).closest('#edit-cond-after').length === 0){
            var item_name = $('#cond-detail-body').find('[data-addcond]')
                                                  .attr('data-addcond');
        }else{
            var item_name = $('#edit-cond-after').find('[data-cond]')
                                                 .attr('data-cond');
        }
        var title_item = cond_detail_items_name[item_name];
        $('#modalSetProducts .modal-title').text('プロダクト検索: ' + title_item);
        if(item_name === 'product-match'){
            $('#another-side-products-label').text(cond_detail_items_name['product-unmatch'] + '条件で、すでに設定済みのプロダクト');
            //$('#modalSetProducts .modal-title').text('プロダクト検索： プロダクト「適合」');
            var selected_products = param['product-unmatch'];
            var selected_item_name = 'product-unmatch'
        }else{
            $('#another-side-products-label').text(cond_detail_items_name['product-match'] + '条件で、すでに設定済みのプロダクト');
            //$('#modalSetProducts .modal-title').text('プロダクト検索： プロダクト「不適合」');
            var selected_products = param['product-match'];
            var selected_item_name = 'product-match'
        }
        if(selected_products){
            const createProductElm = (prod, isLprod) => {
                return '<a class="btn btn-xs mb-1 mr-1 selected-version"' +
                       ' href="#"' +
                       ' data-productid="' + prod['id'] + '"' +
                       ' data-selectedname="' + selected_item_name + '"' +
                       ' data-item="another-side-product"' + (isLprod ? ' data-lproduct="true"' : '') + '>' +
                       '<span>' +
                       escape_html_special_chars(prod['name']) +
                       '</span>' +
                       '</a>';
            };
            const createVersionElm = (ver, isLver) => {
                  return '<a class="btn btn-xs mb-1 mr-1 selected-version"' +
                         ' href="#"' +
                         ' data-versionid="' + ver['id'] + '"' +
                         ' data-selectedname="' + selected_item_name + '"' +
                         ' data-item="another-side-product"' + (isLver ? ' data-lproduct="true"' : '') + '>' +
                         '<span>' +
                         escape_html_special_chars(ver['fullname']) +
                         '</span>' +
                         '</a>';
            };
            $.post(
                 path,
                 { 
                     'product_ids': JSON.stringify(selected_products['product-ids']), 
                     'version_ids': JSON.stringify(selected_products['version-ids']),
                     'lproduct_ids': JSON.stringify(selected_products['lproduct-ids']),
                     'lversion_ids': JSON.stringify(selected_products['lversion-ids']) 
                },
                function(data){
                    var append_products = '';
                    var append_versions = '';
                    if(data['products'].length){
                        data['products'].forEach(prod => append_products += createProductElm(prod, false));
                    }
                    if(data['lproducts'].length){
                        data['lproducts'].forEach(lprod => append_products += createProductElm(lprod, true));
                    }
                    if(append_products.length){
                        $('#another-side-products').html(append_products);
                    }else{
                        $('#another-side-products').html('<div class="font-weight-bold">なし</div>');
                    }    
                    if(data['versions'].length){
                        data['versions'].forEach(ver => append_versions += createVersionElm(ver, false));
                    }
                    if(data['lversions'].length){
                        data['lversions'].forEach(lver => append_versions += createVersionElm(lver, true));
                    }
                    if(append_versions.length){
                         $('#another-side-versions').html(append_versions);
                    }else{
                        $('#another-side-versions').html('<div class="font-weight-bold">なし</div>');
                    }
                    $('#modalSetProducts').modal('show');
            });               
        }else{
            $('#another-side-products, #another-side-versions').html(
                                           '<div class="font-weight-normal">' +
                                           'なし' +
                                           '</div>'
                                       );
            $('#modalSetProducts').modal('show');
        } 
    });

    //プロダクト限定・除外で、もう一方の側で選択されているプロダクトを表示
    $('#show-another-side-products').on('click', function(){
        var button = $(this);
        if(button.hasClass('showing')){
            $('#another-side-products-versions-area').hide();
            button.text('表示').removeClass('showing');
        }else{
            $('#another-side-products-versions-area').show(); 
            button.text('非表示').addClass('showing');
        }
   
    });

    //プロダクト検索modalが消える時にツールチップ消す
    $('#modalSetProducts').on('hide.bs.modal', function(){
        $('.original-tooltip').remove();
    });
    
    //プロダクト検索でsupport終了製品を示すメッセージの初期化・有効化
    function toggle_support_end_products_statement(){
        if($('#m_searched_list').find('[data-supportend]').length > 0 ){
            $('#support-end-products-statement').removeClass('invisible').addClass('visible');
        }else{
            $('#support-end-products-statement').removeClass('visible').addClass('invisible'); 
        }
    }

    //プロダクト検索でローカルプロダクトを示すメッセージの初期化・有効化
    function toggle_lproducts_statement(){
        if($('#m_searched_list').find('[data-lproduct="true"]').length > 0 ){
            $('#search-lproducts-statement').show();
        }else{
            $('#search-lproducts-statement').hide();
        }
    } 

    function create_searched_product_elm(product, versions, is_lproduct){
        var prod_name = escape_html_special_chars( product['name'] );
        var prod_id = product['id'];
        var p_item = '<div class="dropdown" data-elm="product-info"' +
            (product['is_support_end'] ? ' data-supportend="true"' : '') +
        '>';
        p_item += '<button class="btn btn-xs btn-outline-dark mb-1 mr-1 dropdown-toggle' + (is_lproduct ? ' text-primary' : '') + '"' +
            ' type="button"' +
            ' data-toggle="dropdown"' +
            ' aria-haspopup="true"' +
            ' aria-expanded="false">' +
            (product['is_support_end'] ? '<span class="text-dark">&times;</span><span class="text-danger">' + prod_name + '</span>' : prod_name) +
        '</button>';
        var p_menu = '<div class="dropdown-menu scrollable-menu-30vh s-shadow-v1 show-versions-list">';
        p_menu += '<a class="dropdown-item' + (product['is_support_end'] ? ' text-danger' : (is_lproduct ? ' text-primary' : '')) + '"' +
             ' data-productid="' + prod_id + '"' +
             ' data-item="version"' +
             ' data-all="true"' +
             (is_lproduct ? ' data-lproduct="true"' : '') +
             ' data-productname="' + prod_name + '">' +
             (product['is_support_end'] ? '<span class="text-dark">&times;</span><span class="text-danger">全て</span>' : '全て') +
             '</a>';
        if (versions) {
            versions.sort(function(a, b){
                var a_is_support_end = a['data']['is_support_end'] ? -1 : 1;
                var b_is_support_end = b['data']['is_support_end'] ? -1 : 1;

                if( a_is_support_end === b_is_support_end ){
                    if( a['data']['priority'] > b['data']['priority'] ){
                        return 1;
                    }else{
                        return -1;
                    }
                }else if(a_is_support_end > b_is_support_end){
                    return -1;
                }else{
                    return 1;
                }
            });
            for (var j = 0; j < versions.length; j++) {
                var version = versions[j];
                var version_data = version['data'];
                version_data['fullname'] = escape_html_special_chars( version_data['fullname'] );
                var related_hostnames = version['related-host'];
                var p_ver = '<a class="dropdown-item d-flex' + (is_lproduct ? ' text-primary' : '') + '"' +
                    'href="#"' +
                    ' data-versionid="' + version_data['id'] + '"' +
                    ' data-productid="' + prod_id + '"' +
                    ' data-item="version"' +
                    (is_lproduct ? ' data-lproduct="true"' : '') +
                    ((product['is_support_end'] || version_data['is_support_end']) ? ' data-supportend="true"' : '') +
                '>';
                if(!related_hostnames){
                    p_ver += '<span class="mr-1" data-elm="version-name">' +
              　    ((product['is_support_end'] || version_data['is_support_end']) ? '<span class="text-dark">&times;</span><span class="text-danger version-text">' + version_data['fullname'] + '</span>' : version_data['fullname']) +
                    '</span>' +
                    '</a>';
                }else{
                    p_ver += '<span class="mr-1" data-elm="version-name">' +
                    (version_data['is_support_end'] ? '<span>&times;</span><span class="text-danger version-text">' + version_data['fullname'] + '</span>' : version_data['fullname']) +
                    '</span>' +
                    '<span data-originaltooltip="true"' +
                    ' data-content="' + related_hostnames.join(',<br>') + '"' +
                    ' class="btn related-hosts-btn btn-info ml-auto">' +
                    '関連ホスト</span>' +
                    '</a>';
                }
                p_menu += p_ver;
            }
        }
        p_menu += '</div>';
        p_item += p_menu + '</div>';
        return p_item;
    } 

    //プロダクト検索
    $('#m_inc_search').on('input', function() {
        var inc_search_val = $('#m_inc_search').val();
        display_selectable_products(inc_search_val);
    });

    var jqxhr = null;
    function display_selectable_products(keyword){
        $('.original-tooltip').remove();
        var target = $('#m_inc_search');
        if(jqxhr){
            jqxhr.abort();
        }
        if(keyword.length < 2){
            $('#filter-searched-products').addClass('disabled');
            if(keyword.length != 0){
                var original_tooltip = make_original_tooltip('２文字以上入力して下さい', true).attr('id', 'product-search-temp-tooltip');
                $('body').append(original_tooltip);
                var offset = target.offset();
                var offset_left = offset.left + target.outerWidth() / 2 - original_tooltip.outerWidth() / 2;
                var offset_top = offset.top - original_tooltip.outerHeight() - 8;
                original_tooltip.offset({'top': offset_top, 'left': offset_left});
                setTimeout(function(){
                    $('#product-search-temp-tooltip').show();
                },300);
            }
        }else{
            $('#filter-searched-products').removeClass('disabled');
        }
        var path = $('#sidfm_internal_api_get_products_with_hostname_path').val();
        jqxhr = $.post(path, { 'keyword': keyword }, function(data) {
            const products = data.products;
            const lproducts = data.lproducts;
            var pids = Object.keys(products);
            pids.sort(function(a, b){
                var a_is_support_end = products[a]['product']['is_support_end'] ? -1 : 1;
                var b_is_support_end = products[b]['product']['is_support_end'] ? -1 : 1;
                if(a_is_support_end > b_is_support_end){
                     return -1;
                }else{
                     return 1;
                }
            });
            var append_elms = '';
            for (var i = 0; i < pids.length; i++) {
                append_elms += create_searched_product_elm(products[pids[i]]['product'], products[pids[i]]['versions'], false);
            }
            var lpids = Object.keys(lproducts);
            lpids.sort(function(a, b){
                var a_is_support_end = lproducts[a]['lproduct']['is_support_end'] ? -1 : 1;
                var b_is_support_end = lproducts[b]['lproduct']['is_support_end'] ? -1 : 1;
                if( a_is_support_end === b_is_support_end ){
                    if( lproducts[a]['lproduct']['priority'] > lproducts[b]['lproduct']['priority'] ){
                        return 1;
                    }else{
                        return -1;
                    }
                }else if(a_is_support_end > b_is_support_end){
                     return -1;
                }else{
                     return 1;
                }
            });
            for (var i = 0; i < lpids.length; i++) {
                append_elms += create_searched_product_elm(lproducts[lpids[i]]['lproduct'], lproducts[lpids[i]]['versions'], true);
            }
            $('#m_searched_list').html(append_elms);
            toggle_support_end_products_statement();
            toggle_lproducts_statement();
            if(append_elms === '' && keyword.length > 1){
                var original_tooltip = make_original_tooltip('該当するプロダクトはありません', true);
                $('body').append(original_tooltip);
                var offset = target.offset();
                var offset_left = offset.left + target.outerWidth() / 2 - original_tooltip.outerWidth() / 2;
                var offset_top = offset.top - original_tooltip.outerHeight() - 6;
                original_tooltip.offset({'top': offset_top, 'left': offset_left}).show();
            }
       });
    }

    $(document).on('mouseover', '#m_searched_list .related-hosts-btn', function(){
        var button = $(this);
        if(button.length === 1){
            var original_tooltip = make_original_tooltip(button.attr('data-content'), true).addClass('product-search-tooltip');
            $('body').append(original_tooltip);
            var offset = button.offset();
            var offset_left = offset.left + (button.outerWidth() / 2) - (original_tooltip.outerWidth() / 2);
            var offset_top = offset.top - original_tooltip.outerHeight() - 6;
            original_tooltip.offset({'top': offset_top, 'left': offset_left}).show();
        }

        var offset = $(this).offset();
        var offset_top = offset.top
        var bottom = $(this).outerHeight(true) + offset_top;
        var left = offset.left
        var right = left + $(this).outerWidth();

        var set_time = setInterval(function(){
            if(cursor_point_y < offset_top || cursor_point_y > bottom || cursor_point_x < left || cursor_point_x > right 
                || $('#m_searched_list .related-hosts-btn:visible').length === 0){
                $('.product-search-tooltip').remove();
                clearInterval(set_time);
            }
        },100);
    });

    var cursor_point_y;
    var cursor_point_x;
    $('#modalSetProducts').on('mousemove', function(e){
        cursor_point_y = e.pageY;
        cursor_point_x = e.pageX;
    });

    $(document).on('mousemove', '.product-search-tooltip', function(e){
        cursor_point_y = e.pageY;
        cursor_point_x = e.pageX;
    });

    //version名クリックでプロダクトを追加
    $(document).on('click', '.show-versions-list [data-item="version"]', function(){
        var button = $(this);
        var version_id = button.attr('data-versionid');
        var product_id = button.attr('data-productid');
        var is_all_selected = button.attr('data-all');
        var is_lproduct = button.attr('data-lproduct') === 'true';
        if($('#m_selected_list').find('[data-versionid="' + version_id + '"]' + (is_lproduct ? '[data-lproduct="true"]' : '')).length > 0){
            alert('既に選択済みのプロダクトです');
            return false;
        }else if($('#m_selected_list').find('[data-productid="' + product_id + '"]' + (is_lproduct ? '[data-lproduct="true"]' : '')).length > 0){
            alert('既に「全て」が選択済みのプロダクトです');
            return false; 
        }
        var pass = true;
        $('#another-side-products-versions-area').find('[data-item="another-side-product"]')
            .each(function(){
                var elm = $(this);
                var another_side_product_id = elm.attr('data-productid');
                var another_side_version_id = elm.attr('data-versionid');
                var is_lp = elm.attr('data-lproduct') === 'true';
                var name = elm.attr('data-selectedname');
                if(product_id && product_id === another_side_product_id && is_lproduct === is_lp){
                    name = cond_detail_items_name[name];
                    alert('既に「' + name + '」で「全て」が選択済みのプロダクトです');
                    pass = false;
                    return false;
                }else if(version_id && version_id === another_side_version_id && is_lproduct === is_lp){
                    name = cond_detail_items_name[name];
                    alert('既に「' + name + '」で選択済みのプロダクトです');
                    pass = false;
                    return false;
                }
        });
        if(!pass){
            return false;
        }
        if(is_all_selected){
            var name = button.attr('data-productname');
            var selected_item = $('<a>').attr({
                                        'class': 'btn btn-xs mb-1 mr-1 selected-version',
                                        'href': '#',
                                        'data-item': 'selected-version',
                                        'data-productid': product_id,
                                        'data-text': name,
                                }).html(escape_html_special_chars(name));
            if( is_lproduct ){
                selected_item.attr('data-lproduct', 'true');
            }
        }else{
            var version_name_elm = button.find('[data-elm="version-name"]');
            if(version_name_elm.find('.version-text').length > 0){
                var name = version_name_elm.find('.version-text').text();
            }else{
                var name = version_name_elm.text();
            }
            var selected_item = $('<a>').attr({
                                        'class': 'btn btn-xs mb-1 mr-1 selected-version',
                                        'href': '#',
                                        'data-item': 'selected-version',
                                        'data-versionid': version_id,
                                        'data-text': name,
                                }).html(escape_html_special_chars(name));
             if( is_lproduct ){
                 selected_item.attr('data-lproduct', 'true');
             }
        }
        var icon = $('<i>').attr({ 
                           'class': 'fas fa-minus-circle ml-1 delete-version-mark',
                           'data-item': 'delete-version'
                           });
        selected_item.append(icon);
        $('#m_selected_list').append(selected_item);
        $('#execute-add-products').prop('disabled', false);
        $('.original-tooltip').remove();
        toggle_support_end_products_statement();
        toggle_lproducts_statement();
    });

    //version削除
    $(document).on('click', '#m_selected_list [data-item="delete-version"]', function(){
        $(this).closest('[data-item="selected-version"]').remove();
        if($('#m_selected_list').find('[data-item="selected-version"]').length === 0){
            $('#execute-add-products').prop('disabled', true);
        }
    });

    //削除マークhoverで色変える
    $(document).on('mouseenter', '.delete-version-mark', function(){
        var mark = $(this);
        mark.addClass('delete-version-mark-hover');
    });

    $(document).on('mouseleave', '.delete-version-mark', function(){
        var mark = $(this);
        mark.removeClass('delete-version-mark-hover');
    });

    //プロダクトを条件設定modalに追加
    $('#execute-add-products').on('click', function(){
        $('#cond-detail-products-area > *').remove();
        var cnt = 0;
        $('#m_selected_list').find('[data-item="selected-version"]')
        .each(function(){
            var elm = $(this);
            var name = elm.text();
            var version_id = elm.attr('data-versionid');
            var product_id = elm.attr('data-productid');
            var is_lproduct = elm.attr('data-lproduct') === 'true';
            if(product_id){
                var add_elm = $('<div>').attr({
                    'class': 'text-truncate cond-detail-add-products ' + (cnt % 2 === 0 ? 'user-name-bg' : ''),
                    'data-item': 'version',
                    'data-productid': product_id,
                }).text(name);
            }else{
                var add_elm = $('<div>').attr({
                    'class': 'text-truncate cond-detail-add-products ' + (cnt % 2 === 0 ? 'user-name-bg' : ''),
                    'data-item': 'version',
                    'data-versionid': version_id,
                }).text(name);
            }
            if( is_lproduct ){
                add_elm.attr('data-lproduct', 'true');
            }
            $('#cond-detail-products-area').append(add_elm);
            cnt += 1;
        });
        $('#modalSetProducts').modal('hide');
    });

    //登録済みのプロダクトのみを表示
    $('#filter-searched-products').on('click', function(){
        var versions_lists = $('#m_searched_list').find('[data-elm="product-info"]');
        versions_lists.each(function(){
            var elm = $(this);
            var related_hosts = elm.find('[data-originaltooltip]');
            if(related_hosts.length === 0){
                elm.remove();
            }else{
                elm.find('[data-item="version"]')
                   .each(function(){
                       var item = $(this);
                       if(item.find('[data-originaltooltip]').length === 0){
                           item.remove();
                       }
                });
                if(elm.find('[data-item="version"]').length === 0){
                    elm.remove();
                }
            }  
        });
        //bootstrapのtooltipが表示されないので独自のtooltipを使用
        var tooltip_text = '検索結果:' + $('#m_searched_list').find('[data-originaltooltip]')
                                                              .length + '件';
        var original_tooltip = make_original_tooltip(tooltip_text, true);
        $('body').append(original_tooltip);
        var button = $('#filter-searched-products');
        var offset = button.offset();
        var offset_left = offset.left + (button.outerWidth() / 2) - (original_tooltip.outerWidth() / 2);
        var offset_top = offset.top - original_tooltip.outerHeight() - 6;
        original_tooltip.offset({'top': offset_top, 'left': offset_left}).show();
    });
        

    //プロダクト追加modal初期化
    function reset_SetProducts_modal(){
        $('#m_selected_list > *').remove();
        $('#m_searched_list > *').remove();
        $('#m_inc_search').val('');
        $('#m_inc_search_old').val('');
        $('#execute-add-products').prop('disabled',true);
        $('#another-side-products-versions-area').hide();
        $('#show-another-side-products').text('表示').removeClass('showing');
        $('#filter-searched-products').addClass('disabled').tooltip('hide');
        toggle_support_end_products_statement();
        toggle_lproducts_statement();
    }

    //カスタムタグの「カスタムタグがなければ成立」を変えた時
    $(document).on('change', '#match-if-not-ctag', function(){
        const ctagElms = $('#cond-detail-ctags-false-area, #cond-detail-ctags-true-area')
        if( $(this).prop('checked') ){
            ctagElms.addClass('disabled');
        }else{
            ctagElms.removeClass('disabled');
        }
    });

    //カスタムタグ検索modal
    const ruleCtagInfo = { inited: false, fetching : false };
    const ruleSelectedCtags = $('#rule-selected-ctags');
    const ruleSelectedRegistCtags = $('#rule-selected-regist-ctags');
    const MATCH_IF_NO_CTAG_LABEL = 'カスタムタグが1つもなければ成立';
    function selectCtagEnableOption(optionElm){ //is_enableがtrueの選択肢を押下するとき
        const selectedCtags = JSON.parse(ruleSelectedCtags.val());
        const ctagData = optionElm.data('ctagdata');
        const isInSearchedList = optionElm.closest('#rule-ctags-searched-list').length;
        if( selectedCtags.enable.find(ctag => ctag.id == ctagData.id) ){
            alert('選択済みのカスタムタグです。');
            return;
        }
        if( isInSearchedList ){
            if( selectedCtags.not_enable.find(ctag => ctag.id == ctagData.id) ){
                alert('選択済みのカスタムタグです。');
                return;
            }
        }
        if( optionElm.closest('#rule-selected-not-enable-ctags-list').length ){
            selectedCtags.not_enable = selectedCtags.not_enable.filter(ctag => {
                return ctag.id != ctagData.id;
            });
        }
        selectedCtags.enable.push(ctagData);
        ruleSelectedCtags.val(JSON.stringify(selectedCtags));
        if( !isInSearchedList ){
            optionElm.closest('[data-ctagid="' + ctagData.id + '"]').remove();
        }
        const ctagElm = createCtagElm(ctagData, [ 'not_enable', 'delete' ]);
        $('#rule-selected-enable-ctags-list').append(ctagElm);
        if( !$('#rule-ctags-searched-list-area [data-ctagid]').length ){
            $('#rule-ctags-searched-list-area').hide();
        }
        ctagElm.find('[data-select="not_enable"]').on('click', function(){
            selectCtagNotEnableOption($(this));
        });
        ctagElm.find('[data-select="delete"]').on('click', function(){
            unselectCtag($(this));
        });
    }
    function selectCtagNotEnableOption(optionElm){ //is_enableがfalseの選択肢を押下するとき
        const selectedCtags = JSON.parse(ruleSelectedCtags.val());
        const ctagData = optionElm.data('ctagdata');
        const isInSearchedList = optionElm.closest('#rule-ctags-searched-list').length;
        if( selectedCtags.not_enable.find(ctag => ctag.id == ctagData.id) ){
            alert('選択済みのカスタムタグです。');
            return;
        }
        if( isInSearchedList ){
            if( selectedCtags.enable.find(ctag => ctag.id == ctagData.id) ){
                alert('選択済みのカスタムタグです。');
                return;
            }
        }
        if( optionElm.closest('#rule-selected-enable-ctags-list').length ){
            selectedCtags.enable = selectedCtags.enable.filter(ctag => {
                return ctag.id != ctagData.id;
            });
        }
        selectedCtags.not_enable.push(ctagData);
        ruleSelectedCtags.val(JSON.stringify(selectedCtags));
        if( !isInSearchedList ){
            optionElm.closest('[data-ctagid="' + ctagData.id + '"]').remove();
        }
        const ctagElm = createCtagElm(ctagData, [ 'enable', 'delete' ]);
        $('#rule-selected-not-enable-ctags-list').append(ctagElm);
        if( !$('#rule-ctags-searched-list-area [data-ctagid]').length ){
            $('#rule-ctags-searched-list-area').hide();
        }
        ctagElm.find('[data-select="enable"]').on('click', function(){
            selectCtagEnableOption($(this));
        });
        ctagElm.find('[data-select="delete"]').on('click', function(){
            unselectCtag($(this));
        });
    }
    function unselectCtag(optionElm){ //選択から削除
        const selectedCtags = JSON.parse(ruleSelectedCtags.val());
        const ctagData = optionElm.data('ctagdata');
        if( optionElm.closest('#rule-selected-enable-ctags-list').length ){
            selectedCtags.enable = selectedCtags.enable.filter(ctag => {
                return ctag.id != ctagData.id;
            });
        }else{
            selectedCtags.not_enable = selectedCtags.not_enable.filter(ctag => {
                return ctag.id != ctagData.id;
            });
        }
        ruleSelectedCtags.val(JSON.stringify(selectedCtags));
        optionElm.closest('[data-ctagid="' + ctagData.id + '"]').remove();
        /*const searchingText = $('#rule-search-ctags-input').val().toLowerCase();
        if( searchingText && ctagData.name.toLowerCase().indexOf(searchingText) !== -1 ){
            const ctagElm = createCtagElm(ctagData, [ 'enable', 'not_enable' ]);
            $('#rule-ctags-searched-list').append(ctagElm);
            $('#rule-ctags-searched-list-area').show();
            ctagElm.find('[data-select="enable"]').on('click', function(){
                selectCtagEnableOption($(this));
            });
            ctagElm.find('[data-select="not_enable"]').on('click', function(){
                selectCtagNotEnableOption($(this));
            });
        }*/
    }

    function createCtagElm(ctag, options){
        const isCond = $('#modalSearchCtagsTarget').val() === 'cond';
        const styleProps = [];
        styleProps.push('background-color: ' + (ctag.bgcolor ? ctag.bgcolor : '#fff') + ';');
        if( ctag.color ){
            styleProps.push('color: ' + ctag.color + ';');
        }

        const ctagData = JSON.stringify({
            id: ctag.id,
            name: ctag.name,
            color: ctag.color,
            bgcolor: ctag.bgcolor,
        });
        const optionElms = [];
        if( options.includes('enable') ){
            optionElms.push(
                $('<a>').attr({
                    href: '#',
                    class: 'dropdown-item',
                    'data-ctagdata': ctagData,
                    'data-select': 'enable'
                }).text(isCond ? 'あれば成立' : '設定する')
            );
        }
        if( options.includes('not_enable') ){
            optionElms.push(
                $('<a>').attr({
                    href: '#',
                    class: 'dropdown-item',
                    'data-select': 'not_enable',
                    'data-ctagdata': ctagData
                }).text(isCond ? 'なければ成立' : '除外する')
            );
        }
        if( options.includes('delete') ){
            optionElms.push(
                $('<a>').attr({
                    href: '#',
                    class: 'dropdown-item',
                    'data-select': 'delete',
                    'data-ctagdata': ctagData
                }).text('削除')
            );
        }
        const ctagElm = $('<div>').attr({
            class: 'dropdown d-inline-block',
            'data-ctagid': ctag.id
        }).append(
            $('<span>').attr({
                class: 'searched-ctag ml-2 mr-2 mb-1 mt-1',
                style: styleProps.join(' '),
                'data-toggle': 'dropdown',
                'aria-haspopup': 'true',
                'aria-expanded': 'false'
            }).text(ctag.name),
            $('<div>').attr({
                class: 'dropdown-menu'
            }).append(optionElms)
        );
        return ctagElm;
    };

    function createSelectedCtagElm(ctag){
        const styleProps = [];
        styleProps.push('background-color: ' + (ctag.bgcolor ? ctag.bgcolor : '#fff') + ';');
        styleProps.push('color: ' + (ctag.color ? ctag.color : '#212529') + ';'); //colorがない場合は、cssで指定されているデフォルトの文字色
        return $('<div>').attr({
            class: 'selected-cond-ctag',
            style: styleProps.join(' ')
        }).text(ctag.name);
    }

    function bindCtagModalEvents(){
        const searchInput = $('#rule-search-ctags-input');
        const searchedListArea = $('#rule-ctags-searched-list-area');
        const searchedList = $('#rule-ctags-searched-list');
        const inputTooltip = $('#rule-search-ctags-input-tooltip');
        const bindSearchedListEvent = () => {
            const target = searchedList.find('[data-toggle="dropdown"]')
            target.dropdown();
            target.parent().on('show.bs.dropdown', function(e){
                $(e.relatedTarget).addClass('searched-ctag-menu-visible');
            });
            target.parent().on('hide.bs.dropdown', function(e){
                $(e.relatedTarget).removeClass('searched-ctag-menu-visible');
            });
            searchedList.find('[data-select="enable"]').on('click', function(){
                selectCtagEnableOption($(this));
            });
            searchedList.find('[data-select="not_enable"]').on('click', function(){
                selectCtagNotEnableOption($(this));
            });
        }

        const searchPath = $('#api_search_ctags_path').val();

        searchInput.on('input', function(){ //カスタムタグ検索
            const keyword = $(this).val();
            const ruleCategoryId = $('#main-category-id').val();
            if( ruleCtagInfo.fetching ){
                ruleCtagInfo.fetching.abort();
            }
            const paramsData = { keyword };
            if( ruleCategoryId !== 'all' ){
                paramsData.only_category_ids = JSON.stringify([ ruleCategoryId ]);
            }
            ruleCtagInfo.fetching = $.post(
                searchPath,
                paramsData,
                function(ctags){
                    const ctagElms = ctags.map( ctag => createCtagElm(ctag, ['enable', 'not_enable']) );
                    searchedList.html('');
                    if( ctagElms.length ){
                        searchedListArea.show();
                        inputTooltip.tooltip('hide');
                        searchedList.append(ctagElms);
                        bindSearchedListEvent();
                    }else{
                        searchedListArea.hide();
                        inputTooltip.tooltip(keyword.length ? 'show' : 'hide');
                    }
                }
            );
        });

        //カスタムタグを決定
        $('#execute-select-ctags').on('click', function(){
            const isCond = $('#modalSearchCtagsTarget').val() === 'cond';
            const selectedCtags = JSON.parse(ruleSelectedCtags.val());
            ruleSelectedRegistCtags.val(ruleSelectedCtags.val()); 
            const enables = selectedCtags.enable.map(ctag => createSelectedCtagElm(ctag));
            const notEnables = selectedCtags.not_enable.map(ctag => createSelectedCtagElm(ctag));
            $(isCond ? '#cond-detail-ctags-true-area' : '#ope-detail-ctags-true-area').html('').append(enables);
            $(isCond ? '#cond-detail-ctags-false-area' : '#ope-detail-ctags-false-area').html('').append(notEnables);
            $('#modalSearchCtags').modal('hide');
        });
    };

    $(document).on('click', '#cond-detail-search-ctags, #ope-detail-search-ctags', function(){
        //const param = JSON.parse($('#set-cond-param').val());
        const selectedCtags = JSON.parse(ruleSelectedCtags.val());
        const triggerId = $(this).attr('id');
        const isCond = triggerId === 'cond-detail-search-ctags';
        const enableAreaTitle = $('#rule-selected-enable-ctags-list-area-title');
        const notEableAreaTitle = $('#rule-selected-not-enable-ctags-list-area-title');
        const targetName = $('#modalSearchCtagsTarget');
        const selectedEnableList = $('#rule-selected-enable-ctags-list');
        const selectedNotEnableList = $('#rule-selected-not-enable-ctags-list');
        const searchInput = $('#rule-search-ctags-input');
        const searchedListArea = $('#rule-ctags-searched-list-area');
        const searchedList = $('#rule-ctags-searched-list');
        searchInput.val('');
        searchedListArea.hide();
        if( !Object.keys(selectedCtags).length ){
            ruleSelectedCtags.val(JSON.stringify({enable: [], not_enable: []}));
            selectedEnableList.html('');
            selectedNotEnableList.html('');
            ruleSelectedRegistCtags.val(ruleSelectedCtags.val());
        }
        if( isCond ){
            targetName.val('cond');
            enableAreaTitle.text('あれば成立');
            notEableAreaTitle.text('なければ成立');
        }else{
            targetName.val('action');
            enableAreaTitle.text('設定する');
            notEableAreaTitle.text('除外する');
        }

        if( !ruleCtagInfo.inited ){
            bindCtagModalEvents();
            ruleCtagInfo.inited = true;
        }
        $('#modalSearchCtags').modal('show');
    });

    //新規条件modalのホスト検索
    $(document).on('click', '#cond-detail-search-host-btn', function(){
        if($('#host-add-area').length === 0){
            var param = JSON.parse($('#set-cond-param').val());
            var target_category_id = $('#main-category-id').val();
            var wrap_host_add_area = $('<div>').attr('class', 'mb-2');
            var host_add_area = $('<div>').attr({'id': 'host-add-area', 'data-targetcond': $(this).attr('data-targetcond')});
            var usage = $('<div>').attr('class', 'g-font-size-14')
                              .text('対象となるホストを検索し、追加して下さい。決定ボタンを押すとホスト名が対象ホスト欄に追加されます。');
            var table_div = $('<div>').attr({
                                      'class': 'table-responsive scrollable-menu-150 g-font-size-13 g-brd-around g-brd-1 g-brd-ccc',
                                      'id': 'wrap-host-list-table'
                                      });
            var table = $('<table>').attr({
                                    'class': 'table table-sm mb-0',
                                    'id': 'host-list-table',
                                    });
            var thead = $('<thead>').attr('class', 'thead-dark');
            var tr = $('<tr>').attr('class', '');
            var th = $('<th>').attr('class', 'th-sticky-top0 font-weight-normal g-text-shadow-black').text('ホスト一覧');
            table.append(
                   thead.append(tr.append(th))
                  );
            var tbody = $('<tbody>').attr({
                                    'id': 'host-list-tbody',
                                    'class': '',
                                });
            table.append(tbody);
            var tr = $('<tr>').attr({
                               'class': 'list_alternative2',
                               'id': 'ask-add-hosts'
                               }).append(
                                   $('<td>').attr('class', '').text('ホストを追加して下さい')
                               );
            tbody.append(tr);
            table_div.append(table);
            var users_categories = JSON.parse($('#users_selectable_categories').val());
            var system_categories = JSON.parse($('#system-categories').val());
            users_categories = Object.assign(users_categories, system_categories);
            var wrap = $('<div>').attr('class', 'text-right mt-2');
            var div = $('<div>').attr('class', 'd-flex');
            var category_name = $('<span>').attr('class', 'g-font-size-13 text-right w-75 ml-auto text-truncate')
                                       .text('ホストの検索範囲');
            var desc = $('<span>').attr({
                                  'id': 'cond-detail-ask-search-host',
                                  'data-toggle' : 'tooltip',
                                  'data-placement' : 'top',
                                  'data-title' : 'カテゴリ: ' + (target_category_id === 'all' ? '全体' : users_categories[target_category_id]) +
                                                 '<br>※「ルールを適用するカスタムカテゴリ」で選択されているカテゴリがホストの検索範囲となります',
                                  'data-html': 'true',
                                  }).append(
                                        $('<i>').attr('class', 'fas fa-info-circle c_gray_70 ml-1')
                                  );
            div.append(category_name, desc);
            wrap.append(div);
            host_add_area.append(usage, wrap, table_div);
            wrap_host_add_area.append(host_add_area);
            var wrap_div = $('<div>').attr('class', 'w-100 mb-2');
            var div = $('<div>').attr('class', 'mt-4');
            var input = $('<input>').attr({
                                    'type': 'text',
                                    'class': 'form-control form-control-sm mr-2',
                                    'placeholder': 'ホスト名を入力して下さい',
                                    'id': 'cond-detail-search-host',
                                    'data-toggle': 'tooltip',
                                    'data-trigger': 'manual',
                                    'data-categoryid': JSON.stringify([ target_category_id ]),
                                    'autocomplete': 'off'
                                    });
             //var wrap_btn = $('<div>').attr('class', 'float-right position-relative');
             div.append(input);   
             var wrap_select = $('<div>').attr({
                                         'class': 'hostcond-host-selection-area',
                                         'id': 'select-searched-host-area',
                                         }).hide();  
             var select_host = $('<select>').attr({
                                            'multiple': 'true',
                                            'class': 'form-control bottom-radius-0',
                                            'id': 'cond-detail-select-searched-host'
                                            });
             var notice = $('<div>').attr('class', 'hostcond-search-notice').append(
                              $('<small>').html(
                                  '※グレーで表示されるホスト名は、既に設定済のホストです。<br>グレーで表示されるホストがホスト一覧内に表示されていない場合は、' +
                                  ( $(this).data('targetcond') === 'host-match' ? '「ホスト除外」' : '「ホスト限定」' ) +
                                  'で選択されているホストです。'
                              )
                          );
             wrap_select.append(select_host, notice);
             div.append(wrap_select.append(select_host, notice)); 
             wrap_div.append(div);
             var keep_margin = $('<div>').attr('class', 'keep-margin');
             var target_modal = $('#modalSelectItems');
             target_modal.find('.modal-body > *').remove();
             target_modal.find('.modal-body')
                         .addClass('min-height-300px').append(wrap_host_add_area, keep_margin ,wrap_div);
             target_modal.find('.modal-title').text('ホスト検索');
             $('#cond-detail-ask-search-host').tooltip();
             $('#cond-detail-select-searched-host').tooltip();
             target_modal.modal('show');
             target_modal.find('[data-selectitem]').attr('id', 'decide-hosts');
        }else{
            $('#modalSelectItems').find('.modal-body')
                                  .addClass('min-height-300px opacity-01');
            $('#modalSelectItems').modal('show');
        }
    });

    $('#modalSelectItems').on('shown.bs.modal', function(){
        if($('#wrap-host-list-table').length === 1){
            var list_table = $('#wrap-host-list-table');
            list_table.find('[data-item="host"]').width(list_table.width() - 2);
            $(this).find('.modal-body').removeClass('opacity-01');
        }
    });

    $('#modalSelectItems').on('hidden.bs.modal', function(){
        if($(this).find('[data-selectitem]').attr('id') === 'set-main-category'){
            setTimeout(function(){
                show_change_rule_name_message();
            },500);
        }
    });

    //新規条件modalでのホスト名検索
    $(document).on('input', '#cond-detail-search-host', function(){
        $('[data-toggle="tooltip"]').tooltip('hide');
        var input = $(this);
        if(input.val().length === 1){
            input.attr('data-original-title', '2文字以上入力して下さい');
            $('#select-searched-host-area').hide();
        }else{
            input.attr('data-original-title', '検索候補がありません');
            $('#select-searched-host-area').hide();
        }
        var path = $('#sidfm_internal_api_get_hosts_by_category_ids_path').val();
        var category_ids = input.attr('data-categoryid');
        var hostname = input.val();
        var cond_param = JSON.parse($('#set-cond-param').val());
        if($('#host-add-area').attr('data-targetcond') === 'host-match'){
            var target_hostnames = cond_param['host-unmatch'];
            var target_cond_name = cond_detail_items_name['host-unmatch'];
        }else{
            var target_hostnames = cond_param['host-match'];
            var target_cond_name = cond_detail_items_name['host-match'];
        }
        if( target_hostnames ){
            var selected_hostnames = target_hostnames;
        }else{
            var selected_hostnames = [];
        }
        $('#host-list-tbody').find('[data-item="hostname"]').each(function(){
            selected_hostnames.push($(this).attr('data-hostname'))
        });

        $.post(path, {
                        'keyword': hostname,
                        'category_ids': category_ids
                     },
        function(data) {
            $('[data-toggle="tooltip"]').tooltip('hide');
            var elm = $('#cond-detail-select-searched-host');
            $('#cond-detail-select-searched-host > *').remove();
            if(data != null && data.length != 0){
                var options = '';
                for(var i=0; i < data.length; i++){
                    hostdata = data[i];
                    var ext_name = hostdata.ext_name;
                    var hostname = hostdata.hostname;
                    var is_selected = false;
                    if(selected_hostnames.includes(hostname)){
                        is_selected = true;
                    }
                    var regexp = new RegExp(input.val().toLowerCase().replace(/\\/g, '\\\\'));
                    var hitname = (ext_name && ext_name.toLowerCase().match(regexp)) ? ext_name : hostname; 
                    options += '<option class="cond-searched-hostname' + (is_selected ? ' hostcond-selected-hostname' : '') + '" data-hostname="' + hostname + '" data-toggle="tooltip" data-title="ホスト名: ' + hostname + '<br/>表示名: ' + (ext_name ? ext_name : '(未設定)') + '" data-placement="left" data-html="true" data-boundary="window">' +
                               hitname + 
                               '</option>';
                }
                elm.html(options).show();
                elm.find('[data-toggle="tooltip"]').tooltip();
                input.tooltip('hide');
                $('#select-searched-host-area').show();
            }else if($('#cond-detail-search-host').val().length === 0){
                input.tooltip('hide');
            }else{
                input.tooltip('show');
            }
        });
    });


    //ホスト名クリックでホスト名入力
    $(document).on('click', '.cond-searched-hostname', function(){
        if( $(this).hasClass('hostcond-selected-hostname') ){
           return false;
        }
        var hostname = $(this).data('hostname');
        var cond_param = JSON.parse($('#set-cond-param').val());
        var wrap_add_elm = $('<tr>').attr({ 
            'class': 'list_alternative2',
            'data-toggle': 'tooltip',
            'data-title': $(this).attr('data-title'),
            'data-placement': 'left',
            'data-boundary': 'window',
            'data-html': 'true',
            'data-item': 'host'
        });
        var add_elm = $('<td>').width($('#wrap-host-list-table').width());
        var wrap = $('<div>').attr('class', 'd-flex justify-content-between');
        var add_host = $('<div>').attr({
                                     'class': 'pt-1 ml-2 mr-1 text-left',
                                     'data-item': 'hostname',
                                     'data-hostname': hostname
                                 }).append(
                                     $('<span>').attr('class', 'text-truncate hostname-item')
                                                .text(hostname)
                                 );
        var del_btn = $('<div>').attr('class', 'mr-1')
                                .append(
                                    $('<a>').attr({
                                                'class': 'btn btn-danger btn-xs text-nowrap del-add-host g-font-size-12',
                                                'data-hostname': hostname
                                            })
                                            .text('削除')
                                );
        wrap.append(add_host, del_btn);
        add_elm.append(wrap);
        wrap_add_elm.append(add_elm);
        $('#ask-add-hosts').remove();
        $('#host-list-tbody').append(wrap_add_elm);
        $(this).addClass('hostcond-selected-hostname'); 
        wrap_add_elm.tooltip();
    });    

   //ホスト検索で削除ボタンを押した時
   $(document).on('click', '.del-add-host', function(){
       var hostname = $(this).data('hostname');
       $(this).closest('[data-item="host"]').tooltip('hide').remove();
       $('#cond-detail-select-searched-host option[data-hostname="' + hostname + '"]').removeClass('hostcond-selected-hostname');
       if($('#host-list-tbody').find('td').length === 0){
           var tr = $('<tr>').attr({
                                 'class': 'd-block pr-0 pl-0',
                                 'id': 'ask-add-hosts'
                             }).append(
                                 $('<td>').attr('class', 'd-block').text('ホストを追加して下さい')
                             );
           $('#host-list-tbody').append(tr);
       }
    });

    //適合・不適合ホストの決定
    $(document).on('click', '#decide-hosts', function(){
        $('#cond-detail-add-hosts > *').remove();
        var cnt = 0;
        var elm = '';
        $('#host-list-tbody').find('[data-item="hostname"]')
                             .each(function(){
            if(cnt % 2 === 1){
                var div = '<div class="user-name-bg cond-detail-added-hostname" data-hostname="'+ $(this).attr('data-hostname') + '">' + $(this).text() + '</div>';
            }else{
                var div = '<div class="cond-detail-added-hostname" data-hostname="' + $(this).attr('data-hostname') + '">' + $(this).text() + '</div>';
            }
            elm += div;
            cnt += 1;
        });
        $('#cond-detail-add-hosts').append($(elm));
        $('#modalSelectItems').modal('hide');
    });        
    //新規条件modalの入力確認
    function check_value_is_Int(value){
        var num = parseInt(value);
        if(value.match(/[^0-9]+/)){
            return false;
        }else if( num < 0 || isNaN(num)){
            return false;
        }
        return true;
    }
    function check_input_is_correct(input_element, message = null){
        var pass = true;
        var is_blank = false;
        var is_float = false;
        $(input_element).each(function(){
            var value = $(this).val();
            if(value == ''){
                is_blank = true;
            }
            if(value != ''){
                var check = check_value_is_Int(value);
                if(!check){
                    pass = false;
                }
            }
        });
        if(is_blank){
            alert(message ? message : '選択した条件詳細の項目に値を入力して下さい');
            return false;
        }
        if(!pass){
            alert('正数で入力して下さい');
            return false;
        }
        return true;
    }

    function check_cvss_input_is_correct(num){
        if(num === ''){
            alert('選択した条件詳細の項目に値を入力して下さい');
            return false;
        }
        var decimal = num - Math.floor(num);
        var num = parseFloat(num);
        if(isNaN(num) || isNaN(decimal)){
            alert('入力した値に設定できない文字が含まれています。※半角で入力して下さい');
            return false;
        }else if(num > 10.0 || num < 0.0){
            alert('入力したCVSSの値はCVSS値の範囲内でないため設定できません');
            return false;
        }
        return true;
    }

    function check_selected_category_is_one(){
        var selected_cat = $('.optional-cond').find('.detail-value:selected');
        if(selected_cat.length > 1){
            alert('カテゴリは１つのみ選択して下さい');
            return false;
        }else if(selected_cat.length === 0){
            alert('カテゴリを選択して下さい');
            return false;
        }
        return true;
    }

    //上書き確認
    function confirm_update_condition_setting(elm_data, button){
        var button_id = button.attr('id');
        var same_cond = $('#set-condition-tbody').find('[data-cond=' + elm_data + ']');
        if(same_cond.length === 1){
            var conf = button_id === 'execute-edit-condition' ? true : confirm('既に同じ条件が設定されています。上書きしますか？');
           if(conf){
               same_cond.remove();
           }else{
               return false;
           }
       }
       return true;     
    }

    var executed_button_id = null;
    const ctagCondInfo = { dataid: 0 }; //カスタムタグは複数の条件がある可能性があるので、データ/要素を特定するためのidが必要
    const timerangeCondInfo = { dataid: 0 }; //期間指定は複数の条件がある可能性があるので、データ/要素を特定するためのidが必要 
    //追加ボタン押下で条件設定modalに条件追加
    $('#execute-add-cond, #execute-edit-condition').on('click', function(){
       var param = JSON.parse($('#set-cond-param').val());
       var button = $(this);
       var button_id = button.attr('id');
       executed_button_id = button_id;
       if(button_id === 'execute-add-cond'){ 
           var selected_elm = $('#cond-detail').find('[data-addcond]:visible');
           var selected_elm_parent = selected_elm.closest('[data-cond]'); 
           var data = selected_elm_parent.attr('data-cond');
           var elm_data = selected_elm.attr('data-addcond');
       }else{
           var selected_elm = $('#edit-cond-after').find('[data-addcond]');
           var selected_elm_parent = selected_elm.closest('[data-cond]');
           var data = selected_elm_parent.attr('data-cond'); 
           var elm_data = selected_elm.attr('data-addcond');
       }
       var is_update = $('#modalRegisterCondition-table').find('[data-conddetail=' + elm_data + ']').length > 0;
       var item_name = cond_detail_items_name[elm_data]; 
       var pass = true;
       if(data === 'issue-update' || data === 'due-date-approach'){
           var input_elm = selected_elm_parent.find('.detail-value:visible');
           var check = check_input_is_correct(input_elm);
           if(check){
               var conf = confirm_update_condition_setting(elm_data, button);
               if(conf && elm_data === 'issue-update-regi'){
                   var v = input_elm.val();
                   param[elm_data] = v;
               }else if(conf && elm_data === 'issue-update-del'){
                   var v = input_elm.val();
                   param[elm_data] = v;
               }else if(conf && elm_data === 'issue-update-regi-and-del'){
                   var regi_val = $('#issue-update-double-cond-regi').val();
                   var del_val = $('#issue-update-double-cond-del').val();
                   param[elm_data] = {"regi": regi_val, "del": del_val};
               }else if(conf && elm_data === 'due-date-approach'){
                   var v = input_elm.val();
                   param[elm_data] = {'val': v};
               }else{
                   return false;
               }
           }else{
               return false;
           }
       } else if( data === 'history-create' ){
           var input_elm = $('#history-create-cond-history-num');
           if( check_input_is_correct(input_elm, '投稿件数を入力して下さい。') ){
               var flg_exclusion = null;
               if( $('#history-create-cond-select-only').prop('checked') ){
                   flg_exclusion = 'false';
               }else if( $('#history-create-cond-select-except').prop('checked') ){
                   flg_exclusion = 'true';
               }
               var user_ids = JSON.parse($('#history-create-cond-selected-user-ids').val());
               if( flg_exclusion && user_ids.length === 0 ){
                   alert('ユーザを指定して下さい。');
                   return false;
               }
               var conf = confirm_update_condition_setting(elm_data, button);
               if( conf ){
                   param[elm_data] = { 
                       'val': input_elm.val(),
                       'user_ids': (flg_exclusion ? user_ids : []),
                       'flg_exclusion': flg_exclusion
                   };
               }else{
                   return false;
               }
           }else{
               return false;
           }
       }else if(data === 'sri-above' || data === 'sri-below'){
           var conf = confirm_update_condition_setting(elm_data, button);
           if(conf){
               var selected_op = selected_elm_parent.find('.detail-value:selected');
               var text = selected_op.text();
               param[elm_data] = selected_op.val();
           }else{
               return false;
           }
       }else if(data === 'sri-warning' || data === 'cvss-warning' || data === 'cvss-notreg'){
           var conf = confirm_update_condition_setting(elm_data, button);
           var selected_op = selected_elm_parent.find('.detail-value:selected');
           if(conf){
               param[elm_data] = selected_op.val();
           }else{
               return false;
           }
       }else if(data === 'cvss-above' || data === 'cvss-below'){
           var v = selected_elm_parent.find('.detail-value').val();
           var check = check_cvss_input_is_correct(v);
           if(check){
               var conf = confirm_update_condition_setting(elm_data, button);
               if(conf){
                   var exploitable = $('#flg-exploitable option:selected');
                   var notreg = $('#if-include-notreg option:selected');
                   var decimal = ((v * 10) - (Math.floor(v) * 10)) / 10;
                   var p_num = ((v * 10) - (decimal * 10)) / 10;
                   var cvss_value = String(((p_num * 10) + (decimal * 10)) / 10);
                   var dot_place = cvss_value.indexOf('.');
                   if( dot_place != -1 ){
                       cvss_value = parseFloat(cvss_value.slice(0, dot_place + 2)); 
                   }
                   param[elm_data] = {
                       'val': parseFloat(cvss_value), 
                       'include_notreg': notreg.val(), 
                       'flg_exploitable': exploitable.val()
                   };
               }else{
                    return false;
               }
           }else{
               return false;
           }
       }else if( data === 'cvss-vector' ){
           var vectors = JSON.parse($('#cvss-vectors').val());
           if( Object.values(vectors['vectors']).includes(null) ){
               alert('「選択する」ボタンから各項目の値を選択して下さい。');
               return false;
           }
           var conf = confirm_update_condition_setting(elm_data, button);
           if( conf ){
               param[elm_data] = vectors;
           }else{
               return false;
           }
       }else if(data === 'due-date-expired' || data === 'due-date-undef'){
           var conf = confirm_update_condition_setting(elm_data, button);
           if(conf){
               param[elm_data] = elm_data;
           }
       }else if(data === 'category-match' || data === 'category-unmatch'){
           var check = $('#select-cond-cats').find('[data-checkbox]:checked').length > 0;
           if(check){
               var cats = $('#select-cond-cats [data-checkbox]:checked');
               var cat_ids = [];
               var cat_names = [];
               var same_cats = [];
               var check_target = (data === 'category-match' ? param['category-unmatch'] : param['category-match']);
               cats.each(function(){
                   var cat = $(this);
                   var cat_val = cat.val();
                   var cat_name = cat.attr('data-text');
                   cat_ids.push(cat_val);
                   cat_names.push(cat_name);
                   if(check_target && check_target.includes(cat_val)){
                       same_cats.push(cat_name);
                   }
               });
               if(same_cats.length > 0){
                   if(data === 'category-match'){
                       alert('カテゴリ除外の項目に" ' + same_cats.join(', ') + ' "が既に設定されています');
                   }else{
                       alert('カテゴリ限定の項目に" ' + same_cats.join(', ') + ' "が既に設定されています');
                   }
                   return false;
               }
               /*if((param['host-match'] || param['host-unmatch']) && data === 'category-match'){
                   if(param['host-match'] && param['host-unmatch']){
                       var a_text = '「ホスト限定」と「ホスト除外」';
                   }else if(param['host-match']){
                       var a_text = '「ホスト限定」';
                   }else{
                       var a_text = '「ホスト除外」';
                   } 
                   var conf = confirm('ホストの選択にカテゴリが関係しているため、カテゴリを変更すると' + a_text + 'の項目が削除されます。よろしいですか？');
               }
               if(conf === false){
                   return false;
               }*/
               var conf = confirm_update_condition_setting(elm_data, button);
               if(conf){
                   /*var v = cat_names.join(',');
                   if(data === 'category-match'){
                       delete param['host-match']
                       delete param['host-unmatch']
                       $('#set-condition-tbody').find('[data-cond="host-match"], [data-cond="host-unmatch"]').remove();
                   }*/
                   param[elm_data] = cat_ids;
               }else{
                   return false;
               }
           }else{
               alert('カテゴリを選択して下さい');
               return false;
           }
       }else if(data === 'host-match' || data === 'host-unmatch'){
          var selected_hostnames = $('#cond-detail-add-hosts').find('[data-hostname]');
          if(selected_hostnames.length === 0){
               alert('対象ホストがありません');
               return false;
          }else{ 
               var conf = confirm_update_condition_setting(elm_data, button);
               if(conf){
                   var hostnames = [];
                   selected_hostnames.each(function(){
                       hostnames.push($(this).attr('data-hostname'));
                   });
                   param[elm_data] = hostnames;
               }else{
                   return false;
               }
           }
       }else if(data === 'host-syncing'){
           var is_syncing = $('#select-host-syncing').val();
           var conf = confirm_update_condition_setting(elm_data, button);
           if(conf){
                param[elm_data] = is_syncing;
           }else{
               return false;
           }
       }else if(data === 'status-match'){
           var selected_statuses = $('#cond-detail-select-statuses').find('.selected-status');
           if(selected_statuses.length === 0){
               alert('ステータスを選択して下さい');
               return false;
           }
           if(selected_statuses.length < 3){
               var conf = confirm_update_condition_setting(elm_data, button);
               if(conf){
                   var status_ids = [];
                   selected_statuses.each(function(){
                       var elm = $(this);
                       status_ids.push(elm.attr('data-statusid'));
                   }); 
                   param[elm_data] = status_ids;
              }else{
                  return false;
              }
           }else{
               alert('選択可能なステータスは２つまでです');
               return false;
           }
       }else if(data === 'product-match' || data === 'product-unmatch'){
           var selected_products = $('#cond-detail-products-area').find('[data-item="version"]');
           if(selected_products.length === 0){
               alert('対象プロダクトがありません');
               return false;
           }
           var conf = confirm_update_condition_setting(elm_data, button);
           if(conf){
               var product_ids = [];
               var version_ids = [];
               var lproduct_ids = [];
               var lversion_ids = [];
               selected_products.each(function(){
                   var elm = $(this);
                   var prod_id = elm.attr('data-productid');
                   var ver_id = elm.attr('data-versionid');
                   if( elm.attr('data-lproduct') === 'true' ){
                        if(prod_id){
                            lproduct_ids.push(prod_id);
                        }else{
                            lversion_ids.push(ver_id);
                        }
                   }else{
                       if(prod_id){
                           product_ids.push(prod_id);
                       }else{
                           version_ids.push(ver_id);
                       }
                   }
               });
               param[elm_data] = {
                   'product-ids': product_ids,
                   'version-ids': version_ids,
                   'lproduct-ids': lproduct_ids,
                   'lversion-ids': lversion_ids,
               };
           }else{
               return false;
           }
       }else if(data === 'cvss-exploitable'){
           var is_exploitable = $('#cvss-exploitable-is-exploitable option:checked').val();
           var is_change = $('#cvss-exploitable-is-updated').prop('checked') ? 'true' : 'false';
           var conf = confirm_update_condition_setting(elm_data, button);
           if(conf){
               param[elm_data] = {'value': is_exploitable, 'is_change': is_change};
           }else{
               return false;
           }
       }else if(data === 'ctag'){
           var selected_ctags = JSON.parse(ruleSelectedRegistCtags.val());
           var match_if_no_ctag = $('#match-if-not-ctag').prop('checked');
           if( !match_if_no_ctag && (!Object.keys(selected_ctags).length || (!selected_ctags.enable.length && !selected_ctags.not_enable.length)) ){
               alert('カスタムタグを選択するか、「' + MATCH_IF_NO_CTAG_LABEL + '」にチェックを入れてください。');
               return;
           }
           if( match_if_no_ctag ){
               selected_ctags.enable = [];
               selected_ctags.not_enable = [];
           }
           if( !param[elm_data] ) param[elm_data] = [];
           if( button_id === 'execute-edit-condition' ){
               //編集の時は、対象のデータを削除して新しいものを追加
               const edit_dataid = $('#edit-cond-before-data').val();
               param[elm_data] = param[elm_data].filter(ctagData => {
                   return ctagData.dataid != edit_dataid
               });
               $('#set-condition-tbody [data-conddetail="ctag"][data-editdata="' + edit_dataid + '"]').closest('tr[data-cond="ctag"]').remove();
           }
           ctagCondInfo.dataid += 1;
           selected_ctags.dataid = ctagCondInfo.dataid;
           param[elm_data].push(selected_ctags);
       }else if(data === 'timerange'){
           var selected_value = $('#cond-timerange-select-value').val().split('-');
           if( !param[elm_data] ) param[elm_data] = [];
           if( button_id === 'execute-edit-condition' ){
               //編集の時は、対象のデータを削除して新しいものを追加
               const edit_dataid = $('#edit-cond-before-data').val();
               param[elm_data] = param[elm_data].filter(timerange_cond => {
                   return timerange_cond.dataid != edit_dataid
               });
               $('#set-condition-tbody [data-conddetail="timerange"][data-editdata="' + edit_dataid + '"]').closest('tr[data-cond="timerange"]').remove();
           }
           timerangeCondInfo.dataid += 1;
           param[elm_data].push({
               'unit': selected_value[1],
               'value': selected_value[0],
               'except':  $('#cond-timerange-flg-except').prop('checked'),
               'dataid': timerangeCondInfo.dataid
           });
       } 
       update_modalRegisterCondition_items(elm_data, param, is_update);
       $('#set-cond-param').val(JSON.stringify(param));
       toggle_execute_set_btn();
       reset_add_condition_area();
       if(button_id === 'execute-edit-condition'){
           $('#modalEditCondition').modal('hide');
       }
    });

    //条件設定modalに項目を追加もしくは項目をアップデートするfunction
    function update_modalRegisterCondition_items(data_name, parameter, is_update){
        sri_values = { "1" : "低", "2" : "中", "3" : "高", "4" : "重大" };
        var cond_data = parameter[data_name];
        var item_name = cond_detail_items_name[data_name];
        var front_text = '';
        var back_text ='';
        var highlight_text = '';
        var info = null;
        if(data_name === 'issue-update-regi' || data_name === 'issue-update-del'){
            front_text = (data_name === 'issue-update-regi' ? '脆弱性の新規登録が' : '脆弱性影響無への変更が');
            back_text =  '件以上';
            var num = cond_data
            highlight_text = String(num);
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
        }else if(data_name === 'issue-update-regi-and-del'){
            var regi_num = cond_data['regi'];
            var del_num = cond_data['del'];
            front_text = regi_num;
            back_text = del_num;
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
        }else if(data_name === 'due-date-approach'){
            var days = cond_data['val'];
            front_text = '対策予定日が';
            back_text = '日以内';
            highlight_text = String(days);
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
        }else if(data_name === 'history-create'){
            front_text = '履歴の投稿が合計';
            back_text = '件以上';
            var history_num = cond_data['val'];
            highlight_text = String(history_num);
            if( cond_data['flg_exclusion'] ){
                $.post($('#get_users_path').val(), { 'user_ids': JSON.stringify(cond_data['user_ids']), 'is_sidfmadm_selectable': true }, function(res){
                    front_text = '履歴の投稿が合計';
                    back_text = '件以上';
                    highlight_text = String(history_num);
                    var user_names = res.map(function(user){
                        return (user['lastname'] + user['firstname'])
                    }).join(', ');
                    info = { 'flg_exclusion': cond_data['flg_exclusion'], 'user_names': escape_html_special_chars(user_names) };
                    append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info); 
                });
            }else{
                append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, {});
            }
        }else if(data_name === 'sri-above' || data_name === 'sri-below'){
            var sri_name = sri_values[cond_data];
            front_text = 'SRI値が';
            back_text = (data_name === 'sri-above' ? '以上' : '以下');
            highlight_text = sri_name;
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
        }else if(data_name === 'sri-warning' || data_name === 'cvss-warning'){
            highlight_text = ((cond_data === true || cond_data === 'true') ? '要対応条件に一致' : '要対応条件に不一致');
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
        }else if(data_name === 'cvss-notreg'){
            front_text = 'CVSS値が'
            highlight_text = ((cond_data === true || cond_data === 'true') ? '未定義' : '未定義以外');
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
        }else if(data_name === 'cvss-above' || data_name  === 'cvss-below'){
            front_text = 'CVSS値が';
            back_text = (data_name === 'cvss-above' ? '以上' : '以下');
            highlight_text = String(cond_data['val']);
            var exploitable = cond_data['flg_exploitable']
            if(exploitable === 'all'){
                exploitable = '全ての脆弱性が対象';
            }else if(exploitable === 'not_exist'){
                exploitable = '無しの脆弱性が対象';
            }else{
                exploitable = '有りの脆弱性が対象';
            }
            var inc_notreg = (cond_data['include_notreg'] === true || cond_data['include_notreg'] === 'true') ? '未定義の脆弱性を含める' : '未定義の脆弱性を含めない';
            info = '攻撃コード: ' + exploitable + '<br>' +
                   '未定義を含めるかどうか: ' + inc_notreg;
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
        }else if(data_name === 'cvss-vector'){
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, cond_data);
        }else if(data_name === 'due-date-expired'){
            highlight_text = '対策予定日を超過';
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
        }else if(data_name === 'due-date-undef'){
            highlight_text = '対策予定日が未設定';
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
        }else if(data_name === 'category-match' || data_name === 'category-unmatch'){
            var users_categories = JSON.parse($('#users_selectable_categories').val());
            var system_categories = JSON.parse($('#system-categories').val());
            var categories = Object.assign(users_categories, system_categories)
            var category_descs = JSON.parse($('#category-descs').val());
            var cat_names = [];
            for(var i = 0; i < cond_data.length; i++){
                var cat_id = cond_data[i];
                var desc_cat = category_descs[cat_id];
                var cat_elm = "<span data-toggle='tooltip' data-title='" + escape_html_special_chars(desc_cat ? desc_cat.desc : '') + "' data-placement='top' data-boundary='window'>" + escape_html_special_chars(categories[cat_id]) + "</span>";
                cat_names.push(cat_elm);
            }
            highlight_text = cat_names.join(',');
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
        }else if(data_name === 'host-match' || data_name === 'host-unmatch'){
            var hostnames = cond_data;
            var get_display_name_path = $('#sidfm_internal_api_get_hosts_display_name_path').val();
            $.ajax({
                type: 'POST',
                url: get_display_name_path,
                data: {'hostnames': JSON.stringify(hostnames)},
            }).done(function(hostsdata){
                hostnames = hostnames.map(function(h_name){ 
                    var host_elm = "<span data-toggle='tooltip' data-html='true' data-placement='right' data-boundary='window' data-title='" + 'ホスト名: ' + h_name + '<br/>表示名: ' + (hostsdata[h_name] ? hostsdata[h_name] : '(未設定)') + '\'>' + (hostsdata[h_name] ? hostsdata[h_name] : h_name) + '</span>';
                    return host_elm;
                });    
                highlight_text = hostnames.join(', ');
                append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
            });
        }else if(data_name === 'host-syncing'){
            back_text = 'のホスト';
            highlight_text = cond_data === 'true' ? '同期中' : '非同期';
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
        }else if(data_name === 'status-match' || data_name === 'status-unmatch'){
            var status_names = [];
            for(var i = 0; i < cond_data.length; i++){
                var status_id = cond_data[i];
                status_names.push(all_statuses[status_id]);
            }
            front_text = '対象ステータス:&nbsp';
            highlight_text = status_names.join(',');
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
        }else if(data_name === 'product-match' || data_name === 'product-unmatch'){
            const create_cond_elms = async () => {
                const names = await get_products_by_ids({
                    prod_ids: cond_data['product-ids'],
                    ver_ids: cond_data['version-ids'],
                    lprod_ids: cond_data['lproduct-ids'],
                    lver_ids: cond_data['lversion-ids']
                });
                highlight_text = names.join(',<br>');
                append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
            }
            create_cond_elms();
        }else if(data_name === 'cvss-exploitable'){
            front_text = '攻撃コードが';
            highlight_text = cond_data['value'] === 'true' ? '有り' : '無し';
            info = cond_data['is_change'] === 'true' ? '攻撃コードの更新がある脆弱性のみが対象' : null;
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info);
        }else if(data_name === 'ctag'){
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, cond_data); 
        }else if(data_name === 'timerange'){
            append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, cond_data);
        }
    }
  
    function append_cond_elms(is_update, item_name, data_name, front_text, back_text, highlight_text, info){
        const condDataInfo = {}; //カスタムタグ等の、複数あるデータを識別するためのIDを格納するために使う 
        if(data_name === 'issue-update-regi-and-del'){
           var div_name = $('<div>').attr('class', 'set-cond-name g-line-height-1_3')
                                    .text(item_name);
           var div_cond = $('<div>').attr({
                                    'data-addcond': data_name, 'class': 'g-line-height-1_3',
                                    })
                                    .html(
                                        '脆弱性の新規登録が' +
                                        '<span class="font-weight-bold edit-target">' +
                                        front_text +
                                        '</span>' +
                                        '件以上' +
                                        'かつ、 脆弱性影響無への変更が' +
                                        '<span class="font-weight-bold edit-target">' +
                                        back_text +
                                        '</span>' +
                                        '件以上'
                                    );
        }else if(data_name === 'host-match' || data_name === 'host-unmatch' || data_name === 'category-match' || data_name === 'category-unmatch' || data_name === 'product-match' || data_name === 'product-unmatch'){
            var btn_name = '';
            var btn_class = '';
            if(data_name === 'host-match' || data_name === 'host-unmatch'){
                btn_name = '対象ホスト';
                btn_class = 'btn-purple';
            }else if(data_name === 'category-match' || data_name === 'category-unmatch'){
                btn_name = '対象カテゴリ';
                btn_class = 'btn-orange';
            }else if(data_name  === 'product-match' || data_name === 'product-unmatch'){
                btn_name = '対象プロダクト';
                btn_class = 'btn-bluegray';
            }
            var div_name = $('<div>').attr('class', 'set-cond-name')
                                     .text(item_name);
            var inner_elm = $('<a>').attr({
                'class': 'btn btn-xs show-content-btn show-content text-white text-nowrap show-original-popover ' + btn_class,
                'data-item': data_name,
                'data-select': data_name,
                'data-content': highlight_text,
                'data-html': 'true'
            }).html(btn_name);
            var div_cond = $('<div>').attr({
                               'data-addcond': data_name,
                           })
                           .append(inner_elm);
        }else if(data_name === 'cvss-vector'){
            var div_name = $('<div>').attr('class', 'set-cond-name').text(item_name);
            var div_cond = $('<div>').attr({'data-addcond': data_name});
            var elm_html = ''
            //var cnt = 0;
            var obj_size = Object.keys(info).length;
            var cvss = JSON.parse($('#cvss-vectors-data').val());
            var create_cvss_value_elm = function(target_cvss, item, value){
                return $('<div>').attr({
                           'class': 'cvss-v-item selected-cvss-item text-center cvss-v-item-' + item.toLowerCase(),
                           'data-cvss': item,
                           'data-value': value
                        }).html(target_cvss['values'][value]);
            };
            $.each(info['vectors'], function(k, values){
                //cnt += 1;
                var target_cvss = cvss[k];
                var wrap = $('<div>').attr('class', 'd-flex mt-1 mr-1 ml-1');
                var cvss_item_name = $('<div>').attr('class', 'cvss-v-item-name text-center').html(target_cvss['item_name']);
                var selected_items = $('<div>').attr('class', 'selected-cvss-values-list');
                sort_cvss_values(k, values);
                if( k === 'AV' && ((values.length > 2 && values.includes('N') ) || values.length > 3)) {
                    cvss_item_name.attr('style', 'padding-top: 1.35rem;');
                    var elm = $('<div>').attr('class', 'd-flex');
                    var elm2 = $('<div>').attr('class', 'd-flex');
                    $.each(values, function(idx, value){
                        var selected_item = create_cvss_value_elm(target_cvss, k, value);
                        if( idx + 1 <= 2 ){
                            elm.append(selected_item);
                        }else{
                            elm2.append(selected_item);
                        }
                        selected_items.append(elm, elm2);
                    });
                }else{
                    selected_items.addClass('d-flex');
                    $.each(values, function(){
                        selected_items.append(create_cvss_value_elm(target_cvss, k, this));
                    });
                }
                wrap.append(cvss_item_name, selected_items);
                elm_html += $('<div>').append(wrap).html();//html文字列化
            });
            if( info['flg_perfect'] === 'true' ){
                var flg_perfect_elm = create_cvss_vector_flg_perfect_elm(info['flg_perfect']).addClass('mt-1 ml-1 mr-1');
                elm_html += $('<div>').append(flg_perfect_elm).html();//html文字列化
                var flg_ignore_undef_elm = create_cvss_vector_flg_ignore_undef_elm(info['flg_ignore_undef']).addClass('mb-1 mt-1 ml-1 mr-1'); 
                elm_html += $('<div>').append(flg_ignore_undef_elm).html();//html文字列化
            }else{
                var flg_perfect_elm = create_cvss_vector_flg_perfect_elm(info['flg_perfect']).addClass('mb-1 mt-1 ml-1 mr-1');
                elm_html += $('<div>').append(flg_perfect_elm).html();//html文字列化
            }
            var cvss_button = $('<button>').attr({
                'type': 'button',
                'id': 'show-cvss-selected-vectors',
                'class': 'btn btn-info btn-sm',
                'data-vectors': elm_html
            }).html('CVSS指標を表示<i id="show-cvss-selected-vectors-caret" class="fas fa-caret-down ml-2"></i>');
            div_cond.append(cvss_button);
        }else if(data_name === 'ctag'){
            const latestData = info[info.length - 1];
            condDataInfo.dataid = latestData.dataid;
            const empty_message = '選択されたカスタムタグはありません';
            var div_name = $('<div>').attr('class', 'set-cond-name').text(item_name);
            var div_cond = $('<div>').attr({'data-addcond': data_name});
            if( latestData.enable.length || latestData.not_enable.length ){
                const enables = latestData.enable.map(ctag => {
                    return createSelectedCtagElm(ctag);
                });
                const not_enables = latestData.not_enable.map(ctag => {
                    return createSelectedCtagElm(ctag);
                });
                var true_btn = $('<a>').attr({
                    'class': 'mr-2 btn btn-xs show-content-btn show-content text-white text-nowrap show-original-popover btn-success',
                    'data-item': data_name + '-true-' + condDataInfo.dataid,
                    'data-select': data_name + '-true-' + condDataInfo.dataid,
                    'data-content': (enables.length ? $('<div>').append(enables).html() : empty_message), //要素を文字列化
                    'data-html': 'true'
                }).text('あれば成立');
                var false_btn = $('<a>').attr({
                    'class': 'btn btn-xs show-content-btn show-content text-white text-nowrap show-original-popover btn-light-orange',
                    'data-item': data_name + '-false-' + condDataInfo.dataid,
                    'data-select': data_name + '-false-' + condDataInfo.dataid,
                    'data-content': (not_enables.length ? $('<div>').append(not_enables).html() : empty_message), //要素を文字列化
                    'data-html': 'true'
                }).text('なければ成立');
                div_cond.append(true_btn, false_btn);
            }else{
                div_cond.append(
                    $('<div>').text(MATCH_IF_NO_CTAG_LABEL)
                )
            }
        }else{
            highlight_text = highlight_text.replace(/\s+/g, '');
            front_text = front_text.replace(/\s+/g, '');
            back_text = back_text.replace(/\s+/g, '');
            var div_name = $('<div>').attr('class', 'set-cond-name')
                                     .text(item_name);
            if(!info){
                var div_cond = $('<div>').attr({ 'data-addcond': data_name })
                                         .html(front_text + '<span class="font-weight-bold edit-target">' + highlight_text + '</span>' + back_text);
            }else if(data_name === 'timerange'){
                const latest_data = info[info.length - 1];
                var timerange_text = create_timerange_text(latest_data); //末尾のデータが対象
                condDataInfo.dataid = latest_data.dataid;
                var div_cond = $('<div>').attr({ 'data-addcond': data_name })
                                         .html(
                                             '単位:<span class="font-weight-bold edit-target">' + timerange_text['unit'] + '</span>' +
                                             '&nbsp;&nbsp;間隔:<span class="font-weight-bold edit-target">' + timerange_text['value'] + '</span>' +
                                             '<span class="ml-2">' + timerange_text['except'] + '</span>'
                                         );
            }else if(data_name === 'history-create'){
                var div_cond = $('<div>').attr({ 'data-addcond': data_name })
                                         .append(
                                             $('<div>').append(
                                                 front_text,
                                                 $('<span>').attr('class', 'font-weight-bold edit-target').text(highlight_text),
                                                 back_text
                                             )
                                         );
                if( info['flg_exclusion'] ){
                    div_cond.find('div').addClass('mt-1 mb-1');
                    var flg_exclusion_elm = $('<span>').attr({
                        'class': 'font-weight-bold',
                        'style': 'font-size: 96%;'
                    }).text(info['flg_exclusion'] === 'true' ? 'ユーザが作成した履歴以外が対象:' : 'ユーザが作成した履歴のみが対象:');

                    div_cond.append(
                        $('<div>').append(
                            flg_exclusion_elm,
                            $('<a>').attr({
                                'class': 'ml-1 btn btn-xs btn-default show-content-btn show-content text-nowrap show-original-popover',
                                'data-item': data_name,
                                'data-select': data_name,
                                'data-content': escape_html_special_chars(info['user_names']),
                                'data-html': 'true'
                            }).text('対象ユーザ')
                        )
                    );
                }
            }else{
                var div_cond = $('<div>').attr({'data-addcond': data_name })
                                         .html(
                                             front_text +
                                             '<span class="font-weight-bold edit-target">' +
                                             highlight_text +
                                             '</span>' +
                                             back_text +
                                             '<i class="fas fa-info-circle ml-1 c_gray_70" data-detailinfo="true" data-toggle="tooltip" data-placement="right" data-html="true" data-title="' + info + '">'
                                          );
            }
        }
        var div_btns = $('<div>');
        var edit_btn = $('<button>').attr({
                                   'class': 'btn btn-xs btn-edit cond-edit-btn mr-lg-1 mr-0',
                                   'type': 'button',
                                   'data-conddetail': data_name,
                                   }).text('変更');
       if(uneditable_cond_details.includes(data_name)){
           edit_btn.addClass('disabled');
           edit_btn.attr({
                    'data-toggle': 'tooltip',
                    'data-title': '変更可能な項目が無いため、変更はできません',
                    'data-placement': 'left',
                    }).tooltip();
       }
       var del_btn = $('<button>').attr({
                                  'class': 'btn btn-xs btn-danger cond-del-btn',
                                  'type': 'button',
                                  'data-conddetail': data_name,
                                  }).text('削除');
       if(data_name === 'ctag'){ //カスタムタグの条件を削除するときは、複数ある可能性があるので対象のデータを編集/削除する
           edit_btn.attr('data-editdata', condDataInfo.dataid);
           del_btn.attr('data-deldata', condDataInfo.dataid);
       }else if(data_name === 'timerange'){
           edit_btn.attr('data-editdata', condDataInfo.dataid);
           del_btn.attr('data-deldata', condDataInfo.dataid);
       }
       div_btns.append([edit_btn, del_btn]);            
       var tr = $('<tr>').attr({
                         'class': 'list_alternative',
                         'data-cond': data_name,
                         });
       var td1 = $('<td>').attr('class', 'set-condition-t-left g-font-size-12');
       td1.append(div_name);
       tr.append(td1);
       var td2 = $('<td>').attr({
                          'class': 'set-condition-t-mid g-font-size-12',
                          });
       td2.append(div_cond);
       tr.append(td2);
       var td3 = $('<td>').attr({
                          'class': 'set-condition-t-right',
                          });
       td3.append(div_btns);
       tr.append(td3);
       if($('#show-no-cond').length === 1){
           $('#show-no-cond').remove();
       }
       $('#set-condition-tbody').append(tr);
       $('#execute-set-cond').prop('disabled', false);
       $('#set-condition-tbody').find('[data-detailinfo]').tooltip();
       if(is_update != 'prepare_edit'){
           if( executed_button_id === 'execute-edit-condition' ){
               append_original_tooltip('cond', is_update, true);
           }else{
               append_original_tooltip('cond', is_update);
           }
       }
       executed_button_id = null;
       var tbody = $('#modalRegisterCondition-table').get(0);
       tbody.scrollTop = tbody.scrollHeight;
       $('#modalSelectItems').find('.modal-body *').remove();
    }

    //条件追加modalと新規操作modalの条件項目の選択でスクロール
    $('#scroll-to-cond-item, #scroll-to-ope-item').on('change', function(){
        var select = $(this);
        var select_id = select.attr('id');
        select.attr('id', 'selecting-' + select_id);
        var area = $('#wrap-condition-item-btns:visible, #wrap-operation-item-btns:visible');
        var selected_val = select.find('option:selected')
                                  .data('target');
       if(area.attr('id') === 'wrap-condition-item-btns'){
            var target = area.find('[data-condid=' + selected_val + ']');
        }else if(area.attr('id') === 'wrap-operation-item-btns'){
            var target = area.find('[data-opeid=' + selected_val + ']'); 
        }
        var target_pos = target.position();
        var scroll_top = area.scrollTop();
        var scroll_place = target_pos.top + scroll_top;
        area.animate({
            scrollTop: scroll_place
        },500);
        setTimeout(function(){
            select.attr('id', select_id);
        },500);
    });

    //CVSS指標を表示するボタンを押した時
    $(document).on('click', '#show-cvss-selected-vectors', function(e){
        if( $('#selected-cvss-display').length === 0 ){
            $('#original-tooltip').remove();
            var button = $(this);
            var vectors = button.attr('data-vectors');
            var div = $('<div>').attr('id', 'selected-cvss-display').html(vectors);
            var target_offset = button.offset();
            $('body').append(div.hide());
            div.offset({
                'top': target_offset.top + button.outerHeight(),
                'left': target_offset.left
                //'top': target_offset.top,
                //'left': target_offset.left + button.outerWidth()
            }).show();
            return false;
        }else{
            $('#selected-cvss-display').remove();
        }
    });

    //条件追加modalの設定ボタンの押下可能・不可能の切り替え
    function toggle_execute_set_btn(){
        var set_condition_param =  JSON.parse($('#set-cond-param').val());
        var set_operation_param = JSON.parse($('#set-ope-param').val());
        if(Object.keys(set_condition_param).length === 0){
            $('#execute-set-cond').prop('disabled', true);
        }else{
            $('#execute-set-cond').prop('disabled',false);
        }
        if(Object.keys(set_operation_param).length === 0){
            $('#execute-set-ope').prop('disabled', true);
        }else{
            $('#execute-set-ope').prop('disabled', false);
        }
    }

    //条件追加modalと操作追加modalで項目ボタンを押した時、選択したボタンの色を変化
    function color_selecting_item_on_AddModal(button){
        var class_name = button.attr('class');
        var target_class = '';
        if(class_name.indexOf('select-cond') != -1){
            target_class = 'select-cond';
        }else{
            target_class = 'select-ope';
        }
        $('.modal').find('.' + target_class)
                   .removeClass('bg-darkgray')
                   .addClass('btn-light');
        button.addClass('bg-darkgray')
              .removeClass('btn-light');
    }
   
    //条件追加modalと新規操作modalの条件項目のスクロールで選択を切り替え
    $('#wrap-condition-item-btns, #wrap-operation-item-btns').on('scroll',function(){
        var elm = $(this);
        var elm_id = elm.attr('id');
        if(elm_id === 'wrap-condition-item-btns'){
            var items = $(this).find('[data-condid]');
            var data_name = 'condid';
        }else if(elm_id === 'wrap-operation-item-btns'){
            var items = $(this).find('[data-opeid]');
            var data_name = 'opeid';
        }
        items.each(function(){
            var elm = $(this);
            var elm_posi = elm.position();
            var elm_search_zone = elm.outerHeight(true) / 2;
            if(elm_posi.top < elm_search_zone && elm_posi.top > -elm_search_zone){
                var id = elm.data(data_name);
                $('#scroll-to-cond-item:visible, #scroll-to-ope-item:visible')
                    .find('[data-target=' + id + ']')
                    .prop('selected',true);
             }
         });
        
    });


    //条件設定modalで変更ボタンを押した時のアクション
    $(document).on('click', '.cond-edit-btn', function(){
        if($(this).hasClass('disabled')){
            return false;
        }
        $('#edit-cond-before > *').remove();
        $('#edit-cond-after > *').remove();
        $('#modalSelectItems').find('.modal-body > *').remove();
        var cond_detail = $(this).attr('data-conddetail');
        var selected_tr = $(this).closest('tr');
        var selected_cond  = selected_tr.find('[data-addcond]')
                                        .clone()
                                        .attr('class','cond-tr bg-light before-cond-edit-content');
        if( cond_detail != 'history-create' ){
            if(selected_tr.find('.show-content').length > 0){
                if( cond_detail === 'ctag' ){
                    let true_content;
                    let false_content;
                    selected_tr.find('.show-content').each((idx, elm) => {
                        elm = $(elm);
                        if( elm.attr('data-item').indexOf('true') !== -1 ){
                            true_content = elm.attr('data-content');
                        }else{
                            false_content = elm.attr('data-content');
                        }
                    });
                    selected_cond.html('');
                    selected_cond.append(
                        $('<div>').append(
                            $('<div>').attr('class', 'font-weight-bold').text('あれば成立'),
                            $('<div>').attr('class', 'edit-before-ctag-content').html(true_content),
                        ),
                        $('<div>').attr('class', 'edit-content-divider'),
                        $('<div>').append(
                            $('<div>').attr('class', 'font-weight-bold').text('なければ成立'),
                            $('<div>').attr('class', 'edit-before-ctag-content').html(false_content),
                        )
                    );
                }else{
                    selected_cond.html(
                        selected_tr.find('.show-content')
                                   .attr('data-content')
                    );
                }
            }
        }
        var selected_cond_item_name = selected_tr.find('.set-cond-name').text();
        $('#modalEditCondition').find('.modal-title').text('条件変更: ' + selected_cond_item_name); 
        $('#edit-cond-before').append(selected_cond);
        selected_cond.find('[data-toggle="tooltip"]').tooltip();
        if(cond_detail.indexOf('issue-update') != -1){
            var cond_detail_to_get_elm = cond_detail;
            cond_detail = 'issue-update';
        }
        var items = make_cond_detail_item(cond_detail);
        var wrap_items = $('<div>').attr({
                                   'class': 'bg-light after-cond-edit-content',
                                   'data-cond': cond_detail
                                    });
        if(cond_detail.indexOf('issue-update') != -1){
            var item = items.find('[data-addcond="' +  cond_detail_to_get_elm + '"]');
            wrap_items.append(item.show());
        }else{
            wrap_items.append(items);
        }
        if( cond_detail === 'history-create' ){
            wrap_items.find('.history-create-cond-checkbox-select-only-area, .history-create-cond-checkbox-select-except-area').removeClass('mt-2');
        }
        if( cond_detail === 'ctag' || cond_detail === 'timerange' ){ //複数登録可能な条件は、データを特定するための要素を追加
            $('#edit-cond-before').append(
                $('<option>').attr({
                    type: 'hidden',
                    id: 'edit-cond-before-data'
                }).val($(this).data('editdata'))
            )
        }
        $('#edit-cond-after').append(wrap_items);
        $('#cond-detail-ask-search-host').tooltip();
        $('#edit-cond-after [data-toggle="tooltip"]').tooltip();
        $('#edit-cond-after [data-toggle="dropdown"]').dropdown();
        $('#modalEditCondition').modal('show');
    }); 

    //条件設定modalで削除ボタンを押した時のアクション
    $(document).on('click', '.cond-del-btn', function(){
        var button = $(this);
        var del_value = button.data('conddetail');
        var param = JSON.parse($('#set-cond-param').val());
        button.closest('tr').remove();
        if( del_value === 'ctag' || del_value === 'timerange' ){
            const delid = button.data('deldata');
            param[del_value] = param[del_value].filter(c_data => {
                return c_data.dataid != delid;
            });
            if( !param[del_value].length ) delete param[del_value];
        }else{
            delete param[del_value];
        }
        $('#set-cond-param').val(JSON.stringify(param));
        display_no_condition_message();
        toggle_execute_set_btn();
    });

    function reset_add_condition_area(){
        $('#cond-detail-body > *').remove();
        $('#selecting-cond-item').text('- - -');
        $('#wrap-condition-item-btns').find('.bg-darkgray')
                                      .addClass('btn-light')
                                      .removeClass('bg-darkgray');
        $('#execute-add-cond').hide();
    }

    function reset_add_operation_area(){
        $('#ope-detail').find('.ope-ask-check').hide();
        $('#ope-detail-body > *').remove();
        //$('#execute-add-ope').prop('disabled', true);
        $('#execute-add-ope').hide();
        $('#selecting-ope-item').text('- - -');
        $('#wrap-operation-item-btns').find('.bg-darkgray')
                                      .addClass('btn-light')
                                      .removeClass('bg-darkgray');
    }
    //条件変更modalが開くとき
    $('#modalEditCondition').on('show.bs.modal',function(){
        reset_add_condition_area();
    });

    //条件変更moldalが閉じたら要素削除
    $('#modalEditCondition').on('hidden.bs.modal',function(e){
        $('#edit-cond-before > *').remove();
        $('#edit-cond-after > *').remove();
        $('#modalSelectItems').find('.modal-body > *').remove();
    });
      
               
    //タイミング設定画面のラジオボタン切り替え
    $('.set-rule-trigger').on('click', function(){ 
        check_blank_in_trigger_setting_modal();
    });


    $('#search-hostname').on('mouseleave', function(){ 
        $(this).tooltip('hide');
    });

    //ホスト名検索
    $('#search-hostname').on('input', function(){
        $('[data-toggle="tooltip"]').tooltip('hide');
        if(parseInt($('#trigger-category').attr('data-categoryid')) === -1){
            $(this).attr('data-original-title', 'カテゴリを選択して下さい')
                   .tooltip('show');
            return false;
        }else{
            $(this).tooltip('hide');
        }
        var input = $(this);
        if(input.val().length === 1){
            input.attr('data-original-title', '2文字以上入力して下さい');
        }else{
            input.attr('data-original-title', '検索候補がありません');
        }
        var path = $('#sidfm_internal_api_get_hosts_by_category_ids_path').val();
        $.post(path, {
                        'keyword': $(this).val(), 
                        'category_ids': JSON.stringify([$('#trigger-category').attr('data-categoryid')])
                     }, 
        function(data) {
            var elm = $('#select-trigger-hostname');
            elm.find('[data-toggle="tooltip"]').tooltip('hide');
            $('#select-trigger-hostname > *').remove();
            elm.hide();
            if(data != null && data.length != 0){
                var options = '';
                for(var i=0; i < data.length; i++){
                    var hostdata = data[i];
                    var ext_name = hostdata.ext_name;
                    var hostname = hostdata.hostname;
                    var regexp = new RegExp(input.val().toLowerCase().replace(/\\/g, '\\\\'));
                    var hit_name = (ext_name && ext_name.toLowerCase().match(regexp)) ? ext_name : hostname;
                    options += '<option class="modaltrigger-hostname" data-hostname="' + hostname + '" data-extname="' + (ext_name ? ext_name : '') + '" data-toggle="tooltip" data-title="ホスト名: ' + hostname + '<br/>表示名: ' + (ext_name ? ext_name : '(未設定)') + '" data-html="true" data-boundary="window" data-placement="left">' + hit_name + '</option>';
                }
                elm.html(options).show();
                elm.find('[data-toggle="tooltip"]').tooltip();
                input.tooltip('hide');
            }else if($('#search-hostname').val().length === 0){
                input.tooltip('hide');
            }else{
                input.tooltip('show');
            }
        });
        check_blank_in_trigger_setting_modal();
    });

    $(document).on('click', '.modaltrigger-hostname', function(){
        $('#search-hostname').val($(this).text()).attr({
            'data-hostname': $(this).attr('data-hostname'),
            'data-extname': $(this).attr('data-extname')
        });
        $('#select-trigger-hostname').hide();
        check_blank_in_trigger_setting_modal();
    });

    //定期時間の選択時にボタン表示とタイミングの時間の仮パラメータ設定
    $(document).on('click', '#select-trigger-time-area .periodic-time-option', function(){
        var elm = $(this);
        var target_btn_id = elm.data('targetbtn');
        var target_btn = $(target_btn_id);
        target_btn.text(elm.text());
        check_blank_in_trigger_setting_modal();
    });

    //定期実行間隔指定の単位選択
    $('#wrap-unit-of-periodic-interval .select-unit').on('click', function(){
        var button = $(this);
        var unit = button.attr('data-unit');
        $('#unit-of-periodic-interval').text('単位: ' + button.text());
        $('#select-periodic-interval > *').remove();
        var num_of_unit = 0;
        var unit_text = '';
        if(unit === 'hour'){
            num_of_unit = 24;
            unit_text = '時間毎';
        }else if(unit === 'day'){
            num_of_unit = 30;
            unit_text = '日毎';
        }else if(unit === 'week'){
            num_of_unit = 90;
            unit_text = '週毎';
        }else{
            num_of_unit = 12;
            unit_text = 'カ月毎';
        } 
        for(var i=1; i <= num_of_unit; i++){
             $('#select-periodic-interval').append(
                    $('<option>').attr({
                    'class': 'periodic-time-option dropdown-item',
                    'value': i,
                    'data-targetbtn': '#periodic-interval-btn',
                    'data-period' : 'periodic-interval', 
                    'data-unit': unit,
                    }).text(i + unit_text)
             );
        }
        $('#periodic-interval-btn').text('間隔を選択');
        $('#periodic-interval-area').show();     
        check_blank_in_trigger_setting_modal();   
    });  
     
    //タイミングの設定内容を表示するhtml作成
    function trigger_display_content(value, option_text, display_name, ext_name, hostname){
        if(value === 'vulsync'){
            var content = '<dl class="text-left mx-1 g-mb-1">' +
                          '<dt>脆弱性同期時&nbsp実行</dt>' + '</dl>';
                          if(display_name){
                              content += '<div class="mx-1 mt-1 g-brd-around g-brd-ddd p-1 g-bg-lightblue2 s-shadow-inset1">' +
                                         '<div class=" g-text-ul pl-1">' + '対象ホスト: ' + '</div>' +
                                         '<div class="text-truncate pl-1" data-toggle="tooltip" data-html="true" data-placement="left" data-title="' + 'ホスト名: ' + hostname + '<br/>表示名: ' + (ext_name ? ext_name : '(未設定)') + '" data-boundary="window" >' + display_name + '</div>';
                          }else{
                              var category_descs = JSON.parse($('#category-descs').val());
                              var desc = null;
                              for(category_id in category_descs){
                                  if( category_descs[category_id].name === option_text ){
                                      desc = category_descs[category_id].desc;
                                      break;
                                  }
                              }
                              content += '<div class="mx-1 mt-1 g-brd-around g-brd-ddd p-1 g-bg-lightblue2 s-shadow-inset1">' +
                                         '<div class=" g-text-ul pl-1">' + '対象カテゴリ: ' + '</div>' +
                                         '<div class="text-truncate pl-1" data-toggle="tooltip" data-placement="left" data-title="' + escape_html_special_chars(desc ? desc : '') + '" data-boundary="window" >' + escape_html_special_chars(option_text) + '</div>';
                          }
        }else if(value === 'periodic'){
            var content = '<dl class="text-left mx-1 g-mb-1">' +
                          '<dt>毎日実行-時間指定</dt>' + '</dl>' +
                          '<div class="mx-1 mt-1 g-brd-around g-brd-ddd p-1 g-bg-lightblue2 s-shadow-inset1">' +
                          '<div class="g-letter-spacing-0_08 pl-1">' + option_text + '</div>' 
                          ;
        }else if(value === 'interval'){
            var content = '<dl class="text-left mx-1 g-mb-1">' +
                          '<dt>定期実行-間隔指定</dt>' + '</dl>' +
                          '<div class="mx-1 mt-1 g-brd-around g-brd-ddd p-1 g-bg-lightblue2 s-shadow-inset1">' +
                          '<div class="g-letter-spacing-0_08 pl-1">' + option_text + '</div>'
                          ;
        }
        return content;
    }

    //タイミングのパラメータセット
    $('#set-trigger-setting').on('click',function(){ 
        var searched_hostname = $('#search-hostname').val();
        var selected = $('#select-trigger-time-area .set-rule-trigger:checked');
        var pass = false;
        var conds = JSON.parse($('#set-cond-param').val());
        var ope = JSON.parse($('#set-ope-param').val());
        var trigger_param = JSON.parse($('#form-create-trigger').val());
        var del_items = [];
        if(selected.length === 1){
            if(selected.val() === 'vulsync'){ 
                var delete_match_cat_items = false;
                var delete_host_items = false;
                var hosts = $('.modaltrigger-hostname');
                var is_correct_host = false;
                var set_category_id = $('#trigger-category').attr('data-categoryid');
                var del_actions = [];
                hosts.each(function(){
                    if($(this).text() === searched_hostname){
                        is_correct_host = true;
                        return false;
                    }
                });
                if(!is_correct_host && searched_hostname != ''){
                    var conf_text = 'タイミングの対象となるホストは存在しないため、選択したカテゴリに属する全てのホストがタイミングの対象となります。';
                }else if(searched_hostname === ''){
                    var conf_text = 'タイミングの対象となるホストは指定されていないため、選択したカテゴリに属する全てのホストがタイミングの対象となります。';
                }else{
                    var conf_text = 'タイミングの対象となるホストに"' + searched_hostname + '"が設定されます。';
                }
                if( trigger_param['period'] != 'vulsync' ){ //タイミングが違って、proc_emptyがfalseなら処理するデータが変わるので削除
                    for(data_name in ope){
                        if(ope[data_name] && String(ope[data_name].proc_empty) === 'false'){
                            del_items.push('「' + ope_detail_items_name[data_name] + '」');
                            del_actions.push(data_name);
                        }
                    }
                }
                for(data_name in conds){
                    if((data_name === 'category-match' || data_name === 'category-unmatch') && (set_category_id != 'all' || is_correct_host)){
                        delete_match_cat_items = true;
                        del_items.push('「' + cond_detail_items_name[data_name] + '」');
                    }else if((data_name === 'host-match' || data_name === 'host-unmatch') && (set_category_id != 'all' || is_correct_host)){
                        delete_host_items = true;
                        del_items.push('「' + cond_detail_items_name[data_name] + '」');
                    }
                }
                if(delete_match_cat_items || delete_host_items || del_items.length){
                    conf_text += '加えて、' + del_items.join(', ') + 'の項目は、タイミングの設定が関係しているため削除されます。';
                }
                conf_text += 'よろしいですか?';
                var conf = confirm(conf_text);
                if(conf){
                    if(delete_match_cat_items || delete_host_items || del_actions.length){
                        if(del_actions.length){
                            //delete ope['send-mail-category'];
                            //delete ope['send-mail-user'];
                            //var target_selecters = '[data-addope="send-mail-category"], [data-addope="send-mail-user"]';
                            var target_selecters = [];
                            del_actions.forEach(del_ac => {
                                delete ope[del_ac];
                                target_selecters.push('[data-addope="' + del_ac + '"]');
                            });
                            target_selecters = target_selecters.join(', ');
                            $('#form-create-operation, #set-ope-param').val(JSON.stringify(ope));
                            $('#set-operation-tbody').find(target_selecters).remove();
                            $('#rule-setting-item-ope').find(target_selecters).remove();
                            if(Object.keys(ope).length === 0){
                                $('#modalCreateRule-set-operation-btn').removeClass('btn-success set')
                                                                       .addClass('btn-secondary')
                                                                       .text('操作未設定');
                                $('#rule-setting-item-ope').html('未設定').removeClass('text-left');
                            }
                        }
                        if(delete_match_cat_items || delete_host_items){
                            if(delete_match_cat_items){
                                delete conds['category-match'];
                                delete conds['category-unmatch'];
                            }
                            if(delete_host_items){
                                delete conds['host-match'];
                                delete conds['host-unmatch'];
                            }
                            $('#form-create-condition, #set-cond-param').val(JSON.stringify(conds));
                            var target_selecters = '[data-cond="category-match"], [data-cond="category-unmatch"], [data-cond="host-match"], [data-cond="host-unmatch"]';
                            $('#set-condition-tbody').find(target_selecters).remove();
                            $('#rule-setting-item-cond').find(target_selecters).remove();
                            if(Object.keys(conds).length === 0){
                                $('#modalCreateRule-set-cond-btn').addClass('btn-secondary')
                                                                  .removeClass('set btn-success')
                                                                  .text('条件未設定');
                                $('#rule-setting-item-cond').html('未設定').removeClass('text-left');
                            }
                        }
                    }
                
                    var param = {
                        'hostname': is_correct_host ? $('#search-hostname').attr('data-hostname') : '',
                        'period'  : 'vulsync',
                        'value'    : 'vulsync',
                        'unit'    : '',
                        'category': set_category_id,
                    }
                    pass = true;
                    var hostname_of_trigger = (searched_hostname === '' || !is_correct_host) ? null : searched_hostname;
                    var html_content = trigger_display_content('vulsync', $('#trigger-category').text(), hostname_of_trigger, $('#search-hostname').attr('data-extname'), $('#search-hostname').attr('data-hostname'));
                }else{
                    return false;
                }
            }else{
                var del_actions = [];
                var conf_text = '';
                if(conds['issue-update-regi'] || conds['issue-update-del'] || conds['issue-update-regi-and-del'] || (conds['cvss-exploitable'] && conds['cvss-exploitable']['is_change'] === 'true' && trigger_param['period'] === 'vulsync')){
                    var display_conds = [];
                    for(data_name in conds){
                        if(data_name === 'issue-update-regi' || data_name === 'issue-update-del' || data_name === 'issue-update-regi-and-del' || (data_name === 'cvss-exploitable' && trigger_param['period'] === 'vulsync')){
                            display_conds.push('「' + cond_detail_items_name[data_name] + '」') ;
                        }
                    }
                    conf_text += '設定されている条件に「脆弱性同期時関連」の項目があります。このままタイミングを設定すると条件に設定されている' + display_conds.join(', ') + 'の項目が削除されます。';
                }
                if(trigger_param['period'] === 'vulsync' ){ //タイミングが脆弱性同期時になってproc_emptyがfalseなら処理するデータが変わるので削除
                    for(data_name in ope){
                        if(ope[data_name] && String(ope[data_name].proc_empty) === 'false'){
                            del_items.push('「' + ope_detail_items_name[data_name] + '」');
                            del_actions.push(data_name);
                        }
                    }
                    if(del_items.length){
                        if(conf_text != ''){
                            conf_text += '加えて、';
                        }
                        conf_text += ('操作の' + del_items.join(', ') + 'の項目は、タイミングの設定が関係しているため削除されます。');
                    }
                }/*else if(ope['status-change'] || ope['due-date-change']){
                    var ope_items = [];
                    if(conf_text != ''){
                        conf_text += 'また、';
                    }
                    for(data_name in ope){
                        if(data_name === 'status-change' || data_name === 'due-date-change'){
                            ope_items.push('「' + ope_detail_items_name[data_name] + '」');
                        }
                    }
                    conf_text += '操作の「ステータス変更」,「対策予定日設定」の項目は、タイミングに「脆弱性同期時 実行」が設定されている場合にのみ使用可能な機能です。このため、このままタイミングを設定すると操作に設定されている' + ope_items.join(', ') + 'の項目が削除されます。';
                }*/
                /*if(trigger_param['period'] === 'vulsync' && (ope['send-mail-category'] || ope['send-mail-user'])){
                    if(conf_text != ''){ 
                        conf_text += '加えて、';
                    }
                    var ope_items = [];
                    delete_mail_items = true;
                    for(data_name in ope){
                        if(data_name === 'send-mail-category' || data_name === 'send-mail-user'){
                            ope_items.push('「' + ope_detail_items_name[data_name] + '」');
                        }
                    }
                    conf_text += ope_items.join(', ') + 'の項目は、タイミングの設定が関係しているため削除されます。';
                }*/
                if(conf_text != ''){
                   conf_text += 'よろしいですか？';
                   var conf = confirm(conf_text);     
                   if(conf){
                      delete conds['issue-update-regi'];
                      delete conds['issue-update-del'];
                      delete conds['issue-update-regi-and-del'];
                      if(trigger_param['period'] === 'vulsync'){
                          delete conds['cvss-exploitable'];
                      }
                      $('#form-create-condition, #set-cond-param').val(JSON.stringify(conds));
                      var target_selecters = '[data-cond="issue-update-regi"], [data-cond="issue-update-del"], [data-cond="issue-update-regi-and-del"], [data-cond="cvss-exploitable"]';
                      $('#set-condition-tbody').find(target_selecters).remove();
                      $('#rule-setting-item-cond').find(target_selecters).remove();
                      if(Object.keys(conds).length === 0){
                      $('#modalCreateRule-set-cond-btn').addClass('btn-secondary')
                                                        .removeClass('set btn-success')
                                                        .text('条件未設定');
                          $('#rule-setting-item-cond').html('未設定').removeClass('text-left');
                      }
                      if( del_actions.length ){
                          var target_selecters = [];
                          del_actions.forEach(del_ac => {
                              delete ope[del_ac];
                              target_selecters.push('[data-addope="' + del_ac + '"]');
                          });
                          target_selecters = target_selecters.join(', ');
                          $('#form-create-operation, #set-ope-param').val(JSON.stringify(ope));
                          $('#set-operation-tbody').find(target_selecters).remove();
                          $('#rule-setting-item-ope').find(target_selecters).remove();
                          if(Object.keys(ope).length === 0){
                              $('#modalCreateRule-set-operation-btn').removeClass('btn-success set')
                                                                     .addClass('btn-secondary')
                                                                     .text('操作未設定');
                              $('#rule-setting-item-ope').html('未設定').removeClass('text-left');
                          }
                      }/*else{
                          delete ope['status-change'];
                          delete ope['due-date-change'];
                          var target_selecters = '[data-addope="status-change"], [data-addope="due-date-change"]';
                          if(delete_mail_items === true){
                              delete ope['send-mail-category'];
                              delete ope['send-mail-user'];
                              target_selecters += ', [data-addope="send-mail-category"], [data-addope="send-mail-user"]';
                          }
                          $('#form-create-operation, #set-ope-param').val(JSON.stringify(ope));
                          $('#set-operation-tbody').find(target_selecters).remove();
                          $('#rule-setting-item-ope').find(target_selecters).remove();
                          if(Object.keys(ope).length === 0){
                              $('#modalCreateRule-set-operation-btn').removeClass('btn-success set')
                                                                     .addClass('btn-secondary')
                                                                     .text('操作未設定');
                              $('#rule-setting-item-ope').html('未設定').removeClass('text-left');
                          } 
                       }*/
                   }else{
                       return false;
                   }
                }
                var selected_op = selected.closest('[data-wrapping=periodic-time]')
                                          .find('.periodic-time-option:selected');
                if(selected_op.length === 1){
                    var param = {
                        'hostname': '',
                        'period'  : selected_op.attr('data-period'),
                        'value'    : selected_op.val(),
                        'unit'    : selected_op.attr('data-unit') ? selected_op.attr('data-unit') : '', 
                        'category': '',
                    };
                    pass = true;
                    if(selected.val() === 'periodic'){
                        var html_content = trigger_display_content(selected.val(), $('#periodic-hour-btn').text());
                    }else{
                        var html_content = trigger_display_content(selected.val(), $('#periodic-interval-btn').text());
                    }                     
                }else if(selected_op.length > 1){
                    alert('選択欄では１つのみ選択して下さい');
                    return false;
                }
            }
            if(pass){
                $('#form-create-trigger').val(JSON.stringify(param));
                $('#rule-setting-item-trigger').html(html_content).addClass('text-left');
                $('#rule-setting-item-trigger').find('[data-toggle="tooltip"]').tooltip();
                //var expand_btn = $('<i>').attr('class', 'expand-btn fas fa-angle-down mr-1')
                //$('#rule-setting-item-trigger').append(expand_btn);
                $('#modalCreateRule-set-cond-btn, #modalCreateRule-set-operation-btn').removeClass('disabled')
                                                                                      .tooltip('disable');
                setTimeout(function(){
                    if($(window).width() >= 992){
                        $('#form-create-rulename, #select-enable-rule, #modalCreateRule-set-cond-btn, #modalCreateRule-set-operation-btn').each(function(){
                            var target = $(this);
                            var elm_id = target.attr('id');
                            if($('#modalCreateRule-set-cond-btn, #modalCreateRule-set-operation-btn').hasClass('set')){
                                return false;
                            }
                            if(elm_id === 'form-create-rulename' && target.val() != ''){
                                return true;
                            }
                            var original_tooltip = make_original_tooltip('残りの項目を設定して下さい', true);
                            $('body').append(original_tooltip);
                            var offset = target.offset();
                            var offset_left = offset.left + (target.outerWidth() / 2) - (original_tooltip.outerWidth() / 2);
                            var offset_top = offset.top - original_tooltip.outerHeight() - 8;
                            //条件設定ボタンと操作設定ボタンのメッセージの矢印は左寄せにする
                            if(elm_id === 'modalCreateRule-set-cond-btn' || elm_id === 'modalCreateRule-set-operation-btn'){
                                original_tooltip.find('.inner-original-tooltip-down').addClass('setting-btn-arrow');
                            }
                           original_tooltip.offset({'top': offset_top, 'left': offset_left});  
                       });
                       $('.original-tooltip').show();
                    }  
                }, 650);
            }
        }
        check_rule_is_registable();
        
    });

    //クリックでツールチップ消す
    $('#form-create-rulename, #modalCreateRule-set-cond-btn, #modalCreateRule-set-operation-btn').on('click', function(){
        if($('.original-tooltip').length > 0){
            $('.original-tooltip').remove();
        }
    });

    $('#select-enable-rule').on('click',function(){
        if(!$('#modalCreateRule').attr('data-create')){
            if($('.original-tooltip').length > 0){
                $('.original-tooltip').remove();
            }
        }
    });

    //新規操作追加の詳細作成
    function make_operation_detail_area(items, operation){
        var ope = operation;
        var wrap = $('<div>').attr({
                             'data-ope': ope,
                             'class':'position-relative'
                             });
        if(ope.indexOf('send-mail') != -1){
            wrap.attr('id', 'ope-detail-send-mail-area')
                .addClass('font-size-small');
        }else if(ope.indexOf('status-change') != -1){
            wrap.attr({
                'id': 'ope-detail-status-change-area',
                'class': 'bg-f8',
                });
        }else if(ope.indexOf('due-date-change') != -1){
            var is_update = ope.indexOf('update') != -1;
            wrap.attr({
                'id': is_update ? 'edit-due-date-change-area' : 'ope-detail-due-date-change-area',
                'class': 'bg-f8',
            });
            var div = $('<div>').attr({
                                'class':'text-left',
                                'id': is_update ? 'wrap-edit-select-due-date-item' : 'wrap-ope-detail-select-due-date-item',
                                }); 
            var select_item = $('<select>').attr({
                                          'class': 'form-control form-control-sm',
                                          'id': is_update ? 'edit-select-due-date-item' : 'ope-detail-select-due-date-item',
                                          });
            var option1 = $('<option>').attr({
                                       'value': 'date',
                                       }).text('日付で設定');
            var option2 = $('<option>').attr({
                                       'value': 'days',
                                       }).text('日数で設定');
            select_item.append([option1, option2]);
            div.append(select_item);
            if(is_update){
                $('#after-ope-edit-content').append(div);
            }else{
                $('#ope-detail-body').append(div);
            }
        }else if(ope.indexOf('send-syslog') != -1){
            var is_update = ope.indexOf('update') != -1;
            wrap.attr({
                'id': (is_update ? 'edit-send-syslog-area' : 'ope-detail-send-syslog-area'),
                'class': 'bg-f8'
            });
            var log_sample = $('<div>').attr('class', 'text-left' + (is_update ? '' : '')).append(
                $('<button>').attr({
                    'type': 'button',
                    'id': 'show-sample-syslog-message-btn',
                    'class': 'btn btn-edit btn-xs',
                    'data-toggle': 'tooltip',
                    'data-title': '<div style="white-space:nowrap">[フォーマット]<br>月 日 時間 ホスト名 識別子[プロセスID] : メッセージ<br>例：<br>Aug 13 11:30:00 sample_host SIDfmVM[12345]: sample message</div>',
                    'data-html': 'true',
                    'data-trigger': 'click',
                    'data-template': '<div style="max-width: 500px !important" class="tooltip" role="tooltip"><div class="arrow"></div><div style="max-width: 500px !important" class="tooltip-inner"></div></div>' 
                }).text('送信されるログメッセージ')
            );
            if( is_update ){
                $('#after-ope-edit-content').append(log_sample);
            }else{
                $('#ope-detail-body').append(log_sample);
            }
        } 
        $(items).each(function(){
            var item = $(this);
            wrap.append(item);
        });
        return wrap;
    }

    //新規操作追加の項目作成function
    function ope_send_mail(operation, address_type){
        var ope = operation;
        var type = address_type;
        $('#search-address-items-area-tbody > *').remove();
        $('.mail-setting-items-list').remove();
        var setting_btn_div = $('<div>');
        var setting_btn = $('<button>').attr({
                                        'type': 'button',
                                        'id' : 'ope-detail-set-mail-content-btn',
                                        'class':'btn btn-sm btn-edit',
                                        'data-opedetail': ope,
                                        }).text('メールを設定する');
        setting_btn_div.append(setting_btn);
        $('#ope-detail-body').append(setting_btn_div);

        var target_data_div = $('<div>').attr({'class': 'form-group mb-2 ope-detail-item-top'});
        var div = $('<div>').text('対象となるデータ');
        var target_data = $('<input>').attr({
                               'type': 'text',
                               'class': 'd-block w-100 form-control form-control-sm',
                               'id': 'mail-target-data',
                               'readonly': 'true'
                          });
        target_data_div.append(div, target_data);
  
        var sender_div = $('<div>').attr({
                                   'class': 'form-group mb-2 ope-detail-item'
                                   });
        var div = $('<div>').text('送信元');
        var sender = $('<input>').attr({
                                 'type': 'text',
                                 'class' : 'd-block w-100 form-control form-control-sm',
                                 'id': 'ope-detail-readonly-sender',
                                 'readonly': 'true'
                                 });
        sender_div.append(div.append(sender));
                               
        var address_div = $('<div>').attr({
                                     'data-opedetail': ope,
                                     'class': 'form-group mb-2 ope-detail-item', 
                                     });
        var div = $('<div>').attr('id','ope-detail-address-name');
        if(type === 'user'){
            div.attr('data-sendby', 'user').text('送信先ユーザ');
        }else{
            div.attr('data-sendby', 'category').text('送信先カテゴリ');
        }
        address_div.append(div);
        var address_text = $('<input>').attr({
                                       'id': 'ope-detail-readonly-address-type-text',
                                       'type':'text',
                                       'readonly': true,
                                       'class': 'd-block w-100 form-control form-control-sm ope-detail-mail-address-name',
                                       });
        address_div.append(address_text);
        var hidden_store_selected_items = $('<div>').attr({
                                                    'id': 'hidden-store-search-address-items-area-tbody',
                                                    'style': 'display:none;'
                                                    });
        address_div.append(hidden_store_selected_items);
        var mail_title_div = $('<div>').attr({
                                       'data-opedetail': 'mail-title',
                                       'class': 'form-group mb-2 ope-detail-item',
                                       });
        var div = $('<div>').attr('class','ope-mail-title')
                            .text('件名');
        mail_title_div.append(div);
        var title = $('<input>').attr({
                                'id': 'ope-detail-readonly-mail-title',
                                'type': 'text',
                                'readonly':true,
                                'class':'w-100 d-block form-control form-control-sm',
                                }); 
        mail_title_div.append(title);
        var mail_content_div = $('<div>').attr({
                                         'data-opedetail': 'mail-content',
                                         'class': 'ope-detail-item form-group'
                                         });
        var div = $('<div>').text('メール本文');
        mail_content_div.append(div);
        var content_text = $('<textarea>').attr({
                                          'id': 'ope-detail-readonly-mail-content',
                                          'class':'d-block w-100 form-control form-control-sm',
                                          'readonly':true,
                                          });
        mail_content_div.append(content_text);
        var items = [target_data_div, sender_div, address_div, mail_title_div, mail_content_div];
        return items;
    }

    function ope_status_change(){
        var div1 = $('<div>').attr('class','')
                             .html('<div class="g-font-size-14 g-text-ul text-left p-2 mb-2">脆弱性のステータスが</div><div class="d-flex g-font-size-13 justify-content-center"><button type="button" class="mr-1 btn btn-light g-font-size-13  position-relative" id="ope-detail-status-from">選択して下さい<i class="ml-1 fas fa-caret-down"></i></button><div class="pt-2">のものを</div></div>');
        var div2 = $('<div>').attr('class', 'text-center');
        var arrow = $('<i>').attr('class','far fa-arrow-alt-circle-down fa-2x my-3');
        div2.append(arrow);
        var div3 = $('<div>').attr('class', '')
                             .html('<div class="d-flex g-font-size-13 justify-content-center"><button type="button" class="ml-3 mr-1 btn btn-light g-font-size-13 position-relative" id="ope-detail-status-to">選択して下さい<i class="ml-1 fas fa-caret-down"></i></button><div class="pt-2">に変更する</div></div>');
        var div4 = $('<hr>').attr('class', 'my-3 hr-dotted');
        var div5 = $('<div>').attr('class', 'mx-2 mb-1');
        var process_data_select = create_process_data_select_elm();
        process_data_select.removeClass('w-auto');
        div5.append(
            $('<div>').attr('class', 'ml-1').append(
                $('<span>').attr('class', 'font-weight-bold').text('この「操作」で処理する、対象となるデータ'),
                $('<span>').attr('class', 'd-block text-danger g-font-size-11 g-mt-2 mb-2').text('※但し、当該対象データは「条件」に一致するもののみとなります')
            ),
            process_data_select
        )
        var items = [div1, div2, div3, div4, div5];
        return items;
    }
    
    function ope_due_date_change(){
        var div1 = $('<div>').attr('class','g-font-size-14 g-text-ul text-left p-2')
                             .html('<div>脆弱性の対策予定日を</div>');
        var div2 = $('<div>').attr('class', 'd-flex g-font-size-13 justify-content-center')
                             .html('<div class="ml-5 pl-5" data-setting="date"><input type="text" class="mx-1 detail-value text-center w-50" id="ope-detail-due-date-to" data-toggle="datepicker" readonly>に設定する</div><div style="display:none;" class="" data-setting="days">脆弱性検出日の<input type="number" min="0" class="mx-1 g-width-5em detail-value" id="ope-detail-due-date-days">日後に設定する</div>');
        var div3 = $('<div>').attr({'class': 'text-right mt-1', 'id': 'ope-detail-duedate-checkbox-area'})
                             .html('<div class="g-font-size-13 d-inline-block p-2 mt-1"><input id="ope-detail-checkbox-is-overwrite-duedate" style="vertical-align: -0.15em;" type="checkbox"><span class="ml-1">対策予定日上書き</span><span data-toggle="tooltip" data-placement="top" title="既に対策予定日が設定されている際に、対策予定日を上書きする場合はチェックを入れて下さい"><i class="fas fa-question-circle c_gray_70 ml-1"></i></span></span></div>');
        var div4 = $('<hr>').attr('class', 'my-3 hr-dotted');
        var div5 = $('<div>').attr('class', 'mx-2 mb-1');
        var process_data_select = create_process_data_select_elm();
        process_data_select.removeClass('w-auto'); 
        div5.append(
            $('<div>').attr('class', 'ml-1').append(
                $('<span>').attr('class', 'font-weight-bold').text('この「操作」で処理する、対象となるデータ'),
                $('<span>').attr('class', 'd-block text-danger g-font-size-11 g-mt-2 mb-2').text('※但し、当該対象データは「条件」に一致するもののみとなります')
            ),
            process_data_select     
        )                           
        var items = [div1, div2, div3, div4, div5];
        return items; 
    }

    function remove_keyword_items(keyword_button){
        var remove_values = JSON.parse($('#syslog-unusable-keywords').val());
        var remove_menus = ['vulsync-info-all', 'vulsync-info-match'];
        remove_values.forEach(function(v){
            keyword_button.find('[value="' + v + '"]').closest('.original-dropdown-item').remove();
            keyword_button.find('[data-itemid="' + v + '"]').closest('.original-dropdown-item').remove();
        });
        remove_menus.forEach(function(menu_name){
            keyword_button.find('[data-menuname="' + menu_name + '"]').closest('.original-dropdown-item ').remove();
        });
        return keyword_button
    }

    function ope_send_syslog(){
        var div_protocol = $('<div>').attr('class', 'g-font-size-14 text-left p-2 syslog-item');
        div_protocol.append(
            $('<label>').attr('class', 'font-weight-bold mb-1').text('プロトコル'),
            $('<select>').attr({'id': 'select-syslog-protocol', 'class': 'form-control form-control-sm'}).append(
                $('<option>').val('udp').text('UDP'),
                $('<option>').val('tcp').text('TCP')
            )
        );
        var div_address = $('<div>').attr('class', 'g-font-size-14 text-left p-2 syslog-item');
        div_address.append(
            $('<div>').append(
                $('<label>').attr('class', 'font-weight-bold mb-2').append(
                    $('<span>').text('送信先ホスト'),
                ),
                $('<select>').attr({
                    'id': 'select-syslog-address', 'class': 'form-control form-control-sm w-50 d-inline-block ml-1 syslog-select'
                }).append(
                    $('<option>').val('ip-address').text('IPアドレスを指定'),
                    $('<option>').val('hostname').text('ホスト名を指定')
                )
            ),
            $('<div>').append(
                $('<div>').attr('id', 'syslog-hostname-field').append(
                    $('<label>').attr({'class': 'g-width-6em text-center mr-1', 'for': 'syslog-input-hostname'}).text('ホスト名'),
                    $('<input>').attr({
                        'id': 'syslog-input-hostname',
                        'class': 'd-inline-block form-control form-control-sm w-50'
                    })
                ).hide(),
                $('<div>').attr('id', 'syslog-ip-address-field').append(
                    $('<label>').attr({'class': 'g-width-6em text-center mr-1', 'for': 'syslog-input-ip-address'}).text('IPアドレス'),
                    $('<input>').attr({
                        'id': 'syslog-input-ip-address',
                        'class': 'd-inline-block form-control form-control-sm w-50'
                    }).val('172.17.0.1')
                )
            )
        );
        var div_port_number = $('<div>').attr('class', 'g-font-size-14 text-left p-2 syslog-item');
        div_port_number.append(
            $('<label>').attr({'class': 'font-weight-bold mb-1', 'for': 'syslog-port-number'}).text('送信先ポート番号'),
            $('<input>').attr({
                'id': 'syslog-port-number', 
                'class': 'form-control form-control-sm',
                'type': 'number',
                'min': '0'
            }).val('514'),
        );
        var syslog_facilities = JSON.parse($('#syslog-facilities').val());
        syslog_facilities = syslog_facilities.map(function(f){
            return  $('<option>').val(f).text(f);
        });
        var div_facilities = $('<div>').attr('class', 'g-font-size-14 text-left p-2 syslog-item');
        div_facilities.append(
            $('<label>').attr({'class': 'font-weight-bold mb-1'}).append(
                $('<span>').text('ログの種別'),
                $('<i>').attr({
                    'class': 'fas fa-question-circle c_gray_70 ml-1',
                    'data-toggle': 'tooltip',
                    'data-title': 'syslogのfacilityです',
                })
            ),
            $('<select>').attr({
                'id': 'select-syslog-facility',
                'class': 'form-control form-control-sm'
            }).append(
                syslog_facilities
            )
        );
        div_facilities.find('option[value="LOCAL3"]').prop('selected', true);
        var syslog_severities = JSON.parse($('#syslog-severities').val());
        syslog_severities = syslog_severities.map(function(s){
            return $('<option>').val(s).text(s);
        });
        var div_severities = $('<div>').attr('class', 'g-font-size-14 text-left p-2 syslog-item');
        div_severities.append(
            $('<label>').attr({'class': 'font-weight-bold mb-1'}).append(
                $('<span>').text('ログの深刻度'),
                $('<i>').attr({
                    'class': 'fas fa-question-circle c_gray_70 ml-1',
                    'data-toggle': 'tooltip',
                    'data-title': 'syslogのseverityです',
                })
            ),
            $('<select>').attr({
                'id': 'select-syslog-severity',
               'class': 'form-control form-control-sm'
            }).append(
                syslog_severities
            )
        )
        div_severities.find('option[value="INFO"]').prop('selected', true);
        var div_hostname = $('<div>').attr('class', 'g-font-size-14 text-left p-2 syslog-item');
        div_hostname.append(
            $('<label>').attr({'class': 'font-weight-bold mb-1', 'for': 'syslog-input-message-hostname'}).append(
                $('<span>').text('ホスト名'),
                $('<i>').attr({
                    'class': 'fas fa-question-circle c_gray_70 ml-1',
                    'data-toggle': 'tooltip',
                    'data-title': 'ログメッセージの"ホスト名"の部分に出力される内容です',
                })
            ),
            $('<input>').attr({
                'id': 'syslog-input-message-hostname',
                'class': 'form-control form-control-sm'
            }).val('sidfmvmsrv')
        );
        var div_ident = $('<div>').attr('class', 'g-font-size-14 text-left p-2 syslog-item');
        div_ident.append(
            $('<label>').attr({'class': 'font-weight-bold mb-1', 'for': 'syslog-input-ident'}).append(
                $('<span>').text('識別子'),
                $('<i>').attr({
                    'class': 'fas fa-question-circle c_gray_70 ml-1',
                    'data-toggle': 'tooltip',
                    'data-title': 'ログメッセージの"識別子"の部分に出力される内容です',
                })
            ),
            $('<input>').attr({
                'id': 'syslog-input-ident',
                'class': 'form-control form-control-sm'
            }).val('SIDfmVM')
        );
        var div_message = $('<div>').attr('class', 'g-font-size-14 text-left p-2 syslog-item');
        var keywords_inputter = create_keywords_inputter();
        keywords_inputter.find('#keywords-list').attr('data-drop', 'left-side');
        keywords_inputter = remove_keyword_items(keywords_inputter);
        div_message.append(
            $('<label>').attr({'class': 'font-weight-bold mb-1', 'for': 'syslog-input-message'}).append(
                $('<span>').text('メッセージ'),
                $('<span>').attr('class', 'font-size-small').text('※改行不可'),
                $('<i>').attr({
                    'class': 'fas fa-question-circle c_gray_70 ml-1',
                    'data-toggle': 'tooltip',
                    'data-title': 'ログメッセージの"メッセージ"の部分に出力される内容です',
                })
            ),
            $('<div>').attr({'id': 'syslog-popup-target', 'class': 'position-relative'}),
            $('<textarea>').attr({
                'id': 'syslog-input-message',
                'class': 'form-control form-control-sm mb-1'
            }),
            $('<div>').attr('class', 'd-flex').append(keywords_inputter, create_keywords_replacer())
        );
        var divider = $('<hr>').attr('class', 'my-3 hr-dotted');
        var div_process_data = $('<div>').attr('class', 'mx-2 mb-1');
        var process_data_select = create_process_data_select_elm();
        process_data_select.removeClass('w-auto');
        div_process_data.append(
            $('<div>').attr('class', 'ml-1').append(
                $('<span>').attr('class', 'font-weight-bold').text('この「操作」で処理する、対象となるデータ'),
                $('<span>').attr('class', 'd-block text-danger g-font-size-11 g-mt-2 mb-2').text('※但し、当該対象データは「条件」に一致するもののみとなります')
            ),
            process_data_select
        ) 
        var items = [div_protocol, div_address, div_port_number, div_facilities, div_severities, div_hostname, div_ident, div_message, divider, div_process_data];
        return items;
    }

    function ope_set_ctag(){
        var area = $('<div>').attr('class', 'ope-detail-ctag-area');
        var search_btn = $('<div>').append(
                                       $('<button>').attr({
                                           'type': 'button',
                                           'id': 'ope-detail-search-ctags',
                                           'class': 'btn btn-xs btn-primary text-nowrap',
                                       }).text('カスタムタグ検索')
                                   );
        var true_searched_area = $('<div>').attr('class', 'mt-3').append(
            $('<div>').attr('class', 'g-font-size-14 font-weight-bold pb-1').text('設定する'),
            $('<div>').attr({ id: 'ope-detail-ctags-true-area', class: 'detail-ctags-area' })
        );
        var false_searched_area = $('<div>').attr('class', 'mt-2').append(
            $('<div>').attr('class', 'g-font-size-14 font-weight-bold pb-1').text('除外する'),
            $('<div>').attr({ id: 'ope-detail-ctags-false-area', class: 'detail-ctags-area' })
        );
        var div_process_data = $('<div>').attr('class', 'mt-3');
        var process_data_select = create_process_data_select_elm();
        process_data_select.removeClass('w-auto');
        div_process_data.append(
            $('<div>').append(
                $('<span>').attr('class', 'font-weight-bold').text('この「操作」で処理する、対象となるデータ'),
                $('<span>').attr('class', 'd-block text-danger g-font-size-11 g-mt-2 mb-2').text('※但し、当該対象データは「条件」に一致するもののみとなります')
            ),
            process_data_select
        )
        area.append(search_btn, true_searched_area, false_searched_area, div_process_data);
        return [ area ];
    }

    function make_operation_detail_item(operation){
        $('#ope-item-title, #ope-detail-body, #selecting-ope-item-area').show();
        var ope = operation;
        var descripition = $('#ope-detail').find('.ope-ask-check');
        if(ope === 'send-mail-user'){
            var items = ope_send_mail(ope, 'user');
            descripition.text('設定した「条件」に一致する脆弱性あるいはホストが存在する場合、ここで選択するユーザもしくはメールアドレスにメールを送信します。').show();
        }else if(ope === 'send-mail-category'){
            var items = ope_send_mail(ope, 'category');
            descripition.text('設定した「条件」に一致する脆弱性あるいはホストが存在する場合、ここで選択するカテゴリに属しているユーザにメールを送信します。').show();
        }else if(ope === 'status-change'){
            var items = ope_status_change();
            descripition.text('設定条件に一致する脆弱性が存在する場合、その脆弱性のステータスを変更します。').show();
        }else if(ope === 'due-date-change'){
            var items = ope_due_date_change();
            descripition.text('設定条件に一致する脆弱性が存在する場合、その脆弱性の対策予定日を設定します。').show();
        }else if(ope === 'send-syslog'){
            var items = ope_send_syslog();
            descripition.text('設定条件に一致する脆弱性が存在する場合、syslog にログを送信します。').show();
        }else if(ope === 'set-ctag'){
            ruleSelectedCtags.val('{}'); //登録内容をリセット
            ruleSelectedRegistCtags.val('{}');
            var items = ope_set_ctag();
            descripition.text('設定条件に一致する脆弱性が存在する場合、指定したカスタムタグを設定・除外します。').show();
        }
        var item = make_operation_detail_area(items, ope);
        $('#ope-detail-body').append(item); 
        $('#ope-detail-due-date-to').datepicker();
        $('#switch_due-date-menu').dropdown();
        $('#ope-detail-body').find('[data-toggle="tooltip"]').tooltip();
    }
   
    // 操作を選択するとき
    $('#operation-item-area [data-ope]').on('click', function(){
        var button = $(this);
        if(button.hasClass('disabled')){
            return false;
        }
        if(button.attr('data-ope') === $('#ope-detail-body').find('[data-ope]').attr('data-ope')){
            return false;
        }
        var pass = check_move_to_other_ope_item();
        if(pass){
            $('#ope-detail-body > *').remove();
            var ope = button.data('ope');
            var item_name = ope_items_name[ope];
            var selecting = $('#selecting-ope-item');
            selecting.text(item_name);
            selecting.attr('data-selecting', ope);
            var items_div = make_operation_detail_item(ope);
            $('#ope-detail-body').append(items_div);
            $('#execute-add-ope').show();
            color_selecting_item_on_AddModal(button);
            //メール設定modal初期化
            if(ope === 'send-mail-user' || ope === 'send-mail-category'){
                $('#mail-setting-users-tbody > *').remove();
                reset_mail_select_type();
                $('.mail-setting-items-list').remove();
                $('#search-address-items-area-tbody > *').remove();
                $('#add-address-area-tbody > *').remove();
                $('#email-address-list-tbody > *').remove();
                $('#send-mail-title').val('');
                $('#send-mail-content').val('');
                $('#send-mail-users-table input[type=checkbox]').prop('checked', false);
                $('#mail-select-process-data-area > *').remove();
                $('#modalOperationMailSetting').find('[data-processflg]').prop('checked',false);  
                $('[name="mail-setting-process-data-radio"]').prop('checked', false);
                if(ope === 'send-mail-user'){
                    var search_btn = $('#search-address-items-area').find('[data-searchbtn]'); 
                    search_btn.attr('id', 'modalOperationMailSetting-search-users-btn')
                              .text('送信先ユーザ検索');
                    $('#search-address-type-items-title').text('送信先ユーザ');
                    $('#switch-address-email-or-user-area').show();
                }else if(ope === 'send-mail-category'){
                    var search_btn = $('#search-address-items-area').find('[data-searchbtn]');
                    search_btn.attr('id', 'modalOperationMailSetting-search-category-btn')
                              .text('カテゴリ検索');
                    $('#search-address-type-items-title').text('送信先カテゴリ');
                    $('#switch-address-email-or-user-area').hide();
               }
            } 
        }else{
            return false;
        }
    });

    function check_move_to_other_ope_item(){
        var pass = true;
        if($('#ope-detail-send-mail-area').length === 1){
            if($('#ope-detail-readonly-sender').val() != '' || $('#ope-detail-readonly-address-type-text').val() != '' || $('#ope-detail-readonly-mail-title').val() != '' | $('#ope-detail-readonly-mail-content').val() != ''){
                pass = false;
            }
        }else if($('#ope-detail-status-change-area').length === 1){
            if($('#ope-detail-status-change-area').find('[data-statusids]').length > 0){
                pass = false;
            }
        }else if($('#ope-detail-due-date-change-area').length === 1){
            if($('#ope-detail-due-date-to:visible, #ope-detail-due-date-days:visible').val() != ''){
                pass = false;
            }
        }
        if(!pass){
            return confirm('操作詳細の内容は破棄されます。よろしいですか?');
        }else{
            return true;
        }
    }   

    //操作設定modal
    $('#modalRegisterOperation').on('show.bs.modal',function(e){
        toggle_execute_set_btn();
        var button = $(e.relatedTarget);
        if($('#set-operation-tbody').find('tr').length === 0){
            display_no_operation_message();
        }
        $('#hidden_modalRegisterOperation-zone *').remove();
        var clone_tbody = $('#set-operation-tbody').clone()
                                                   .attr('id', 'clone-set-operation-tbody');
        $('#hidden_modalRegisterOperation-zone').append(clone_tbody);
        reset_add_operation_area();
        var param = JSON.parse($('#form-create-trigger').val());
        if(param['period'] != 'vulsync'){
            var desc = '脆弱性同期時に行う操作です。タイミング設定で「脆弱性同期時 実行」を選択した場合にのみ使用可能な機能です。';
            $('#wrap-operation-item-btns').find('[data-ope="status-change"], [data-ope="due-date-change"]')
                                          .addClass('disabled')
                                          .attr({
                                              'data-toggle': 'tooltip',
                                              'data-title': desc,
                                          }).tooltip('enable');
       }else{
           var btns = $('#wrap-operation-item-btns').find('[data-ope="status-change"], [data-ope="due-date-change"]');
           btns.removeClass('disabled');
           if(btns.attr('data-toggle') === 'tooltip'){
               btns.tooltip('disable');
           }
       } 
    });
    
    $('#modalRegisterOperation').on('shown.bs.modal',function(){
        //操作追加ボタンと被らないようにスペースを確保
        $('#keep-margin-for-ope').height($('#execute-add-ope').outerHeight(true) + 5);
    });

    $('#modalRegisterOperation').on('hide.bs.modal',function(){
        $('#modalRegisterOperation-table').scrollTop(0);
    });

    $('#modalRegisterOperation').on('hidden.bs.modal',function(){
        var param = JSON.parse($('#form-create-operation').val());
        if(Object.keys(param).length > 0){
            $('#modalCreateRule-set-operation-btn').removeClass('btn-secondary')
                                                   .addClass('btn-success set')
                                                   .text('操作設定済み');
        }else{
            $('#modalCreateRule-set-operation-btn').removeClass('btn-success set')
                                                   .addClass('btn-secondary')
                                                   .text('操作未設定');
        }
        check_rule_is_registable();
        setTimeout(function(){
            show_change_rule_name_message();
        },500);
    });

    //操作設定modalキャンセルで閉じる時
    $('#close-modalRegisterOperation-btn, #cancel-modalRegisterOperation-btn').on('click',function(){
        var conf = confirm_close_RegisterModal('modalRegisterOperation');
        if(conf){
            $('#modalRegisterOperation').modal('hide');
        }
        $('.original-tooltip').remove();
    });

    //操作設定modalで設定ボタンを押した時
    $('#execute-set-ope').on('click', function(){
        if($('#ope-detail-body').find('[data-ope]').length > 0){
            var conf = confirm('編集中の操作詳細項目があります。このまま操作を設定しますか？');
            if(!conf){
                return false;
            }
        }
        $('#form-create-operation').val($('#set-ope-param').val());
        $('.original-tooltip').remove();
        create_ope_display_content(JSON.parse($('#set-ope-param').val()));
        $('#modalRegisterOperation').modal('hide');
        
    });

    //操作設定modalの「処理するデータ」の切り替え時
  /*  $('[data-id="select-process-data"]').on('change', function(){
        var reset_elms = [];
        var reset_names = [];
        var keyword_btn = $('#mail-setting-keyword-btn:visible');
        var before_changed = $('#selected-process-data-before').val();
        var ope_param = JSON.parse($('#set-ope-param').val());
        var selected_process_data = $(this).find('option:selected').val();
        var process_values_list = JSON.parse($('#process-values-list').val());
        var check_elms = $('#send-mail-title:visible, #send-mail-content:visible, #mail-change-new-title:visible, #mail-change-new-content:visible');
        var check_is_necessary = false;
        if(selected_process_data != "5" && $(this).find('[data-valuefor="vulsync"]').length > 0){
            var target_keywords = all_keywords['vulsync-info-match'];
            check_is_necessary = true;
        }else if(selected_process_data == process_values_list[0]){
            var target_keywords = all_keywords['prod-info-match'].concat(all_keywords['host-info-match']);
            check_is_necessary = true;
        }else if(selected_process_data == process_values_list[1]){
            var target_keywords = all_keywords['vul-info-match'];
            check_is_necessary = true;
        }
        if(check_is_necessary){
            check_elms.each(function(){
                var elm = $(this);
                for(var i = 0; i < target_keywords.length; i++){
                    var keyword = target_keywords[i];
                    var sliced_keyword = keyword.slice(0, -1);
                    var regexp = new RegExp(keyword + '|' + sliced_keyword + ':C%' + '|' + sliced_keyword + ':C-[0-9]+%' + '|' + sliced_keyword + ':T[0-9]+%' + '|' + sliced_keyword + ':B[0-9]+%');
                    if(elm.val().match(regexp)){
                        var elm_id = elm.attr('id');
                        reset_elms.push('#' + elm_id);
                       if(elm_id === 'send-mail-title' || elm_id === 'mail-change-new-title'){
                            var reset_name = '「件名」';
                        }else{
                            var reset_name = '「メール本文」';
                        }   
                        reset_names.push(reset_name);
                    }
                }
            });
        }
        if(reset_elms.length > 0){
            var conf = confirm('この項目の選択に関連するキーワードが含まれているため、このまま変更すると' + reset_names.join(', ') + 'をリセットします。よろしいですか?');
            if(conf){
                $(reset_elms.join(', ')).val('');
            }else{
                $(this).val(before_changed);
                return false
            }
        }
        $('#selected-process-data-before').val($(this).val());
        if(process_values_list[2] == selected_process_data){
            keyword_btn.find('[data-itemid="%VULSYNC_RENEW_MATCH%"], [data-itemid="%VULSYNC_DEL_MATCH%"]').prop('disabled', true);
            keyword_btn.find('[data-itemid="%VULSYNC_REGI_MATCH%"]').prop('disabled', false);
        }else if(process_values_list[3] == selected_process_data ){
            keyword_btn.find('[data-itemid="%VULSYNC_REGI_MATCH%"], [data-itemid="%VULSYNC_DEL_MATCH%"]').prop('disabled', true);
            keyword_btn.find('[data-itemid="%VULSYNC_RENEW_MATCH%"]').prop('disabled', false);
        }else if(process_values_list[4] == selected_process_data){
            keyword_btn.find('[data-itemid="%VULSYNC_REGI_MATCH%"], [data-itemid="%VULSYNC_RENEW_MATCH%"]').prop('disabled', true);
            keyword_btn.find('[data-itemid="%VULSYNC_DEL_MATCH%"]').prop('disabled', false);
        }else if(selected_process_data == "5"){
            $('#mail-setting-keyword-btn:visible').find('[data-itemid="%VULSYNC_REGI_MATCH%"], [data-itemid="%VULSYNC_RENEW_MATCH%"], [data-itemid="%VULSYNC_DEL_MATCH%"]').prop('disabled', false);
        }
      //  toggle_process_data_related_keywords(keyword_btn, selected_process_data);
    });*/
   
    //「処理するデータ」の選択に関連するキーワードボタンの活性・非活性の切り替え 
   /* function toggle_process_data_related_keywords(keyword_btn, selected_process_data){
        var process_values_list = JSON.parse($('#process-values-list').val());
        if(selected_process_data == process_values_list[0]){
            var host_info_btn = keyword_btn.find('[data-menuname="host-info-match"], [data-menuname="prod-info-match"]');
            host_info_btn.addClass('disabled');
            var hots_info_wrap = host_info_btn.closest('.original-dropdown-item');
            hots_info_wrap.attr({'data-toggle': 'tooltip',
                                  'data-title': '「処理するデータ」で「' + process_values_with_names[process_values_list[1]] + '」を選択している場合に使用できるキーワードです'                                     }).tooltip('enable');
            var vul_info_btn = keyword_btn.find('[data-menuname="vul-info-match"]');
            vul_info_btn.removeClass('disabled');
            if(vul_info_btn.closest('.original-dropdown-item').attr('data-toggle') === 'tooltip'){
                vul_info_btn.closest('.original-dropdown-item').tooltip('disable');
            }
        }else if(selected_process_data == process_values_list[1]){
            var vul_info_btn = keyword_btn.find('[data-menuname="vul-info-match"]');
            vul_info_btn.addClass('disabled');
            var vul_info_wrap = vul_info_btn.closest('.original-dropdown-item');
            vul_info_wrap.attr({'data-toggle': 'tooltip',
                                'data-title': '「処理するデータ」で「' + process_values_with_names[process_values_list[0]] + '」を選択している場合に使用できるキーワードです'
                          }).tooltip('enable');
            var host_info_btn = keyword_btn.find('[data-menuname="host-info-match"], [data-menuname="prod-info-match"]');
            host_info_btn.removeClass('disabled');
            if(host_info_btn.closest('.original-dropdown-item').attr('data-toggle') === 'tooltip'){
                host_info_btn.closest('.original-dropdown-item').tooltip('disable');
            }
        }
    }*/

    const all_statuses = {
      '1': '未対策',
      '2': '対策中',
      '3': '対策済',
      '4': '影響無',
    };

    function make_statuses_menu_for_status_change(button){
        var trigger = button;
        var trigger_id = trigger.attr('id');
        var div = '<div class="ope-detail-wrap-statuses s-shadow-v7">';
        div += '<div class="d-flex justify-content-center">';
        for(status_id in all_statuses){
            var status_name = all_statuses[status_id];
            var button = '<button type="button" class="mr-1 text-nowrap btn btn-sm unselected-status" data-status="from" data-statusid="' + status_id + '">' + status_name + '</button>';
            div += button;
        }
        div += '</div>';
        var btns_field = '<div class="d-flex mt-2">';
        var caution = '<div class="float-left ml-2 font-size-smaller caution-select-statuses">※2つまで選択可</div>';
        btns_field += caution;
        var wrap_btns = '<div class="float-right ml-auto">';
        var cancel_btn = '<button type="button" class="btn btn-xs btn-secondary text-nowrap" id="ope-detail-cancel-select-status-from">キャンセル</button>';
        var ok_btn = '<button type="button" class="btn btn-xs btn-primary ml-1 text-nowrap" id="ope-detail-decide-status-from">決定</button>';
        wrap_btns += cancel_btn;
        wrap_btns += ok_btn;
        wrap_btns += '</div>';
        btns_field += wrap_btns;
        btns_field += '</div>';
        div += btns_field;
        div += '</div>';
        div = $(div);
        if(trigger_id === 'ope-detail-status-to'){
            div.find('[data-status]').attr('data-status', 'to');
            div.find('#ope-detail-cancel-select-status-from').attr('id', 'ope-detail-cancel-select-status-to');
            div.find('#ope-detail-decide-status-from').attr('id', 'ope-detail-decide-status-to');
            div.find('.caution-select-statuses').text('※１つのみ選択');
        }else if(trigger_id === 'edit-status-change-from-menu'){
            div.attr('class', 'edit-statuses-wrap-statuses s-shadow-v7');
            div.find('#ope-detail-cancel-select-status-from').attr('id', 'edit-status-change-cancel-select-status-from');
            div.find('#ope-detail-decide-status-from').attr('id', 'edit-status-change-decide-status-from');
        }else if(trigger_id === 'edit-status-change-to-menu'){
            div.attr('class', 'edit-statuses-wrap-statuses s-shadow-v7');
            div.find('[data-status]').attr('data-status', 'to');
            div.find('#ope-detail-cancel-select-status-from').attr('id', 'edit-status-change-cancel-select-status-to');
            div.find('#ope-detail-decide-status-from').attr('id', 'edit-status-change-decide-status-to');
            div.find('.caution-select-statuses').text('※１つのみ選択');
        }       
        trigger.prepend(div);

    }
  
    //ステータス変更対象の選択ステータス表示
    $(document).on('click', '#ope-detail-status-from', function(){
        if($('#ope-detail-decide-status-from').length === 0){
            make_statuses_menu_for_status_change($(this));
        }
    }); 

    $(document).on('click', '#ope-detail-status-to', function(){
        if($('#ope-detail-decide-status-to').length === 0){
            make_statuses_menu_for_status_change($(this));
        }
    });

   //変更対象ステータスの選択
   $(document).on('click', '#ope-detail-status-from [data-status=from]', function(){
       var button = $(this);
       color_clicked_status_from_btn('ope-detail-status-from', button);
   });

   function color_clicked_status_from_btn(box_id, button){
       var order = [];
       var selector = '#' + box_id;
       $(selector + ' [data-status=from]').each(function(){
           var btn = $(this);
           var num = btn.attr('data-order');
           if(num){
               order.push(parseInt(num));
           }
       });
       order.sort(function(a,b){
           if(a < b) return -1;
           if(a > b) return 1;
           return 0;
       });
       if(order.length > 1){
         $(selector + ' [data-order=' + order[0] + ']').removeClass('selected-status').removeAttr('data-order');
       }
       button.addClass('selected-status').attr('data-order', order.length === 0 ? 1 : order[order.length -1] + 1);
   }

   $(document).on('click', '#ope-detail-status-to [data-status=to]',function(){
       color_clicked_status_to_btn('ope-detail-status-to', $(this));
   });

   function color_clicked_status_to_btn(box_id, button){
       var selector = '#' + box_id;
       $(selector + ' [data-status=to]').removeClass('selected-status');
       button.addClass('selected-status');
   }
      
   $(document).on('click', '#ope-detail-decide-status-from, #ope-detail-decide-status-to', function(e){ 
       var button = $(this);
       var id = button.attr('id');
       var selected_status_names = [];
       var selected_status_ids = [];
       if(id === 'ope-detail-decide-status-from'){
           var target = $('#ope-detail-status-from');
       }else if(id === 'ope-detail-decide-status-to'){
           var target = $('#ope-detail-status-to');
       }
       target.find('.selected-status')
           .each(function(){
               var button = $(this);
               var status_id = button.data('statusid');
               var status_name = button.text();
               selected_status_names.push(status_name);
               selected_status_ids.push(status_id);
       });
       if(selected_status_names.length > 0 && selected_status_ids.length >0){
           target.attr('data-statusids', JSON.stringify(selected_status_ids))
                                       .text(selected_status_names.join(','));
       }
       button.closest('.ope-detail-wrap-statuses').remove();
       e.stopPropagation();
   });

   $(document).on('click', '#ope-detail-cancel-select-status-from, #ope-detail-cancel-select-status-to', function(e){
       $(this).closest('.ope-detail-wrap-statuses').remove();
       e.stopPropagation();
   });

   $(document).on('change', '#ope-detail-select-due-date-item', function(){
       var target = $(this).val();
       $('#ope-detail-due-date-change-area').find('[data-setting]').hide();
       $('#ope-detail-due-date-change-area').find('[data-setting=' + target + ']').show();
   });

   //syslogの送信先ホストの切り替え
   $(document).on('change', '#select-syslog-address', function(){
       if( $(this).val() === 'hostname' ){
           $('#syslog-ip-address-field').hide();
           $('#syslog-hostname-field').show();
       }else{
           $('#syslog-hostname-field').hide(); 
           $('#syslog-ip-address-field').show();
       }
   });

   //syslogのメッセージには改行は入れられないようにする
   //syslogのメッセージ入力でメッセージ変換後の表示をするポップアップ削除
   $(document).on('input', '#syslog-input-message', function(){
       $('#syslog-keyword-preview-popup').remove();
       var v = $(this).val();
       if( v.includes('\n') ){
           alert('メッセージに改行は使用できません。');
           $(this).val(v.replace('\n', ''));
           return false;
       }
   });

   //syslogに使用できないキーワードが含まれていないかチェック
   function check_syslog_message(message){
       var is_pass = true;
       var unusable_keywords = JSON.parse($('#syslog-unusable-keywords').val());
       unusable_keywords.forEach(function(kwd){
           if( message.includes(kwd) ){
               is_pass = false;
           }  
       });
       return is_pass;
   }

   //syslogの項目名と内容
   function create_syslog_name_and_content(syslog_data){
       var data = {}
       data['protocol'] = {'name': 'プロトコル', 'content': syslog_data['protocol'].toUpperCase()};
       data['address'] = {
           'name': '送信先', 
           'content': '送信先ホスト: ' + escape_html_special_chars(syslog_data['address']) + '<br>' + '送信先ポート番号: ' + syslog_data['port_number']
       };
       data['facility'] = {'name': 'ログの種別', 'content': syslog_data['facility']};
       data['severity'] = {'name': 'ログの深刻度', 'content': syslog_data['severity']};
       data['message'] = {
           'name': 'メッセージ情報', 
           'content':  'ホスト名: ' + escape_html_special_chars(syslog_data['hostname']) + '<br>' +
                       '識別子: ' + escape_html_special_chars(syslog_data['ident']) + '<br>' +
                       'メッセージ: ' + escape_html_special_chars(syslog_data['message'])
       }; 
       return data
   }

   //ホスト名チェック
   function validate_hostname(hostname, prefix){
       var max_hostname_byte_length = $('#max-hostname-byte-length').val();
       if(hostname.length === 0){
           return prefix + 'を入力して下さい。';
       }else if(!hostname.match(/^[\x20-\x7e]*$/)){
           return prefix + 'は半角英数字で入力して下さい。';
       }else if(hostname.match(/\s/)){
           return prefix + 'に空白文字は使用できません。';
       }else if(!hostname.match(/^[0-9a-zA-Z._-]*$/)){
           return prefix + 'に使用できない文字が含まれています。';
       }else if(encodeURIComponent(hostname).replace(/%../g,"x").length > max_hostname_byte_length){
           return (prefix + 'に' + max_hostname_byte_length + 'バイトを超えるホスト名は登録できません。');
       }
       return null;
   }
   //syslogのparameter作成
   function create_syslog_param(){
       var action_target = $('[data-id="select-process-data"]:visible option:selected').val();
       var protocol = $('#select-syslog-protocol').val();
       var address_hostname = $('#syslog-input-hostname:visible').val();//address_hostnameとip_addressのどちらかはundefinedになる
       var ip_address = $('#syslog-input-ip-address:visible').val();
       var address = (address_hostname === undefined ? ip_address :address_hostname);
       var port_number = $('#syslog-port-number').val();
       var facility = $('#select-syslog-facility').val();
       var severity = $('#select-syslog-severity').val();
       var hostname = $('#syslog-input-message-hostname').val();
       var ident = $('#syslog-input-ident').val();
       var message = $('#syslog-input-message').val();
       var alert_msgs = [];
       if( port_number.length === 0 ){
           alert_msgs.push('送信先ポート番号を入力して下さい。');
           $('#syslog-port-number').val('');//type=numberでも半角英字のeが打ち込める(valueとしては解釈されない)。このため、数>字でない場合はvalueを空にする
       }else if( !port_number.match(/[0-9]/) ){
           alert_msgs.push('送信先ポート番号は数字で入力して下さい。');
       }else if( port_number < 0 || port_number > 65535 ){
           alert_msgs.push('送信先ポート番号は0以上65535以下の数値を入力して下さい。');
       }
       if( address.length === 0 ){
           alert_msgs.push('送信先ホストを入力して下さい。');
       }else{
           if( address_hostname ) {
               var err_message = validate_hostname(address_hostname, '送信先ホスト');
               if( err_message ){
                   alert_msgs.push(err_message);
               }
           }else{
               if( !ip_address.match(/^((25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.){3}(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])$/) ){
                   alert_msgs.push('IPアドレスは、n.n.n.n(nは0-255)で入力して下さい。');
               }    
           }
       }
       var hostname_err_message = validate_hostname(hostname, 'ホスト名');
       if( hostname_err_message ){
           alert_msgs.push(hostname_err_message);
       }
       if( ident.length === 0 ){
           alert_msgs.push('識別子を入力して下さい。');
       }
       if( message.length === 0 ){
           alert_msgs.push('メッセージを入力して下さい。');
       }else{
           var is_message_pass = check_syslog_message(message);
           if( !is_message_pass ){
                alert_msgs.push('メッセージに使用できないキーワードが含まれています。');
           }
       }
       if( alert_msgs.length > 0 ){
           alert(alert_msgs.join('\n'));
           return false;
       }
       var data = {
           'protocol'        : protocol,
           'address'         : (address_hostname === undefined ? ip_address : address_hostname),
           'selected_address': (address_hostname === undefined ? 'ip_address' : 'hostname'),
           'port_number'     : port_number,
           'facility'        : facility,
           'severity'        : severity,
           'hostname'        : hostname,
           'ident'           : ident,
           'message'         : message,
           'target'          : String(action_target),
           'proc_empty'      : 'false'
       };
       return data;
   }

   //操作設定modalへappend
   function append_modalRegisterOperation_items(items, selecting_item, is_update, is_edit){
       var div_name = items[0];
       var div_ope = items[1];
       var tr = $('<tr>').attr({
                         'class': '',
                         'data-addope': selecting_item,
                         });
       var td1 = $('<td>').attr('class', 'set-operation-t-left');
       td1.append(div_name);
       tr.append(td1);
       var td2 = $('<td>').attr({
                          'class': 'set-operation-t-mid g-line-height-1_3',
                          'data-item': 'ope-content',
                          });
       td2.append(div_ope);
       tr.append(td2);
       var td3 = $('<td>').attr('class', 'set-operation-t-right');
       var btns_field = $('<div>');
       var edit_btn = $('<button>').attr({
                                   'type':'button',
                                   'class': 'btn btn-xs btn-edit edit-ope mr-lg-1 mr-sm-0',
                                   'data-opedetail': selecting_item,
                                   }).text('変更');
       btns_field.append(edit_btn);
       var del_btn = $('<button>').attr({
                                  'type':'button',
                                  'class': 'btn btn-xs btn-danger del-ope',
                                  'data-opedetail': selecting_item,
                                  }).text('削除');
       btns_field.append(del_btn);
       td3.append(btns_field);
       tr.append(td3); 
       if($('#show-no-ope').length === 1){
           $('#show-no-ope').remove();
       }
       $('#set-operation-tbody').append(tr);
       var tbody = $('#set-operation-tbody').get(0);
       tbody.scrollTop = tbody.scrollHeight;
       $('#execute-set-ope').prop('disabled', false);
       $('[data-info="overwrite-dudate"]').tooltip();
       reset_add_operation_area();
       var tbody = $('#modalRegisterOperation-table').get(0);
       tbody.scrollTop = tbody.scrollHeight;
       $('#modalSelectItems').find('.modal-body *').remove();
       if(is_update != 'prepare_edit'){
           if(selecting_item === 'send-syslog'){
               setTimeout(function(){ append_original_tooltip('ope', is_update, is_edit)}, 300);
           }else{
               setTimeout(function(){ append_original_tooltip('ope', is_update, is_edit)}, 100);
           }
       }
   }
 
   //操作設定modalへ追加もしくは変更内容を作成
   function update_modalRegisterOperation_items(selecting_item, param, is_update, is_edit){
       var data = param[selecting_item];
       $('#set-operation-tbody').find('[data-addope=' + selecting_item + ']').remove();
       var item_name = ope_detail_items_name[selecting_item];
       var div_name = $('<div>').attr('class', 'set-ope-name').text(item_name);
       var div_ope = $('<div>');
       var action_target = data['target'];
       var action_flg = data['proc_empty'];
       var mail_users_req = null;
       if(selecting_item === 'send-mail-user' || selecting_item === 'send-mail-category'){
           if(action_flg === 'false'){
               var process_data_info = '※データの指定有り: 「' + process_values_with_names[action_target] + '」が無い場合、メールを送信しない';
           }else{
               var process_data_info = 'データの指定無し';
           }
           div_ope.attr('class', 'mail-contents-field');
           var sendby = data['sendby'];
           var title = data['title'];
           var content = data['content'];
           var sender = data['sender'];
           if(sendby['ids']){
               var ids = sendby['ids'];
               var path = $('#sidfm_internal_api_get_data_for_mail_setting_path').val();
               var mail_users_req = $.ajax({
                   type: 'POST',
                   url: path,
                   data: {'info_name': selecting_item, 'ids': JSON.stringify(ids)},
               }).done(function(data){ 
                   var sender_btn = $('<a>').attr({
                                            'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 mr-1 edit-mail-item show-original-popover',
                                            'data-item': 'sender',
                                            'data-content': sender['username'] ? sender['username'] : sender['address'] ,
                                            'data-mailsender': sender['username'] ? 'user' : 'address',
                                            'data-select': selecting_item,
                                            }).text('送信元');
                   var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
                   sender_btn.append(mark);
                   div_ope.append(sender_btn);
                   var names = [];
                   if( ids[0] === 'all' ){
                       names = [ '全体' ];
                   }else if( selecting_item === 'send-mail-category' ){
                       var category_descs = JSON.parse($('#category-descs').val());
                       var cat_descs = {}
                       Object.values(category_descs).forEach(function(cat){
                           cat_descs[cat.name] = cat.desc
                       });
                       names = data.map(function(name){
                           return "<span data-toggle='tooltip' data-title='" + escape_html_special_chars(cat_descs[name] ? cat_descs[name] : '') + "' data-placement='top' data-boundary='window'>" + escape_html_special_chars(name) + '</span>';
                       });
                   }else{
                       names = data.map(function(name){ return escape_html_special_chars(name) });
                   }
                   var address = $('<a>').attr({
                                         'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 mr-1 edit-mail-item show-original-popover', 
                                         'data-item': 'address',
                                         'data-content': names.join(', '),
                                         'data-mailids': ids,
                                         'data-html': 'true',
                                         'data-select': selecting_item,
                                         }).html('送信先');
                   var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
                   address.append(mark);
                   div_ope.append(address);
                   var title_btn = $('<a>').attr({
                                           'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 edit-mail-item mr-1 show-original-popover',
                                           'data-item': 'title',
                                           'data-content': title,
                                           'data-mailtitle': 'true',
                                           'data-select': selecting_item,
                                           }).text('件名');
                   var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
                   title_btn.append(mark);
                   div_ope.append(title_btn);
                   var content_btn = $('<a>').attr({
                                             'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 edit-mail-item mr-1 show-original-popover',
                                             'data-item': 'message',
                                             'data-content': content,
                                             'data-mailcontent': 'true',
                                             'data-select': selecting_item,
                                             }).text('メール本文');
                   var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
                   content_btn.append(mark);
                   div_ope.append(content_btn);
                   var process_data_btn = $('<a>').attr({
                                                'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 edit-mail-item show-original-popover',
                                                'data-processinfo': 'true',  
                                                'data-item': 'processinfo',
                                                'data-select': selecting_item,
                                                'data-content': process_data_info,
                                           }).text('対象データ');
                   var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
                   process_data_btn.append(mark);
                   div_ope.append(process_data_btn);
              });      
           }else{
               var addresses = sendby['address'];
               var sender_btn = $('<a>').attr({
                                        'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 mr-1 edit-mail-item show-original-popover',
                                        'data-item': 'sender',
                                        'data-content': sender['username'] ? sender['username'] : sender['address'] ,
                                        'data-mailsender': sender['username'] ? 'user' : 'address',
                                        'data-select': selecting_item,
                                        }).text('送信元');
               var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
               sender_btn.append(mark);
               div_ope.append(sender_btn);
               var address = $('<a>').attr({
                                     'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 mr-1 edit-mail-item show-original-popover',
                                     'data-item': 'address',
                                     'data-content': addresses.join(','), 
                                     'data-mailaddresses': 'true',
                                     'data-select': selecting_item,
                                     }).html('送信先');
               var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
               address.append(mark);
               div_ope.append(address);
               var title_btn = $('<a>').attr({
                                       'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 edit-mail-item mr-1 show-original-popover',
                                       'data-item': 'title',
                                       'data-content': title,
                                       'data-mailtitle': 'true',
                                       'data-select': selecting_item,
                                       }).text('件名');
               var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
               title_btn.append(mark);
               div_ope.append(title_btn);
               var content_btn = $('<a>').attr({
                                         'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 edit-mail-item show-original-popover',
                                         'data-item': 'message',
                                         'data-content': content,
                                         'data-mailcontent': 'true',
                                         'data-select': selecting_item,
                                         }).text('メール本文');
               var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
               content_btn.append(mark);
               div_ope.append(content_btn);
               var process_data_btn = $('<a>').attr({
                                          'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 edit-mail-item show-original-popover',
                                          'data-processinfo': 'true',
                                          'data-item': 'processinfo',
                                          'data-select': selecting_item,
                                          'data-content': process_data_info,
                                      }).text('対象データ');
               var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
               process_data_btn.append(mark);
               div_ope.append(process_data_btn);
           }                      
       }else if(selecting_item === 'status-change'){
           var process_data_info = '処理するデータ: ' + process_values_with_names[action_target];
           var all_statuses_ids_names = {};
           var status_ids_from = data['status_ids_from'];
           var status_ids_to = data['status_ids_to'];
           var from_name = [];
           var to_name = [];
           for(var i = 0; i < status_ids_from.length; i++){
               var id = status_ids_from[i];
               var name = all_statuses[id];
               from_name.push(name);
           }
           for(var i = 0; i < status_ids_to.length; i++){
               var id = status_ids_to[i];
               var name = all_statuses[id];
               to_name.push(name);
           }
           var content = $('<div>').html(
                             '<span class="font-weight-bold edit-target">' + from_name.join(',') + '</span>' +
                             'の場合 ' + '<i class="fas fa-arrow-right text-danger"></i>' +
                             '<span class="font-weight-bold edit-target">' + to_name.join(',') + '</span>' +
                             '<span>に変更</span>' +
                             '<a class="btn btn-xs btn-outline-dark2 g-font-size-12 show-original-popover ml-1"' +
                             ' data-processinfo="true"' +
                             ' data-content="' + process_data_info + '"' +
                             ' data-item="processinfo"' +
                             ' data-select="' + selecting_item + '">' +
                             '対象データ<i class="ml-1 fas fa-caret-up edit-mail-item-caret-up"></i></a>'
                         );
           div_ope.append(content); 
       }else if(selecting_item === 'due-date-change'){
           var process_data_info = '処理するデータ: ' + process_values_with_names[action_target];
           var content = $('<div>');
           var is_overwrite = data['permit-update'] === 'true' ? '上書き' : '上書きしない';
           if(data['days']){
               var days = data['days'];
               content.attr({
                      'data-days' : days,
                      }).html(
                              '対策予定日を脆弱性検出日の' + 
                              '<span class="font-weight-bold edit-target">' + days + '</span>' + 
                              '<span>日後に設定する</span>' +
                              '<i class="fas fa-info-circle ml-1 c_gray_70" data-info="overwrite-dudate" data-toggle="tooltip" data-placement="right" data-title="上書き設定: ' +
                              is_overwrite + '"></i>' +
                              '<a class="btn btn-xs btn-outline-dark2 g-font-size-12 show-original-popover ml-1"' +
                              ' data-processinfo="true"' +
                              ' data-content="' + process_data_info + '"' +
                              ' data-item="processinfo"' +
                              ' data-select="' + selecting_item + '">' +
                              '対象データ<i class="ml-1 fas fa-caret-up edit-mail-item-caret-up"></i></a>'
                      );  
           }else if(data['date']){
               date = data['date'];
               content.attr({
                      'data-date': date,
                      }).html(
                              '対策予定日を' +
                              '<span class="font-weight-bold edit-target">' + date + '</span>' +
                              '<span>に設定する</span>' +
                              '<i class="fas fa-info-circle ml-1 c_gray_70" data-info="overwrite-dudate" data-toggle="tooltip" data-placement="right" data-title="上書き設定: ' +
                              is_overwrite + '"></i>' +
                              '<a class="btn btn-xs btn-outline-dark2 g-font-size-12 show-original-popover ml-1"' +
                              ' data-processinfo="true"' +
                              ' data-content="' + process_data_info + '"' +
                              ' data-item="processinfo"' +
                              ' data-select="' + selecting_item + '">' +
                              '対象データ<i class="ml-1 fas fa-caret-up edit-mail-item-caret-up"></i></a>'
                      );
           }
           div_ope.append(content);
       }else if(selecting_item === 'send-syslog'){
           var syslog_text = create_syslog_name_and_content(data);
           var content = $('<div>');
           var protocol = $('<a>').attr({
                              'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 mr-1 show-original-popover',
                              'data-item': 'protocol',
                              'data-content': syslog_text['protocol']['content'],
                              'data-select': selecting_item,
                          }).text(syslog_text['protocol']['name']); 
           var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
           protocol.append(mark);
           div_ope.append(protocol);
           var address_info = $('<a>').attr({
                                  'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 mr-1 show-original-popover',
                                  'data-item': 'address',
                                  'data-content': syslog_text['address']['content'],
                                  'data-select': selecting_item,
                                  'data-html': 'true'
                              }).text(syslog_text['address']['name']);
           var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
           address_info.append(mark);
           div_ope.append(address_info);
           var facility = $('<a>').attr({
                              'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 mr-1 show-original-popover',
                              'data-item': 'facility',
                              'data-content': syslog_text['facility']['content'],
                              'data-select': selecting_item,
                          }).text(syslog_text['facility']['name']);
           var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
           facility.append(mark);
           div_ope.append(facility);
           var severity = $('<a>').attr({
                              'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 mr-1 show-original-popover',
                              'data-item': 'severity',
                              'data-content': syslog_text['severity']['content'],
                              'data-select': selecting_item,
                          }).text(syslog_text['severity']['name']);
           var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
           severity.append(mark);
           div_ope.append(severity);
           var message_info = $('<a>').attr({
                                  'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 show-original-popover',
                                  'data-item': 'message_info',
                                  'data-content': syslog_text['message']['content'],
                                  'data-select': selecting_item,
                                  'data-html': 'true'
                              }).text(syslog_text['message']['name']);
           var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
           message_info.append(mark);
           div_ope.append(message_info);
           var process_data_info = '処理するデータ: ' + process_values_with_names[action_target]; 
           var process_data_btn = $('<a>').attr({
                                      'class': 'btn btn-xs btn-outline-dark2 g-font-size-12 show-original-popover',
                                      'data-processinfo': 'true',
                                      'data-item': 'processinfo',
                                      'data-select': selecting_item,
                                      'data-content': process_data_info,
                                   }).text('対象データ');
           var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
           process_data_btn.append(mark);
           content.append(protocol, address_info, facility, severity, message_info, process_data_btn);
           div_ope.append(content);
       }else if(selecting_item === 'set-ctag'){
           const empty_message = '選択されたカスタムタグはありません';
           const enables = data.enable.map(ctag => {
               return createSelectedCtagElm(ctag);
           });
           const not_enables = data.not_enable.map(ctag => {
               return createSelectedCtagElm(ctag);
           });
           const prosessInfo = '処理するデータ: ' + process_values_with_names[action_target];
           const processBtn = $('<a>').attr({
               'class': 'ml-2 btn btn-xs btn-outline-dark2 g-font-size-12 show-original-popover',
               'data-processinfo': 'true',
               'data-item': 'processinfo',
               'data-select': selecting_item,
               'data-content': prosessInfo,
           }).text('対象データ');
           var mark = $('<i>').attr('class', 'ml-1 fas fa-caret-up edit-mail-item-caret-up');
           processBtn.append(mark);
           var true_btn = $('<a>').attr({
               'class': 'mr-2 btn btn-xs show-content-btn show-content text-white text-nowrap show-original-popover btn-success',
               'data-item': selecting_item + '-true',
               'data-select': selecting_item + '-true',
               'data-content': (enables.length ? $('<div>').append(enables).html() : empty_message), //要素を文字列化
               'data-html': 'true'
           }).text('設定する');
           var false_btn = $('<a>').attr({
               'class': 'btn btn-xs show-content-btn show-content text-white text-nowrap show-original-popover btn-light-orange',
               'data-item': selecting_item + '-false',
               'data-select': selecting_item + '-false',
               'data-content': (not_enables.length ? $('<div>').append(not_enables).html() : empty_message), //要素を文字列化
               'data-html': 'true'
           }).text('除外する');
           div_ope.append(true_btn, false_btn, processBtn);
       }
       var items = [div_name, div_ope];
       if( mail_users_req ){
           $.when(mail_users_req).done(function(){
               append_modalRegisterOperation_items(items, selecting_item, is_update, is_edit);
           });
       }else{
           append_modalRegisterOperation_items(items, selecting_item, is_update, is_edit);
       }                    
   }

    //操作追加modalで追加ボタンを押した時
    $('#execute-add-ope').on('click', function(){
        var selecting_item = $('#selecting-ope-item').attr('data-selecting');
        var param = JSON.parse($('#set-ope-param').val());
        if(selecting_item ===  'send-mail-user' || selecting_item === 'send-mail-category'){
            var send_by = $('#ope-detail-address-name').attr('data-sendby');
            if(send_by === 'user' || send_by === 'category'){
                var tbody = $('#clone-search-address-items-area-tbody');
            }else{
                var tbody = $('#clone-add-address-area-tbody');
            }
            var sender = $('#ope-detail-readonly-sender');
            var sender_val = sender.val();
            var title = $('#ope-detail-readonly-mail-title').val();
            var content = $('#ope-detail-readonly-mail-content').val();
            var target_data = $('#mail-target-data').val();
            if(sender_val != '' && tbody.length === 1 && title != '' && content != '' && target_data != ''){
                var selected_ids = [];
                tbody.find('[data-addresstype]').each(
                    function(){
                        var item = $(this);
                        var item_id = item.data('addresstypeid');
                        selected_ids.push(item_id);
                });
                var added_address = [];
                tbody.find('[data-address]').each(function(){
                    added_address.push($(this).text());
                });
                if(sender.attr('data-sender') === 'user'){
                    var sender_info = {'userid': sender.attr('data-userid'), 'username': sender_val}
                }else{
                    var sender_info = {'address': sender_val}
                }
                var action_flg = $('#mail-setting-process-data-radio-no-execute').prop('checked') ? 'false' : 'true';
                if(action_flg === 'false'){
                    var action_target = $('#mail-select-process-data-area').find('[data-id="select-process-data"] option:selected').val();
                }else{
                    var action_target = '0';
                }
                if(selected_ids.length === 0 && added_address.length === 0){
                    alert('入力されていない項目があります');
                    return false;
                }
                var data = {
                    'sendby': selected_ids.length > 0 ? {'ids': selected_ids} : {'address': added_address},
                    'title' : title,
                    'content': content,
                    'sender': sender_info,
                    'target': String(action_target),
                    'proc_empty': action_flg,
                }
            }else{
                alert('入力または選択されていない項目があります');
                return false;
            } 
        }else if(selecting_item === 'status-change'){
            var target_status_ids = $('#ope-detail-status-from').data('statusids');
            var change_to_status_id = $('#ope-detail-status-to').data('statusids');
            if(!target_status_ids){
                alert('変更対象のステータスを選択して下さい');
                return false;
            }
            if(!change_to_status_id){
                alert('変更後のステータスを選択して下さい');
                return false;
            }
            var action_target = $('#ope-detail-body').find('[data-id="select-process-data"] option:selected').val();
            var data = {
                'status_ids_from': target_status_ids,
                'status_ids_to'  : change_to_status_id,
                'target'         : String(action_target),
                'proc_empty'     : 'false'
            }
        }else if(selecting_item === 'due-date-change'){
            var action_target = $('#ope-detail-body').find('[data-id="select-process-data"] option:selected').val();
            var is_overwrite = String($('#ope-detail-checkbox-is-overwrite-duedate').prop('checked'));
            if($('#ope-detail-due-date-days').is(':visible')){ 
                var days = $('#ope-detail-due-date-days').val();
                if(days === ''){
                    alert('日数を入力して下さい');
                    return false;
                }else if(String(days).match(/[^0-9]+/)){
                    alert('半角数字で入力してください');
                    return false;
                }
                var data = {
                    'days': days, 
                    'permit-update': is_overwrite,
                    'target'         : String(action_target),
                    'proc_empty'     : 'false'
                };
            }else if($('#ope-detail-due-date-to').is(':visible')){
                var date = $('#ope-detail-due-date-to').val();
                if(date === ''){
                    alert('日付を入力して下さい');
                    return false;
                }
                alert('設定する日付を超過後も同じ日付が設定され続けます。日付を超過後は日付の設定を変更して下さい。');
                var data = {
                    'date': date,
                    'permit-update': is_overwrite,
                    'target'       : String(action_target),
                    'proc_empty'   : 'false'
                };
            }
        }else if(selecting_item === 'send-syslog'){
            var data = create_syslog_param();
            if( !data ){
                return false;
            }
        }else if(selecting_item === 'set-ctag'){
            var selected_ctags = JSON.parse(ruleSelectedRegistCtags.val());
            if( !Object.keys(selected_ctags).length || (!selected_ctags.enable.length && !selected_ctags.not_enable.length) ){
                alert('カスタムタグが選択されておりません。');
                return;
            }
            const selected_data_target = $('[data-id="select-process-data"]:visible option:selected').val();
            if( selected_data_target == 0 ){
                alert(`処理するデータで、「${process_values_with_names[selected_data_target]}」が選択されています。\nこの場合、条件に一致する全ての脆弱性に対してカスタムタグが設定・除外されますので、ご注意ください。`);
            }else if( selected_data_target == 1 ){
                alert(`処理するデータで、「${process_values_with_names[selected_data_target]}」が選択されています。\nこの場合、条件に一致する全てのホストの、全ての脆弱性に対してカスタムタグが設定・除外されますので、ご注意ください。`);
            }
            var data = {
               target: String(selected_data_target),
               proc_empty: 'false',
               ...selected_ctags
            }
        }
        var is_update = $('#modalRegisterOperation-table').find('[data-opedetail=' + selecting_item + ']').length > 0;
        if(is_update){
            var conf = confirm('既に同じ操作が存在します。上書きしますか?');
            if(!conf){
                return false;
            }
        }
        param[selecting_item] = data;   
        $('#set-ope-param').val(JSON.stringify(param));
        reset_add_operation_area();
        var update = update_modalRegisterOperation_items(selecting_item, param, is_update);
        return update;
    });

    //操作設定modalでメールのボタン・条件設定modalで対象ホストボタンを押したとき、内容を表示する
    $(document).on('click', '.show-original-popover, .triangle', function(){
            var button = $(this);
            var detail = button.attr('data-item');
            var item = button.attr('data-select');
            var popover = $('.rule-setting-original-popover[data-select=' + item + ']');
            var popover_item_name = popover.attr('data-itemname');
            $('[data-toggle="tooltip"]').tooltip('hide');
            popover.remove();
            if(popover_item_name === detail){
                return false;
            }
            var content = $('<div>').attr({
                                    'class': 'rule-setting-original-popover',
                                    'data-select': item,
                                    'data-itemname': detail,
                                    });
            var inner_content = $('<div>').attr('class', 'original-popover-inner');
            var text = button.attr('data-content');
            button.attr('data-html') ? inner_content.html(text) : inner_content.text(text);
            content.append(inner_content);
            var triangle_target = $('<div>').attr({
                                             'class': 'position-relative triangle',
                                             'data-select': item,
                                             'data-itemname': detail,
                                             });
            if(button.attr('class').indexOf('rule-detail-info') != -1){
                content.addClass('rule-detail-popover');
                inner_content.addClass('rule-detail-popover-inner');
                triangle_target.addClass('rule-detail-popover-triangle');
            }
            content.append(triangle_target);
            content.css('opacity', 0);
            $('body').append(content);
            var btn_width = button.outerWidth(true);
            var content_width = content.outerWidth(true);
            var offset = button.offset();
            var offset_top = offset.top - content.outerHeight(true) - 5;
            var offset_left = offset.left + (btn_width / 2) - (content_width / 2);
            content.offset({
                'top': offset_top,
                'left': offset_left
            }); 
            content.css({
                'opacity': ''
            });
            content.find('[data-toggle="tooltip"]').tooltip();
            $('.original-popover-inner:not(.rule-detail-popover-inner').on('scroll', function(){
                $('.rule-detail-popover').remove();
            });
    });

    //画面リサイズでtooltip・popover消去・条件設定modalの条件詳細欄・操作設定詳細欄のstyleが壊れないようにする
    $(window).on('resize',function(){
        $('[data-toggle="tooltip"]').tooltip('hide');
        $('.rule-setting-original-popover').remove();
        $('.rule-detail-popup').remove();
        $('.original-tooltip').remove();
        $('#keep-margin-for-cond:visible').height($('#execute-add-cond').outerHeight(true) + 5);
        $('#keep-margin-for-ope:visible').height($('#execute-add-ope').outerHeight(true) + 5);
        $('#selected-cvss-display').remove();
    });

    //画面のスクロール時にpopoverの位置が壊れないようにする
    $(window).on('scroll', function(){
        $('[data-toggle="tooltip"]').tooltip('hide');
        if( $('.modal:visible').length ){
            $('.rule-setting-original-popover').each(function(){
                var elm = $(this);
                var target_name = elm.attr('data-select');
                var item_name = elm.attr('data-itemname');
                var target = $('[data-select=' + target_name + ']');
                if(target.length > 0){
                    var offset_top = target.offset().top - elm.outerHeight(true) - 5;
                    elm.offset({'top': offset_top});
                }
            });
        }else{
            $('.rule-setting-original-popover').remove();
        }
    });
   
    $(document).on('scroll', function(){
        if( $('#modalRegisterCondition-table:visible, #modalRegisterOperation-table:visible').length > 0 ){
            $('.rule-setting-original-popover').remove();
            $('#selected-cvss-display').remove();
        } 
    });
  
    //画面クリックでpopover消去・dropdown消去
    $(document).on('click', function(e){
        if(e.target){
            var t_elm = $(e.target);
            if(!t_elm.hasClass('show-original-popover') && t_elm.closest('.show-original-popover').length === 0 && !t_elm.hasClass('triangle')){
                $('[data-toggle="tooltip"]').tooltip('hide');
                $('.rule-setting-original-popover').remove();
                $('.rule-detail-popup').remove();
            }
            if(t_elm.attr('data-originaldropdown') != 'true' && t_elm.closest('#keyword').length === 0){
                $('#keyword').find('.keywords-menu').hide();
                if( !$('#keyword-main-menu').is(':visible') ){
                    $('#keyword-main-menu').tooltip('hide');
                }
            }
            if( t_elm.attr('id') != 'show-cvss-selected-vectors' && t_elm.attr('id') != 'show-cvss-selected-vectors-caret' ){
                $('#selected-cvss-display').remove();
            }
        }
    });

    $('#mail-preset-list').on('show.bs.dropdown',function(){
        $('#keyword').find('.keywords-menu').hide();
    });

   
    //条件設定画面のツールチップ削除
    $(document).on('click', '#close-modalRegisterCondition-btn, #cancel-modalRegisterCondition-btn, .cond-edit-btn, .cond-del-btn', function(){
        $('.original-tooltip').remove();
    });
   
    //操作設定modalで変更ボタンを押した時、対象がステータス変更だった場合のfunction
    function make_edit_menu_for_status_change(button){
          var tr = button.closest('[data-addope]');
          var before = $('<div>').attr('class', 'pl-3 status-change-edit-before')
                                 .html(
                                      tr.find('[data-item="ope-content"]').html()
                                 );
          $('#before-ope-edit-content').append(before);
          var after = $('<div>').attr('class', 'status-change-edit-after');
          var div1 = $('<div>').attr('class','')
                               .html('<div class="g-font-size-14 g-text-ul text-left p-2 mb-2">脆弱性のステータスが</div>' + 
                                     '<div class="d-flex g-font-size-13 justify-content-center">' +
                                     '<button type="button" ' + 
                                     'class="ml-1 mr-1 btn btn-light g-font-size-13 position-relative" ' +
                                     'id="edit-status-change-from-menu">' + 
                                     '選択して下さい' + 
                                     '<i class="ml-1 fas fa-caret-down"></i>' + 
                                     '</button>' +
                                     '<div class="pt-2">のものを</div></div>'
                               );
          var div2 = $('<div>').attr('class', 'text-center');
          var arrow = $('<i>').attr('class','far fa-arrow-alt-circle-down fa-2x my-3');
          div2.append(arrow);
          var div3 = $('<div>').attr('class', '')
                               .html('<div class="d-flex g-font-size-13 justify-content-center">' +
                                     '<button type="button" ' +
                                     'class="ml-3 mr-1 btn btn-light g-font-size-13 position-relative" ' +
                                     'id="edit-status-change-to-menu">選択して下さい' +
                                     '<i class="ml-1 fas fa-caret-down"></i>' +
                                     '</button>' +
                                     '<div class="pt-2">に変更する</div></div>'
                               );
          var div4 = $('<hr>').attr('class', 'my-3 hr-dotted');
          var div5 = $('<div>').attr('class', 'mx-2 mb-1');
          var process_data_select = create_process_data_select_elm();
          process_data_select.removeClass('w-auto');
          div5.append(
              $('<div>').append(
                $('<span>').attr('class', 'font-weight-bold').text('この「操作」で処理する、対象となるデータ'),
                $('<span>').attr('class', 'd-block text-danger g-font-size-11 g-mt-2 mb-1').text('※当該対象データは別途指定しているる「条件」に一致するもののみとなります')
              ),
              process_data_select
          )
          var divider = $('<div>').attr('class', '');
          after.append([divider, div1, div2, div3, div4, div5]);
          $('#after-ope-edit-content').append(after);
          $('#after-ope-edit-content').attr('data-select', button.attr('data-opedetail'));
          $('#modalEditOperation').modal('show'); 
    }
    
    //操作設定modalでステータスの選択
    $(document).on('click', '#edit-status-change-from-menu', function(){
        if($('#edit-status-change-decide-status-from').length === 0){
            make_statuses_menu_for_status_change($(this));
        }
    });

    $(document).on('click', '#edit-status-change-to-menu', function(){
        if($('#edit-status-change-decide-status-to').length === 0){
            make_statuses_menu_for_status_change($(this));
        }
    });

    //操作設定modalでステータスボタンを押した時
    $(document).on('click', '#after-ope-edit-content [data-status=from]',function(){
        color_clicked_status_from_btn('after-ope-edit-content', $(this));
    });

    $(document).on('click', '#after-ope-edit-content [data-status=to]',function(){
        color_clicked_status_to_btn('after-ope-edit-content', $(this));
    });

    //操作設定modalでステータス決定するとき
    $(document).on('click', '#edit-status-change-decide-status-from, #edit-status-change-decide-status-to',function(e){
        var btn_id = $(this).attr('id');
        var selected = $(this).closest('.edit-statuses-wrap-statuses').find('.selected-status');
        if(selected){
            var status_names = [];
            var status_ids = [];
            selected.each(function(){
                var elm = $(this);
                var status_id = elm.attr('data-statusid');
                var status_name = elm.text();
                status_names.push(status_name);
                status_ids.push(status_id);
            });
            var target_menu = btn_id === 'edit-status-change-decide-status-from' ? '#edit-status-change-from-menu' : '#edit-status-change-to-menu';
            $(target_menu).attr('data-statusids', JSON.stringify(status_ids))
                          .text(status_names.join(','));
        }
        e.stopPropagation();
                    
    });

    //操作設定modalでステータス変更キャンセルのとき
    $(document).on('click', '#edit-status-change-cancel-select-status-from, #edit-status-change-cancel-select-status-to', function(e){
        $(this).closest('.edit-statuses-wrap-statuses').remove();
        e.stopPropagation();
    });    

    //操作設定modalで変更ボタンを押した時、対象が対策予定日変更だった場合のfunction 
    function make_edit_menu_for_due_date_change(button){
        var content = button.closest('[data-addope]').find('.set-operation-t-mid').html();
        var before = $('<div>').attr('class', 'pl-3').html(content);
        $('#before-ope-edit-content').append(before);
        var div1 = $('<div>').attr('class','g-font-size-14 g-text-ul text-left p-2')
                             .html('<div>脆弱性の対策予定日を</div>');
        var div2 = $('<div>').attr('class', 'd-flex g-font-size-13 justify-content-center')
                             .html('<div class="" data-setting="date">' +
                                   '<input type="text" class="mx-1 text-center detail-value w-50" id="edit-due-date-to" data-toggle="datepicker" readonly>' + 
                                   'に設定する' +
                                   '</div>' +
                                   '<div style="display:none;" class="" data-setting="days">' + 
                                   '脆弱性検出日の' + 
                                   '<input type="number" class="mx-1 g-width-5em detail-value" id="edit-due-date-days" min="0">' + 
                                   '日後に設定する' + 
                                   '</div>'
                               );
        var div3 = $('<div>').attr({
                             'id': 'due-date-change-due-date-change-area',
                             'class': 'text-right'
                             }) 
                             .html( 
                                   '<div class="g-font-size-13 d-inline-block pr-2">' +
                                   '<input id="due-date-change-checkbox-is-overwrite-duedate" type="checkbox" style="vertical-align: -0.15em;">' + 
                                   '<span class="ml-1">' + '対策予定日上書き' + '</span>' +
                                   '<span data-toggle="tooltip" data-placement="left" title="既に対策予定日が設定されている際に、対策予定日を上書きする場合はチェックを付けて下さい">' +
                                   '<i class="fas fa-question-circle c_gray_70"></i>' +
                                   '</span>' + 
                                   '</div>'
                             );
        var div4 = $('<hr>').attr('class', 'mt-2 mb-3 hr-dotted');
        var div5 = $('<div>').attr('class', 'mx-2 mb-1');
        var process_data_select = create_process_data_select_elm();
        process_data_select.removeClass('w-auto');
        div5.append(
            $('<div>').attr('class', '').append(
                $('<span>').attr('class', 'font-weight-bold').text('この「操作」で処理する、対象となるデータ'),
                $('<span>').attr('class', 'd-block text-danger g-font-size-11 g-mt-2 mb-1').text('※当該対象データは別途指定している「条件」に一致するもののみとなります')
            ),
            process_data_select
        )
        var items = [div1, div2, div3, div4, div5];
        var after =  make_operation_detail_area(items, 'due-date-change-update');
        $('#after-ope-edit-content').attr('data-select', button.attr('data-opedetail'));
        $('#after-ope-edit-content').append(after);
        $('#edit-due-date-to').datepicker();
        $('#modalEditOperation').modal('show');
    }

    //操作変更modalで対策予定日変更メニューの切り替え
    $(document).on('change', '#edit-select-due-date-item', function(){
       var target = $(this).val();
       $('#edit-due-date-change-area').find('[data-setting]').hide();
       $('#edit-due-date-change-area').find('[data-setting=' + target + ']').show();
   });

    //操作設定modalで変更ボタンを押した時、対象がsyslog送信だった場合のfunction
    function make_edit_menu_for_syslog(button){
        var param_name = button.attr('data-opedetail');
        var ope_param = JSON.parse($('#set-ope-param').val());
        var data = ope_param[param_name];
        var before = $('<div>');
        var syslog_text = create_syslog_name_and_content(data);
        var items = [ 
            '<span class="font-weight-bold mr-1 d-inline-flex g-width-8em">' + syslog_text['protocol']['name'] + ':</span>' + syslog_text['protocol']['content'],
            '<span class="font-weight-bold mr-1 d-inline-flex g-width-8em">' + syslog_text['address']['name'] + ':</span>' + syslog_text['address']['content'].replace(/<br>/g, ', '),
            '<span class="font-weight-bold mr-1 d-inline-flex g-width-8em">' + syslog_text['facility']['name'] + ':</span>' + syslog_text['facility']['content'],
            '<span class="font-weight-bold mr-1 d-inline-flex g-width-8em">' + syslog_text['severity']['name'] + ':</span>' + syslog_text['severity']['content'],
            '<span class="font-weight-bold mr-1 d-inline-flex g-width-8em">' + syslog_text['message']['name']  + ':</span>' + syslog_text['message']['content'].replace(/<br>/g, ', '),
            '<span class="font-weight-bold mr-1 d-inline-flex g-width-8em">' + '処理するデータ:</span>' + process_values_with_names[data['target']]
        ];
        before.html(items.join('<br>'));
        $('#before-ope-edit-content').append(before);
        var edit_items = ope_send_syslog();
        var after =  make_operation_detail_area(edit_items, 'send-syslog-update');
        after.find('#select-syslog-protocol').val(data['protocol']);
        if( data['selected_address'] === 'ip_address' ){
            after.find('#select-syslog-address').val('ip-address');
            after.find('#syslog-ip-address-field').show();
            after.find('#syslog-hostname-field').hide();
            after.find('#syslog-input-ip-address').val(data['address']);
        }else{
            after.find('#select-syslog-address').val('hostname');
            after.find('#syslog-ip-address-field').hide();
            after.find('#syslog-hostname-field').show();
            after.find('#syslog-input-hostname').val(data['address']);
        }
        after.find('#syslog-port-number').val(data['port_number']);
        after.find('#select-syslog-facility').val(data['facility']);
        after.find('#select-syslog-severity').val(data['severity']);
        after.find('#syslog-input-message-hostname').val(data['hostname']);
        after.find('#syslog-input-ident').val(data['ident']);
        after.find('#syslog-input-message').val(data['message']);
        after.find('[data-id="select-process-data"]').val(data['target']);
        $('#after-ope-edit-content').attr('data-select', param_name);
        $('#after-ope-edit-content').append(after);
        $('#modalEditOperation').modal('show');
    }

    //操作設定modalで変更ボタンを押した時、対象がメール送信だった場合のfunction
     function make_select_menu_for_edit_mail(button, item_name){
         var div = $('<div>').attr({
                             'id': 'edit-mail-select-menu',
                             'class': 'd-flex'
                             }); 
         var contents = $('<div>').attr('class', 'mail-contents-area');
         var target_area = button.closest('[data-addope]')
                                 .find('.mail-contents-field');
         var sender = target_area.find('[data-mailsender]');
         if(target_area.find('[data-mailids]').length === 1){
             var address_type = target_area.find('[data-mailids]');
         }else if(target_area.find('[data-mailaddresses]').length === 1){
             var address_type = target_area.find('[data-mailaddresses]');
         }
         var title = target_area.find('[data-mailtitle]');
         var message = target_area.find('[data-mailcontent]');
         var process_info = target_area.find('[data-processinfo]');
         var sender_btn = $('<a>').attr({
                                  'class': 'btn btn-sm btn-outline-dark2 mr-3 mb-1 edit-mail-item show-original-popover',
                                  'data-mailsender': sender.attr('data-mailsender'),
                                  'data-content': sender.attr('data-content'),
                                  'data-itemname': item_name,
                                  }).text('送信元');
         var address = $('<a>').attr({
                               'class': 'btn btn-sm btn-outline-dark2 mr-3 mb-1 edit-mail-item show-original-popover',
                               'data-content': address_type.attr('data-content'),
                               'data-itemname': item_name,
                               }).text('送信先');
         if(address_type.attr('data-mailids')){
             address.attr('data-mailids', address_type.attr('data-mailids'));
         }else if(address_type.attr('data-mailaddresses')){
             address.attr('data-mailaddresses', 'true');
         }
         var title_btn = $('<a>').attr({
                                 'class': 'btn btn-sm btn-outline-dark2 mb-1 edit-mail-item show-original-popover',
                                 'data-mailtitle': 'true',
                                 'data-content': title.attr('data-content'),
                                 'data-itemname': item_name,
                                 }).text('件名');
         var content_btn = $('<a>').attr({
                                   'class': 'btn btn-sm btn-outline-dark2 ml-3 mb-1 edit-mail-item show-original-popover',
                                   'data-mailcontent': 'true',
                                   'data-content': message.attr('data-content'), 
                                   'data-itemname': item_name,
                                  }).text('メール本文');
         var process_info_btn = $('<a>').attr({
                                    'class': 'btn btn-sm btn-outline-dark2 ml-3 mb-1 edit-mail-item show-original-popover',
                                    'data-mailprocess': 'true',
                                    'data-content': process_info.attr('data-content'),
                                    'data-itemname': item_name,
                                }).text('対象となるデータの指定');
         contents.append([sender_btn, address, title_btn, content_btn, process_info_btn]);
         div.append(contents);
         $('#modalEditOperation-body-main').hide();
         $('#modalEditOperation-edit-mail-menu').append(div);
         $('#modalEditOperation-edit-mail-menu').show();
         $('#modalEditOperation').modal('show');             
    }

    //カスタムタグの操作変更
    function make_edit_menu_for_set_ctag(button){
        const empty_message = '選択されたカスタムタグはありません';
        var param_name = button.attr('data-opedetail');
        var ope_param = JSON.parse($('#set-ope-param').val());
        var data = ope_param[param_name];
        var true_ctag_elms = data.enable.length?
            data.enable.map(ctag => createSelectedCtagElm(ctag) ) :
            [ $('<div>').text(empty_message) ]
        ;
        var false_ctag_elms = data.not_enable.length ?
            data.not_enable.map(ctag => createSelectedCtagElm(ctag) ) :
            [ $('<div>').text(empty_message) ]
        ;
        $('#before-ope-edit-content').append(
            $('<div>').append(
                $('<div>').attr('class', 'font-weight-bold').text('設定する'),
                $('<div>').attr('class', 'edit-before-ctag-content').append(true_ctag_elms),
            ),
            $('<div>').attr('class', 'edit-content-divider'),
            $('<div>').append(
                $('<div>').attr('class', 'font-weight-bold').text('設定しない'),
                $('<div>').attr('class', 'edit-before-ctag-content').append(false_ctag_elms),
            ),
            $('<div>').attr('class', 'mt-2 font-weight-bold')
                       .text('処理するデータ: ' + process_values_with_names[data['target']])
        );
        var after = ope_set_ctag();
        $('#after-ope-edit-content').attr('data-select', param_name);
        $('#after-ope-edit-content').append(after);
        $('#modalEditOperation').modal('show');
    }

    //操作設定modalで変更ボタンを押した時
    $(document).on('click', '#set-operation-tbody .edit-ope', function(){
        $('#modalEditOperation-body-main').show();
        $('#modalEditOperation-edit-mail-menu').hide();
        $('#before-ope-edit-content *').remove();
        $('#after-ope-edit-content *').remove();
        var button = $(this);
        var item_name = button.attr('data-opedetail');
        $('#modalEditOperation').find('.modal-title').text('操作変更: ' + ope_detail_items_name[item_name]);
        if(item_name === 'send-mail-user' || item_name === 'send-mail-category'){
            $('#modalEditOperation .modal-dialog').addClass('modal-lg');
            $('#edit-mail-select-menu').remove();
            $('#add-address-area-tbody > *').remove();
            $('#mail-setting-users-tbody > *').remove();
            make_select_menu_for_edit_mail(button, item_name);
        }else if(item_name === 'status-change'){
            make_edit_menu_for_status_change(button);
        }else if(item_name === 'due-date-change'){
            make_edit_menu_for_due_date_change(button);
        }else if(item_name === 'send-syslog'){
            make_edit_menu_for_syslog(button);
        }else if(item_name === 'set-ctag'){
            ruleSelectedCtags.val('{}'); //登録内容をリセット
            ruleSelectedRegistCtags.val('{}');
            make_edit_menu_for_set_ctag(button);
        }
        $('#modalEditOperation [data-toggle="tooltip"]').tooltip();
        reset_add_operation_area();
        $('.original-tooltip').remove(); 
    });

    //操作変更modalが開くとき
    $('#modalEditOperation').on('show.bs.modal',function(e){
        $('#modalSelectItems').find('.modal-body > *').remove();
        reset_add_operation_area();
        selected_textarea = null;
    });

    //操作変更modalが閉じた時
    $('#modalEditOperation').on('hidden.bs.modal',function(e){
        var button = $(e.relatedTarget);
        if(button.attr('data-item') === 'execute-ope-change'){
            append_tooltip_to_modalRegisterOperation('ope', true, true);
        }
    });

    //タイミングの設定によって使用可能なキーワードを活性・非活性にする
    function toggle_keyword_btns(keyword_button, trigger_param){
        //タイミングが定期実行・毎日実行の場合
        if(trigger_param && trigger_param['period'] != 'vulsync'){
            var elms = keyword_button.find('[data-menuname="vulsync-info-all"], [data-menuname="vulsync-info-match"]');
            elms.addClass('disabled');
            elms.closest('.original-dropdown-item').attr({
                'data-toggle': 'tooltip',
                'data-title': 'タイミング設定で「脆弱性同期時 実行」が選択されている時のみ使用可能なキーワードです'
            });
        }else{
            //タイミングが脆弱性同時期 実行の場合
            var elm_names = [
                '[data-menuname="host-info-all"]',
                '[data-menuname="vul-info-all"]',
                '[data-menuname="prod-info-all"]',
                '[data-menuname="host-info-match"]',
                '[data-menuname="vul-info-match"]',
                '[data-menuname="prod-info-match"]'
            ].join(', ');
            var elms = keyword_button.find(elm_names);
            elms.addClass('disabled');
            elms.closest('.original-dropdown-item').attr({
                'data-toggle': 'tooltip',
                'data-title': 'タイミング設定で「定期実行」または「毎日実行」が選択されている時のみ使用可能なキーワードです'
            });
        }
    }
    //操作変更modalでメールのメニューを選択した時
    $(document).on('click', '#edit-mail-select-menu .edit-mail-item', function(){
        var button = $(this);
        var item = button.text();
        var item_name = button.attr('data-itemname');
        var content = button.attr('data-content');
        $('#before-ope-edit-content > *').remove();
        $('#after-ope-edit-content > *').remove();
        var item_div = $('<div>').attr('class', 'font-weight-bold pt-1 pl-1').text(item + ':');
        var content_div = $('<div>').attr('class', 'mail-change-content').html((button.attr('data-mailids') ? content : escape_html_special_chars(content)).replace(/\n/g, '<br>'));
        $('#before-ope-edit-content').append([item_div, content_div]);
        $('#before-ope-edit-content [data-toggle="tooltip"]').tooltip();
        var item_div = $('<div>').attr('class', 'font-weight-bold text-left mt-1 pl-1 mb-2').text(item + ':');
        $('#after-ope-edit-content').append(item_div);
        if(button.attr('data-mailsender')){
            var select_sender_type = $('<select>').attr({
                                                  'class': 'form-control form-control-sm mb-0 g-font-size-13',
                                                  'id': 'mail-change-select-sender-type'
                                                  });
            var op1 = $('<option>').text('メールアドレスを入力').val('manual');
            var op2 = $('<option>').text('ユーザ選択').val('user-select');
            select_sender_type.append([op1, op2]);
            $('#after-ope-edit-content').append(
                                            $('<div>').attr('class', 'text-left')
                                                      .append(select_sender_type)
                                         ); 
            var wrap = $('<div>').attr({
                                 'class': 'w-100 position-relative',
                                 'id': 'mail-change-sender-select-area',
                                  }); 
            var wrap_area = $('<div>').attr('id', 'mail-change-sender-user-select-area').hide();
            var user_select_area = $('<div>').attr({
                                              'class': 'd-flex',
                                              });
            var input = $('<input>').attr({
                                    'type': 'text', 
                                    'class': 'form-control float-left mr-1 w-75 g-font-size-14',
                                    'id': 'mail-change-sender-select-input',
                                    'readonly': 'true',
                                     });   
            var search_user_btn = $('<div>').attr('class', 'position-relative float-right')
                                            .append(
                                                $('<button>').attr({ 
                                                             'type': 'button',
                                                             'class': 'btn btn-primary text-nowrap position-absolute bottom-0 left-0',
                                                             'id': 'mail-change-search-sender-user'
                                                }).text('ユーザ検索')
                                             );
            user_select_area.append([input, search_user_btn]);
            wrap_area.append(user_select_area);
            var input_manual = $('<input>').attr({
                                           'type': 'text',      
                                           'class': 'form-control g-font-size-14',      
                                           'id': 'mail-change-sender-manual-input',
                                           'placeholder': 'メールアドレスを入力'
                                           });
            wrap.append([wrap_area, input_manual]);
            $('#wrap-send-mail-sender-users-list').remove();
            $('#after-ope-edit-content').append(wrap);
        }else if(button.attr('data-mailids') || button.attr('data-mailaddresses')){
            var wrap = $('<div>').attr('class', 'd-flex');
            var name_area = $('<div>').attr({
                                      'id':'mail-change-name-area',
                                      });
            var wrap_search_btn = $('<div>').attr({
                                            'class': 'float-right position-relative ml-auto w-25 position-relative',
                                            'id': 'mail-change-wrap-search-btn'
                                            });
            var search_btn = $('<button>').attr({
                                          'type': 'button',
                                          'class': 'btn btn-primary text-nowrap',
                                          });
            if(item_name === 'send-mail-user'){
                var select_address = $('<select>').attr({
                                                  'class': 'form-control form-control-sm',
                                                  'id': 'mail-change-select-address-email-or-user'
                                                  });
                var op1 = $('<option>').text('ユーザ選択').val('user-select');
                var op2 = $('<option>').text('メールアドレス設定').val('input-address');
                select_address.append([op1, op2]);
                search_btn.text('ユーザ検索').attr('id', 'mail-change-user-search');
                var address_area = $('<div>').attr({
                                      'id':'mail-change-add-address-area',
                                      });
                $('#after-ope-edit-content').append(
                                                 $('<div>').attr('class', 'text-left mb-0 ml-1')
                                                           .append(select_address)
                                            );
                wrap.append(address_area.hide());   
            }else if(item_name === 'send-mail-category'){
                search_btn.text('カテゴリ検索').attr('id', 'mail-change-category-search');
            }
            wrap_search_btn.append(search_btn);
            wrap.append([name_area, wrap_search_btn]);
            $('#after-ope-edit-content').append(wrap); 
        }else if(button.attr('data-mailtitle')){
            $('#wrap-keyword').remove();
            $('#LoadingTarget').remove();
            var new_title_area  = $('<div>').append(
                                      $('<input>').attr({
                                          'class': 'form-control g-font-size-14',
                                          'id'   : 'mail-change-new-title',
                                      }).val(content)
                                  );
            var wrap_elm_keyword_button = $('<div>').attr('class', 'd-flex mr-auto mt-4');
            var elm_keyword_button = create_keywords_inputter();
            var elm_keyword_replace_btn = create_keywords_replacer();
            var elm_create_mail_change_reset_btn = create_mail_change_reset_btn('title');
            var trigger_param = JSON.parse($('#form-create-trigger').val());

            toggle_keyword_btns(elm_keyword_button, trigger_param)
            wrap_elm_keyword_button.append(elm_keyword_button, elm_keyword_replace_btn,elm_create_mail_change_reset_btn);
            new_title_area.append(wrap_elm_keyword_button);
            $('#after-ope-edit-content').append(new_title_area);
            wrap_elm_keyword_button.find('[data-toggle="tooltip"]').tooltip();
            var message = $('#edit-mail-select-menu').find('[data-mailcontent]').attr('data-content');
            var store_message = $('<input>').attr({
                                    'type': 'hidden',
                                    'id': 'store-message',
                                }); 
            new_title_area.append(store_message);
            var path = $('#sidfm_internal_api_replace_mail_keywords_path').val();
            $.post(path, {
                'title': '',
                'message': message
            },function(data){
                var replaced_message = data['message'];
                $('#store-message').val(replaced_message);
            });
        }else if(button.attr('data-mailcontent')){
           content_div.addClass('white-space-pre-wrap font-family-monospace');
           $('#wrap-keyword').remove();
           $('#LoadingTarget').remove();
           var new_content_area = $('<div>').append(
                                      $('<textarea>').attr({
                                          'class': 'form-control h-150px g-font-size-14 font-family-monospace',
                                          'id'   : 'mail-change-new-content',
                                      }).val(content)
                                  );
           var wrap_elm_keyword_button = $('<div>').attr('class', 'd-flex mr-auto mt-4');
           var elm_keyword_button = create_keywords_inputter();
           var elm_keyword_replace_btn = create_keywords_replacer();
           var elm_create_mail_change_reset_btn = create_mail_change_reset_btn('message');
           var trigger_param = JSON.parse($('#form-create-trigger').val());

           toggle_keyword_btns(elm_keyword_button, trigger_param);
           wrap_elm_keyword_button.append(elm_keyword_button, elm_keyword_replace_btn, elm_create_mail_change_reset_btn);
           new_content_area.append(wrap_elm_keyword_button);
           $('#after-ope-edit-content').append(new_content_area);
           wrap_elm_keyword_button.find('[data-toggle="tooltip"]').tooltip();
           var title = $('#edit-mail-select-menu').find('[data-mailtitle]').attr('data-content');
           var store_title = $('<input>').attr({
                                 'type': 'hidden',
                                 'id': 'store-title',
                             });
           new_content_area.append(store_title);
           var path = $('#sidfm_internal_api_replace_mail_keywords_path').val();
            $.post(path, {
                'title': title,
                'message': ''
            },function(data){
                var replaced_title = data['title'];
                $('#store-title').val(replaced_title);
            });
        }else if(button.attr('data-mailprocess')){
            var selection_area = $('<div>').attr('id', 'mail-change-process-data-area');
            var option1 = $('<div>').attr('class', 'text-left ml-3').append(
                              $('<input>').attr({
															 	 'class':'align-middle',
                                  'type': 'radio', 
                                  'name': 'process-data-radio',
                                  'id': 'process-data-radio-execute'
                              }), 
                              $('<label>').attr({'for': 'process-data-radio-execute', 'class': 'mb-0'})
                                          .append(
                                              $('<span>').attr('class', 'ml-1').text('データの指定無し'),
                                              $('<span>').attr('class', 'ml-1 font-size-small').text('※データの存在に関わらずメールを送信する')
                                          )
                          );
            var option2 = $('<div>').attr('class', 'mt-2 text-left ml-3');
            var radio = $('<input>').attr({
														'class':'align-middle',
                            'type': 'radio',
                            'name': 'process-data-radio',
                            'id': 'process-data-radio-no-execute'
                        });
            var label = $('<label>').attr({'for': 'process-data-radio-no-execute', 'class': 'd-inline mb-0'});
            label.append(
                         $('<span>').attr('class', 'ml-1').text('データの指定有り'),
                         $('<span>').attr('class', 'ml-1 font-size-small').text('※以下のデータが存在しない場合はメールを送信しない')
            );
            option2.append([radio, label]);             
            var select = create_process_data_select_elm().removeAttr('data-id')
                                                         .addClass('mt-1')
                                                         .prop('disabled', true)
                                                         .attr('id', 'mail-change-process-data');
            var exclamation = $('<span>').attr({
                                  'class': 'ml-1',
                                  'data-toggle': 'tooltip',
                              }).append($('<i>').attr('class', 'fas fa-exclamation-circle c_gray_70'));
            //見た目のためにselect要素の表示位置を変える
            var wrap_select = $('<div>').attr({ 'class': 'text-left', 'style': 'margin-left:32px;'});
            exclamation.attr('data-title', '選択肢から対象データを指定して下さい。※但し、当該対象データは別途指定している「条件」に一致するデータのみとなります');
            label.append(exclamation);
            wrap_select.append(select);
            selection_area.append([option1, option2]);
            $('#after-ope-edit-content').append([selection_area, wrap_select]);
            $('#after-ope-edit-content').find('[data-toggle="tooltip"]').tooltip();
        }
        $('#after-ope-edit-content').attr({
                                     'data-select': item_name,
                                     });
        $('#modalEditOperation-content')
            .animate({
                'opacity': 0.1
             },400, 'swing',function(){
                  $('#modalEditOperation-edit-mail-menu').hide();
                  $('#edit-mail-select-menu').remove();
        });
        setTimeout(function(){
            $('#modalEditOperation-body-main').show()
            $('#modalEditOperation-content')
            .animate({
                'opacity': 1
            },600);
       },400);
    });

    //操作変更modalでの「データの指定」の活性・非活性切り替え
    $(document).on('change', '[name="process-data-radio"]', function(){
        if($(this).prop('checked') && $(this).attr('id') === 'process-data-radio-no-execute'){
            $('#mail-change-process-data').prop('disabled', false);
        }else{
            $('#mail-change-process-data').prop('disabled', true);
        }
    });

    //操作変更modalでの送信先の切り替え
    $(document).on('change', '#mail-change-select-address-email-or-user', function(){
        var val = $(this).find('option:selected').val();
        if(val === 'input-address'){
            $('#mail-change-user-search').attr('id', 'mail-change-input-email-address')
                                         .text('アドレス設定');
            $('#mail-change-name-area').hide();
            $('#mail-change-add-address-area').show();
  
        }else{
            $('#mail-change-input-email-address').attr('id', 'mail-change-user-search')
                                                 .text('ユーザ検索');
            $('#mail-change-add-address-area').hide();
            $('#mail-change-name-area').show();
        }
    });    

    //操作変更modalで送信元の選択
    $(document).on('change', '#mail-change-select-sender-type', function(){
        var type = $(this).val();
        if(type === 'manual'){
            $('#mail-change-sender-user-select-area').hide();
            $('#mail-change-sender-manual-input').show();
        }else{
            $('#mail-change-sender-manual-input').hide();
            $('#mail-change-sender-user-select-area').show();
        }
    });


    //変更ユーザ決定する時
    $(document).on('click', '#decide-change-users', function(){
        $('#mail-change-name-area *').remove();
        var cnt = 0;
        $('#mail-setting-users-tbody').find('.chk-box:checked')
            .each(function(){
                var checked = $(this);
                var id = checked.val();
                var tr = checked.closest('tr');
                var name = $('<div>').text(tr.find('[data-name]').text())
                                     .attr({
                                         'class': 'min-h-30px pt-1 pl-1 g-font-size-14',
                                         'data-mailid': id,
                                     });
                if(cnt % 2 === 0){
                    name.addClass('user-name-bg');
                }
                cnt += 1;
                $('#mail-change-name-area').append(name);
         });
         $('#modalSearchUsers').modal('hide');
     });

     //操作変更modalでカテゴリ検索した時
     $(document).on('click', '#mail-change-category-search',function(){
         var button = $(this);
         make_category_selection_menu(button);
     });   

     //操作変更modalでカテゴリを選択
     $(document).on('click', '#set-temp-mail-change-category', function(){
         var op = $('#select-mail-change-category .c-category-radio:checked') 
         if(op.length === 0){
             alert('カテゴリを選択して下さい');
             return false;
         }else{
             $('#mail-change-name-area *').remove();
             var div = $('<div>').attr({
                                 'class': 'min-h-30px pt-1 pl-1 g-font-size-14',
                                 'data-mailid': op.val(),
                                 }).text(op.attr('data-text'));
             $('#mail-change-name-area').append(div);
         }
         $('#modalSelectItems').modal('hide');
    });
   
    //操作変更modalで変更ボタンを押した時
    $('#execute-ope-change').on('click', function(){
        var param = JSON.parse($('#set-ope-param').val());
        var change_item = $('#after-ope-edit-content').attr('data-select');
        var sender_user = $('#mail-change-sender-select-input:visible');
        var sender_address = $('#mail-change-sender-manual-input:visible');
        var name_area = $('#mail-change-name-area:visible');
        var add_mail_area = $('#mail-change-add-address-area:visible'); 
        var title_area = $('#mail-change-new-title');
        var message_area = $('#mail-change-new-content');
        var status_from = $('#edit-status-change-from-menu');
        var status_to = $('#edit-status-change-to-menu');
        var due_date = $('#edit-due-date-to:visible, #edit-due-date-days:visible');
        var mail_process_data = $('#mail-change-process-data-area:visible');
        var syslog_area = $('#edit-send-syslog-area:visible');
        var pass = false;
        if(sender_user.length === 1){
            if(sender_user.val() === ''){
                alert('送信元ユーザを選択して下さい');
                return false;
            }else{
                param[change_item]['sender'] = {'userid': sender_user.attr('data-userid'), 'username': sender_user.val()};
                pass = true;
            }
        }else if(sender_address.length === 1){
            var address = sender_address.val();
            if(address === ''){
                alert('送信元メールアドレスを入力して下さい');
                return false;
            }else if(!address.match(/^([^@\s]+)@((?:[-a-z0-9]+\.)+[a-z]{2,})$/i)){
                alert('メールアドレスに誤りがあります');
                return false;
            }else{
                param[change_item]['sender'] = {'address': address};
                pass = true;
            }
        }else if(name_area.length === 1){
            new_ids = [];
            $('#after-ope-edit-content').find('[data-mailid]')
                .each(function(){ 
                    var id = $(this).attr('data-mailid');
                    new_ids.push(id == 'all' ? 'all' : parseInt(id));
            }); 
            if(new_ids.length > 0){
                param[change_item]['sendby'] = {'ids':new_ids};
                pass = true;
            }
        }else if(add_mail_area.length === 1){
            new_addresses = [];
            $('#mail-change-add-address-area').find('[data-address]')
                                              .each(function(){
                new_addresses.push($(this).text());
            });
            if(new_addresses.length > 0){
                param[change_item]['sendby'] = {'address':new_addresses};
                pass = true;
            }
        }else if(title_area.length === 1 && title_area.val() != ''){
             param[change_item]['title'] = title_area.val();
             pass = true;
        }else if(message_area.length === 1 && message_area.val() != ''){
             param[change_item]['content'] = message_area.val();
             pass = true;
        }else if(mail_process_data.length > 0){
            var action_flg_elm = mail_process_data.find('[name="process-data-radio"]:checked');
            if(action_flg_elm.length === 0){
                alert('項目が選択されていません');
                return false;
            }
            var action_flg = action_flg_elm.attr('id') === 'process-data-radio-execute' ? 'true' : 'false';
            if(action_flg === 'false'){
                var action_target = $('#mail-change-process-data option:selected').val();
            }else{
                var action_target = "0";
            }
            param[change_item]['target'] = action_target;
            param[change_item]['proc_empty'] = action_flg;
            pass = true;
        }else if((status_from.length === 1 && status_to.length === 1) && (status_from.attr('data-statusids') && status_to.attr('data-statusids'))){
             var action_target = $('#after-ope-edit-content').find('[data-id="select-process-data"] option:selected').val();
             param['status-change']['status_ids_from'] = JSON.parse(status_from.attr('data-statusids'));
             param['status-change']['status_ids_to'] = JSON.parse(status_to.attr('data-statusids'));
             param['status-change']['target'] = String(action_target);
             pass = true;
        }else if(due_date.length === 1 && due_date.val() != ''){
             var change_name = due_date.attr('id') === 'edit-due-date-to' ? 'date' : 'days';
             var is_overwrite = String($('#due-date-change-checkbox-is-overwrite-duedate').prop('checked'));
             var action_target = $('#after-ope-edit-content').find('[data-id="select-process-data"] option:selected').val();
             param['due-date-change'] = {
                 [change_name]: due_date.val(),
                 'permit-update': is_overwrite,
                 'target': String(action_target),
                 'proc_empty': 'false'
             };
             pass = true;
        }else if(syslog_area.length == 1){
            var data = create_syslog_param();
            if( !data ){
                return false;
            }
            param['send-syslog'] = data;
            pass = true;
        }else if( $('#after-ope-edit-content[data-select="set-ctag"]:visible').length ){
            var selected_ctags = JSON.parse(ruleSelectedRegistCtags.val());
            if( !Object.keys(selected_ctags).length || (!selected_ctags.enable.length && !selected_ctags.not_enable.length) ){
                alert('カスタムタグが選択されておりません。');
                return;
            }
            const selected_data_target = $('#after-ope-edit-content').find('[data-id="select-process-data"] option:selected').val();
            if( selected_data_target == 0 ){
                alert(`処理するデータで、「${process_values_with_names[selected_data_target]}」が選択されています。\nこの場合、条件に一致する全ての脆弱性に対してカスタムタグが設定・除外されますので、ご注意ください。`);
            }else if( selected_data_target == 1 ){
                alert(`処理するデータで、「${process_values_with_names[selected_data_target]}」が選択されています。\nこの場合、条件に一致する全てのホストの、全ての脆弱性に対してカスタムタグが設定・除外されますので、ご注意ください。`);
            }
            param['set-ctag'] = {
                'target': selected_data_target,
                'proc_empty': 'false',
                ...selected_ctags
            };
            pass = true;
        }
        if(pass){
            $('#set-ope-param').val(JSON.stringify(param));
            update_modalRegisterOperation_items(change_item, param,'update', 'edit');
            $('#modalEditOperation').modal('hide');
        }else{
            alert('未設定の項目があります');
            return false;
        }
    });

    //操作設定modalで削除ボタンを押した時
    $(document).on('click', '#set-operation-tbody .del-ope', function(){
        var button = $(this);
        var del_item = button.attr('data-opedetail')
        button.closest('[data-addope=' + del_item + ']')
              .remove();
        var param = JSON.parse($('#set-ope-param').val());
        delete param[del_item];
        $('#set-ope-param').val(JSON.stringify(param));
        display_no_operation_message();
        $('.original-tooltip').remove();
        toggle_execute_set_btn();
    });    
             
    //操作詳細でのメール設定
    $(document).on('click', '#ope-detail-set-mail-content-btn',function(){
        $('#modalOperationMailSetting').modal('show');
    });

    $('#modalOperationMailSetting').on('show.bs.modal', function(){
        $('#wrap-send-mail-sender-users-list').remove();
        $('#wrap-keyword').remove();
        $('#LoadingTarget').remove();
        var trigger_param = JSON.parse($('#form-create-trigger').val());
        if($('#mail-select-process-data-area').find('[data-id="select-process-data"]').length === 0){
            var process_data_select = create_process_data_select_elm().prop('disabled', true);
            $('#mail-select-process-data-area').append(process_data_select);
        }
        $('#selected-process-data-before').val($('#mail-select-process-data-area').find('[data-id="select-process-data"]').val());
        var elm_keyword_inputter = create_keywords_inputter();
        var elm_keyword_replacer = create_keywords_replacer();

        toggle_keyword_btns(elm_keyword_inputter, trigger_param);
        if(trigger_param && trigger_param['period'] != 'vulsync'){
            $('#mail-preset-list').find('.mail-preset-item').each(function(){
                var elm = $(this);
                if(elm.attr('data-for') === 'vulsync'){
                   elm.hide();
                }else{
                   elm.show();
                } 
            });
        }else{
            $('#mail-preset-list').find('.mail-preset-item').each(function(){
                var elm = $(this);
                if(elm.attr('data-for') === 'vulsync'){
                   elm.show();
                }else{
                   elm.hide();
                }
            });
        }
        $('#mail-keyword-area').prepend(elm_keyword_inputter);
        $('#mail-preset-list').after(elm_keyword_replacer);
        $('#mail-keyword-area').find('[data-toggle="tooltip"]').tooltip();
        if($('#clone-search-address-items-area-tbody').length === 1){
            var copied_tbody = $('#clone-search-address-items-area-tbody').clone(true);
            $('#search-address-items-area-tbody').replaceWith(copied_tbody);
            copied_tbody.attr('id', 'search-address-items-area-tbody');
        }
        if($('#clone-add-address-area-tbody').length === 1){
            var copied_tbody = $('#clone-add-address-area-tbody').clone(true);
            $('#add-address-area-tbody').replaceWith(copied_tbody);
            copied_tbody.attr('id', 'add-address-area-tbody');
        }
        var sender = $('#ope-detail-readonly-sender');
        if(sender.attr('data-sender') === 'user'){
            $('#select-mail-sender-type').val('user-select');
            $('#send-mail-sender-manual-input').hide();
            $('#send-mail-sender-select-input').attr({
                                                'data-sender': 'user',
                                                'data-userid': sender.attr('data-userid'),
                                               }).val(sender.val());
            $('#send-mail-sender-select-area').show();
        }else{
            $('#select-mail-sender-type').val('manual');
            $('#send-mail-sender-select-area').hide();
            $('#send-mail-sender-manual-input').val(sender.val()).show(); 
        }                                                  
        $('#send-mail-title').val($('#ope-detail-readonly-mail-title').val());
        $('#send-mail-content').val($('#ope-detail-readonly-mail-content').val());
        if(sender.val() === ''){
            $('#send-mail-sender-manual-input').val($('#setting-address').val()); 
        }
    });

    //メール設定画面を開いたときにプリセットを使用できることをガイドするツールチップを表示する
    $('#modalOperationMailSetting').on('shown.bs.modal',function(){
        if($('#send-mail-title').val() === '' || $('#send-mail-content').val() === ''){
            var original_tooltip = make_original_tooltip('プリセットを使用して、メールの設定ができます。※送信元と送信先は、個々に設定して下さい', true).attr('id', 'mail-preset-tooltip');
            $('body').append(original_tooltip);
            var target = $('#mail-preset-list');
            var offset = target.offset();
            original_tooltip.offset({
                'top': offset.top - original_tooltip.outerHeight() - 8,
                'left':( offset.left - original_tooltip.outerWidth() / 2) + target.outerWidth() / 2
            }).show();
        }
    });

    $('#modalOperationMailSetting').on('click', function(){
        $('#mail-preset-tooltip').remove();
    });

    $('#modalOperationMailSetting').on('hidden.bs.modal', function(){
        var target_btn = $('#ope-detail-set-mail-content-btn');
        var sender = $('#ope-detail-readonly-sender').val();
        var address_type_text = $('#ope-detail-readonly-address-type-text').val();
        var title = $('#ope-detail-readonly-mail-title').val();
        var content = $('#ope-detail-readonly-mail-content').val();
        var target_data = $('#mail-target-data').val();
            
        if(sender != '' && address_type_text != '' && title != '' && content != '' && target_data != ''){
            target_btn.removeClass('btn-primary')
                      .addClass('btn-success setting-done')
                      .text('設定済み');
        }else{
             target_btn.removeClass('btn-success setting-done')
                     .addClass('btn-primary')
                     .text('メールを設定する');
        }
        
    }); 

    //スクロールでpopover削除
    $('#wrap-send-mail-users-table').on('scroll', function(){
        $('.rule-setting-original-popover').remove();
    }); 

    //「対象となるデータの指定」の選択可能・不可能の切り替え
    $('[name="mail-setting-process-data-radio"]').on('change', function(){
        if($('#mail-setting-process-data-radio-no-execute').prop('checked')){
            $('#mail-select-process-data-area [data-id="select-process-data"]').prop('disabled', false);
        }else{
            $('#mail-select-process-data-area [data-id="select-process-data"]').prop('disabled', true)
        }
    });
  
    //キーワード内の決定ボタンhoverで色変える
    $(document).on('mouseenter', '.decide-narrow-down-option', function(){
        $(this).closest('.wrap-narrow-item-option').addClass('pointed-option');
    });
    $(document).on('mouseleave', '.decide-narrow-down-option', function(){
        $(this).closest('.wrap-narrow-item-option').removeClass('pointed-option');
    });

    //メールのプリセットボタンを押した時にメールを設定
    $('#mail-preset-list [data-preset]').on('click', function(){
        var preset_data = JSON.parse($(this).attr('data-value'));
        var proc_empty = preset_data['proc_empty'];
        var target = preset_data['target'];
        if(!proc_empty){
            //$('#process-data-flg').prop('checked', true);
            $('#mail-setting-process-data-radio-no-execute').prop('checked', true);
            $('#mail-setting-process-data-radio-execute').prop('checked', false); 
            $('#mail-select-process-data-area [data-id="select-process-data"]').prop('disabled', false)
                                                                               .val(target);
        }else{
            //$('#process-data-flg').prop('checked', false);
            $('#mail-setting-process-data-radio-no-execute').prop('checked', false);
            $('#mail-setting-process-data-radio-execute').prop('checked', true);
            $('#mail-select-process-data-area [data-id="select-process-data"]').prop('disabled', true)
        }
        $('#send-mail-title').val(preset_data['subject']);
        $('#send-mail-content').val(preset_data['message']);
        $('#send-mail-content').scrollTop(0);
    });

    $('#execute-set-ope-detail-mail').on('click', function(){
        $('#hidden-store-search-address-items-area-tbody > *').remove();
        var clone_tbody = $('#search-address-items-area-tbody').clone(true)
                                                               .attr('id', 'clone-search-address-items-area-tbody');
        var clone_mail_tbody = $('#add-address-area-tbody').clone(true)
                                                           .attr('id', 'clone-add-address-area-tbody');
        $('#hidden-store-search-address-items-area-tbody').append(clone_tbody, clone_mail_tbody);
        var address_type_field = $('#ope-detail-readonly-address-type-text');
        var address_type_item_names = [];
        var sender_field = $('#ope-detail-readonly-sender');
        var title_field = $('#ope-detail-readonly-mail-title');
        var content_field = $('#ope-detail-readonly-mail-content');
        if($('#search-address-items-area').is(':visible')){
            var selected_address_type_items = $('#search-address-items-area-tbody').find('[data-addresstype]');
            selected_address_type_items.each(
                function(){
                    var item = $(this); 
                    var item_name = item.text().trim();
                    address_type_item_names.push(item_name);
            });
            if($('#ope-detail-address-name').attr('data-sendby') != 'category'){
                $('#ope-detail-address-name').attr('data-sendby', 'user').text('送信先ユーザ');
            }
        }else{
            var add_addresses = $('#add-address-area-tbody').find('[data-address]');
            add_addresses.each(function(){
                var add_address = $(this).text();
                address_type_item_names.push(add_address);
            });
            if($('#ope-detail-address-name').attr('data-sendby') != 'category'){
                $('#ope-detail-address-name').attr('data-sendby', 'address').text('送信先アドレス');
            }
        }                           
        address_type_field.val(address_type_item_names.join(','));
        if($('#send-mail-sender-manual-input:visible').length === 1){
            var address = $('#send-mail-sender-manual-input:visible').val();
            if(address != '' && !address.match(/^([^@\s]+)@((?:[-a-z0-9]+\.)+[a-z]{2,})$/i)){
                alert('メールアドレスに誤りがあります');
                return false;
            }
            sender_field.val(address)
                        .attr('data-sender', 'address');
        }else{ 
            var sender = $('#send-mail-sender-select-input:visible');
            sender_field.val(sender.val())
                        .attr({
                            'data-sender': 'user',
                            'data-userid': sender.attr('data-userid'),
                        });
            
        }
        title_field.val($('#send-mail-title').val());
        content_field.val($('#send-mail-content').val());
        if($('#mail-setting-process-data-radio-execute').prop('checked')){
            var target_data = 'データの指定無し';
        }else if($('#mail-setting-process-data-radio-no-execute').prop('checked')){
            var selected_value_text = $('#mail-select-process-data-area').find('[data-id="select-process-data"] option:selected').text();
            var target_data = selected_value_text;
        }else{ 
            var target_data = '';
        }
        $('#mail-target-data').val(target_data);
        $('#modalOperationMailSetting').modal('hide');                  
    });

    //キーワードボタンを押した時に、データの指定があればそのデータのキーワードボタンに色を付ける
    function color_keyword_btns() {
        $('#keyword').find('.bg-selected-mail-data-keyword').removeClass('bg-selected-mail-data-keyword');
        if( $('#mail-setting-process-data-radio-no-execute').prop('checked') || $('#modalEditOperation').is(':visible') ){
            var selected_val = $('[data-id="select-process-data"]:visible').val();
            if( !selected_val ){
                //メール設定変更の時
                var ope_param = JSON.parse($('#set-ope-param').val());
                if( $('#after-ope-edit-content').attr('data-select') === 'send-mail-category' ) {
                    if( ope_param['send-mail-category']['proc_empty'] === 'true' ){
                        return false;
                    }
                    var selected_val = ope_param['send-mail-category']['target'];
                }else{
                    if( ope_param['send-mail-user']['proc_empty'] === 'true' ){
                        return false;
                    }
                    var selected_val = ope_param['send-mail-user']['target'];
                } 
            }     
            if( selected_val == 0 ){
                var elm = $('#keyword').find('[data-menuname="vul-info-match"]');
                elm.addClass('bg-selected-mail-data-keyword');
                elm.closest('[data-item="keyword-menu-btn"]')
                   .find('[data-itemid]')
                   .addClass('bg-selected-mail-data-keyword');
            }else if( selected_val == 1 ){
                var elm = $('#keyword').find('[data-menuname="host-info-match"]');
                elm.addClass('bg-selected-mail-data-keyword');
                elm.closest('[data-item="keyword-menu-btn"]')
                   .find('[data-itemid], [value="%OS_RATE_MATCH%"]')
                   .addClass('bg-selected-mail-data-keyword');
            }else if( selected_val == 2 ){
                var target_elms = [ '[data-menuname="vulsync-info-match"]',
                                    '[data-itemid="%VULSYNC_REGI_MATCH%"]'
                                   ].join(', ');    
                $('#keyword').find(target_elms).addClass('bg-selected-mail-data-keyword');
            }else if( selected_val == 3 ){
                var target_elms = [ '[data-menuname="vulsync-info-match"]',
                                    '[data-itemid="%VULSYNC_RENEW_MATCH%"]'
                                   ].join(', ');
                $('#keyword').find(target_elms).addClass('bg-selected-mail-data-keyword');
            }else if( selected_val == 4 ){
                var target_elms = [ '[data-menuname="vulsync-info-match"]',
                                    '[data-itemid="%VULSYNC_DEL_MATCH%"]'
                                  ].join(', ') ;
                $('#keyword').find(target_elms).addClass('bg-selected-mail-data-keyword');
            }else if( selected_val == 5 ){
                var elm = $('#keyword').find('[data-menuname="vulsync-info-match"]');
                elm.addClass('bg-selected-mail-data-keyword');
                elm.closest('[data-item="keyword-menu-btn"]')
                   .find('[data-itemid]')
                   .addClass('bg-selected-mail-data-keyword');
            }
            $('#keyword-main-menu:visible').tooltip('show');
        }else{
            $('#keyword-main-menu:visible').tooltip('hide');
        }
    }
   
    $(document).on('click', '#keyword', function(){
        color_keyword_btns();
    });

    //キーワードボタン内のdropdown
    $(document).on('click', '#keyword [data-originaldropdown]', function(){
        var elm = $(this);
        var target_menu = elm.siblings('.original-dropdown-menu');
        var display_position = elm.attr('data-display');
        var elm_parent_menu = elm.closest('[data-item="keyword-menu-btn"]').get(0);
        if(elm.attr('id') != 'keywords-list'){
            $('.keywords-menu [data-item="keyword-menu-btn"]').each(function(){
                var menu_btn = $(this);
                if(menu_btn.get(0) != elm_parent_menu){
                    menu_btn.find('.original-dropdown-menu').hide();
                }
            });
        }
        if(!target_menu.is(':visible')){
            if(display_position === 'up'){
                target_menu.css({
                    'bottom': '101%',
                });
                if( elm.attr('data-drop') === 'left-side' ){
                    target_menu.css('right', '0');
                }else{
                    target_menu.css('left', '0');
                }
            }else if(display_position === 'right'){
                target_menu.css({
                    'top': '0',
                    'left': '101%'
                });
            }else if(display_position === 'right-up'){
                target_menu.css({
                    'bottom': '0',
                    'left': '101%'
                });
            }else if(display_position === 'narrow-down-option'){
                target_menu.css({
                    'bottom': '-5px',
                    'left': '101%',
                });
            }
            $('.narrow-down-option').hide();
            target_menu.show();
        }else{
            target_menu.hide();
        }
    });

    var selected_textarea = null;
    $(document).on('click', '#send-mail-title, #send-mail-content, #mail-change-new-title, #mail-change-new-content, #syslog-input-message', function(){
        selected_textarea = $(this);
    });

    $(document).on('click','.put-keyword', function(){
        var button = $(this);
        var textarea = (selected_textarea ? selected_textarea : $('#send-mail-content:visible, #mail-change-new-content:visible, #mail-change-new-title:visible, #syslog-input-message'));
        var pos = textarea.get(0).selectionStart;
        var len = textarea.val().length;
        var content =  textarea.val();
        var before = content.substr(0, pos);
        var after = content.substr(pos);
        if(button.hasClass('decide-narrow-down-option')){
            var target = button.closest('.narrow-down-option-item').find('[data-narrow]');
            var target_val = target.val();
            var narrow_val = target.attr('data-narrow');
            var target_type = target.attr('type');
            if(narrow_val != ''){
                var keyword = target.attr('data-keyword').slice(0, -1) + narrow_val + target_val + '%';
            }else{
                var keyword = target.attr('data-keyword');
            }
            if(target_type === 'number' && target_val.length === 0){
                alert('値を入力して下さい');
                return false;
            }else if(target_type === 'number' && parseInt(target_val) <= 0){
                alert('正数で入力して下さい');
                return false;
            }     
        }else{
            var keyword = button.val()
        }
        content = before + keyword + after;
        textarea.val(content);
        $('#keyword-main-menu:visible').tooltip('hide');
        $('#keywords-list').click();
        selected_textarea = null;
        if($('#syslog-keyword-preview-popup').length === 1){
            $('#syslog-keyword-preview-popup').remove();
        }
    });

    //送信元メールアドレスを手入力かユーザ選択かでinputを切り替え
    $('#select-mail-sender-type').on('change', function(){
        $('#wrap-send-mail-sender-users-list').remove();
        var val = $(this).val();
        if(val === 'manual'){
            $('#send-mail-sender-select-area').hide();
            $('#send-mail-sender-manual-input').show();
        }else{
            $('#send-mail-sender-select-area').show();
            $('#send-mail-sender-manual-input').hide();
        }
    });

    //送信元メールアドレスがユーザ選択の時、inputをクリックで選択可能なユーザ一覧表示
    $('#search-sender-user').on('click', function(){
        make_sender_users_list($(this));
    });
    $(document).on('click', '#mail-change-search-sender-user', function(){
        make_sender_users_list($(this));
    });

    const SENDER_USERS_LIST_STATE = {
        creating_user_elms: false,
    };
    function make_sender_users_list(clicked_elm){
        if( !SENDER_USERS_LIST_STATE.creating_user_elms ){ 
            SENDER_USERS_LIST_STATE.creating_user_elms = true;
            var target_modal = $('#modalSelectItems');
            target_modal.find('.modal-body > *').remove();
            var clicked_elm_id = clicked_elm.attr('id'); 
            var path = $('#sidfm_internal_api_get_data_for_mail_setting_path').val();
            var category_id = $('#main-category-id').val();
            var wrap = '<div id="wrap-send-mail-sender-users-list">';
            var div = '<div id="send-mail-sender-users-list">';
            $.post(path, {'info_name': 'get_users_by_category_id', 'ids': JSON.stringify([category_id])},
            function(data){
                if(data.length > 0){
                    for(var i=0; i < data.length; i++){
                        var info = data[i];
                        var user_id = info[0];
                        var username = escape_html_special_chars(info[1]);
                        var address = info[2];
                        var userinfo_div = '<div class="sender-username list_alternative-r2" data-info="sender-name" data-userid="' + user_id + '">' +
                                  '<div class="float-left mr-2 username">' + username + '</div>' + 
                                  '<div class="float-right ml-auto">' +
                                  '<button type="button" class="text-nowrap btn btn-xs sender-address" data-item="sender" data-toggle="tooltip" data-title="' + address + '">' +
                                  'アドレス' +
                                  '</button></div>' +
                                  '</div>'; 
                       div += userinfo_div;
                    }
                    div += '</div>';
                    wrap = $(wrap += div + '</div>');
                    target_modal.find('.modal-body')
                                .append([
                                    $('<div>').attr('class', 'mb-2')
                                              .html('<div class="font-weight-bold">送信元ユーザを選択して下さい</div>' +
                                                    '<div class="text-danger mt-1 g-font-size-12">※表示される選択可能なユーザは、「ルールを適用するカスタムカテゴリ」で選択されているカテゴリのユーザのみです。</div>'),
                                    wrap
                                ]);
               }else{
                   target_modal.find('.modal-body')
                                .append(
                                    $('<div>').attr('class', 'font-weight-bold mb-2 text-center user-name-bg')
                                              .text('選択可能なユーザは存在しません')
                                );
               }
               target_modal.find('[data-selectitem]').attr('id', clicked_elm_id === 'search-sender-user' ? 'decide-sender-user' : 'decide-change-sender-user');
               target_modal.find('.modal-title').text('ユーザ選択');
               target_modal.modal('show');
               $('[data-item=sender]').tooltip();
               SENDER_USERS_LIST_STATE.creating_user_elms = false;
           });
        }
    }

    //送信元ユーザ選択時
    $(document).on('click', '#send-mail-sender-users-list [data-info=sender-name]', function(){
        var button = $(this);
        $('#send-mail-sender-users-list [data-info=sender-name]').removeClass('selected-sender selected');
        button.addClass('selected-sender selected');
    }); 

    //送信元ユーザ決定
    $(document).on('click', '#decide-sender-user', function(){
        var selected_user = $('#send-mail-sender-users-list').find('.selected');
        var user_name = selected_user.find('.username');
        if(user_name.length === 0){
            alert('ユーザを選択して下さい');
            return false;
        }
        $('#send-mail-sender-select-input').val(user_name.text())
                                           .attr('data-userid', selected_user.attr('data-userid'));
        $('#modalSelectItems').modal('hide');
    });   

    //メールの変更送信元ユーザ決定
    $(document).on('click', '#decide-change-sender-user', function(){
        var selected_user = $('#send-mail-sender-users-list').find('.selected');
        var user_name = selected_user.find('.username');
        if(user_name.length === 0){
            alert('ユーザを選択して下さい');
            return false;
        }
        $('#mail-change-sender-select-input').val(user_name.text())
                                             .attr('data-userid', selected_user.attr('data-userid'));
        $('#modalSelectItems').modal('hide');
    });
  
    //キーワードを変換ボタンを押した時
    $(document).on('click', '#change-keywords', function(){
        var button = $('#LoadingTarget');
        var button_height = button.outerHeight();
        var button_width = button.outerWidth();
        if($('#modalOperationMailSetting').is(':visible')){
            var title = $('#send-mail-title').val();
            var message = $('#send-mail-content').val();
        }else if($('#mail-change-new-content').is(':visible')){
            var title = '';
            var message = $('#mail-change-new-content').val();
        }else if($('#mail-change-new-title').is(':visible')){
            var title = $('#mail-change-new-title').val(); 
            var message = '';
        }else if($('#syslog-input-message').is(':visible')){
            $('#syslog-keyword-preview-popup').remove();
            if($('#syslog-input-message').val().length > 0){
                var title = '';
                var message = $('#syslog-input-message').val();
                var is_pass = check_syslog_message(message);
                if( !is_pass ){
                    alert('メッセージに使用できないキーワードが含まれています');
                    return false;
                }
            }else{
                $(this).attr({
                    'data-toggle': 'tooltip',
                    'data-title': 'メッセージを入力して下さい',
                    'data-trigger': 'manual'
                }).tooltip('show');
                setTimeout(function(){
                    $('#change-keywords').tooltip('hide');
                },1500);
                return false;
            }
        }
        $('#keyword .keywords-menu').hide();
            var target_offset = button.offset();
            var keyword_loading = $('<div>').attr({
                                      'id': 'keyword_loading',
                                  }).css({
                                      'position': 'absolute',
                                      'width': button_width,
                                      'height': button_height,
                                      'bottom': (button_height / 2) + 'px', 
                                      'left': '0',
                                  });
            var hidden_loading = $('<div>').attr({
                                     'id': 'hidden_loading',
                                     'class': '_loading mx-auto',
                                 });
            keyword_loading.append(hidden_loading);
            button.append(keyword_loading);
            var path = $('#sidfm_internal_api_replace_mail_keywords_path').val(); 
            $.post(path, {
                'title': (title ? title : ''), 
                'message': (message ? message : '')
            },function(data){
                var replaced_title = data['title'];
                var replaced_message = data['message'];
                $('#keyword_loading').remove();
                if($('#syslog-input-message').is(':visible')){
                    var popup = $('<div>').attr({
                                    'id': 'syslog-keyword-preview-popup',
                                }).css({
                                    'position': 'absolute',
                                    'width': $('#syslog-input-message').outerWidth(),
                                    //'height': $('#syslog-input-message').outerHeight(),
                                    'bottom': 'calc(100% + 10px)',
                                    'left': '0',
                                });
                    var inner = $('<div>').attr({
                                    'id': 'syslog-keyword-preview-popup-inner',
                                });
                    var close_btn = $('<span>').attr({
                        'class': 'ml-auto pl-1 pr-1 text-white syslog-close-popup',
                        'data-close': 'syslog-keyword-preview-popup'
                    }).html('&times;');
                    var header = $('<div>').attr('class', 'd-flex p-1 syslog-keyword-preview-header').append(
                                     $('<span>').append(
                                         $('<span>').attr('class', 'text-white').text('キーワード変換後'),
                                         $('<span>').attr('class', 'ml-1 font-weight-bold font-size-small text-white').text('※注意事項'),
                                         $('<i>').attr({
                                             'class': 'fas fa-question-circle c_gray_70 ml-1',
                                             'data-toggle': 'tooltip',
                                             'data-boundary': 'window',
                                             'data-title': '※「キーワード」の指定により置き換えられた場合の見本を表示しています。ログの送信時に、当該キーワードの内容は実際の対象データに変換されます。'
                                         })
                                     ),
                                     close_btn
                                 );
                    var content = $('<div>').attr('class','syslog-popup-content')
                                                 .html(replaced_message.replace(/\n/g, '<br>'));
                    inner.append(header, content);
                    popup.append(inner);
                    $('#syslog-popup-target').append(popup);
                    popup.find('[data-toggle="tooltip"]').tooltip();
                    $('[data-close="syslog-keyword-preview-popup"]').on('click', function(){
                        $('#syslog-keyword-preview-popup').remove();
                    });
                }else{
                    var title_preview = $('#mail-title-of-preview');
                    var message_preview = $('#mail-content-of-preview');
                    if($('#modalOperationMailSetting').is(':visible')){
                        title_preview.val(replaced_title);
                        message_preview.val(replaced_message);
                    }else if($('#mail-change-new-content').is(':visible')){
                        var stored_title = $('#store-title').val();
                        title_preview.val(stored_title);
                        message_preview.val(replaced_message); 
                    }else if($('#mail-change-new-title').is(':visible')){
                        var stored_message = $('#store-message').val();
                        title_preview.val(replaced_title);
                        message_preview.val(stored_message);
                    }
                    $('#modalMailPreview').modal('show');
                }
            });
    });

    //プレビュー画面を閉じる前にメール本文のtextareaのscrollを一番上にする
    $('#modalMailPreview').on('hide.bs.modal', function(){
        $('#mail-content-of-preview').scrollTop(0); 
    });

    //mail設定modal画面の設定
    const USERS_DIALOG_STATE = {
        creating_user_elms: false,
    };
    $(document).on('click', '#modalOperationMailSetting-search-users-btn, #mail-change-user-search, #history-create-cond-search-user-button', function(){
        var target_tbody = $('#mail-setting-users-tbody');
        var button = $(this);
        var main_category_id = $('#main-category-id').val();
        var trigger_id = button.attr('id');
        if( trigger_id === 'history-create-cond-search-user-button' ){
            $('#modalSearchUsers-description-of-displayed-users').text('※表示される選択可能なユーザは、「ルールを適用するカスタムカテゴリ」で選択されているカテゴリのユーザと、' + (main_category_id === 'all' ? 'SIDfmVM管理者権限' : 'ログイン名「sidfmadm」') + 'のユーザのみです。');
        }else{
            $('#modalSearchUsers-description-of-displayed-users').text('※表示される選択可能なユーザは、「ルールを適用するカスタムカテゴリ」で選択されているカテゴリのユーザ' + (main_category_id === 'all' ? 'と、SIDfmVM管理者権限のユーザ' : '') + ' のみです。');
        }
        if(target_tbody.find('tr').length === 0){
            $('#send-mail-all-users-check').prop('checked', false);
            var path = $('#api_get_rule_selectable_users_path').val();
            var param = { 'category_id': main_category_id };
            if( trigger_id === 'history-create-cond-search-user-button' ){
                param['is_sidfmadm_selectable'] = 'true'
            }
            if( !USERS_DIALOG_STATE.creating_user_elms ){
                USERS_DIALOG_STATE.creating_user_elms = true;
                $.post(path, param, function(users_info){
                    if(users_info.length === 0){
                        var tr = $('<tr>');
                        tr.append(
                            $('<td>').attr({'colspan': '5', 'class': 'text-center'})
                                     .text('選択可能なユーザは存在しません')
                        );
                        target_tbody.append(tr);
                    }else{
                        var category_descs = JSON.parse($('#category-descs').val());
                        var catname_descs = {};
                        for(category_id in category_descs){
                            var cat = category_descs[category_id];
                            catname_descs[cat.name] = cat.desc;
                        }
                        var cnt = 1;
                        for(var i = 0; i < users_info.length;  i++){
                            var user_info = users_info[i];
                            var { user_id, login, firstname, lastname, categories } = user_info;
                            var mail_address = user_info['address'];
                            var tr = $('<tr>');
                            var chk_box_td = $('<td>').attr('class', 'modal-mail-chkbox text-center')
                                                      .append(
                                                          $('<input>').attr({
                                                              'type' : 'checkbox',
                                                              'class' : 'chk-box',
                                                              'value' : user_id,
                                                              'id': 'mail-setting-users-tbody-user-' + user_id
                                                          })
                                                      );
                            var login_td = $('<td>').attr('class', 'login')
                                                    .append(
                                                        $('<label>').attr({
                                                            'class': 'login text-truncate mb-0 d-block',
                                                            'for': 'mail-setting-users-tbody-user-' + user_id
                                                        }).text(login)
                                                    );
                            var name_td = $('<td>').attr('class', 'name')
                                                   .append(
                                                       $('<label>').attr({
                                                           'class': 'name text-truncate mb-0 d-block',
                                                           'data-name': 'true',
                                                           'for': 'mail-setting-users-tbody-user-' + user_id
                                                       }).text(lastname + ' ' + firstname)
                                                   );
                            var address_td = $('<td>').attr('class', 'address')
                                                      .append( 
                                                          $('<label>').attr({
                                                              'class': 'address text-truncate mb-0 d-block',
                                                              'for': 'mail-setting-users-tbody-user-' + user_id
                                                          }).text(mail_address)
                                                      );
                            var category_td = $('<td>').attr('class', 'category');
                            if(categories.length > 1){
                                var categories_content = '';
                                for(var j = 0; j < categories.length; j++){
                                    categories_content += "<div data-toggle='tooltip' data-title='" + escape_html_special_chars(catname_descs[categories[j]] ? catname_descs[categories[j]] : '') + "' data-placement='right' data-boundary='window'>" + escape_html_special_chars(categories[j]) + '</div>';
                                }
                                var category_btn = $('<div>').attr({
                                                       'class' : 'mail-btn show-original-popover btn btn-xs btn-light-nb text-nowrap w-100 max-width-160px',
                                                       'data-item' : 'mail-popover-category' + cnt,
                                                       'data-select' : 'mail-popover-category',
                                                       'data-html' : 'true',
                                                       'data-content' : categories_content,
                                                   });
                                var category_inner = $('<div>').attr('class', 'd-flex')
                                                               .append(
                                                                   $('<span>').attr({
                                                                       'class': 'text-left mr-1 inline-block text-truncate w-100'
                                                                   }).text(categories[0]),
                                                                   $('<span>').attr({
                                                                       'class' : 'position-static ml-auto inline-block badge badge-pill badge-gray-999'
                                                                   }).text(categories.length)
                                                               );
                                category_btn.append(category_inner);
                                category_td.append(category_btn);
                            }else{
                                var category_name = $('<label>').attr({
                                    'class': 'g-pl-5 g-font-size-0_89x text-truncate max-width-160px mb-0 d-block',
                                    'for': 'mail-setting-users-tbody-user-' + user_id
                                });
                                if( categories.length === 0 ){
                                    category_name.addClass('text-center').text('---');
                                }else{
                                    category_name.attr({
                                        'data-toggle': 'tooltip',
                                        'data-title': (catname_descs[categories[0]] ? catname_descs[categories[0]] : ''),
                                        'data-placement': 'top',
                                        'data-boundary': 'window'
                                    }).text(categories[0]);
                                }
                                category_td.append(category_name);
                            }                                  
                            tr.append(chk_box_td, login_td, name_td, address_td, category_td);
                            target_tbody.append(tr);
                            tr.find('[data-toggle="tooltip"]').tooltip();
                            cnt += 1;         
                        }
                    }
                    USERS_DIALOG_STATE.creating_user_elms = false;
                });
            }
        }

        selected_textarea = null;
        var target_button = $('[data-buttonid="modalSearchUsers-decide-button"]');
        if(trigger_id === 'modalOperationMailSetting-search-users-btn'){
            target_button.attr('id', 'decide-users');  
        }else if(trigger_id === 'mail-change-user-search'){
            target_button.attr('id', 'decide-change-users');
        }else{
            target_button.attr('id', 'decide-history-create-cond-users');
        }
        setTimeout(function(){
            $('#modalSearchUsers').modal('show');
        },150);
    });

    //ユーザ検索でユーザを選択したとき
    $(document).on('change', '#wrap-send-mail-users-table .chk-box',function(){
        var chk_box_count = $('#wrap-send-mail-users-table').find('.chk-box').length;
        var checked_count = $('#wrap-send-mail-users-table').find('.chk-box:checked').length;
        if(chk_box_count === checked_count){
            $('#send-mail-all-users-check').prop('checked', true);
        }else{
            $('#send-mail-all-users-check').prop('checked', false);
        }
    });
 
    $(document).on('change', '#send-mail-all-users-check', function(){
        var all_check = $(this);
        var is_checked = all_check.prop('checked');
        var chk_boxes = $('#wrap-send-mail-users-table').find('.chk-box');
        if(is_checked){ 
            chk_boxes.prop('checked', true);
        }else{
            chk_boxes.prop('checked', false);
        }
    });

    $(document).on('click', '#decide-users', function(){
        var items = [];
        $('#send-mail-users-table').find('.chk-box:checked')
            .each(function(){
                items.push($(this).closest('tr'));
        });
        append_address_items(items, 'user');
        $('#modalSearchUsers').modal('hide');
    });                            

    $('#modalSearchUsers').on('hide.bs.modal',function(){
        $('#mail-setting-users-tbody').scrollTop(0);
    });
 
    //mail設定modalでのカテゴリ検索
    $(document).on('click','#modalOperationMailSetting-search-category-btn', function(){
        var button = $(this); 
        make_category_selection_menu(button);
    });

    function make_category_selection_menu(button){
        var target_modal = $('#modalSelectItems');
        target_modal.find('.modal-body > *').remove();
        var is_mail_change = (button.attr('id')).indexOf('mail-change') != -1;
        var wrap = $('<div>').attr('class', ' d-flex mail-setting-items-list')
        var div1 = $('<div>').attr('class', 'mx-auto');
        var id_name = is_mail_change ? 'select-mail-change-category' : 'select-mail-category';
        var selections = make_selectable_categories(id_name, true);
        div1.append(selections);
        wrap.append([div1]);
        target_modal.find('.modal-body')
                    .append([
                        $('<div>').attr('class', 'font-weight-bold mb-3')
                                  .text('カテゴリを選択して下さい'),
                        wrap
                    ]);
        target_modal.find('.modal-title').text('送信先カテゴリ選択');
        target_modal.find('[data-selectitem]').attr('id', is_mail_change ? 'set-temp-mail-change-category': 'set-temp-category');
        target_modal.find('[data-toggle="tooltip"]').tooltip();
        target_modal.modal('show');
    }

    //メール設定modalの宛先ユーザ・カテゴリの項目をappend
    function append_address_items(items, address_type){
        $('#search-address-items-area-tbody > *').remove();
        var type = address_type;
        var cnt = 0;
        $(items).each(function(){
            var item = $(this);
            switch(type){
            case 'user':
                var id = item.find('.chk-box').val();
                var name = item.find('[data-name]').text();
                var tr = $('<tr class="list_alternative2">').attr({
                                  'data-addresstype':'user',
                                  'data-addresstypeid':id,
                                  });
                if(cnt % 2 === 0){
                    tr.addClass('');
                }
                cnt += 1;
                var td = $('<td class="text-left">').text(name);
                tr.append(td);
                $('#search-address-items-area-tbody').append(tr);
                break;
            case 'category':
                var id = item.val();
                var item_name = item.attr('data-text');
                var div = $('<div>').attr({
                                    'class': 'mail-setting-cat-td',
                                    'data-addresstype': 'category',
                                    'data-addresstypeid': id,
                                    }).text(item_name);
                $('#search-address-items-area-tbody').append(div);
                break;
            }    
        });  
    }

    //メール設定modalのカテゴリ設定時の確認
    $(document).on('click', '#set-temp-category',function(){
        var selected_op = $('#select-mail-category').find('.c-category-radio:checked');
        if(selected_op.length === 0){
            alert('カテゴリを選択して下さい');
            return false;
        }else{ 
            append_address_items(selected_op, 'category');
            $('#modalSelectItems').modal('hide');
        }
            
    });

   //アドレス設定ボタンを押した時
   $(document).on('click', '#input-email-address-btn, #mail-change-input-email-address',function(){
       var target_modal = $('#modalSelectItems');
       var btn_id = $(this).attr('id');
       target_modal.find('.modal-body > *').remove();
       var desc = $('<div>').attr('class', 'mb-3 g-font-size-14').html('必要なメールアドレスを全て追加して下さい。決定ボタンを押下するとメールアドレスが送信先欄に設定されます。');
       target_modal.find('.modal-body').append(desc);
       var wrap = $('<div>').attr({
                             'class': 'mb-2',
                             'data-setmailid': btn_id
                            }); 
       var wrap_table = $('<div>').attr({
                                   'class': 'table-responsive scrollable-menu-150 g-font-size-13 g-brd-around g-brd-1 g-brd-ccc',
                                   'id' : 'wrap-email-address-list-table'
                                   });
       var table = $('<table>').attr({
                                'class': 'table table-sm mb-0',
                                'id': 'email-address-list-table'
                               });
       var thead = $('<thead>').attr('class', 'thead-dark');
       var tr = $('<tr>').attr('class', '');
       var th = $('<th>').attr('class', 'th-sticky-top0 font-weight-normal g-text-shadow-black').text('メールアドレス一覧');
       tr.append(th);
       thead.append(tr);
       var tbody = $('<tbody>').attr({'class': '', 'id': 'email-address-list-tbody'});
       var show_no_address = $('<tr>').attr({'class': 'list_alternative2', 'id': 'ask-add-emails'})
                                      .append(
                                          $('<td>').attr({
                                                    'class': '',
                                           }).text('メールアドレスを追加して下さい')
                                       );
       tbody.append(show_no_address);
       table.append([thead, tbody]);
       wrap_table.append(table);
       var wrap_div = $('<div>').attr('class', 'position-absolute bottom-0 w-100 mb-2');
       var div = $('<div>').attr('class', 'd-flex mt-4');
       wrap_div.append(div);
       var input = $('<input>').attr({
                                'type': 'text',
                                'class': 'form-control form-control-sm float-left mr-2 w-72',
                                'id': 'email-address-to-add',
                                'placeholder': 'メールアドレスを入力して下さい',
                               });
       var btn = $('<div>').attr('class', 'float-right position-relative')
                           .append(
                                $('<button>').attr({
                                              'type': 'button',
                                              'class': 'btn btn-sm btn-primary text-nowrap position-absolute bottom-0',
                                              'id': 'add-email-address',
                                 }).text('アドレス追加')
                            );
        div.append([input, btn]);
        wrap_table.append(table);
        var keep_margin = $('<div>').attr('class', 'keep-margin');
        wrap.append([wrap_table, keep_margin, wrap_div]);
        target_modal.find('.modal-body').append(wrap);
        target_modal.find('modal-title').text('メールアドレス追加');
        target_modal.find('[data-selectitem]').attr('id', 'execute-add-email-addresses');
        target_modal.find('.modal-body').addClass('min-height-300px');
        target_modal.find('.modal-title').text('メールアドレス設定'); 
        if(btn_id === 'mail-change-input-email-address'){
            var param = JSON.parse($('#set-ope-param').val());
            var current_addresses = param['send-mail-user']['sendby']['address'];
            var inputing_addresses = $('#mail-change-add-address-area').find('[data-address]');
            if(inputing_addresses.length > 0){
                current_addresses = [];
                inputing_addresses.each(function(){
                    current_addresses.push($(this).text());
                });
            }
        }else if(btn_id === 'input-email-address-btn'){
            var current_addresses = [];
            $('#add-address-area-tbody').find('[data-address]')
                                        .each(function(){
                current_addresses.push($(this).text());
            });
        }
        if(current_addresses && current_addresses.length > 0){
            for(var i = 0; i < current_addresses.length; i++){
                var address = current_addresses[i];
                var add_elm_tr = $('<tr>').attr({ 'class': 'list_alternative2'} );
                var add_elm = $('<td>').attr({
                                       'class': '',
                                       'data-item': 'add-address'}
																				);
                var wrap = $('<div>').attr('class', 'd-flex justify-content-between');
                var add_address = $('<div>').attr({
                                            'class': 'pt-1 ml-2 mr-1 text-truncate',
                                            'data-item': 'address-text'
                                            }).text(address);
                var del_btn = $('<div>').attr('class', 'mr-1')
                                        .append(
                                            $('<a>').attr('class', 'btn btn-danger btn-xs g-font-size-12 text-nowrap del-add-mailaddress')
                                            .text('削除')
                                        );
                wrap.append(add_address, del_btn);
                add_elm.append(wrap);
                add_elm_tr.append(add_elm);
                $('#ask-add-emails').remove();
                $('#email-address-list-tbody').append(add_elm_tr);
            }
            $('#wrap-email-address-list-table').addClass('opacity-01');
        }             
        target_modal.modal('show');
   });

   //メールアドレス設定modalでメールアドレスをappendした後にwidthを入れて整形する
   $('#modalSelectItems').on('shown.bs.modal',function(){
       if($('#wrap-email-address-list-table').hasClass('opacity-01')){
           $('#email-address-list-tbody').find('[data-item="add-address"]')
                                         .width($('#wrap-email-address-list-table').width() -2);
           $('#wrap-email-address-list-table').removeClass('opacity-01');
       }
   });

   //メールアドレス設定でアドレス追加ボタンを押した時
   $(document).on('click','#add-email-address',function(){
       var input = $('#email-address-to-add');
       var address = input.val();
       var is_same_address = false;
       $('#email-address-list-tbody').find('[data-item="address-text"]')
                                     .each(function(){
           if(address === $(this).text()){
               is_same_address = true;
           }
       });    
       if(address === ''){
           alert('メールアドレスを入力して下さい');
           return false;
       }else if(!address.match(/^([^@\s]+)@((?:[-a-z0-9]+\.)+[a-z]{2,})$/i)){
           alert('メールアドレスに誤りがあります');
           return false;
       }else if(is_same_address){
           alert('既に同じメールアドレスが存在します');
           return false;
       }
       var add_elm_tr = $('<tr>').attr({ 'class': 'list_alternative2'} );
       var add_elm = $('<td>').attr({
                              'class': '',
                              'data-item': 'add-address'
                              }).width($('#wrap-email-address-list-table').width());
       var wrap = $('<div>').attr('class', 'd-flex justify-content-between');
       var add_address = $('<div>').attr({
                                   'class': 'pt-1 ml-2 mr-1 text-truncate',
                                   'data-item': 'address-text'
                                   }).text(address);
       var del_btn = $('<div>').attr('class', 'mr-1')
                               .append(
                                   $('<a>').attr('class', 'btn btn-danger btn-xs g-font-size-12 text-nowrap del-add-mailaddress')
                                           .text('削除')
                               );
       wrap.append(add_address, del_btn);
       add_elm.append(wrap);
       add_elm_tr.append(add_elm);
       $('#ask-add-emails').remove();
       $('#email-address-list-tbody').append(add_elm_tr);
       input.val('');
   });

   //resizeでメールアドレス設定modal・ホスト検索modal内のテーブル表示が崩れないようにする
   $(window).on('resize', function(){
       $('#email-address-list-tbody').find('[data-item="add-address"]')
                                     .width($('#wrap-email-address-list-table').width() - 2);
       $('#host-list-tbody').find('[data-item="host"]')
                            .width($('#wrap-host-list-table').width() - 2);
       //ホスト名検索欄のtop調整
       $('#select-searched-host-area').css('top', $('#cond-detail-search-host').outerHeight() + 'px');
   });

   //メールアドレス設定で削除ボタンを押した時
   $(document).on('click', '#modalSelectItems .del-add-mailaddress', function(){
       $(this).closest('[data-item="add-address"]').remove();
       if($('#email-address-list-tbody').find('[data-item="add-address"]').length === 0){
           var show_no_address = $('<tr>').attr({'class': 'font-weight-bold d-block pr-0 pl-0', 'id': 'ask-add-emails'})
                                          .append(
                                              $('<td>').attr({
                                                       'class': 'd-block',
                                              }).text('メールアドレスを追加して下さい')
                                          );
           $('#email-address-list-tbody').append(show_no_address);
       }
   });

   //メールアドレス設定で決定ボタンを押した時
   $(document).on('click', '#execute-add-email-addresses', function(){
       var cnt = 0;
       $('#mail-change-add-address-area > *').remove();
       $('#add-address-area-tbody > *').remove();
       if($('#modalEditOperation').is(':visible')){
           var input_addresses = $('#email-address-list-tbody').find('[data-item="address-text"]');
           if(input_addresses.length === 0){
               alert('変更するメールアドレスがありません');
               return false;
           }
           input_addresses.each(function(){
               var add_elm = $('<div>').attr({
                                       'class': 'min-h-30px pt-1 pl-1 g-font-size-14',
                                       'data-address': 'true',
                                       }).text($(this).text());
               if(cnt % 2 === 1){
                   add_elm.addClass('user-name-bg');
               }
               $('#mail-change-add-address-area').append(add_elm);
               cnt += 1;
           });
       }else{
           $('#email-address-list-tbody').find('[data-item="address-text"]')
                                         .each(function(){
               var add_elm = $('<tr>').append(
                                          $('<td class="text-left">').attr('data-address', 'true')
                                                   .text($(this).text())
                                       );
               if(cnt % 2 === 1){
                   add_elm.addClass('user-name-bg');
               }
               $('#add-address-area-tbody').append(add_elm);
               cnt += 1;
           });
       }
       $('#modalSelectItems').modal('hide');
   }); 
 
   //メニュー項目を追加するmodalが閉じた時
   $('#modalSelectItems').on('hidden.bs.modal',function(){
       var modal = $(this);
       modal.find('.modal-body').removeClass('min-height-300px');
       modal.find('[data-selectitem]').prop('disabled', false);
   });

   //メール設定modalの宛先ユーザの表示項目切り替え
   $('#switch-address-email-or-user').on('change', function(){
       var val = $(this).find('option:selected').val();
       if(val === 'input-address'){
           $('#search-address-type-items-title').text('送信先アドレス');
           $('#search-address-items-area').hide();
           $('#wrap-add-address-area').show();
       }else{
           $('#search-address-type-items-title').text('送信先ユーザ');
           $('#wrap-add-address-area').hide();
           $('#search-address-items-area').show();   
       }
           
    });

    //メール設定modalの変更内容破棄確認 
    $('.close-modalOperationMailSetting').on('click', function(){
        var sender = $('#ope-detail-readonly-sender').val();
        var address_type_items = $('#search-address-items-area-tbody').find('[data-addresstype]');
        var title = $('#send-mail-title').val();
        var content = $('#send-mail-content').val();
        if(sender != '' || address_type_items.length > 0 || title != '' || content != ''){ 
            var conf = confirm('変更した内容は破棄されます。よろしいですか？');
            if(conf){
                $('#modalOperationMailSetting').modal('hide');
            }else{
                return false;
            }
        }
        $('#search-address-items-area-tbody > *').remove();
        $('#modalOperationMailSetting').modal('hide');
    });

    //サマリーテーブルの操作popoverの詳細ボタンを押した時、その内容を表示
    $(document).on('click', '.open-action-detail', function(e){
        var button = $(this);
        var action_name = button.attr('data-select');
        var displayed_detail_popup_action_name = $('.rule-detail-popup').attr('data-select');
        $('.rule-detail-popup').remove();
        if(action_name === displayed_detail_popup_action_name){
            return false;
        }
        var data = JSON.parse(button.attr('data-detailinfo'));
        var div = $('<div>').attr({
                            'class': 'rule-detail-popup',
                            'data-select': action_name,
                            });
        var content = $('<div>').attr('class', 'original-popover-inner text-left max-height-170px max-width-250px');
        if(action_name === 'send-mail-user' || action_name === 'send-mail-category'){
            var sender = $('<div>').text('送信元: ' + data['sender']);
            var address = $('<div>').text('送信先: ' + data['address']);
            var title = $('<div>').text('件名: ' + data['title']);
            var message = $('<div>').text('メール本文: ' + data['content']);
            content.append([sender, address, title, message]);
        }else if(action_name === 'status-change'){
            var statuses = {};
            var status_from = $('<div>').text(
                                            (
                                             data["status_from"].map(function(status_id){
                                                return all_statuses[status_id]
                                             }).join(',')
                                             ) + 'のステータスを'
                                         );
           var status_to = $('<div>').text(
                                      (all_statuses[data["status_to"]]) + 
                                      'に変更'
                                     );
           content.append([status_from, status_to]);
       }else if(action_name === 'due-date-change'){
           if(data["date"]){
               var date = $('<div>').text('対策予定日を' + data["date"].replace(/[-]/g, '/') + 'に設定する: 設定済みの場合' + (data["permit-update"] ? '上書きする' : '上書きしない'));
               content.append(date);
           }else{
               var days = $('<div>').text('対策予定日を脆弱性検出日の' + data["days"] + '日後に設定する: 設定済みの場合' + (data["permit-update"] ? '上書きする' : '上書きしない'));
               content.append(days);
               }
            
       }
       div.append(content);
       $('body').append(div.css('opacity', 0));
       var offset = button.offset();
       div.offset({
                'top': offset.top,
                'left': offset.left + button.outerWidth() + 3
       }).css({
           'opacity': ''
       });
      
       e.stopPropagation();
    });      

    function confirm_close_RegisterModal(id){
        if(id === 'modalRegisterCondition'){
            var form_param = JSON.parse($('#form-create-condition').val());
            var temp_param = JSON.parse($('#set-cond-param').val());
        }else{
            var form_param = JSON.parse($('#form-create-operation').val());
            var temp_param = JSON.parse($('#set-ope-param').val());
        }
        var form_param_len = Object.keys(form_param).length;
        var temp_param_len = Object.keys(temp_param).length;
        if(form_param_len === 0 && temp_param_len === 0){
            return true;
        }else{
            var pass = true;
                for(key in temp_param){
                    var data = form_param[key];
                    if(!data || data != temp_param[key]){
                         if(typeof data === "object"){
                             if(JSON.stringify(data) != JSON.stringify(temp_param[key])){
                                 pass = false;
                             }
                         }else{
                             pass = false;
                         }
                    } 
                }
                for(key in form_param){
                    var data = temp_param[key];
                    if(!data || data != form_param[key]){
                        if(typeof data === "object"){
                            if(JSON.stringify(data) != JSON.stringify(form_param[key])){
                                pass = false;
                            }
                        }else{
                            pass = false;
                        }
                    }
                }
            if(!pass){
                var conf = confirm('今回の変更は破棄されます。よろしいですか?(変更を反映するには設定ボタンをクリックして下さい)');
                if(conf){
                    if(id === 'modalRegisterCondition'){
                        $('#set-cond-param').val(JSON.stringify(form_param)); 
                        $('#set-condition-tbody').replaceWith($('#clone-set-condition-tbody'));
                        $('#clone-set-condition-tbody').attr('id','set-condition-tbody');
                        $('#set-condition-tbody').find('[data-toggle="tooltip"]').tooltip();
                    }else{
                        $('#set-ope-param').val(JSON.stringify(form_param));
                        $('#set-operation-tbody').replaceWith($('#clone-set-operation-tbody'));
                        $('#clone-set-operation-tbody').attr('id', 'set-operation-tbody');
                        $('#set-operation-tbody').find('[data-toggle="tooltip"]').tooltip();
                    }                    
                }else{
                     return false;
                }  
            }
        }
        if($('#cond-detail-add-cond-area:visible').find('[data-cond]').length > 0){
            var conf = confirm('編集中の条件詳細項目があります。このまま条件設定を終了してもよろしいですか?');
            if(!conf){
                return false;
            }
         }else if($('#ope-detail-add-ope-area:visible').find('[data-ope]').length > 0){
            var conf = confirm('編集中の操作詳細項目があります。このまま操作設定を終了してもよろしいですか?');
            if(!conf){
                return false;
            }
        }
        return true;
    }

    function display_no_operation_message(){
        if($('#set-operation-tbody').find('tr, #show-no-ope').length === 0){
             var div = $('<div>').attr({
                                 'class': 'font-weight-bold text-center p-1',
                                 'id':'show-no-ope'
                                  })
                                  .text('操作を追加して下さい');
             $('#set-operation-tbody').append(div);
             $('#execute-set-ope').prop('disabled', true);
        }
        var param = JSON.parse($('#set-ope-param').val());
        if(Object.keys(param).length === 0){
            $('#execute-set-ope').prop('disabled', true);
        }else{
            $('#execute-set-ope').prop('disabled', false);
        }
    }

    function display_no_condition_message(){
        if($('#set-condition-tbody').find('tr, #show-no-cond').length === 0){
             var div = $('<div>').attr({
                                 'class': 'font-weight-bold text-center p-1',
                                 'id':'show-no-cond'
                                  })
                                  .text('条件を追加して下さい');
             $('#set-condition-tbody').append(div);
        }
        var param = JSON.parse($('#set-cond-param').val());
        if (Object.keys(param).length === 0){     
            $('#execute-set-cond').prop('disabled', true);
        }else{
            $('#execute-set-cond').prop('disabled', false);
        }
    } 

    function check_blank_in_trigger_setting_modal(){
        var selected = $('#select-trigger-time-area input:checked');
        if(selected.val() === 'vulsync'){
            var category_id = $('#trigger-category').attr('data-categoryid');
            var selected_val = category_id === '-1' ? null : category_id;
        }else{
            var selected_val = selected.closest('[data-wrapping=periodic-time]')
                                       .find('option:selected').val();
        }
        if(selected_val){
            $('#set-trigger-setting').prop('disabled', false);
        }else{
            $('#set-trigger-setting').prop('disabled', true);
        }
    }

    function reset_mail_select_type(){
        $('#wrap-send-mail-sender-users-list').remove();
        $('#select-mail-sender-type').val('manual');
        $('#send-mail-sender-manual-input').val('').show();
        $('#send-mail-sender-select-input').val('');
        $('#send-mail-sender-select-area').hide();
        $('#switch-address-email-or-user').val('user-select');
        $('#wrap-add-address-area').hide();
        $('#search-address-items-area').show();
        $('#add-address-area-tbody > *').remove();
        $('#email-address-to-add').val('');
    }
 
    function make_original_tooltip(text, is_arrow_down){
        var original_tooltip = $('<div>').attr({
                                   'id': 'original-tooltip',
                                   'class': 'position-fixed original-tooltip',
                               }).hide();
        var inner = $('<div>').attr('class', is_arrow_down ? 'inner-original-tooltip-down' : 'inner-original-tooltip ')
                              .html(text);
        original_tooltip.append(inner);
        return original_tooltip;  
    }

    function append_original_tooltip(data, is_update, is_edit){
        $('#original-tooltip:visible').remove();
        var text = is_update ? '変更しました!' : '追加しました!';
        var original_tooltip = make_original_tooltip(text);
        $('body').append(original_tooltip);
        if(!is_edit){
            var target_elm = data === 'ope' ? $('#modalRegisterOperation-table') : $('#modalRegisterCondition-table');
            var target_offset = target_elm.offset();
            var offset_top = target_offset.top + target_elm.outerHeight() + 8;
            var offset_left_target = (data === 'ope' ? target_elm.find('.set-operation-t-mid') : target_elm.find('.set-condition-t-mid'));
            var offset_left = offset_left_target.offset().left + (offset_left_target.outerWidth() / 2) - (original_tooltip.outerWidth() / 2);
            original_tooltip.offset({'top': offset_top, 'left': offset_left}).show();
            var set_timer = setTimeout(function(){
                                $('#original-tooltip:visible').remove();
                            }, 3000);
        }else{
            setTimeout(function(){
                var target_elm = data === 'ope' ? $('#modalRegisterOperation-table') : $('#modalRegisterCondition-table');
                var target_offset = target_elm.offset();
                var offset_top = target_offset.top + target_elm.outerHeight() + 8;
                var offset_left_target = data === 'ope' ?  target_elm.find('.set-operation-t-mid') : target_elm.find('.set-condition-t-mid');
                var offset_left = offset_left_target.offset().left + (offset_left_target.outerWidth() / 2) - (original_tooltip.outerWidth() / 2);
                $('#original-tooltip').offset({'top': offset_top, 'left': offset_left}).show();
                var set_timer = setTimeout(function(){
                                    $('#original-tooltip:visible').remove();
                                }, 3000);
            },650);
        }  
    }

    //カテゴリの選択肢作成
    function make_selectable_categories(unique_name, is_all_necessary, is_condition_elm = false){
        var categories = JSON.parse($('#users_selectable_categories').val());
        var category_descs = JSON.parse($('#category-descs').val());
        var item = function(id_name, category_name, category_id, is_enable){     
                       var category_div = $('<div>').attr({
                           'class': 'category input-group g-brd-around g-brd-eee g-mb-2 g-bg-white g-font-size-13'
                       });
                       var content = $('<div>').attr({
                                         'class': 'input-group-prepend',
                                     });
                       var radio_section = $('<div>').attr({
                                             'class': 'input-group-text'
                                           }).append(
                                               $('<input>').attr({
                                                   'type': 'radio',
                                                   'class': 'c-category-radio',
                                                   'id': id_name,
                                                   'name': 'a-users-category',
                                                   'data-text': category_name,
                                               }).val(category_id).prop('disabled', !is_enable)
                                           );
                       var desc_cat = category_descs[category_id];
                       var label = $('<label>').attr({
                           'class': 'mb-0 px-1 form-control-plaintext form-control-sm text-truncate g-width-26em font-weight-normal',
                           'for': id_name,
                           'data-toggle': 'tooltip',
                           'data-title': (desc_cat ? desc_cat.desc : ''),
                           'data-placement': 'right',
                           'data-boundary': 'window'
                       }).text(category_name)
                       content.append(radio_section, label);
                       return category_div.append(content);
                   };
        var wrap = $('<div>').attr('id', unique_name);
        if(is_all_necessary){
            var user_auth = JSON.parse($('#user_auth').val());
            if(user_auth.includes('admin')){
                wrap.append(item(unique_name + '-all', '全体', 'all', true))
            }
        }
        var cats = [];
        for(category_id in categories){
            cats.push( {'id': category_id, 'name': categories[category_id] });
        }
        cats.sort(function(a, b){
            if( a.name.toLowerCase() < b.name.toLowerCase() ){
                return -1;
            }else{
                return 1;
            }             
        });
        if( is_condition_elm ){
            var trigger = JSON.parse($('#form-create-trigger').val());
            if( trigger['period'] === 'vulsync' && (trigger['category'] != 'all' || trigger['hostname'] != '') ){
                var is_custom_categories_needed = false;
            }else{
                var is_custom_categories_enable = $('#main-category-id').val() === 'all';
            }
            var system_categories = JSON.parse($('#system-categories').val());
            var system_cat_ids = Object.keys(system_categories).sort();
            for(i=0; i < system_cat_ids.length; i++){
                let target_cat_id = system_cat_ids[i];
                wrap.append(item(unique_name + target_cat_id, system_categories[target_cat_id], target_cat_id, true));
            }
            cats.forEach(function(cat){
                wrap.append(item(unique_name + cat.id, cat.name, cat.id, is_custom_categories_enable));
            });
        }else{
            cats.forEach(function(cat){
                wrap.append(item(unique_name + cat.id, cat.name, cat.id, true));
            });
        }
        return wrap;
    }

    function check_rule_is_registable(){
        var has_set_class = true;
        $('.open-register-modal').each(function(){
            if(!$(this).hasClass('set')){
                has_set_class = false;
                return false;
            }
        });
        if($('#form-create-rulename').val().length === 0 || !has_set_class){
            $('#execute-register-rule').prop('disabled', true);
        }else{
            $('#execute-register-rule').prop('disabled', false);
        }
    }

    function cond_display_content(data_name, item_name, content){
        return $('<dl>').attr({
            'class': 'text-left mx-1 g-mb-1 d-flex justify-content-between',
            'data-cond': data_name,
        }).append(
            $('<dt>').text(item_name),
            $('<dd>').attr({
                'class': 'badge badge-info mb-0 rule-detail-info g-text-shadow-title show-original-popover',
                'data-html': 'true',
                'data-item': data_name,
                'data-select': data_name,
                'data-content': content
            }).text('詳細')
        );
    }

    function create_process_data_select_elm(){
        var trigger_param = JSON.parse($('#form-create-trigger').val());
        var clone_select = $('#hidden-select-process-data-area').find('[data-id="select-process-data"]').clone(true);
        if(trigger_param['period'] === 'vulsync'){
            clone_select.find('[data-valuefor="periodic"]').remove();
        }else{
            clone_select.find('[data-valuefor="vulsync"]').remove();
        }
        return clone_select;
    }

    function createCtagDetail(ctagData, isCond){
        if( !ctagData.enable.length && !ctagData.not_enable.length ){
            return '<div>' + MATCH_IF_NO_CTAG_LABEL + '</div>';
        }
        const enableElms = ctagData.enable.length ?
            $('<div>').append(
                $('<div>').attr('class', 'font-weight-bold').text(isCond ? 'あれば成立' : '設定する'),
                $('<div>').append(
                    ctagData.enable.map(ctag => createSelectedCtagElm(ctag) )
                )
            ) :
            ''
        ;
        const notEableElms = ctagData.not_enable.length ?
            $('<div>').attr('class', 'mt-1').append(
                $('<div>').attr('class', 'font-weight-bold').text(isCond ? 'なければ成立' : '除外する'),
                $('<div>').append(
                    ctagData.not_enable.map(ctag => createSelectedCtagElm(ctag) )
                )
            ) :
            ''
        ;
        return $('<div>').append(enableElms, notEableElms).html(); //文字列化
    }

    function createCtagCondDisplayElm(ctagData){
        var ctagContent = createCtagDetail(ctagData, true);
        return cond_display_content('ctag' + ctagData.dataid, cond_detail_items_name['ctag'], ctagContent);
    }

    function create_cond_display_content(parameter, action_name, popover_elm, button){
        if(action_name === 'open-detail'){
            var cond_area = $('<div>');
        }else if(action_name === 'modify'){
            var cond_area = $('#rule-setting-item-cond');
        }else{
            var cond_area = $('#rule-setting-item-cond');
            cond_area.html('');
        }
        var sri_values = { "1" : "低", "2" : "中", "3" : "高", "4" : "重大" };
        var cond_cnt = 0;
        var cond_size = Object.keys(parameter).length;
        if( popover_elm ){
            popover_elm.addClass('invisible');
        }
        for(key in parameter){
            var cond_data = parameter[key];
            var cond_item_name = cond_detail_items_name[key];
            if(key === 'issue-update-regi' || key === 'issue-update-del'){
                var label = (key === 'issue-update-regi' ? '脆弱性の新規登録が' : '脆弱性影響無への変更が');
                var setting_content = cond_display_content(key, cond_item_name, label + cond_data + '件以上');
                cond_area.append($(setting_content));
                cond_cnt += 1;
            }else if(key === 'issue-update-regi-and-del'){
                var regi_num = cond_data['regi'];
                var del_num = cond_data['del'];
                var cond_content = '脆弱性の新規登録が' + regi_num + '件以上且つ' +
                                   '脆弱性影響無への変更が' + del_num + '件以上';
                var setting_content = cond_display_content(key, cond_item_name, cond_content);
                cond_area.append($(setting_content));
                cond_cnt += 1;
            }else if(key === 'due-date-approach'){
                var days = cond_data["val"];
                var setting_content = cond_display_content(key, cond_item_name, '対策予定日が' + days + '日以内');
                cond_area.append($(setting_content));
                cond_cnt += 1;
            }else if(key === 'history-create'){
                var history_create_display_content = '履歴の投稿が合計' + cond_data["val"] + '件以上';
                var history_create_display_content_item_name = cond_item_name;
                var history_create_display_content_key = key;
                var history_create_display_content_flg_exclusion = cond_data['flg_exclusion'];
                if( history_create_display_content_flg_exclusion ){
                    $.post($('#get_users_path').val(), { 'user_ids': JSON.stringify(cond_data['user_ids']), 'is_sidfmadm_selectable': true }, function(res){
                        var user_names = res.map(function(user){
                            return (user['lastname'] + user['firstname'])
                        }).join(', ');
                        history_create_display_content += '<br>' + (history_create_display_content_flg_exclusion === 'true' ? 'ユーザが作成した履歴以外が対象' : 'ユーザが作成した履歴のみが対象') + '<br>' + '対象ユーザ: ' + escape_html_special_chars(user_names);
                        cond_area.append(
                            $(
                                cond_display_content(
                                    history_create_display_content_key, history_create_display_content_item_name, history_create_display_content
                                )
                            )
                        );
                        cond_cnt += 1;
                    });
                }else{
                    cond_area.append(
                        $(
                            cond_display_content(
                                history_create_display_content_key, history_create_display_content_item_name, history_create_display_content
                            )
                        )
                    );
                    cond_cnt += 1;
                }
            }else if(key === 'sri-above' || key === 'sri-below'){
                var level = (key === 'sri-above' ? '以上' : '以下');
                var sri_name = sri_values[cond_data];
                var setting_content = cond_display_content(key, cond_item_name, 'SRI値が' + sri_name + level);
                cond_area.append($(setting_content));
                cond_cnt += 1;
            }else if(key === 'sri-warning' || key === 'cvss-warning'){
                var cond_content = (cond_data === 'true' ? '要対応条件に一致' : '要対応条件に不一致');
                var setting_content = cond_display_content(key, cond_item_name, cond_content);
                cond_area.append($(setting_content));
                cond_cnt += 1;
            }else if(key === 'cvss-notreg'){
                var cond_content = (cond_data === 'true' ? 'CVSS値が未定義' : 'CVSS値が未定義以外');
                var setting_content = cond_display_content(key, cond_item_name, cond_content);
                cond_area.append($(setting_content));
                cond_cnt += 1;
            }else if(key === 'cvss-above' || key === 'cvss-below'){
                var v = cond_data["val"]
                var exploitable = cond_data["flg_exploitable"]
                if(exploitable === 'all'){
                    exploitable = '全ての脆弱性が対象';
                }else if(exploitable === 'not_exist'){
                    exploitable = '無しの脆弱性が対象';
                }else{
                    exploitable = '有りの脆弱性が対象';
                }
                var inc_notreg = cond_data["include_notreg"] === 'true' ? '未定義の脆弱性を含める' : '未定義の脆弱性を含めない';
                var additional_info = '攻撃コード: ' + exploitable + '<br>' +
                                     '未定義を含めるかどうか: ' + inc_notreg;
                var level = (key === 'cvss-above' ? '以上' : '以下');
                var cond_content = 'CVSS値が' + v + level + '<br>' +
                                    additional_info;
                var setting_content = cond_display_content(key, cond_item_name, cond_content);
                cond_area.append($(setting_content));
                cond_cnt += 1;
            }else if(key === 'cvss-vector'){
                var cvss = JSON.parse($('#cvss-vectors-data').val());
                var cond_content = ''
                $.each(cond_data['vectors'], function(k, values){
                    var target_cvss = cvss[k];
                    cond_content += target_cvss['item_name'] + ':&nbsp;'
                    sort_cvss_values(k, values)  
                    $.each(values, function(idx, v){
                        cond_content += target_cvss['values'][v];
                        if( idx === values.length - 1 ) {
                            cond_content += '<br>';
                        }else{
                            cond_content += ', ';
                        }
                    });
                });
                var cvss_vectors_opts = JSON.parse($('#cvss-vectors-opts').val());
                cond_content += cvss_vectors_opts['flg_perfect']['label'] + ': ' +
                                cvss_vectors_opts['flg_perfect']['options'][cond_data['flg_perfect']];
                if( cond_data['flg_perfect'] == 'true' ){
                    cond_content += '<br>';
                    cond_content += cvss_vectors_opts['flg_ignore_undef']['label'] + ': ' +
                                    cvss_vectors_opts['flg_ignore_undef']['options'][cond_data['flg_ignore_undef']];
                }
                //cond_content = cond_content.slice(0, -4); //末端の改行(<br>)を削除
                var setting_content = cond_display_content(key, cond_item_name, cond_content);
                cond_area.append($(setting_content));
                cond_cnt += 1;
            }else if(key === 'due-date-expired'){
                var setting_content = cond_display_content(key, cond_item_name,'対策予定日を超過');
                cond_area.append($(setting_content));
                cond_cnt += 1;
            }else if(key === 'due-date-undef'){
                var setting_content = cond_display_content(key, cond_item_name,'対策予定日が未設定');
                cond_area.append($(setting_content));
                cond_cnt += 1;
            }else if(key === 'category-match' || key === 'category-unmatch'){
                var users_categories = JSON.parse($('#users_selectable_categories').val());
                var system_categories = JSON.parse($('#system-categories').val());
                var categories = {...users_categories, ...system_categories}
                var category_descs = JSON.parse($('#category-descs').val());
                var cat_names = [];
                for(var i = 0; i < cond_data.length; i++){
                    var cat_id = cond_data[i];
                    var desc_cat = category_descs[cat_id];
                    var cat_elm = "<span data-toggle='tooltip' data-title='" + escape_html_special_chars(desc_cat ? desc_cat.desc : '') + "' data-placement='top' data-boundary='window'>" + escape_html_special_chars(categories[cat_id]) + '</span>';
                    cat_names.push(cat_elm);
                }
                var setting_content = cond_display_content(key, cond_item_name, cat_names.join(','));
                cond_area.append($(setting_content));
                cond_cnt += 1;
            }else if(key === 'host-match'){
                var hostnames = cond_data;
                var get_display_name_path = $('#sidfm_internal_api_get_hosts_display_name_path').val();
                var match_host_item_name = cond_item_name;
                var match_host_key = key;
                $.ajax({
                    type: 'POST',
                    url: get_display_name_path,
                    data: {'hostnames': JSON.stringify(hostnames)}
                }).done(function(hostsdata){
                    hostnames = hostnames.map(function(h_name){
                        var match_host_display_name = hostsdata[h_name];
                        var host_elm = "<span data-toggle='tooltip' data-html='true' data-placement='right' data-boundary='window' data-title='" + 'ホスト名: ' + h_name + '<br/>表示名: ' + (match_host_display_name ? match_host_display_name : '(未設定)') + '\'>' + (match_host_display_name ? match_host_display_name : h_name) + '</span>';
                        return host_elm;
                    });
                    var setting_match_host_content = cond_display_content(match_host_key, match_host_item_name, hostnames.join(', '));
                    cond_area.append($(setting_match_host_content));
                    cond_cnt += 1;
                });
            }else if(key === 'host-unmatch'){
                var unmatch_hostnames = cond_data;
                var get_display_name_path = $('#sidfm_internal_api_get_hosts_display_name_path').val();
                var unmatch_host_item_name = cond_item_name;
                var unmatch_host_key = key;
                $.ajax({
                    type: 'POST',
                    url: get_display_name_path,
                    data: {'hostnames': JSON.stringify(unmatch_hostnames)}
                }).done(function(unmatch_hostsdata){
                    unmatch_hostnames = unmatch_hostnames.map(function(unmatch_h_name){
                        var unmatch_host_display_name = unmatch_hostsdata[unmatch_h_name];
                        var unmatch_host_elm = "<span data-toggle='tooltip' data-html='true' data-placement='right' data-boundary='window' data-title='" + 'ホスト名: ' + unmatch_h_name + '<br/>表示名: ' + (unmatch_host_display_name ? unmatch_host_display_name : '(未設定)') + '\'>' + (unmatch_host_display_name ? unmatch_host_display_name : unmatch_h_name) + '</span>';
                        return unmatch_host_elm;
                    });
                    var setting_unmatch_host_content = cond_display_content(unmatch_host_key, unmatch_host_item_name, unmatch_hostnames.join(', '));
                    cond_area.append($(setting_unmatch_host_content));
                    cond_cnt += 1;
                });
            }else if(key === 'host-syncing'){
                var cond_content = (cond_data === 'true' ? '同期中': '非同期') + 'のホスト';
                var setting_content = cond_display_content(key, cond_item_name, cond_content);
                cond_area.append($(setting_content));
                cond_cnt += 1;
            }else if(key === 'status-match' || key === 'status-unmatch'){
                var status_names = [];
                for(var i = 0; i < cond_data.length; i++){
                    var status_id = cond_data[i];
                    status_names.push(all_statuses[status_id]);
                }
                var cond_content = '対象ステータス:&nbsp' + status_names.join(',');
                var setting_content = cond_display_content(key, cond_item_name, cond_content);
                cond_area.append($(setting_content));
                cond_cnt += 1;
            }else if(key === 'product-match'){
                const match_prod_key = key;
                const match_prod_item_name = cond_item_name;
                const create_cond_elm = async () => {
                    const match_prod_names = await get_products_by_ids({
                        prod_ids: cond_data['product-ids'],
                        ver_ids: cond_data['version-ids'],
                        lprod_ids: cond_data['lproduct-ids'],
                        lver_ids: cond_data['lversion-ids']
                    });
                    const setting_prod_content = cond_display_content(match_prod_key, match_prod_item_name, match_prod_names.join(',<br>'));
                    cond_area.append($(setting_prod_content));
                    cond_cnt += 1;
                };
                create_cond_elm();
            }else if( key === 'product-unmatch' ){
                const unmatch_prod_key = key;
                const unmatch_prod_item_name = cond_item_name;
                const create_cond_elm = async () => {
                    const unmatch_prod_names = await get_products_by_ids({
                        prod_ids: cond_data['product-ids'],
                        ver_ids: cond_data['version-ids'],
                        lprod_ids: cond_data['lproduct-ids'],
                        lver_ids: cond_data['lversion-ids']
                    });
                    const setting_unmatch_prod_content = cond_display_content(unmatch_prod_key, unmatch_prod_item_name, unmatch_prod_names.join(',<br>'));
                    cond_area.append($(setting_unmatch_prod_content));
                    cond_cnt += 1;
                };
                create_cond_elm();
            }else if(key === 'cvss-exploitable'){
                var exploitable_content = (cond_data['value'] === 'true' ? '攻撃コードが有り' : '攻撃コードが無し');
                if(cond_data['is_change'] === 'true'){
                    exploitable_content += '<br>※攻撃コードの更新がある脆弱性のみが対象';
                }
                var setting_prod_content = cond_display_content(key, cond_item_name, exploitable_content);
                cond_area.append($(setting_prod_content));
                cond_cnt += 1;
            }else if(key === 'ctag'){
                cond_data.forEach((ctagData, idx) => { //カスタムタグは複数ある可能性がある配列
                    //var ctagContent = createCtagDetail(ctagData, true);
                    //var setting_prod_content = cond_display_content(key + idx, cond_item_name, ctagContent);
                    cond_area.append($(createCtagCondDisplayElm(ctagData)));
                });
                cond_cnt += 1; //条件の数は異なるが、カスタムタグは１にまとめてあるのでこれで動く
            }else if(key === 'timerange'){
                cond_data.forEach((timerange_data) => { //期間指定は複数ある可能性がある配列
                    var timerange_text = create_timerange_text(timerange_data);
                    var cond_content = '単位:' + timerange_text['unit'] + '&nbsp;&nbsp;間隔:' + timerange_text['value'] + '&nbsp;&nbsp;' + timerange_text['except']; 
                    var setting_content = cond_display_content(key + timerange_data.dataid, cond_item_name, cond_content);
                    cond_area.append($(setting_content));
                });
                cond_cnt += 1; //条件の数は異なるが、期間指定は１にまとめてあるのでこれで動く
            }
        }
        var timer_to_display_cond_area = setInterval(function(){
            if( cond_cnt === cond_size ){
                if( popover_elm ){
                    popover_elm.removeClass('invisible');
                    var btn_width = button.outerWidth(true);
                    var popover_width = popover_elm.outerWidth(true);
                    var offset = button.offset();
                    var offset_top = offset.top - popover_elm.outerHeight(true) - 5;
                    var offset_left = offset.left + (btn_width / 2) - (popover_width / 2);
                    popover_elm.offset({
                        'top': offset_top,
                        'left': offset_left
                    });
                }
                clearInterval(timer_to_display_cond_area);
            }
        });
        if(action_name === 'open-detail'){
            return cond_area;
        }
    }

    function ope_display_content(item_name, data_name, content){
         return $('<dl>').attr({
             'class': 'text-left mx-1 g-mb-1 d-flex justify-content-between',
             'data-addope': data_name,
         }).append(
             $('<dt>').attr('class', 'font-weight-bold mr-1').text(item_name),
             $('<dd>').attr({
                 'class': 'badge badge-info mb-0 rule-detail-info g-text-shadow-title show-original-popover',
                 'data-html': 'true',
                 'data-item': data_name,
                 'data-select': data_name,
                 'data-content': content
             }).text('詳細')
         )
     }

    function create_ope_display_content(parameter, action_name, popover_elm, button){
        var users_categories = JSON.parse($('#users_selectable_categories').val());
        if(action_name === 'open-detail'){
            var ope_area = $('<div>');
        }else if(action_name === 'modify-send-mail-ope' || action_name === 'modify-set-ctag'){
            var ope_area = $('#rule-setting-item-ope');
        }else{
            var ope_area = $('#rule-setting-item-ope');
            ope_area.html('');
        }
        if( popover_elm ){
            popover_elm.addClass('invisible');
        }
        var ope_cnt = 0;
        var ope_size = Object.keys(parameter).length;
        for(key in parameter){
            var ope_item_name = ope_detail_items_name[key];
            var ope_data = parameter[key];
            var process_data_info = process_values_with_names[ope_data['target']];
            if(key === 'send-mail-user'){
                var sendby = ope_data['sendby'];
                var title = escape_html_special_chars(ope_data['title']);
                var message = escape_html_special_chars(ope_data['content']);
                var sender = ope_data['sender'];
                var send_mail_user_content = [
                    '送信元: ' + escape_html_special_chars(sender['username'] ? sender['username'] : sender['address']),
                    '送信先:',
                    '件名: ' + title,
                    'メール本文: ' + message,
                    (ope_data['proc_empty'] === 'false' ? '※データの指定有り: 「' + process_data_info + '」' + 'が無い場合、メールを送信しない' : '※データの指定無し')
                ];
                if(sendby['ids']){
                    var send_mail_user_key = key;
                    var send_mail_user_item_name = ope_item_name;
                    var get_users_path = $('#sidfm_internal_api_get_data_for_mail_setting_path').val();
                    $.ajax({
                        type: 'POST',
                        url: get_users_path,
                        data: {'info_name' : 'user', 'ids': JSON.stringify(sendby['ids'])},
                    }).done(function(usernames){
                        send_mail_user_content[1] = '送信先: ' + usernames.map(function(name){ return escape_html_special_chars(name) }).join(',');
                        var setting_mail_content = ope_display_content(send_mail_user_item_name, send_mail_user_key, send_mail_user_content.join('<br>'));
                        ope_area.append($(setting_mail_content));
                        ope_cnt += 1;
                    });
                    
                }else{
                    send_mail_user_content[1] = '送信先:' + sendby['address'].map(function(address){ return escape_html_special_chars(address) }).join(',');
                    var setting_content = ope_display_content(ope_item_name, key, send_mail_user_content.join('<br>'));
                    ope_area.append($(setting_content));
                    ope_cnt += 1;
                }
            }else if(key === 'send-mail-category'){
                var title = escape_html_special_chars(ope_data['title']);
                var message = escape_html_special_chars(ope_data['content']);
                var sender = ope_data['sender'];
                var send_mail_category_id = ope_data['sendby']['ids'][0];
                var category_descs = JSON.parse($('#category-descs').val());
                var send_mail_category_content = [
                    '送信元: ' + escape_html_special_chars(sender['username'] ? sender['username'] : sender['address']),
                    '送信先: ' + (send_mail_category_id === 'all' ? '全体' : (users_categories[send_mail_category_id] ? "<span data-toggle='tooltip' data-title='" + escape_html_special_chars(category_descs[send_mail_category_id] ? category_descs[send_mail_category_id].desc : '') + "' data-placement='top' data-boundary='window'>" + escape_html_special_chars(users_categories[send_mail_category_id]) + '</span>' : '表示できないカテゴリです')),
                    '件名: ' + title,
                    'メール本文: ' + message,
                    (ope_data['proc_empty'] === 'false' ? '※データの指定有り: 「' + process_data_info + '」' + 'が無い場合、メールを送信しない' : '※データの指定無し')
                ].join('<br>');
                var setting_content = ope_display_content(ope_item_name, key, send_mail_category_content);
                ope_area.append($(setting_content));
                ope_cnt += 1;
            }else if(key === 'status-change'){
                var status_ids_from = ope_data['status_ids_from'];
                var status_ids_from_names = [];
                for(var i = 0; i < status_ids_from.length; i ++){
                    var status_id = status_ids_from[i];
                    status_ids_from_names.push(all_statuses[status_id]);
                }
                var status_id_to = ope_data['status_ids_to'][0];
                var ope_content = status_ids_from_names.join(',') + 'の場合 ' + "<i class='fas fa-arrow-right mr-1'></i>" +
                                  all_statuses[status_id_to] + 'に変更' +
                                  '<br>※処理するデータ: ' + process_data_info
                ;
                var setting_content = ope_display_content(ope_item_name, key, ope_content);
                ope_area.append($(setting_content));
                ope_cnt += 1;
            }else if(key === 'due-date-change'){
                var permit_update = (ope_data['permit-update'] === 'true') ? '上書き設定: 上書き' : '上書き設定: 上書きしない';
                if(ope_data['date']){
                    var ope_content = '対策予定日を' + ope_data['date'] + 'に設定する' + '<br>' +
                                      permit_update + 
                                      '<br>※処理するデータ: ' + process_data_info
                    ;
                }else if(ope_data['days']){
                    var ope_content = '対策予定日を脆弱性検出日の' + ope_data['days'] + '日後に設定する' + '<br>' +
                                      permit_update + 
                                      '<br>※処理するデータ: ' + process_data_info
                    ;
                }
                var setting_content = ope_display_content(ope_item_name, key, ope_content);
                ope_area.append($(setting_content));
                ope_cnt += 1;
            }else if(key === 'send-syslog'){
                var syslog_text = create_syslog_name_and_content(ope_data);
                var ope_content = [
                    syslog_text['protocol']['name'] + ': ' + syslog_text['protocol']['content'],
                    syslog_text['address']['name'] + ': ' + syslog_text['address']['content'].replace(/<br>/g, ', '),
                    syslog_text['facility']['name'] + ': ' + syslog_text['facility']['content'],
                    syslog_text['severity']['name'] + ': ' + syslog_text['severity']['content'],
                    syslog_text['message']['name']  + ':<br>  ' + syslog_text['message']['content'].replace(/<br>/g, ', '),
                    '※処理するデータ: ' + process_data_info
                ]
                var setting_content = ope_display_content(ope_item_name, key, ope_content.join('<br>'));
                ope_area.append($(setting_content));
                ope_cnt += 1;
            }else if(key === 'set-ctag'){
                var ope_content = createCtagDetail(ope_data, false);
                ope_content += '<div class="mt-1">※処理するデータ: ' + process_data_info + '</div>';
                var setting_content = ope_display_content(ope_item_name, key, ope_content);
                ope_area.append($(setting_content));
                ope_cnt += 1;
            }
        }
        var timer_to_display_ope_elms = setInterval(function(){
            if( ope_size === ope_cnt ){
                if( popover_elm ){
                    popover_elm.removeClass('invisible');
                    var btn_width = button.outerWidth(true);
                    var popover_width = popover_elm.outerWidth(true);
                    var offset = button.offset();
                    var offset_top = offset.top - popover_elm.outerHeight(true) - 5;
                    var offset_left = offset.left + (btn_width / 2) - (popover_width / 2);
                    popover_elm.offset({
                        'top': offset_top,
                        'left': offset_left
                    });
                }
                clearInterval(timer_to_display_ope_elms);
            }
        });
        if(action_name === 'open-detail'){
            return ope_area;
        }else if(action_name === 'modify-send-mail-ope'){
            return true;
        }else{
            //ope_area.removeClass('expanded-detail').addClass('unexpanded-detail');
            //ope_area.append($('<i>').attr('class', 'expand-btn fas fa-angle-down mr-1'));
            //ope_area.addClass('text-left');
        }
    }

    //キーワードを入力するdropdownの作成
    //キーをここで管理して、キーを元にボタンのテキスト・キーワード・ツールチップの説明を作成する
    //all_keywords_menuとall_keywordsはUIで表示したい順番に並べる
    const all_keywords_menus = {
        'analysis'          : '分析結果を出力',
        'page-info'         : 'ページ情報を出力',
        'rule-info'         : 'ルール情報を出力',
        'host-info-all'     : 'ホスト情報を出力',
        'vul-info-all'      : '脆弱性情報を出力',
        'prod-info-all'     : 'プロダクト情報を出力',
        'host-info-match'   : 'ホスト情報を出力(条件一致)',
        'vul-info-match'    : '脆弱性情報を出力(条件一致)',
        'prod-info-match'   : 'プロダクト情報を出力(条件一致)',
        'vulsync-info-all'  : '脆弱性同期時情報を出力',
        'vulsync-info-match': '脆弱性同期時情報を出力(条件一致)',
    };
    
    const all_keywords = {
        'analysis-sri'      : ['%RISK_RATE_SRI_ALL%', '%UNHANDLED_RATE_SRI_ALL%', '%RISK_RATE_SRI_MATCH%', '%UNHANDLED_RATE_SRI_MATCH%'],
        'analysis-cvss'     : ['%RISK_RATE_CVSS_ALL%', '%UNHANDLED_RATE_CVSS_ALL%', '%RISK_RATE_CVSS_MATCH%', '%UNHANDLED_RATE_CVSS_MATCH%'],
        'analysis-common'   : ['%HANDLING_RATE_ALL%', '%HANDLING_RATE_MATCH%'],
        'page-info'         : ['%URL%', '%TITLE%'],
        'host-info-match'   : ['%OS_RATE_MATCH%','%OS_MATCH%', '%HOST_MATCH%', '%DISPLAY_NAME_MATCH%', '%HOSTINFO_MATCH%', '%S_CATEGORY_MATCH%', '%C_CATEGORY_MATCH%', '%HOST_LINKS_MATCH%'],
        'host-info-all'     : ['%OS_RATE_ALL%','%OS_ALL%', '%HOST_ALL%', '%DISPLAY_NAME_ALL%', '%HOSTINFO_ALL%', '%S_CATEGORY_ALL%', '%C_CATEGORY_ALL%', '%HOST_LINKS_ALL%'],
        'vul-info-match'    : ['%VULINFO_MATCH%', '%VULTITLE_MATCH%', '%VUL_LINKS_MATCH%', '%SIDfmID_MATCH%', '%SIDfm_LINKS_MATCH%', '%KB_MATCH%', '%MS_LINKS_MATCH%'],
        'vul-info-all'      : ['%VULINFO_ALL%', '%VULTITLE_ALL%', '%VUL_LINKS_ALL%', '%SIDfmID_ALL%', '%SIDfm_LINKS_ALL%', '%KB_ALL%', '%MS_LINKS_ALL%'],
        'prod-info-match'   : ['%APP_MATCH%', '%PRODUCT_MATCH%', '%LPRODUCT_MATCH%'],
        'prod-info-all'     : ['%APP_ALL%', '%PRODUCT_ALL%', '%LPRODUCT_ALL%'],
        'rule-info'         : ['%RULE_NAME%', '%RULE_CATEGORY%', '%RULE_TIMING%', '%RULE_COND%', '%RULE_ACTION%', '%RULE_DATE%'],
        'vulsync-info-all'  : [
            '%VULSYNC_ALL_HOST%', '%VULSYNC_ALL_DISPLAY_NAME%',
            '%VULSYNC_REGI_ALL%', '%VULSYNC_INFO_REGI_ALL%', '%VULSYNC_REGI_HOST_ALL%', '%VULSYNC_REGI_DISPLAY_NAME_ALL%', 
            '%VULSYNC_RENEW_ALL%', '%VULSYNC_INFO_RENEW_ALL%', '%VULSYNC_RENEW_HOST_ALL%', '%VULSYNC_RENEW_DISPLAY_NAME_ALL%',
            '%VULSYNC_DEL_ALL%', '%VULSYNC_INFO_DEL_ALL%', '%VULSYNC_DEL_HOST_ALL%', '%VULSYNC_DEL_DISPLAY_NAME_ALL%'
        ],
        'vulsync-info-match': [
            '%VULSYNC_REGI_MATCH%', '%VULSYNC_INFO_REGI_MATCH%', '%VULSYNC_REGI_HOST_MATCH%', '%VULSYNC_REGI_DISPLAY_NAME_MATCH%',
            '%VULSYNC_RENEW_MATCH%','%VULSYNC_INFO_RENEW_MATCH%', '%VULSYNC_RENEW_HOST_MATCH%', '%VULSYNC_RENEW_DISPLAY_NAME_MATCH%',
            '%VULSYNC_DEL_MATCH%', '%VULSYNC_INFO_DEL_MATCH%', '%VULSYNC_DEL_HOST_MATCH%', '%VULSYNC_DEL_DISPLAY_NAME_MATCH%'
        ],
    };
    const all_keywords_item_text_and_description = {
        '%RISK_RATE_SRI_MATCH%'            : ['危険度の分布(条件一致)', '条件に一致する脆弱性のSRI値を元に危険度の割合を出力するキーワードを入力します。'],
        '%RISK_RATE_SRI_ALL%'              : ['危険度の分布', '全てのホストの脆弱性のSRI値を元に危険度の割合を出力するキーワードを入力します。'],
        '%UNHANDLED_RATE_SRI_MATCH%'       : ['未対処率(条件一致)', '条件に一致する脆弱性のSRI値を元に未対処の脆弱性の割合を出力するキーワードを入力します。'],
        '%UNHANDLED_RATE_SRI_ALL%'         : ['未対処率', '全てのホストの脆弱性のSRI値を元に未対処の脆弱性の割合を出力するキーワードを入力します。'],
        '%RISK_RATE_CVSS_MATCH%'           : ['危険度の分布(条件一致)', '条件に一致する脆弱性のCVSS値を元に危険度の割合を出力するキーワードを入力します。'],
        '%RISK_RATE_CVSS_ALL%'             : ['危険度の分布', '全てのホストの脆弱性のCVSS値を元に危険度の割合を出力するキーワードを入力します。'],
        '%UNHANDLED_RATE_CVSS_MATCH%'      : ['未対処率(条件一致)', '条件に一致する脆弱性のCVSS値を元に未対処の脆弱性の割合を出力するキーワードを入力します。'],
        '%UNHANDLED_RATE_CVSS_ALL%'        : ['未対処率', '全てのホストの脆弱性のCVSS値を元に未対処の脆弱性の割合を出力するキーワードを入力します。'],
        '%HANDLING_RATE_MATCH%'            : ['対処状況の分布(条件一致)', '条件に一致する脆弱性の対策状況の割合を出力するキーワードを入力します。'],
        '%HANDLING_RATE_ALL%'              : ['対処状況の分布', '全てのホストの脆弱性の対策状況の割合を出力するキーワードを入力します。'],
        '%URL%'                            : ['URL', '本ページのURLを出力するキーワードを入力します。'],
        '%TITLE%'                          : ['ページタイトル', '本ページのタイトルを出力するキーワードを入力します。'],
        '%OS_MATCH%'                       : ['OS名一覧(条件一致)', '条件に一致するホストのOS名一覧を出力するキーワードを入力します。'],
        '%OS_ALL%'                         : ['OS名一覧', '全てのホストのOS名一覧を出力するキーワードを入力します。'],
        '%HOST_MATCH%'                     : ['ホスト名一覧(条件一致)', '条件に一致するホストのホスト名一覧を出力するキーワードを入力します。'],
        '%HOST_ALL%'                       : ['ホスト名一覧', '全てのホスト名一覧を出力するキーワードを入力します。'],
        '%DISPLAY_NAME_MATCH%'             : ['表示名一覧(条件一致)', '条件に一致するホストの表示名一覧を出力するキーワードを入力します。'], 
        '%DISPLAY_NAME_ALL%'               : ['表示名一覧', '全てのホストの表示名一覧を出力するキーワードを入力します。'],
        '%HOSTINFO_MATCH%'                 : ['ホスト情報一覧(条件一致)', '条件に一致するホストのホスト情報一覧を出力するキーワードを入力します。'],
        '%HOSTINFO_ALL%'                   : ['ホスト情報一覧', '全てのホストのホスト情報一覧を出力するキーワードを入力します。'],
        '%S_CATEGORY_MATCH%'               : ['システムカテゴリ一覧(条件一致)', '条件に一致するホストのシステムカテゴリ一覧を出力するキーワードを入力します。'],
        '%S_CATEGORY_ALL%'                 : ['システムカテゴリ一覧', '全てのホストのシステムカテゴリ一覧を出力するキーワードを入力します。'],
        '%C_CATEGORY_MATCH%'               : ['カスタムカテゴリ一覧(条件一致)', '条件に一致するホストのカスタムカテゴリ一覧を出力するキーワードを入力します。'], 
        '%C_CATEGORY_ALL%'                 : ['カスタムカテゴリ一覧', '全てのホストのカスタムカテゴリ一覧を出力するキーワードを入力します。'],
        '%HOST_LINKS_MATCH%'               : ['ホスト詳細リンク一覧(条件一致)', '条件に一致するホストの「ホスト詳細」ページへのリンク一覧を出力するキーワードを入力します。'],
        '%HOST_LINKS_ALL%'                 : ['ホスト詳細リンク一覧', '全てのホストの「ホスト詳細」ページへのリンク一覧を出力するキーワードを入力します。'],
        '%VULTITLE_MATCH%'                 : ['脆弱性タイトル一覧(条件一致)', '条件に一致する脆弱性の脆弱性タイトル一覧を出力するキーワードを入力します。'],
        '%VULTITLE_ALL%'                   : ['脆弱性タイトル一覧', '全ての脆弱性の脆弱性タイトル一覧を出力するキーワードを入力します。'],
        '%VULINFO_MATCH%'                  : ['脆弱性情報一覧(条件一致)', '条件に一致する脆弱性の脆弱性情報一覧を出力するキーワードを入力します。'],
        '%VULINFO_ALL%'                    : ['脆弱性情報一覧', '全ての脆弱性の脆弱性情報一覧を出力するキーワードを入力します。'],
        '%VUL_LINKS_MATCH%'                : ['脆弱性詳細リンク一覧(条件一致)', '条件に一致する脆弱性の「脆弱性詳細」ページへのリンク一覧を出力するキーワードを入力します。'],
        '%VUL_LINKS_ALL%'                  : ['脆弱性詳細リンク一覧', '全ての脆弱性の「脆弱性詳細」ページへのリンク一覧を出力するキーワードを入力します。'],
        '%OS_RATE_MATCH%'                  : ['利用OS率(条件一致)', '条件に一致するホストで利用されているOSの割合を出力するキーワードを入力します。'],
        '%OS_RATE_ALL%'                    : ['利用OS率', '全てのホストで利用されているOSの割合を出力するキーワードを入力します。'],
        '%SIDfmID_MATCH%'                  : ['SIDfmID番号一覧(条件一致)', '条件に一致する脆弱性に関連しているSIDfm情報のID番号一覧を出力するキーワードを入力します。'],
        '%SIDfmID_ALL%'                    : ['SIDfmID番号一覧', '全ての脆弱性に関連するSIDfm情報のID番号一覧を出力するキーワードを入力します。'],
        '%SIDfm_LINKS_MATCH%'              : ['SIDfm情報リンク一覧(条件一致)', '条件に一致する脆弱性に関連している「脆弱性情報の詳細」ページへのリンク一覧を出力するキーワードを入力します。'],
        '%SIDfm_LINKS_ALL%'                : ['SIDfm情報リンク一覧', '全ての脆弱性に関連する「脆弱性情報の詳細」ページへのリンク一覧を出力するキーワードを入力します。'],
        '%MS_LINKS_MATCH%'                 : ['MSカタログリンク一覧(条件一致)', '条件に一致する脆弱性に関連している「Microsoft Update カタログ」へのリンク一覧を出力するキーワードを入力します。'],
        '%MS_LINKS_ALL%'                   : ['MSカタログリンク一覧', '全ての脆弱性に関連する「Microsoft Update カタログ」へのリンク一覧を出力するキーワードを入力します。'],
        '%KB_MATCH%'                       : ['KB番号一覧(条件一致)', '条件に一致する脆弱性に関連しているKB番号一覧を出力するキーワードを入力します。'],
        '%KB_ALL%'                         : ['KB番号一覧', '全ての脆弱性に関連するKB番号一覧を出力するキーワードを入力します。'],
        '%APP_MATCH%'                      : ['検出アプリケーション一覧（条件一致)', '条件に一致するホストの検出アプリケーションの名前とパスの一覧を出力するキーワードを入力します。'],
        '%APP_ALL%'                        : ['検出アプリケーション一覧', '全てのホストの検出アプリケーションの名前とパスの一覧を出力するキーワードを入力します。'],
        '%PRODUCT_MATCH%'                  : ['手動登録プロダクト名一覧(条件一致)', '条件に一致するホストの手動登録プロダクト名一覧を出力するキーワードを入力します。'],
        '%PRODUCT_ALL%'                    : ['手動登録プロダクト名一覧', '全てのホストの手動登録プロダクト名一覧を出力するキーワードを入力します。'],
        '%LPRODUCT_MATCH%'                 : ['ローカルプロダクト名一覧（条件一致)', '条件に一致するホストのローカルプロダクト名一覧を出力するキーワードを入力します。'],
        '%LPRODUCT_ALL%'                   : ['ローカルプロダクト名一覧', '全てのホストのローカルプロダクト名一覧を出力するキーワードを入力します。'],
        '%RULE_NAME%'                      : ['ルール名', 'ルール名を出力するキーワードを入力します。'],
        '%RULE_CATEGORY%'                  : ['ルールのカテゴリ', 'ルールのカスタムカテゴリを出力するキーワードを入力します。'],
        '%RULE_TIMING%'                    : ['ルールのタイミング', 'ルールに設定されているタイミングの設定を出力するキーワードを入力します。'],
        '%RULE_COND%'                      : ['ルールの条件', 'ルールに設定されている条件の設定を出力するキーワードを入力します。'],
        '%RULE_ACTION%'                    : ['ルールの操作', 'ルールに設定されている操作の設定を出力するキーワードを入力します。'],
        '%RULE_DATE%'                      : ['ルールの実行日', 'ルールを実行した日付を出力するキーワードを入力します。'],
        '%VULSYNC_ALL_HOST%'               : ['同期対象のホスト名一覧', '条件の一致・不一致に関わらず、同期対象となった全てのホストのホスト名を出力するキーワードを入力します。'], 
        '%VULSYNC_ALL_DISPLAY_NAME%'       : ['同期対象のホストの表示名一覧', '条件の一致・不一致に関わらず、同期対象となった全てのホストの表示名を出力するキーワードを入力します。'],
        '%VULSYNC_INFO_REGI_ALL%'          : ['登録された脆弱性の情報一覧', '脆弱性同期時に新たに登録された脆弱性の情報一覧を出力するキーワードを入力します'],
        '%VULSYNC_INFO_RENEW_ALL%'         : ['更新された脆弱性の情報一覧', '脆弱性同期時に値が更新された脆弱性の情報一覧を出力するキーワードを入力します'],
        '%VULSYNC_INFO_DEL_ALL%'           : ['削除された脆弱性の情報一覧', '脆弱性同期時にステータスが「影響無」に更新された脆弱性の情報一覧を出力するキーワードを入力します'],
        '%VULSYNC_REGI_ALL%'               : ['登録された脆弱性一覧', '脆弱性同期時に新たに登録された脆弱性の一覧を出力するキーワードを入力します'],
        '%VULSYNC_RENEW_ALL%'              : ['更新された脆弱性一覧', '脆弱性同期時に値が更新された脆弱性の一覧を出力するキーワードを入力します'],
        '%VULSYNC_DEL_ALL%'                : ['削除された脆弱性一覧', '脆弱性同期時にステータスが「影響無」に更新された脆弱性の一覧を出力するキーワードを入力します'],
        '%VULSYNC_REGI_HOST_ALL%'          : ['登録された脆弱性のホスト名一覧',  '脆弱性同期時に新たに登録された脆弱性のホストのホスト名一覧を出力するキーワードを入力します'],
        '%VULSYNC_REGI_DISPLAY_NAME_ALL%'  : ['登録された脆弱性のホストの表示名一覧', '脆弱性同期時に新たに登録された脆弱性のホストの表示名一覧を出力するキーワードを入力します'],
        '%VULSYNC_RENEW_HOST_ALL%'         : ['更新された脆弱性のホスト名一覧', '脆弱性同期時に値が更新された脆弱性のホストのホスト名一覧を出力するキーワードを入力します'],
        '%VULSYNC_RENEW_DISPLAY_NAME_ALL%' : ['更新された脆弱性のホストの表示名一覧', '脆弱性同期時に値が更新された脆弱性のホストの表示名一覧を出力するキーワードを入力します'],
        '%VULSYNC_DEL_HOST_ALL%'           : ['削除された脆弱性のホスト名一覧', '脆弱性同期時にステータスが「影響無」に更新された脆弱性のホストのホスト名一覧を出力するキーワードを入力します'],
        '%VULSYNC_DEL_DISPLAY_NAME_ALL%'   : ['削除された脆弱性のホストの表示名一覧', '脆弱性同期時にステータスが「影響無」に更新された脆弱性のホストの表示名一覧を出力するキーワードを入力します'],
        '%VULSYNC_INFO_REGI_MATCH%'        : ['登録された脆弱性の情報一覧(条件一致)', '脆弱性同期時に新たに登録された脆弱性の内、条件に一致する脆弱性の情報一覧を出力するキーワードを入力します'],
        '%VULSYNC_INFO_RENEW_MATCH%'       : ['更新された脆弱性の情報一覧(条件一致)', '脆弱性同期時に値が更新された脆弱性の内、条件に一致する脆弱性の情報一覧を出力するキーワードを入力します'],
        '%VULSYNC_INFO_DEL_MATCH%'         : ['削除された脆弱性の情報一覧(条件一致)', '脆弱性同期時にステータスが「影響無」に更新された脆弱性の内、条件に一致する脆弱性の情報一覧を出力するキーワードを入力します'],
        '%VULSYNC_REGI_MATCH%'             : ['登録された脆弱性一覧(条件一致)', '脆弱性同期時に新たに登録された脆弱性の内、条件に一致する脆弱性の一覧を出力するキーワードを入力します'],
        '%VULSYNC_RENEW_MATCH%'            : ['更新された脆弱性一覧(条件一致)', '脆弱性同期時に値が更新された脆弱性の内、条件に一致する脆弱性の一覧を出力するキーワードを入力します'],
        '%VULSYNC_DEL_MATCH%'              : ['削除された脆弱性一覧(条件一致)', '脆弱性同期時にステータスが「影響無」に更新された脆弱性の内、条件に一致する脆弱性一覧を出力するキーワードを入力します'],
        '%VULSYNC_REGI_HOST_MATCH%'        : ['登録された脆弱性のホスト名一覧(条件一致)', '脆弱性同期時に新たに登録された脆弱性のホストの内、条件に一致するホストのホスト名一覧を出力するキーワードを入力します'],
        '%VULSYNC_REGI_DISPLAY_NAME_MATCH%': ['登録された脆弱性のホストの表示名一覧(条件一致)', '脆弱性同期時に新たに登録された脆弱性のホストの内、条件に一致するホストの表示名一覧を出力するキーワードを入力します'],
        '%VULSYNC_RENEW_HOST_MATCH%'        : ['更新された脆弱性のホスト名一覧(条件一致)', '脆弱性同期時に値が更新された脆弱性のホストの内、条件に一致するホストのホスト名一覧を出力するキーワードを入力します'], 
        '%VULSYNC_RENEW_DISPLAY_NAME_MATCH%': ['更新された脆弱性のホストの表示名一覧(条件一致)', '脆弱性同期時に値が更新された脆弱性のホストの内、条件に一致するホストの表示名一覧を出力するキーワードを入力します'],
        '%VULSYNC_DEL_HOST_MATCH%'          : ['削除された脆弱性のホスト名一覧(条件一致)', '脆弱性同期時にステータスが「影響無」に更新された脆弱性のホストの内、条件に一致するホストのホスト名一覧を出力するキーワードを入力します'],
        '%VULSYNC_DEL_DISPLAY_NAME_MATCH%'  : ['削除された脆弱性のホストの表示名一覧(条件一致)', '脆弱性同期時にステータスが「影響無」に更新された脆弱性のホストの内、条件に一致するホストの表示名一覧を出力するキーワードを入力します']
    };

    const keywords_with_narrow_down_option = [
      '%OS_MATCH%', '%OS_ALL%', '%HOST_MATCH%', '%HOST_ALL%', '%HOSTINFO_MATCH%', '%HOSTINFO_ALL%','%DISPLAY_NAME_MATCH%', '%DISPLAY_NAME_ALL%',
      '%S_CATEGORY_MATCH%','%S_CATEGORY_ALL%', '%C_CATEGORY_MATCH%','%C_CATEGORY_ALL%','%HOST_LINKS_MATCH%', '%HOST_LINKS_ALL%',
      '%VULTITLE_ALL%','%VULTITLE_MATCH%','%VULINFO_MATCH%','%VULINFO_ALL%','%VUL_LINKS_MATCH%','%VUL_LINKS_ALL%',
      '%SIDfmID_MATCH%', '%SIDfm_LINKS_MATCH%', '%MS_LINKS_MATCH%', '%SIDfmID_ALL%', '%SIDfm_LINKS_ALL%', '%MS_LINKS_ALL%', '%KB_MATCH%', '%KB_ALL%',
      '%APP_MATCH%', '%PRODUCT_MATCH%', '%LPRODUCT_MATCH%', '%APP_ALL%', '%PRODUCT_ALL%', '%LPRODUCT_ALL%',
      '%VULSYNC_ALL_HOST%', '%VULSYNC_ALL_DISPLAY_NAME%',
      '%VULSYNC_INFO_REGI_ALL%', '%VULSYNC_INFO_RENEW_ALL%', '%VULSYNC_INFO_DEL_ALL%','%VULSYNC_REGI_ALL%', '%VULSYNC_RENEW_ALL%', '%VULSYNC_DEL_ALL%',
      '%VULSYNC_INFO_REGI_MATCH%', '%VULSYNC_INFO_RENEW_MATCH%', '%VULSYNC_INFO_DEL_MATCH%', '%VULSYNC_REGI_MATCH%', '%VULSYNC_RENEW_MATCH%', '%VULSYNC_DEL_MATCH%',
      '%VULSYNC_REGI_HOST_ALL%', '%VULSYNC_REGI_DISPLAY_NAME_ALL%','%VULSYNC_RENEW_HOST_ALL%', '%VULSYNC_RENEW_DISPLAY_NAME_ALL%','%VULSYNC_DEL_HOST_ALL%', '%VULSYNC_DEL_DISPLAY_NAME_ALL%',
      '%VULSYNC_REGI_HOST_MATCH%', '%VULSYNC_REGI_DISPLAY_NAME_MATCH%','%VULSYNC_RENEW_HOST_MATCH%', '%VULSYNC_RENEW_DISPLAY_NAME_MATCH%','%VULSYNC_DEL_HOST_MATCH%', '%VULSYNC_DEL_DISPLAY_NAME_MATCH%'
    ];

    const keywords_analysis_items = {
        'analysis-sri'      : '分析結果を出力(SRI)',
        'analysis-cvss'     : '分析結果を出力(CVSS)',
        'analysis-common'   : '分析結果を出力(SRI/CVSS共通)',
    };

    function create_keywords_inputter(){
        var wrapper = $('<div>').attr('id', 'wrap-keyword');
        var wrap_inputter = $('<div>').attr({
                                'class': 'dropup position-relative',
                                'id': 'keyword'
                            });
        var keyword_button = $('<button>').attr({
                                 'type': 'button',
                                 'class': 'btn btn-info btn-sm mr-1 dropdown-toggle text-nowrap',
                                 'id': 'keywords-list',
                                 'data-originaldropdown': 'true',
                                 'data-display': 'up',
                             }).text('キーワード');
        var original_dropdown_menu = $('<div>').attr({
                                         'id': 'keyword-main-menu',
                                         'class':'original-dropdown-menu bg-f3 position-absolute keywords-menu',
                                         'data-toggle': 'tooltip',
                                         'data-trigger': 'manual',
                                         'data-html': 'true',
                                         'data-title': '緑色の項目は、<br>' +
                                                       '「メール送信操作で対象となるデータの指定」で指定されたデータです。<br>'+
                                                       '緑色の項目以外のキーワードを使用した場合は、キーワードの内容が出力されない可能性がありますので、予めご了承下さい。'  
                                     }).hide();
        for(key in all_keywords_menus){
            var original_dropdown_item  = $('<div>').attr({
                                              'class': 'original-dropdown-item dropright flex-fill position-relative',
                                              'data-item': 'keyword-menu-btn'
                                          }).append(
                                              $('<a>').attr({
                                                  'data-menuname': key,
                                                  'data-display': (key === 'vulsync-info-all' || key == 'vulsync-info-match') ? 'right-up' : 'right',
                                                  'data-originaldropdown': 'true',
                                                  'class': 'btn btn-sm btn-outline-dark2 dropdown-toggle w-100 text-center text-nowrap',
                                               }).text(all_keywords_menus[key])
                                          );
            var keywords_menu = $('<div>').attr({
                                     'class': 'original-dropdown-menu bg-f3 position-absolute'
                                }).hide();
            if(key === 'analysis'){
                for(a_key in keywords_analysis_items){
                    var keywords = all_keywords[a_key];
                    var button_name = keywords_analysis_items[a_key]; 
                    var keyword_a_item = $('<div>').attr({
                                           'class': 'original-dropdown-item dropright flex-fill position-relative',
                                       });
                    var kwd_btn = $('<a>').attr({
                                        'class': 'dropdown-toggle btn btn-sm btn-outline-dark2 w-100 text-center text-nowrap',
                                        'data-originaldropdown': 'true',
                                        'data-display': 'right',
                                  }).text(button_name);
                    var narrow_item_menu = $('<div>').attr({
                                               'class': 'original-dropdown-menu bg-f3 position-absolute narrow-down-option',
                                           }).hide(); 
                    for(var j = 0; j < keywords.length; j++){   
                        var keyword = keywords[j];
                        var text_and_description = all_keywords_item_text_and_description[keyword];  
                        var keyword_item = $('<div>').attr({
                                               'class': 'original-dropdown-item position-relative',
                                           }).append(
                                               $('<button>').attr({
                                                   'type': 'button',
                                                   'class': 'btn btn-sm btn-outline-dark2 put-keyword w-100 text-center text-nowrap',
                                                   'data-toggle': 'tooltip',
                                                   'data-title': text_and_description[1],
                                                   'data-trigger': 'hover',
                                                   'data-delay': '{ "show": 300}',
                                                   'value': keyword
                                               }).text(text_and_description[0])
                                           ); 
                         narrow_item_menu.append(keyword_item);    
                     }
                     keyword_a_item.append(kwd_btn, narrow_item_menu);
                     keywords_menu.append(keyword_a_item);
                }
            }else{
                var keywords = all_keywords[key];
                for(var j = 0; j < keywords.length; j++){
                    var keyword = keywords[j];
                    var text_and_description = all_keywords_item_text_and_description[keyword];
                    var button_name = text_and_description[0];
                    var tooltip_description = text_and_description[1];
                    if(keywords_with_narrow_down_option.indexOf(keyword) != -1){
                        var keyword_item = $('<div>').attr({
                                               'class': 'original-dropdown-item dropright flex-fill position-relative',
                                           });
                        if( keyword.match(/^%VULSYNC_.*$/) ){
                            //脆弱性同時期のキーワードは選択式
                            //デフォルトは「全ての脆弱性」を選択
                            var all_val = 'all';
                            var regi_val = 'regi';
                            var renew_val = 'renew';
                            var del_val = 'del';
                            if( keyword.match(/.+_REGI_.+/) ){
                                keyword_item.attr({
                                    'data-vulsync': regi_val,
                                    'style': key === 'vulsync-info-all' ? 'display: none;' : ''
                                });
                            }else if( keyword.match(/.+_RENEW_.+/) ){
                                keyword_item.attr('data-vulsync', renew_val).hide();       
                            }else if( keyword.match(/.+_DEL_.+/) ){
                                keyword_item.attr('data-vulsync', del_val).hide();
                            }else{
                                keyword_item.attr('data-vulsync', all_val);
                            }
                            if( j === 0 ){
                                if( key === 'vulsync-info-all' ){
                                    var select = $('<div>').attr('class', 'original-dropdown-item flex-fill').append(
                                        $('<select>').attr({
                                            'id': 'select-vulsync-keyword',
                                            'class': 'form-control form-control-sm font-weight-bold'
                                        }).append(
                                            $('<option>').val(all_val).text('全ての脆弱性'),
                                            $('<option>').val(regi_val).text('登録された脆弱性'),
                                            $('<option>').val(renew_val).text('更新された脆弱性'),
                                            $('<option>').val(del_val).text('削除された脆弱性'),
                                        )
                                    );
                                }else{
                                    var select = $('<div>').attr('class', 'original-dropdown-item flex-fill').append(
                                        $('<select>').attr({
                                            'id': 'select-vulsync-keyword',
                                            'class': 'form-control form-control-sm font-weight-bold'
                                        }).append(
                                            $('<option>').val(regi_val).text('登録された脆弱性'),
                                            $('<option>').val(renew_val).text('更新された脆弱性'),
                                            $('<option>').val(del_val).text('削除された脆弱性'),
                                        )
                                    );
                                }
                                keywords_menu.append(select);
                            }
                        }
                        var kwd_btn = $('<button>').attr({
                                            'type': 'button',
                                            'class': 'dropdown-toggle btn btn-sm btn-outline-dark2 w-100 text-center text-nowrap',
                                            'data-display': 'narrow-down-option',
                                            'data-toggle': 'tooltip',
                                            'data-trigger': 'hover',
                                            'data-delay': '{ "show": 300}',
                                            'data-title': tooltip_description,
                                            'data-originaldropdown': 'true',
                                            'data-itemid': keyword,
                                      }).text(button_name);
                        var narrow_down_option = $('<div>').attr({
                                                     'class': 'original-dropdown-menu bg-f3 position-absolute narrow-down-option',
                                                 }).hide();
                        var no_narrow_item = $('<div>').attr({
                                                 'class': 'original-dropdown-item position-relative narrow-down-option-item',
                                             }).append(
                                                 $('<div>').attr({
                                                     'class': 'wrap-narrow-item-option d-flex justify-content-between',
                                                 }).append(
                                                     $('<div>').attr('class', 'mb-0').append(
                                                         $('<input>').attr({
                                                             'type': 'hidden',
                                                             'data-keyword': keyword,
                                                             'data-narrow': '',
                                                             'class': 'keyword-option-value',
                                                         }),
                                                         $('<span>').text('一覧を全件出力する'),
                                                         $('<span>').attr({'class': 'd-block g-mt-2',}).text('※最大100件')
                                                    ),
                                                    $('<div>').attr('class', 'my-auto').append(
                                                         $('<button>').attr({
                                                             'type': 'button',
                                                             'class': 'btn btn-xs btn-primary put-keyword ml-auto text-nowrap decide-narrow-down-option',
                                                             'data-listbtn': 'true',
                                                             'data-listkeyword': keyword,
                                                         }).text('決定')
                                                    )
                                                 ) 
                                             );
                          var narrow_item1 =  $('<div>').attr({
                                                 'class': 'original-dropdown-item position-relative narrow-down-option-item',
                                              }).append(
                                                  $('<div>').attr('class', 'wrap-narrow-item-option d-flex justify-content-between').append(
                                                      $('<div class="mb-0">' +
                                                          '先頭から' +
                                                          '<input type="number" min="1" class="mx-1 keyword-option-value" data-keyword="' + keyword + '" data-narrow=":T">' +
																													'件' +
                                                        	'<span class="d-block g-mt-2">' + '出力する ※最大100件' + '</span>' +
                                                          '</div>'
                                                      ),
                                                      $('<div>').attr('class', 'my-auto').append(   
                                                          $('<button>').attr({
                                                              'type': 'button',
                                                              'class': 'decide-narrow-down-option btn btn-xs btn-primary put-keyword ml-auto text-nowrap',
                                                          }).text('決定')
                                                      )
                                                  )
                                              );
                          var narrow_item2 = $('<div>').attr({
                                                 'class': 'original-dropdown-item position-relative narrow-down-option-item',
                                             }).append(
                                                 $('<div>').attr('class', 'wrap-narrow-item-option d-flex justify-content-between').append(
                                                     $('<div class="mb-0">' +
                                                         '末尾から' +
                                                         '<input type="number" min="1" class="mx-1 keyword-option-value" data-keyword="' + keyword + '" data-narrow=":B">' +
																												 '件' +
                                                        	'<span class="d-block g-mt-2">' + '出力する ※最大100件' + '</span>' +
                                                         '</div>'
                                                      ),
                                                      $('<div>').attr('class', 'my-auto').append(             
                                                          $('<button>').attr({
                                                              'type': 'button',
                                                              'class': 'decide-narrow-down-option btn btn-xs btn-primary put-keyword ml-auto text-nowrap',
                                                          }).text('決定')
                                                      )
                                                  )
                                             );
                          
                          var narrow_item3 =  $('<div>').attr({
                                                  'class': 'original-dropdown-item position-relative narrow-down-option-item',
                                              }).append(
                                                  $('<div>').attr('class', 'wrap-narrow-item-option d-flex justify-content-between').append(
                                                        $('<div>').attr('class', 'mb-0 my-auto').append(
                                                            $('<input>').attr({
                                                                'type': 'hidden',
                                                                'data-keyword': keyword,
                                                                'data-narrow': ':C',
                                                                'class': 'keyword-option-value',
                                                            }),
                                                            $('<span>').attr({
                                                            }).text('一覧の件数を出力する')
                                                       ),
                                                       $('<div>').attr('class', 'my-auto').append(
                                                           $('<button>').attr({
                                                               'type': 'button',
                                                               'class': 'btn btn-xs btn-primary put-keyword ml-auto text-nowrap decide-narrow-down-option',
                                                           }).text('決定')
                                                       )
                                                  )
                                              );
    
                          var narrow_item4 = $('<div>').attr({
                                                 'class': 'original-dropdown-item position-relative narrow-down-option-item',
                                             }).append(
                                                 $('<div>').attr('class', 'wrap-narrow-item-option d-flex justify-content-between').append(
                                                     $('<div class="mb-0">' +
                                                        '一覧の合計の件数から' +
                                                        '<input type="number" min="1" class="mr-1 g-mt-2 keyword-option-value" data-keyword="' + keyword + '" data-narrow=":C-">' +
																												'を引いた' +
                                                       	'<span class="d-block g-mt-2">' + '件数を出力する' + '</span>' +
                                                        '</div>'
                                                     ),
                                                     $('<div>').attr('class', 'my-auto').append(
                                                            $('<button>').attr({
                                                                'type': 'button',
                                                                'class': 'decide-narrow-down-option btn btn-xs btn-primary put-keyword ml-auto text-nowrap',
                                                            }).text('決定')
                                                     )
                                                 )
                                             );                       
                       narrow_down_option.append(no_narrow_item, narrow_item1, narrow_item2, narrow_item3, narrow_item4);
                       keyword_item.append(kwd_btn, narrow_down_option);                      
                    }else{
                        var keyword_item = $('<div>').attr({
                                              'class': 'original-dropdown-item position-relative',
                                           }).append(
                                               $('<button>').attr({
                                                   'type': 'button',
                                                   'class': 'btn btn-sm btn-outline-dark2 put-keyword w-100 text-center text-nowrap',
                                                   'data-toggle': 'tooltip',
                                                   'data-title': text_and_description[1],
                                                   'data-trigger': 'hover',
                                                   'data-delay': '{ "show": 300}',
                                                  'value': keyword
                                               }).text(text_and_description[0])
                                           );
                    }
                    keywords_menu.append(keyword_item);
                }
            }
            original_dropdown_item.append(keywords_menu);
            original_dropdown_menu.append(original_dropdown_item);
        }
        wrap_inputter.append(keyword_button, original_dropdown_menu);
        wrapper.append(wrap_inputter);
        return wrapper;
    }

    function create_keywords_replacer(){
        var keyword_replacer = $('<div>').attr({              
                                   'id': 'LoadingTarget',
                                   'class': 'position-relative',
                               }).append(
                                   $('<button>').attr({
                                       'type': 'button',
                                       'id': 'change-keywords',
                                       'class': 'btn btn-light btn-sm text-nowrap text-dark',
                                   }).text('プレビュー')
                                );
         return keyword_replacer
    }

    function create_mail_change_reset_btn(item_name){
        var mail_change_reset_btn = $('<div>').append(
                                        $('<button>').attr({
                                            'type': 'button',
                                            'class': 'btn btn-danger btn-sm ml-1 text-nowrap',
                                            'id': (item_name === 'title' ? 'reset-mail-change-title' : 'reset-mail-change-content'),
                                            'data-toggle': 'tooltip',
                                            'data-title': 'クリックすると入力した内容がリセットされます。',
                                            'data-placement': 'top',
                                            'data-trigger': 'hover',
                                            'data-delay': '{ "show": 300}',
                                        }).text('リセット') 
                                     ); 
        return mail_change_reset_btn;
    }

    $('#reset-mail-title-and-message').on('click', function(){
        $('#send-mail-title').val('');
        $('#send-mail-content').val('');
    });     

    $(document).on('click', '#reset-mail-change-title', function(){
        $('#mail-change-new-title').val('');
    });

    $(document).on('click', '#reset-mail-change-content', function(){
        $('#mail-change-new-content').val('');
    });

    $(document).on('change', '#select-vulsync-keyword', function(){
        $(this).closest('.original-dropdown-menu').find('[data-vulsync]:not([data-vulsync="' + $(this).val() + '"])').hide();
        $(this).closest('.original-dropdown-menu').find('[data-vulsync="' + $(this).val() + '"]').show();      
    });

    function setRulePositionArea(position){
        $('#rule-position-field').val(position);
    };

    //ルール編集&既存のルールを元に新規作成
    $(document).on('click', '[data-edit="rule-setting"], #create-copied-rule-btn', function(){
        reset_all_modals_related_create_rule();
        var button = $(this);
        var button_id = button.attr('id');
        if(button_id === 'create-copied-rule-btn'){
            var checked = JSON.parse($('#pages_checked').val());
            if(button.hasClass('disabled')){
                return false;
            }else if(checked.length > 1){
                alert('ルールは１つのみ選択して下さい');
                return false;
            }else if(checked.length === 0){
                alert('ルールを１つ選択して下さい');
                return false;
            }
            $('#modalCreateRule-title').text('ルール登録');
            $('#form-create-rule').attr('action', $('#rule_setting_create_path').val());
            var rule_id = checked[0];
            var rule_info = JSON.parse($('#rule-param-' + rule_id).val());
            setRulePositionArea(rule_info.rule_position);
        }else{
            $('#form-create-rule').attr('action', $('#rule_setting_modify_path').val());
            $('#execute-register-rule').text('変更');
            $('#modalCreateRule-title').text('ルール編集');
            var rule_id = button.attr('data-ruleid');
            var rule_info = JSON.parse($('#rule-param-' + rule_id).val());
            setRulePositionArea(rule_info.rule_position);    
 
            $('#rule-updated-at').val(rule_info['rule_updated_at']);
            $('#rule_id_for_modify_rule').val(rule_id);
            //ページネーションの値を保持する
            $('#form-offset').val($('#pages_offset').val());
            $('#form-per-page').val($('#pages_per_page').val());
            $('#form-filters').val($('#pages_filters').val());
            $('#form-order').val($('#pages_order').val());
        }
        $('.rule-setting-content').addClass('text-left');
        $('#modalCreateRule-set-trigger-btn').attr('class', 'btn btn-xs rule-modal-setting-btn open-register-modal btn-success set')
                                             .tooltip('disable')
                                             .text('タイミング設定済み');
        $('#modalCreateRule-set-cond-btn').attr('class', 'btn btn-xs rule-modal-setting-btn open-register-modal btn-success set')
                                          .text('条件設定済み');
        $('#modalCreateRule-set-operation-btn').attr('class', 'btn btn-xs open-register-modal btn-success set')
                                               .text('操作設定済み');
        
        var category_descs = JSON.parse($('#category-descs').val());
        var rule_name = rule_info['rule_name'];
        var flg_enable = rule_info['is_enable'];
        var category_id = rule_info['category'];
        var trigger = rule_info['trigger'];
        var condition = rule_info['condition'];
        var operation = rule_info['action'];
        var users_categories = JSON.parse($('#users_selectable_categories').val());
        var category_name = (category_id === 'all' ? '全体' : users_categories[category_id]);
        $('#form-create-rulename').val(rule_name);
        $('#main-category-id').val(category_id)
                              .attr('data-text', category_name);
        $('#rule-main-category').attr({
           'data-toggle': 'tooltip',
           'data-original-title': (category_descs[category_id] ? category_descs[category_id].desc : ''),
           'data-boundary': 'window'
        }).text(category_name).tooltip();
        $('#select-enable-rule').val(flg_enable ? 'enable' : 'disable');
        prepare_edit_trigger(trigger, users_categories);
        prepare_edit_condition(condition);
        prepare_edit_operation(operation);
        if(button_id === 'create-copied-rule-btn'){
            $('#modalCreateRule').attr('data-create', rule_name);
        }
        $('#modalCreateRule').modal('show');
    });
    //編集時のタイミングの要素設定・パラメータ設定
    function prepare_edit_trigger(trigger, users_categories){
        $('#form-create-trigger').val(JSON.stringify(trigger));
        $('#set-trigger-setting').prop('disabled', false);
        var period = trigger['period'];
        if(period === 'vulsync'){
            $('#input-set-vulsync').prop('checked', true);
            var hostname = trigger['hostname'];
            var category_id = trigger['category'];
            if(category_id === 'all'){
                var category_name = '全体';
                var selections = make_selectable_categories('select-trigger-category', true);
                selections.find('[value="all"]').prop('checked', true);
                $('#trigger-category-menu').append(selections);
                $('#trigger-category-menu [data-toggle="tooltip"]').tooltip();
                $('#trigger-category').addClass('dropdown-toggle')
                                      .prop('disabled', false);
            }else{
                var category_name = users_categories[category_id];
            }
            $('#trigger-category').attr('data-categoryid', category_id)
                                  .text(category_name);
            if(hostname != ''){
                var display_name;
                var get_display_name_path = $('#sidfm_internal_api_get_hosts_display_name_path').val();
                $.ajax({
                    type: 'POST',
                    url: get_display_name_path,
                    data: {'hostnames': JSON.stringify([hostname])}
                }).done(function(hostdata){
                    var ext_name = hostdata[hostname] ? hostdata[hostname] : '';
                    display_name = (ext_name.length > 0 ? ext_name : hostname);
                    var setting_content = trigger_display_content('vulsync', category_name, display_name, ext_name, hostname);
                     $('#rule-setting-item-trigger').html(setting_content);
                     $('#rule-setting-item-trigger').find('[data-toggle="tooltip"]').tooltip();
                     $('#search-hostname').attr({
                         'data-hostname': hostname,
                         'data-extname': ext_name
                     }).val(display_name);
                     var get_hosts_path = $('#sidfm_internal_api_get_hosts_by_category_ids_path').val();
                     $.post(get_hosts_path,
                        {
                             'keyword': display_name,
                             'category_ids': JSON.stringify([category_id])
                         },function(data) {
                             $('#select-trigger-hostname > *').remove();
                             var options = '';
                             for(var i=0; i < data.length; i++){
                                 var hostdata = data[i];
                                 var ext_name = hostdata.ext_name;
                                 var hostname = hostdata.hostname;
                                 var regexp = new RegExp(display_name.toLowerCase().replace(/\\/g, '\\\\'));
                                 var hit_name = (ext_name && ext_name.toLowerCase().match(regexp)) ? ext_name : hostname;
                                 options += '<option class="modaltrigger-hostname" data-hostname="' + hostname + '" data-extname="' + (ext_name ? ext_name : '') + '" data-toggle="tooltip" data-title="ホスト名: ' + hostname + '<br/>表示名: ' + (ext_name ? ext_name : '(未設定)') + '" data-html="true" data-boundary="window" data-placement="left">' + hit_name + '</option>';
                              }
                              $('#select-trigger-hostname').html(options).show();
                              $('#select-trigger-hostname').find('[data-toggle="tooltip"]').tooltip();
                          }
                     );
                 });
            }else{
                hostname = null;
                var setting_content = trigger_display_content('vulsync', category_name, hostname);
                $('#rule-setting-item-trigger').html(setting_content);
                $('#rule-setting-item-trigger').find('[data-toggle="tooltip"]').tooltip();
            }
        }else if(period === 'periodic-hour'){
            var time = trigger['value'];
            $('#input-set-periodic-hour').prop('checked', true);
            $('#select-periodic-hour [value="' + time + '"]').click().prop('selected', true);
            var setting_content = trigger_display_content('periodic', $('#periodic-hour-btn').text());
            $('#rule-setting-item-trigger').html(setting_content);
        }else if(period === 'periodic-interval'){
            var unit = trigger['unit'];
            $('#input-set-periodic-interval').prop('checked', true);
            $('#wrap-unit-of-periodic-interval [data-unit="' +  unit + '"]').click();
            $('#select-periodic-interval [value="' + trigger['value'] + '"]').click().prop('selected', true);
            var setting_content = trigger_display_content('interval', $('#periodic-interval-btn').text());
            $('#rule-setting-item-trigger').html(setting_content);
        }
        //$('#rule-setting-item-trigger').append($('<i>').attr('class', 'expand-btn fas fa-angle-down mr-1'));
    }       
                
    //編集時の条件の要素作成・パラメータ設定
    function prepare_edit_condition(condition){
        $('#form-create-condition, #set-cond-param').val(JSON.stringify(condition));
        create_cond_display_content(condition, 'prepare_edit');
        for(key in condition){
           if( key === 'ctag' || key === 'timerange' ){ //複数ある(可能性のある)条件
               const cp_conds = [...condition[key]];
               condition[key] = [];
               cp_conds.forEach(condData => { //カスタムタグは、配列の最後にあるものが生成されるため、このようにする
                   condition[key].push(condData);
                   if( key === 'ctag' ){
                       ctagCondInfo.dataid = condData.dataid;
                   }else if(key === 'timerange' ){
                       timerangeCondInfo.dataid = condData.dataid;
                   }
                   update_modalRegisterCondition_items(key, condition, 'prepare_edit');
               });
           }else{
               update_modalRegisterCondition_items(key, condition, 'prepare_edit');
           }
        } 
    }

    function prepare_edit_operation(operation){
        $('#form-create-operation, #set-ope-param').val(JSON.stringify(operation));
        create_ope_display_content(operation, 'prepare_edit');
        for(key in operation){
            update_modalRegisterOperation_items(key, operation, 'prepare_edit');
        }
    }

    //ルール名変更を促すツールチップを表示（既存のルールを元に新規登録機能使用時にのみ表示）
    function show_change_rule_name_message(){
        if($('#modalCreateRule').attr('data-create') === $('#form-create-rulename').val()){
            var original_tooltip = make_original_tooltip('ルール名の変更をして下さい', true);
            $('body').append(original_tooltip);
            var target = $('#form-create-rulename');
            var offset = target.offset();
            original_tooltip.offset({
                'left': offset.left + (target.outerWidth() / 2) - (original_tooltip.outerWidth() / 2),
                'top': offset.top - original_tooltip.outerHeight() - 8
            }).show();
        }
    }

    //プリセットmodal
    $('#create-rule-by-preset').on('click', function(){
        var users_categories = JSON.parse($('#users_selectable_categories').val());
        var user_auth = JSON.parse($('#user_auth').val());
        if(Object.keys(users_categories).length === 0 && !user_auth.includes("admin")){
            alert('カテゴリに属していないユーザは登録できません');
            return false;
        }
        $('#modalCreateRule-title').text('ルール登録');
        $('#presets-list-area').find('[name="preset-item"]').prop('checked', false);
        $('#decide-preset-rule').prop('disabled', true);
        $('#modalPresetList').modal('show');
    });

    $('[name="preset-item"]').on('change', function(){
        if($('[name="preset-item"]:checked').length === 1){
            $('#decide-preset-rule').prop('disabled', false);
        }
    });
    
    $('#decide-preset-rule').on('click', function(){
        reset_all_modals_related_create_rule();
        $('#form-create-rule').attr('action', $('#rule_setting_create_path').val());
        $('.rule-setting-content').addClass('text-left');
        $('#modalCreateRule-set-trigger-btn').attr('class', 'btn btn-xs rule-modal-setting-btn open-register-modal btn-success set')
                                             .tooltip('disable')
                                             .text('タイミング設定済み');
        $('#modalCreateRule-set-cond-btn').attr('class', 'btn btn-xs rule-modal-setting-btn open-register-modal btn-success set')
                                          .text('条件設定済み');
        $('#modalCreateRule-set-operation-btn').attr('class', 'btn btn-xs open-register-modal btn-success set')
                                               .text('操作設定済み');
        var preset = $('#presets-list-area').find('[name="preset-item"]:checked');
        if(preset.length === 0){
            alert('プリセットを選択して下さい');
            return false;
        }
        var preset_info = JSON.parse(preset.val());
        var rule_name = preset_info['rule_name'];
        var flg_enable = preset_info['is_enable'];
        var category_id = preset_info['category'];
        var trigger = preset_info['trigger'];
        var condition = preset_info['condition'];
        var operation = preset_info['action'];
        var users_categories = JSON.parse($('#users_selectable_categories').val());
        var user_auth = JSON.parse($('#user_auth').val());
        if(Object.keys(users_categories).length === 0 && !user_auth.includes("admin")){
            alert('カテゴリに属していないユーザは登録できません');
            return false;
        }
        var category_name = category_id === 'all' ?  '全体' : users_categories[category_id];
        $('#form-create-rulename').val(rule_name);
        $('#main-category-id').val(category_id)
                              .attr('data-text', category_name);
        $('#rule-main-category').text(category_name);
        $('#select-enable-rule').val(flg_enable ? 'enable' : 'disable');
        prepare_edit_trigger(trigger, users_categories);
        prepare_edit_condition(condition);
        prepare_edit_operation(operation);
        $('#modalPresetList').modal('hide');
        $('#modalCreateRule').modal('show'); 
    });
    //ルール一括削除
    $('#bulk-delete-rules').on('click', function(){
        var checked = JSON.parse($('#pages_checked').val());
        if(checked.length > 0){
            var rule_names = [];
            for(var i = 0; i < checked.length; i ++){
                var rule_id = checked[i];
                var rule_info = JSON.parse($('#rule-param-' + rule_id).val());
                rule_names.push('「' + rule_info['rule_name'] + '」');
            }
            var conf = confirm(rule_names.join(' ,') + 'を削除します。よろしいですか？');
            if (conf){
                dynamicSummaryTable.executeProcess({
                    paramsData: { delete_ids: JSON.stringify(checked) },
                    path: $('#bulk-delete-form').attr('action'),
                    additionalAlertMessage: '削除に失敗しました。'
                });
                return false;
                //$('#bulk-delete-rule-ids').val(JSON.stringify(checked));
                //$('#bulk-delete-form').submit();
            }else{
                return false;
            }
        }
    });

   //クリップボードjs
   const clipboard = new ClipboardJS('.copybtn');
   const clipboard_supported = ClipboardJS.isSupported();
   //クリップボードがサポートされていなければアラートを表示
   $(document).on('mousedown', '.copybtn', function(e) {
       if(clipboard_supported === false) {
           alert('このブラウザではコピー機能はサポートされておりません');
           return false;
       }
   }); 

   clipboard.on('success', function(e) {
        var copy_btn = $(e.trigger);
        var parent_elm_has_tooltip = copy_btn.closest('.wrap-copybtn[data-toggle="tooltip"]');
        if(parent_elm_has_tooltip){
            parent_elm_has_tooltip.tooltip('hide');
        }
        copy_btn.tooltip('show');
        setTimeout(function(){
            copy_btn.tooltip('hide');
            e.clearSelection();
        },1000);
   });

   clipboard.on('error', function(e) {
       alert('コピーに失敗しました');
       e.clearSelection();
   });

   //dropdown表示時にoriginal-popover削除
   $('#SummaryTable .dropdown').on('show.bs.dropdown', function(){
       $('.rule-setting-original-popover').remove();
   });

   //enterでformが送信されるのを防ぐ
   $('.modal input').on('keydown', function(e) {
       if( (e.which && e.which === 13) || (e.keyCode && e.keyCode === 13) ){
           return false;
       } else {
           return true;
       }
   });

   //有効/無効一括変更
   $('#bulk-update-rule-flg').on('click', function(){
       var checked = JSON.parse($('#pages_checked').val());
       $('#bulk-update-rule-ids').val($('#pages_checked').val());
       $('#modalBulkUpdateRuleFlg-tbody > *').remove();
       var elms = [];
       checked.forEach(function(rule_id){
           var rule_info = JSON.parse($('#rule-param-' + rule_id).val());
           elms.push( 
               $('<tr>').append(
                   $('<td>').attr('class', 'text-left').text(rule_info['rule_name'])
               )
           );
       });
       elms.sort(function(a, b){
           if( a.text() < b.text()){
               return -1;
           }else{
               return 1;
           }
       });
       $('#bulk-update-rule-wrap-table').scrollTop(0);
       $('#modalBulkUpdateRuleFlg-tbody').append(elms);
       $('#select-bulk-update-rule-flg option:first').prop('selected', true);
       $('#modalBulkUpdateRuleFlg').modal('show');
   });
  
   $('#modalBulkUpdateRuleFlg').on('hide.bs.modal', function(){
       $('#bulk-update-rule-wrap-table').scrollTop(0);
   });
 
   $('#form-bulk-update-rule-flg').on('submit', function(e){
       e.preventDefault();
       const paramsData = {
           ids: $('#bulk-update-rule-ids').val(),
           flg: $('#select-bulk-update-rule-flg').val()
       };
       dynamicSummaryTable.executeProcess({
           paramsData,
           path: $(this).attr('action'),
           additionalAlertMessage: '更新に失敗しました。'
       });
       return false;
       /*$('#bulk-update-rule-offset').val($('#pages_offset').val());
       $('#bulk-update-rule-perpage').val($('#pages_per_page').val());
       $('#bulk-update-rule-order').val($('#pages_order').val());
       $('#bulk-update-rule-filters').val($('#pages_filters').val());*/
   });
           
}); 
