#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

class CreateSidfmdHostViews < ActiveRecord::Migration

  def up
    # ホスト一覧参照用テーブル
    create_table :sidfmd_host_views do |t|

      # :kb
      # 現在は未使用(と思われます)
      t.boolean :kb, null: false

      # :sri_cvss_status_ids
      # sidfmd_hv_scsidsテーブルへ

      # :sidfmd_host_id
      t.references :sidfmd_host, foreign_key: true, index: true, null: false

      # :hostname
      t.text :hostname, null: false, index: true

      # :ext_name
      t.text :ext_name, index: true

      # :display_name
      t.text :display_name, null: false, index: true

      # :last_checked_at
      t.datetime :last_checked_at

      # :os
      t.text :osname, index: true
      # OSの詳細データは以下の2種類で、どちらかしか無い（データが削除され
      # る等でどちらも存在しないケースもあります）ので、予め想定して利用
      # すること。
      # 意図的にrefernecesしてません。利用する際は利用する箇所で自分で取得
      # します。なお、利用時には値に対応するデータが存在しない場合がある
      # ことに注意してください。
      # SidfmdQueueDist#id - 以前の:distの場合
      t.integer :queue_dist_id
      # SIDfmdProduct#pid, SidfmdVersion#vid - appの場合
      t.integer :product_id
      t.integer :version_id

      # :exploitable
      t.boolean :exploitable

      # :cvss
      # sidfmd_hv_cvssesテーブルへ

      # :max_sri
      t.integer :max_sri, index: true

      # :softek
      # sidfmd_hv_risks, sidfmd_hv_areasテーブルへ

      # :due_dates
      # sidfmd_hv_due_datesテーブルへ

      # :due_dates_unfinished
      # sidfmd_hv_unf_due_datesテーブルへ

      # :created_ons
      # sidfmd_hv_created_onsテーブルへ

      # 対応予定日設定率
      t.float :duedates_percent, index: true

      # 対処率
      t.float :fixed_percent, index: true

      # :system_category
      t.integer :system_category_id, index: true
      t.text :system_category_name, index: true

      # :no_system_categories
      # sidfmd_hv_categoriesテーブルへ

      # :max_area_id
      t.integer :max_area_id, index: true

      # :status_ids
      # sidfmd_hv_scsidsテーブルから取得。
      # ex. self.sidfmd_hv_scsids.map(&:status_id)

      # :min_status_id
      t.integer :min_status_id, index: true

      # :max_status_id
      t.integer :max_status_id, index: true

      # :max_cvss
      t.boolean :max_cvss_exploitable, index: true
      t.float :max_cvss_score, index: true
      t.text :max_cvss_label, index: true

      # :warning_orig
      t.boolean :warning_orig_sri
      t.boolean :warning_orig_cvss

      # :warning
      t.boolean :warning_sri, index: true
      t.boolean :warning_cvss, index: true

      # :fixed_warning
      t.boolean :fixed_warning_sri
      t.boolean :fixed_warning_cvss

      # :not_fixed_warning
      t.boolean :not_fixed_warning_sri
      t.boolean :not_fixed_warning_cvss

      # 以下は保存したデータは、アクセス時には正確なデータではなくなっているため注意。
      # 最終的にはcreated_onsからSQLで計算して取得してください。
      # 以下は計算例です。
      # item[:elapsed_days] = -1
      # min_elapsed_created_on = item[:elapsed_created_ons].min
      # if min_elapsed_created_on then
      #   elapsed_days = ((Time.current - min_elapsed_created_on) / 24.0 / 60.0 / 60.0).to_i
      #   if item[:elapsed_days] < elapsed_days then
      #     item[:elapsed_days] = elapsed_days
      #   end
      # end
      # :elapsed_created_ons - Issue作成日時からの経過時間 - elapsed_days算出に使用
      #   min_elapsed_created_onsしか使わないので、それだけ保存しています。
      #   再計算する際は、この値と現在時刻で計算してください。
      t.datetime :min_elapsed_created_on
      # :elapsed_days - Issue作成日時からの経過日数。
      # SidfmdHostViewの更新時点のものなので、表示時点では古くなっている可能性が高いです。
      # テスト表示用のみで利用して、最終的にはmin_elapsed_created_onから計算した値を
      # 出力してください。
      t.integer :elapsed_days

      # :tl_issue_ids - 履歴(Journal)のあるIssue#idの一覧 - 現在未使用

      t.timestamps null: false
    end

    # :sri_cvss_status_ids
    # status_id と cvss と sri をセットで保持しているテーブル。
    # 「特定のstatus_idについて、cvssの最高値を取得」のような時に利用する。
    # FIXME: status_idsやmax_css, max_sri等はこれから生成しているので、集約していく予定。
    create_table :sidfmd_hv_scsids do |t|
      t.references :sidfmd_host_view, foreign_key: true, index: true, null: false

      # ステータスID(1-4)
      t.integer :status_id, null: false, index: true

      # CVSS値(0.0-10.0)
      t.float :cvss, null: false, index: true

      # SRI値(1-4)
      t.integer :sri, null: false, index: true
    end

    # :cvss
    # cvss情報を保持しているテーブル。
    # labelはSidfmdIssueCveの定数を参照。
    create_table :sidfmd_hv_cvsses do |t|
      t.references :sidfmd_host_view, foreign_key: true, index: true, null: false

      t.boolean :exploitable
      t.float :score
      t.text :label
    end

    # :softek
    # risk_id(ソフテックの脆弱性危険度。脆弱性自体の危険度で、sriとは異なるので注意）の
    # 配列を保持しているテーブル。
    create_table :sidfmd_hv_risks do |t|
      t.references :sidfmd_host_view, foreign_key: true, index: true, null: false
      t.integer :risk_id
    end

    # :softek
    # area_id(影響範囲)の配列を保持しているテーブル
    create_table :sidfmd_hv_areas do |t|
      t.references :sidfmd_host_view, foreign_key: true, index: true, null: false
      t.integer :area_id
    end

    # :no_system_categories
    # カスタムカテゴリを保持しているテーブル
    # システムカテゴリは、SidfmdHostView#system_category_id/nameを参照。
    create_table :sidfmd_hv_categories do |t|
      t.references :sidfmd_host_view, foreign_key: true, index: true, null: false
      t.integer :sidfmd_category_id, index: true
      t.text :name
    end

    # :created_ons
    # 関連付けられているIssueのcreated_onを保持しているテーブル
    # 経過日数を計算するために利用される。
    create_table :sidfmd_hv_created_ons do |t|
      t.references :sidfmd_host_view, foreign_key: true, index: true, null: false
      t.datetime :dt
    end

    # :due_dates
    # 対策予定日の一覧
    create_table :sidfmd_hv_due_dates do |t|
      t.references :sidfmd_host_view, foreign_key: true, index: true, null: false
      t.date :d
    end

    # :due_dates_unfinished
    # ステータスが終了状態になっていない脆弱性の対策予定日の一覧
    create_table :sidfmd_hv_unf_due_dates do |t|
      t.references :sidfmd_host_view, foreign_key: true, index: true, null: false
      t.date :d
    end
  end

  def down
    drop_table :sidfmd_hv_unf_due_dates
    drop_table :sidfmd_hv_due_dates
    drop_table :sidfmd_hv_created_ons
    drop_table :sidfmd_hv_categories
    drop_table :sidfmd_hv_risks
    drop_table :sidfmd_hv_areas
    drop_table :sidfmd_hv_cvsses
    drop_table :sidfmd_hv_scsids
    drop_table :sidfmd_host_views
  end

end
